#include "formula.h"


int Formula::AD = 0;
 
template <class T>
FormulaList<T>::FormulaList() {
  capacity = 0;
  lsize = 0;
  list = NULL;
}

template <class T>
void FormulaList<T>::print() {
  for (unsigned i = 0; i < lsize; i++)
    cout << list[i] << " ";
  cout << endl;
}

template <class T>
FormulaList<T>::FormulaList(int size, T f) : capacity(size), lsize(size) {
  assert(lsize >= 0);
  if (lsize > 0) {
    list = new T[lsize];
    list[0] = f;
  } else
    list = NULL;
}

template <class T>
T FormulaList<T>::operator[](unsigned i) {
  return list[i];
}

template <class T>
void FormulaList<T>::operator = (FormulaList<T> &that) {
  resize(that.size());
  memcpy(list, that.list, sizeof(T) * lsize);
}

template <class T>
T FormulaList<T>::at(unsigned i) {
  assert(i < lsize);
  return list[i];
} 

template <class T>
void FormulaList<T>::clear() {
  delete [] list;
  list = NULL;
  lsize = 0;
  capacity = 0;
}

template <class T>
void FormulaList<T>::set(unsigned i, T f) {
  assert(i <= lsize && lsize > 0);
  if (i == lsize)
    push(f);
  list[i] = f;
}

template <class T>
void FormulaList<T>::set(unsigned i, T f, int times) {
  assert(i < lsize && lsize > 0);
  assert(sizeof(T) == sizeof(char));
  memset(list + i, (int)f, times);
}

template <class T>
void FormulaList<T>::push(T f) {
  if (capacity == 0)
    capacity = 2;
  if (list == NULL)
    list = new T[capacity];
  else if (lsize == capacity)
    grow(capacity + capacity);
  list[lsize++] = f;
}  

template <class T>
void FormulaList<T>::push(FormulaList<T> &t) {
  int tsize = t.size();
  if (lsize + tsize > capacity)
    grow(lsize + tsize);
  memcpy(list + lsize, t.list, sizeof(T) * tsize);
  lsize += tsize;
}

template <class T>
T FormulaList<T>::pop() {  
  assert(lsize > 0);
  lsize--;
  return list[lsize];
}

template <class T>
void FormulaList<T>::erase(unsigned ind) {
  assert(ind < lsize);
  lsize--;
  if (lsize == 0)
    return;
  list[ind] = list[lsize];
}

template <class T>
void FormulaList<T>::grow(unsigned newsize) {
  capacity = newsize;
  T* tlist = list;
  list = new T[capacity];
  if (lsize > 0)
    memcpy(list, tlist, sizeof(T) * lsize);
  if (tlist != NULL)
    delete [] tlist;
}

template <class T>
unsigned FormulaList<T>::size() {
  return lsize;
}

template <class T>
void FormulaList<T>::resize(unsigned newsize) {	
  if (newsize > capacity)
    grow(newsize);
  lsize = newsize;
}

template <class T>
bool FormulaList<T>::operator == (FormulaList &t) {
  if (lsize != t.size()) 
    return false;
  return memcmp(list, t.list, sizeof(T) * lsize) == 0;
}

template <class T>
FormulaList<T>::~FormulaList() {
  if (list != NULL)
    delete [] list;
}

template <class T>
void FormulaList<T>::sort() {
  ::sort(list, list + lsize);
}

template class FormulaList<Formula*>;
template class FormulaList<int>;
template class FormulaList<bool>;
template class FormulaList<FaustInferenceRules>;
template class FormulaList<VARLIST>;
template class FormulaList<FeatureIndexNode*>;

//----------------------------------------------------------------------

Formula::Formula() {
  id = this;
  parent = 0; 
  term = NULL;
  vars = NULL;
  signature = NULL; 
  depth = 0; 
  weight = 1;
  crappiness = 1.0;
  symmetric = false;
  ftype = _generic;
  arity = 0;
  ADlocal = 1;
  bflatterm = NULL;
}

Formula::Formula(pIndex p, FaustFormulaType _type, int wKBO, int pKBO, int weight,
		 FormulaTerm *newTerm, VARLIST_LIST newMap, int arity, bool sym) 
  : id(NULL), parent(p), term(newTerm), vars(newMap), signature(NULL),
    depth(0), weight(weight), symmetric(sym), wKBO(wKBO), pKBO(pKBO),
    arity(arity), ftype(_type) {
  id = this;
  ADlocal = AD++;
  double t = (double)(weight -1) * 1.2;
  crappiness = t/(double)arity + (double)((arity - 3)*.5);
  bflatterm = NULL;
}

Formula::~Formula() {
  if (bflatterm)
    delete bflatterm;
  if (signature) 
    delete signature;
  if (vars) {
    for (unsigned i = 0; i < vars->size(); i++)
      delete (*vars)[i];
    delete vars;
  }
  for (vector<FInference*>::iterator i = ancestor.begin(); i != ancestor.end(); i++)
    delete *i;
}

FormulaSignature *Formula::symbolCount() {
  if (!signature) {
    signature = new FormulaSignature;
    signature->add(this);
  }
  return signature;
}

Formula* Formula::getId() {
  while (id != id->id)
    id = id->id;
  return id;
}

//called for particular literal. given the VARLIST for that literal
int Formula::getFlattermHelper(TERMLIST flatterm, VARLIST vmap, int index){	
  for (FormulaTerm* t = term; t && t->isTerm(); t = t->prevTerm())
    index = t->f->getFlattermHelper(flatterm, vmap, index);
  //carefull!!! usage of this "variable" overshadows the Datatypes.h 
  if (this->ftype == _universal)
    flatterm->set(index, (Formula*)(vmap->at(depth)));
  else
    flatterm->set(index, this);
  return --index;
}

void Formula::makeFlatterm() {
  if (arity == 0 || bflatterm != NULL)
    return;
  //make it as big as it needs to be (-1 because no need in OR symbol)
  bflatterm = new TermList(weight-1, NULL);
  int index = weight - 2;
  //each i->second is vmap of a particular literal
  int i = 0;	  
  for (FormulaTerm* t = term; t && t->isTerm(); t = t->prevTerm()){    
    index = t->f->getFlattermHelper(bflatterm, (*vars)[i++], index);
  }
}

//takes an empty flatterm, goes over each entry in VARMAP and fills flattern
//stores initial location of literals in lit_loc
//does not replaces the data in flatterm and lit_loc, but appends it
void Formula::getFlatterm(TermList &flatterm, VarList &lit_loc) {
  makeFlatterm();
  //store locations of literals
  for(int index = 0; index < weight-1; index += bflatterm->at(index)->weight)
    lit_loc.push(index + flatterm.size());
  //append the literals to flatterm
  flatterm.push(*bflatterm);
}

inline FormulaTerm* Formula::firstTerm() {
  if (ftype == _negation)
    return term->f->term;
  return term;
}


//----------------------------------------------------------------------------
void FormulaSignature::print() {
/*  for (map<pIndex, float>::iterator i = count.begin(); i != count.end(); i++)
    cout << (*i).first->NameSymbol << "/" << (*i).first->Arity 
	 << "/" << (*i).second << "  ";
  cout << endl;
*/
}

//FIX
void FormulaSignature::add(Formula* formula) {
  add(formula->parent);
  for (FormulaTerm* t = formula->term; t && t->isTerm(); t = t->prevTerm())
    add(t->f->symbolCount());	  
}

void FormulaSignature::add(pIndex new_symbol, float c){
  map<pIndex, float>::iterator i = count.find(new_symbol);
  if (i == count.end())
    count[new_symbol] = c;
  else
    (*i).second += c;
}

void FormulaSignature::add(FormulaSignature* merger) {
  for (map<pIndex, float>::iterator i = merger->count.begin();
       i != merger->count.end(); i++)
    add((*i).first, (*i).second);
}

float FormulaSignature::find(pIndex target) {
  map<pIndex, float>::iterator i = count.find(target);
  if (i != count.end()) return (*i).second;
  return 0;
}

void FormulaSignature::erase(pIndex old_symbol) {
  count.erase(old_symbol);
}

void FormulaSignature::clear() {
  count.clear();
}

//----------------------------------------------------------------------------

HB::HB(int name, SYMBOLNODE _data, FaustFormulaType ft) : name(name),
  Arity(_data->Arity), data(_data), type(ft), NumberOfUses(0), _equality(0) {
  symmetric = false;
  KBOweight = 1;
  NameSymbol = "";
  NameSymbol.append(data->NameSymbol);
}

HB::HB(int name,char* _NameSymbol, int Arity, FaustFormulaType ft) 
  : name(name), Arity(Arity), data(NULL), type(ft), NumberOfUses(0), 
    _equality(0) {
  symmetric = false;
  KBOweight = 1;
  NameSymbol = "";
  NameSymbol.append(_NameSymbol);
}

HB::HB() { data = NULL; };

void HB::set(SYMBOLNODE _data, FaustFormulaType ft) {
  Arity = _data->Arity;
  data = _data;
  type = ft;
  NumberOfUses = 0; 
  _equality = false;
  NameSymbol = "";
  NameSymbol.append(data->NameSymbol);
}

void HB::set(char* _NameSymbol, int _Arity, FaustFormulaType ft) {
  Arity = _Arity;
  data = NULL; 
  type = ft;
  NumberOfUses = 0; 
  _equality = false;  
  NameSymbol = "";
  NameSymbol.append(_NameSymbol);  
}

void HB::print() {
  cout << NameSymbol << endl;
}

//do NOT try to delete content of storage!!! it is MANAGED by problem.C
HB::~HB () { }

void HB::operator = (HB &source) {
  name = source.name;
  Arity = source.Arity;
  data = source.data;
  type = source.type;
  NumberOfUses = source.NumberOfUses;
  _equality = source._equality;
  NameSymbol = source.NameSymbol;
}

//--------------------------------------------------------------------

Formula* FormulaTerm::add(pIndex p, vector<HB> &s) {
  if (!principal)
    principal = new pmap;
  else if (principal->find(p) != principal->end())
    return (*principal)[p];
    Formula* newF = new Formula(p, s[p].type, s[p].KBOweight + KBO, 
		    s[p].KBOweight, 1+weight, this, NULL, depth,s[p].symmetric);
  (*principal)[p] = newF;
  return newF;
}

Formula* FormulaTerm::add(pIndex p, vector<HB> &s, VARLIST_LIST vars) {
  if (!disj)
    disj = new TermList;
  if (!vars)
    vars = new VarList_List;
    Formula* newF = new Formula(p, s[p].type, s[p].KBOweight + KBO, 
		    s[p].KBOweight, 1 + weight, this, vars, depth, 
		    s[p].symmetric);
  disj->push(newF);
  return newF;
}

FormulaTerm* FormulaTerm::nextNode(Formula* g) {
  assert(g);
  tmap* st;
  if (!next) {
    next = new smap;
    (*next)[g->parent] = st = new tmap;
  } else {
    smap::iterator si = next->find(g->parent);
    if (si == next->end())
      (*next)[g->parent] = st = new tmap;
    else
      st = si->second;
  }
  FormulaTerm* newTerm;
  tmap::iterator ti = st->find(g->term);
  if (ti == st->end()) 
    (*st)[g->term] = newTerm = new FormulaTerm(g, this, depth+1, KBO, weight);
  else
    newTerm = ti->second;
  return newTerm;    
}

FormulaTerm::FormulaTerm(Formula* f, FormulaTerm* prev, int depth, 
			 int prevKBO, int prevWeight) 
  : prev(prev), next(0), principal(0), disj(0), depth(depth), 
    KBO(prevKBO + f->wKBO), weight(prevWeight + f->weight), f(f) { }

FormulaTerm::~FormulaTerm() {    
  if (principal) {
    //delete the formulas, if this is a leaf
    pmap::iterator i = principal->begin();
    while (i != principal->end()) {
      if (i->second->ftype != _universal)  //special case
	delete i->second;
      i++;
    }
    delete principal;
  }
  if (next) {
    //go over all branches, delete them
    smap::iterator i = next->begin();
    while (i != next->end()) {
      tmap::iterator j = i->second->begin();
      while (j != i->second->end()) {
	delete j->second;
	j++;
      }
      delete i->second;
      i++;
    }
    delete next;
  }
  if (disj) {
    for (unsigned i = 0; i < disj->size(); i++)
      delete disj->at(i); 
    delete disj;
  }
}

//----------------------------------------------

FeatureIndexNode::~FeatureIndexNode() {
  for (fin::iterator i = next.begin(); i != next.end(); i++)
    delete i->second;
  if (flist)
    delete flist;
}

void FeatureIndex::add(Formula* f) {
  FormulaSignature* s = f->symbolCount();
  FeatureIndexNode* n = &root;
  for (int i = from; i < to; ) {
    int occ = (int) s->find(i);
    fin::iterator j = n->next.find(occ);
    if (j != n->next.end()) 
      n = j->second;
    else
      n = n->next[occ] = new FeatureIndexNode;
    i++;
    if (i == uniStart)
      i += MAX_UNIVERSAL_VARS;		  
  }
  if (!n->flist) 
    n->flist = new TermList;
  n->flist->push(f);
}

void FeatureIndex::query(bool fb, int arity, FormulaSignature &fsig, ARITYq &fset) {
  iterstack.clear();
  indexstack.resize(0);
  nstack.resize(0);
  fin::iterator k, knext;
  int i = from;
  FeatureIndexNode *n = &root;

  do {
    //1.
    if (n->flist == NULL) {
      //if this is not the end of path, find correct branch
      if (fb)
	k = n->next.lower_bound((int) fsig.find(i));
      else 
	k = n->next.upper_bound((int) fsig.find(i));
    } 
    if (n->flist || (fb && k == n->next.end()) || (!fb && k == n->next.begin())) {
      //   if this is end of path || if there is no branch, pop one from stack
      if (iterstack.size() == 0)
	return;
      k = iterstack.back();
      iterstack.pop_back();
      i = indexstack.pop();
      n = nstack.pop();
    }
    //2.
    if (fb) {
      //push the one above it (for backward)
      knext = k; knext++;
      if (knext != n->next.end()) {
	iterstack.push_back(knext);
	indexstack.push(i);
	nstack.push(n);
      }      
    } else {
      // push the current one (will be decremented later) (for forward)
      k--;
      knext = k;      
      if (knext != n->next.begin()) {
	iterstack.push_back(knext);
	indexstack.push(i);      
	nstack.push(n);
      }      
    }
    //3. go deeper into the tree
    n = k->second;
    i++;
    //   jump over universal variables (to skoeloms)
    //    if ((*symbols)[i].type == _universal) 
    if (i == uniStart)
      i += MAX_UNIVERSAL_VARS;
    //4. if brach ended, suck out the formulae
    if (n->flist) {
      int fsize = n->flist->size();
      for (int j = 0; j < fsize; ) {
	Formula* f = n->flist->at(j);
	//the formula was subsumed. delete it from the index
	//(the subsumer will eventually show up too)
	if (f->isSubsumed()) {
	  n->flist->erase(j);
	  if (f->bflatterm) {
	    delete f->bflatterm;
	    f->bflatterm = NULL;
	  }
	  fsize--;
	  continue;
	  //the formula is new. take it
	} else if ((fb && f->arity >= arity) || (!fb && f->arity <= arity)) {
	  fset.push_back(f);
	}
	j++;
      }
    }      
  } while (true);
}


