/*---------------------------------------------------------------------------*
 * FILE: problem.h
 * AUTHOR: Yury Puzis
 *---------------------------------------------------------------------------*/
#ifndef _PROBLEM_H_
#define _PROBLEM_H_

#define jjf FormulaUnion.BinaryFormula
#define jjnext AnnotatedFormula->AnnotatedFormulaUnion.AnnotatedTSTPFormula.FormulaWithVariables->Formula

#include "ConvertToECNF.h"
#include "formula.h"
#include "unifier.h"
#include <map>

#define CONNECTIVES_NUM 4

#define f_and symbols.begin()
#define f_for (symbols.begin() + 1)
#define f_or (symbols.begin() + 2)
#define f_forall (symbols.begin() + 3)

#define CON_EQL "equal"
#define CON_IMP "imply"
#define CON_NOT "~"
#define CON_AND "and"
#define CON_ALL "all"
#define CON_SOM "some"
#define CON_ORR "or"
#define CON_FOR "or"
//FOR == FOF OR, disjunctions from the original problem

#define termChar(i)  (i >= MAX_VARS_IN_FORMULA || i <= -MAX_VARS_IN_FORMULA)

struct VARSTACK {
  Formula* fnode;
  SYMBOLNODE jjnode;
};

class Problem {
 private:
  char* filename;          //file name of the problem (optional)
  SIGNATURE Signature;     //used to work with JJParser
  VARSTACK* stack;      //holds "currently" quant. vars (when parsing,printing)
  int last_in_stack;       //last position in stack (unoccupied)
  Formula univars[MAX_UNIVERSAL_VARS];   //memory slots for universal variables

  //the entry point to repository of all the formulae
  FormulaTerm dataStore[4];

  //those containers are set up during the initialization of the problem
  TERMLIST
    facts,		//list of non-conjectures
    questions,          //list of conjectures
    contra,	        //list of negated conjectures
    formulae;	        //list of all the formulas (in original order)

  vector<StatusType> status; 	//list of statuses correspoding to formulae
  vector<char*> cname;		//list of pointers to names of formulae
  
  LISTNODE Head;           //used to work with JJParser
  unsigned int LS;         //total number of entries in symbols[]
  unsigned int predSymbols;//number of unsigned unique predicate symbols
  unsigned int funcSymbols;//number of original (not schoeloms) unique symbols
  bool TrySubsumption;  
  Formula empty;
  pIndex cnfStore, andStore, forStore, ancStore;

  //----------------------------------------------------------------------
  pIndex find(SYMBOLNODE s);

  //parsing the original JJParser format
  Formula* parse (FORMULA f);
  FormulaTerm* parse (FormulaTerm* t, FORMULA left, FORMULA right, 
		      ConnectiveType cont);
  Formula* parse (TERM t);
  Formula* handleArgument(TERM arg);
  
  //parents printing functions
  void constructSymbol(pIndex p, char* buffer, int &i);
  void constructLogic(Formula* f, char* buffer, int &i);
  void constructFlatLogic(Formula* f, char* buffer, int &i);
  void constructParents(Formula* f, char* buffer, int &i);
  void makeDerivationTree(Formula* f, FSet &tree);
//  void makeDerivationTreeOptimized(Formula* f, FSet &tree);
//  void makeDerivationTreeOptimizedHelper(Formula* f, vector<FSet *> &variant);
  //when formula is not a disjunction, return disjunction instead
  Formula* getDisjunction(Formula* n);
  Formula* makeFOFDisjunction(Formula* f);

  void printFunctorDetails(int i);
  void printPredicateDetails(int i);
  int printFormulaName(Formula* f, char* buffer);
  int printFormulaInference(Formula* f, char* buffer);

  //database manipulation
  void addConjunction(Formula &f, Formula* parent, TermList &out);
  Formula* addFlatterm(pIndex parent, int &i, TermList &flatterm, 
		       VarList_List &varset);

  VARLIST_LIST mirrorVars(FormulaTerm *t);
  VARLIST mirrorVarsHelper(Formula* f, VARLIST v);

public:
  pIndex equalsIndex;
  Formula theoryequality;
  FeatureIndex *featureIndex;
  vector<HB> symbols;
  //indexing structure for subsumption
  
  Problem(char* filename, bool SplitConjunctions, bool deepParsing);
  ~Problem();

  //true if f1 and f2 are opposites of each other
  bool are_opposite(pIndex f1, pIndex f2);
  pIndex getOpposite(pIndex h);
  FaustFormulaType getType(pIndex h);

  //adds arbitrary information to the formula in the original problem
  void storeString(int ClauseNumber, char *symbol);

  //information retrieval routines
  bool problemSolved();
  int problemSize() { return formulae->size(); }
  StatusType getClauseStatus(unsigned ClauseNumber);
  char* getClauseName(unsigned ClauseNumber);
  Formula& getClause(unsigned ClauseNumber);
  void getProblem(TermList &list);	//original clauses
  void getFacts(TermList &list);	//only facts
  void getConjecture(TermList &list, bool sign = true); 

  //printing routines
  void printFormula(Formula* f, bool extra = false);
  void printFormulaList(TERMLIST l, bool extra = false);
  void printDerivationTree(Formula* l = NULL); //if null then from empty clause
  void print();  //print everything about the problem

  //from local database - result of resolution or simplification
  Formula* addFlatClause(TermList &flatterm, VarList_List &varset, 
			 FInference* ancestors = NULL);
  //from different database
  Formula* addForeign(Formula* f, Formula* ancestor = NULL);
  void addAncestors(Formula* f, TERMLIST ancestors);

  void setDepth(Formula* f, int newDepth);

  struct FormulaBag {
   private:
    void queryHelper(FSet &curset, FSet &fout, FSet &ftbu);
    inline bool detectInEquals(Formula* t);
	   
   public: 
    vector<FSet> *vset;
    vector<FSet> *eset;    //all Functors inside $equals (but not ~$equals) go Only here

    Problem* p;
    
    FormulaBag(Problem* problem);    
    ~FormulaBag();
    void add(Formula* f);
    void query(Formula* f, FSet &fout, FSet &ftbu);
    void queryEquality(Formula* f, FSet &fout, FSet &ftbu);	
    void erase(Formula* f);
  };

  FormulaBag* newBag() { return new FormulaBag(this); }

  struct FormulaPair {
    Formula* f;
    Formula* g;
    int hash;

    FormulaPair() : f(0), g(0), hash(0) {}

    FormulaPair(Formula *f, Formula* g);

    bool operator== (const FormulaPair &that) const {
      return f == that.f && g == that.g;
    }

    int operator() (const FormulaPair &pair) const {
      return pair.hash;
    }

  };  

  struct RootTable {
    //hash_map<FormulaPair, TERMLIST, FormulaPair> roots;
    hash_map<FormulaPair, FSET, FormulaPair> roots;
    
    //will return NULL TERMLIST if no resolution has been made yet
    //will return empty TERMLIST if resolution failed
    //will return non-empty TERMLIST if resolvants are known
    //TERMLIST get(Formula* f, Formula* g);
    //    void add(Formula* f, Formula* g, TERMLIST t);
    FSET get(Formula* f, Formula* g);
    void add(Formula* f, Formula* g, FSET t);

    ~RootTable();

  } fullResolutionTable, paramodulationTable;

  struct SoloTable {
    hash_map<Formula*, FSET, FormulaHash> roots;

    FSET get(Formula* f);
    void add(Formula* f, FSET t);

    ~SoloTable();

  } equalityResolutionTable, equalityFactoringTable;

  Formula* doSubsumption(Unifier* u, FInference *ancestors, 
			 TermList &subsumed_clauses);

};


#endif
