#ifndef __STRATEGY_H_
#define __STRATEGY_H_

#include "formula.h"
#include "problem.h"
#include "sorting.h"
#include <unistd.h>
#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>

using namespace std;

template <class R>
class BasicSOS {
protected:
  OneState state_machine[MAX_STATES];
  TermList ToBeUsed;
  TermList subsumed_clauses;      
  R * TBU;
  R * CBUsubset;
  Problem::FormulaBag* CBUset;
  Unifier* unifier;
  Formula* first_chosen;
  Formula* second_chosen;

  BasicSOS() { }

  void addToActive(Formula* f);
  Formula* generate(FaustInferenceRules infrule, bool &stop_all);

  /*		STORING RESULTS			*/
  inline virtual bool putInTBU(Formula *f);
  void putInCBU(Formula *f);
  FSET addStored(FaustInferenceRules infrule);
  void storeAdded(FSET new_clauses, FaustInferenceRules infrule);

public:
  Problem* problem;
  FormulaList<bool> activePreds;

  int saved_resolve;

  BasicSOS(Problem* problem);  
  virtual ~BasicSOS();

  /*		PURE LITERAL DELETION		*/
  //true if f is pure with respect to 'list'
  bool notPureClause(Formula* f);
  //assumes that activePreds was initiated with something
  //removes from ToBeUsed all pure clauses
  void removePureClauses();
  
  /*		GETTING RESOLVANTS		*/
  bool first();
  void queryForMatch(FaustInferenceRules infrule);
  inline void setFirst(Formula* f = NULL);
  inline void setMatch(Formula* f = NULL);

  /*            MAIN LOOP                       */
  bool doResolution(FaustInferenceRules infrule);
  bool doExtraction(FaustInferenceRules infrule);
  bool satisfiable();  
};

template<class R>
class SimpleSOS : public BasicSOS<R> {
  SimpleSOS() { }
public:
  SimpleSOS(Problem* problem);
};

template <class R>
class ModelSOS : public BasicSOS<R> {
protected:	
  ModelSOS() { }
  bool interpretation(Formula* f);
public:
  //move all the formulas with at least one 'sign' literal to CBU
  ModelSOS(Problem* problem, bool my_sign);
  //
  ModelSOS(Problem* problem, Formula* seed);
};

template <class R>
class UnitResolutionSOS : public BasicSOS<R> {
protected:	
  inline bool putInTBU(Formula* f);
public:
  TermList CBUsubset2;
  ~UnitResolutionSOS() {   CBUsubset2.clear();  }
  void doResolution();
  /*
   * find first non unit clause in stored CBUsubset, make it first_chosen. if can't call first()
   * exctract (and store localy) CBUsubset match(), and get the first non-unit
   */
  bool first();
  //move all the formulas with arity 1 into TBU and the rest into 
  UnitResolutionSOS(Problem* problem);
  
  void queryForMatch(FaustInferenceRules infrule);
};

template <class R>
class SemanticResolutionSOS : public ModelSOS<R> {
protected:	
  R TBUstack;
  inline bool putInTBU(Formula *f);

public:  
  SemanticResolutionSOS(Problem* problem, Formula* seed);
  void queryForMatch(FaustInferenceRules infrule);
  void doResolution();
};

#endif
