#include "unifier.h"

//set just the base literal, initialize arrays
//there are no states initially (vgroup, vmap, pattern are invalid)
Unifier::Unifier (Formula* f, vector<HB> *symbols, OneState *sm) 
  : symbols(symbols), sm(sm) {
  ttotal = new ThinClause;
  tclause1 = new ThinClause;
  tclause2 = new ThinClause;
  com_state = state = -1;
  setStatePointers();
  subsumption_on = false;
  f->getFlatterm(pool, lit_location);
  ttotal->vnum = tclause1->vnum = maxVarName(0, 0);
  ttotal->lits = tclause1->lits = f->arity;  
  ttotal->size = tclause1->size = pool.size();
  tclause2->lits = 0;
  tclause2->size = 0;
  combination = NULL;
  eL = 0;
  T1 = 0;
  TT = 0;
  L = 0;
  eLi = 0;
  nek = 0;
  nel = 0;
  flush();
}

Unifier::~Unifier() {
  delete ttotal;
  delete tclause1;
  delete tclause2;
  restore_state(-1);
  if (combination)
    delete [] combination;
}

//replaces/adds the var at the second level
void Unifier::setPeer(Formula* f) {
  //remove previouse clause
  pool.resize(tclause1->size);
  lit_location.resize(tclause1->lits);
  //get new clause
  f->getFlatterm(pool, lit_location);
  ttotal->vnum = maxVarName(tclause1->size, tclause1->vnum);
  tclause2->size = pool.size() - tclause1->size;
  tclause2->lits = f->arity;
  ttotal->lits = tclause1->lits + tclause2->lits;
  ttotal->size = tclause1->size + tclause2->size;
  //if this is a subsumption, keep the old # + new one
  //else reset completely - cancel all droppped literals
  if (subsumption_on)
    sm[com_state].act_lits += f->arity;
  else if (state > -1)
    sm[com_state].act_lits = ttotal->lits;
}

/*
  unify literals at l1 and l2 and later call getResolvant() and then factor() 
  to obtain even more logical consequences
  l1 and l2 are indexes of literals in the original clauses (not local)
  if "opposite" then the unification will fail if literals are not opposite
  (same plurality option is for subsumption and factorization)
*/
bool Unifier::unify(int l1, int l2, bool opposite) {
  assert(l1 < tclause1->lits);
  assert(l2 < tclause2->lits);
  //get location of the literals
  int i1 = lit_location[l1];
  int i2 = lit_location[l2 + tclause1->lits];
  //check literals are opposite (if nessesary) and increment index
  //can't be var - only ~ or pred.
  assert(term((int) (pool[i1])) && term((int) (pool[i2])));
  if (opposite) {	    
    bool nt1 = (pool[i1]->ftype == _negation);
    bool nt2 = (pool[i2]->ftype == _negation);
    if (nt1 == nt2)       
      return false;
    i1 += (int) nt1;
    i2 += (int) nt2;
  }
  //store the current state of unification, or create new, discard resolving f
  push_state();
  dropLiteral(l1);
  dropLiteral(l2 + tclause1->lits);
  //map the two resolving literals to each other
  if (mapper(i1, i2, false)) 
    return true;
  pop_state();
  return false;
}

//if both are on original_i1 side, mapper(false)
//if both are on original_i2 side, same()
//else mapper(true)
bool Unifier::mapper_or_same(int i1, int original_i1, 
			     int i2, int original_i2, bool one_to_two){
  if (!one_to_two)
    return mapper(i1, i2, false);
  
  if ((i1 < tclause1->size) == (i2 < tclause1->size) && 
      (i2 < tclause1->size) == (original_i1 < tclause1->size))
    return mapper(i1, i2, false);

  if ((i1 < tclause1->size) == (i2 < tclause1->size) && 
      (i2 < tclause1->size) == (original_i2 < tclause1->size))
    return same(i1, i2);

  if ((i1 < tclause1->size) == (original_i1 < tclause1->size))
    return mapper(i1, i2, true);

  return mapper(i2, i1, true);    
}

//unify stuff directly located at i1 and i2
//will never be called for vars - those are dealt with on spot
//one_to_two means only vars from side 1 can be mapped to side 2 
//(for subsumption and tautology detection e.t.c) [i.e. can #1 mimic #2 ?]
bool Unifier::mapper(int i1, int i2, bool one_to_two) {
  assert(state > -1);
  //if one of the terms is absent, or they are virtually the same...
  if(i1 < 0 || i2 < 0 || i1 == i2)
    return true;
  //the very first element cannot be a var
  //  assert(term((int)pool[i1]) && term((int)pool[i2]));
  //loop until the end of this term
  int end = endpos(i1);
  while (i1 < end) {
    int val1 = (int) (pool[i1]);
    int val2 = (int) (pool[i2]);
    bool vterm1 = term(val1);
    bool vterm2 = term(val2);
    if (i1 == i2) {
      return true;	    
//      i1 += tweight(i1);
//      i2 += tweight(i2);
    } else if (vterm1 && vterm2) {      
      if (i1 != term_mapping[i1] || i2 != term_mapping[i2]) {
	if (!mapper(term_mapping[i1], term_mapping[i2], one_to_two))
	  return false;
	i1 += pool[i1]->weight;
	i2 += pool[i2]->weight;
      } else {
	if (pool[i1]->parent != pool[i2]->parent) 
	  return false;
	i1++; i2++;
      }
    } else if (vterm1 && !vterm2) {
      int map2 = vmap[vgroup[val2]];      
      if (map2 != i1) {
	if (map2 == -1 && one_to_two)
	  return false;
	if (occures(vgroup[val2], term_mapping[i1]))
	  return false;
	if (map2 != -1 && !mapper_or_same(term_mapping[i1], term_mapping[i1], 
					  map2, i2, one_to_two))
	  return false;	
      }
      //reset map only if no map yet (otherwise it is already mapped anyways)
      if (map2 == -1)
	setMap(val2, i1);
      i1 += pool[i1]->weight; i2++;
    } else if (!vterm1 && vterm2) {
      int map1 = vmap[vgroup[val1]];      
      //this var is already mapped to something else (side two). can't map more
      //(but what if that something _is_ side 2??? FIX
      if (one_to_two && gsize[vgroup[val1]] > 1)
	return false;
      if (map1 != i2) {
	if (occures(vgroup[val1], term_mapping[i2]))
	  return false;
	if (map1 != -1 && !mapper_or_same(map1, i1, term_mapping[i2], 
					  term_mapping[i2], one_to_two))
	  return false;
      }
      //reset map only if no map yet (otherwise it is already mapped anyways)
      if (map1 == -1)
	setMap(val1, i2);
      i2 += pool[i2]->weight; i1++;
    } else if (vgroup[val1] == vgroup[val2]) {	   //already unified vars
      i1++; i2++;      
    } else {		     //i.e. both are not yet unified vars / groups
      //do occurence check and unification check
      int map1 = vmap[vgroup[val1]];
      int map2 = vmap[vgroup[val2]];
      if (map1 != -1 && map2 != -1 && map1 != map2)
	return false;
      //make sure there are no vars from side 2 in group 1
      if (one_to_two) {
	//can't allow mapping side 2 to a term
	if (map1 != -1) 
	  return false;
	//a variable may map to another group only once
	if (gsize[vgroup[val1]] > 1)
	  return false;
      }	  
      //do other checks
      if (occures(vgroup[val1], map2) || occures(vgroup[val2], map1))
	return false;
      if (map1 != -1 && map2 != -1 && 
	  !mapper_or_same(map1, i1, map2, i2, one_to_two))
	return false;      
      //combine groups (reassign group 1 as group 2)
      if (map1 != -1 && map2 == -1)
	setMap(val2, map1);
      //	vmap[vgroup[val2]] = map1;
      else if (map2 != -1 && map1 == -1)
	setMap(val1, map2);
      //	vmap[vgroup[val1]] = map2;
      /*     
      if (gsize[vgroup[val1]] == 1) {
	vgroup[val1] = vgroup[val2];
	gsize[vgroup[val2]]++;
      } else if (gsize[vgroup[val2]] == 1) {
	vgroup[val2] = vgroup[val1];
	gsize[vgroup[val1]]++;
      } else
      */
      reassign(vgroup[val1], vgroup[val2]);
	
      i1++; i2++;
    }
  }
  return true;
}

bool Unifier::occures(int group, int term_index) {
  if (term_index < 0) return false;
  term_index = term_mapping[term_index];
  int end = term_index + pool[term_index]->weight;
  while (term_index < end) {
    int val = (int) (pool[term_index]);
    /* if member of group of the mapping variable occures inside the
       term it maps to, fail the occurence check
       OR
       if member of different group maps to somewhere were the group
       occures,  fail the occurence check
    */
    if (!term(val) &&(vgroup[val] == group || occures(group,vmap[vgroup[val]])))
      return true;
    term_index++;
  }
  return false;
}

//reassign group 1 as group 2
void Unifier::reassign(int from_group, int to_group) {
  if (sm[com_state].com_state_size > 1)
    create_state();
  for (int i = 0; i < ttotal->vnum; i++) {
    if (vgroup[i] == from_group) {
      vgroup[i] = to_group;
      gsize[to_group]++;
    }
  }
}

void Unifier::subsumptionON() {  
  assert(!subsumption_on);
  ttemp = tclause1;
  tclause1 = ttotal;
  ttotal = new ThinClause;
  ttotal->set(tclause1);
  subsumption_on = true;
}

//return true if was able to subsume f
bool Unifier::backSubsume(Formula* f) {
  assert(subsumption_on);
  //second clause is empty. no one got subsumed
  if (f->arity == 0)
    return false;
  setPeer(f);
  int old_state = state;
  push_state();
  create_state();
  for (int i = 0; i < ttotal->vnum; i++)
    gsize[i] = 1;

  bool result = false;  
  if (subsume(0, tclause1->lits, tclause1->lits, ttotal->lits,  -1))
    result = true;  

  //now restore original state of the object
  pool.resize(tclause1->size);
  lit_location.resize(tclause1->lits);
  ttotal->set(tclause1);
  restore_state(old_state);
  sm[com_state].act_lits -= tclause2->lits;
  return result;
}

//returns true if f subsumed this
bool Unifier::forwSubsume(Formula* f, TERMLIST ancestors) {
  assert(subsumption_on);
  //second clause is empty. no one got subsumed
  if (f->arity == 0)
    return false;
  setPeer(f);
  assert(state > -1);
  for (int i = 0; i < ttotal->vnum; i++)
    gsize[i] = 1;

  bool result = false;
  //forward subsumption
  push_state(); ///
  create_state(); ///
  if (subsume(tclause1->lits, ttotal->lits, 0, tclause1->lits, -1))
    result = true;  
  pop_state(); ///
  //subsumption - resolution

  if (!result) {
    
    //resetting the datastructure for the new f (without changing state)
    //in order not to ruin the data in pattern[]
    for (int i = tclause1->vnum; i < ttotal->vnum; i++) {
      sm[com_state].vgroup.set(i, i);
      sm[com_state].vmap.set(i, -1);
      sm[com_state].gsize.set(i, 1);
    }
    //=-=-=-=-=-=-=-==-=-=-=-=-=->>>
    for (int i = tclause1->lits; i < ttotal->lits; i++)
      sm[com_state].pattern.set(i, true);
    for (int i = tclause1->size; i < ttotal->size; i++)
      sm[com_state].term_mapping.set(i, i);  


    bool pushed = false; 
    for (int i = 0; i < tclause1->lits; i++) {
      //=-=-=-=-=-=-=-==-=-=-=-=-=->>>
      if (!pattern[i])
	continue;      
      //going over all literals in subsumer - looking for one to "resolve" w/
      for (int m = tclause1->lits; m < ttotal->lits; m++) {
	int i2 = lit_location[i];
	int i1 = lit_location[m];
	//it must be of opposite polarity
	bool nt1 = (pool[i1]->ftype == _negation);
	bool nt2 = (pool[i2]->ftype == _negation);
	if (nt1 == nt2)
	  continue;
	i1 += (int) nt1;
	i2 += (int) nt2;
	push_state();
	bool killed_lteral = 
	  mapper(i1, i2, true) && 
	  subsume(tclause1->lits, ttotal->lits, 0, tclause1->lits, i);
	pop_state();
	//killed literal i. drop it and keep trying to use m to drop more i's
	if (killed_lteral) {
	  dropLiteral(i);
	  if (!pushed)
	    ancestors->push(f);
	  pushed = true;
	}
      }
    }        
  }
  //now restore original state of the object
  pool.resize(tclause1->size);
  lit_location.resize(tclause1->lits);
  ttotal->set(tclause1);
  sm[com_state].act_lits -= tclause2->lits;
  return result;  
}

void Unifier::subsumptionOFF() {
  assert(subsumption_on);
  delete ttotal;
  ttotal = tclause1;
  tclause1 = ttemp;
  tclause2->size = ttotal->size - tclause1->size;
  tclause2->lits = ttotal->lits - tclause1->lits;
  subsumption_on = false;
}

//  algorithm: every literal in subsuming clause must subsume at least 1 
//  literal in second clause. no variable from the subsumed clause is mapped 
//  to anything (from same clause or another)
//  TAKES: locations of literals of both clauses
//  backward=true ==> the original clause is the subsumer
bool Unifier::subsume(int subsumer_start, int subsumer_end,
		      int subsumed_start, int subsumed_end,
		      int reversed_literal) {

  assert(tclause2->lits > 0);
  int subsumer_size = subsumer_end - subsumer_start;

  //subsumer can't be bigger then the subsumed
  int original_lits = sm[com_state].act_lits - tclause2->lits;
  if ((subsumed_start < subsumer_start && original_lits < tclause2->lits) ||
      (subsumed_start > subsumer_start && original_lits > tclause2->lits))
    return false;

  //index i in subsumed_stack holds the index of the next literal that should
  //be attempted by i'th subsuming literal
  int subsumed_stack[subsumer_size];
  for (int i = 0; i < subsumer_size; i++)
    subsumed_stack[i] = subsumed_start;

  //try next literal map[li1]. if fails, drop back to prev. literal, prev stage
  int subsumer_index = 0;
  bool success;
  int store_state = state;
  while (subsumer_index < subsumer_size && subsumer_index >-1){
    if (!pattern[subsumer_index+subsumer_start])
      success = true;
    else
      success = false;

    while (subsumed_stack[subsumer_index] < subsumed_end) {
      if (success)
	break;

      if (!pattern[subsumed_stack[subsumer_index]]) {
	subsumed_stack[subsumer_index]++;
	continue;
      }
       
      //i is the location of the subsumed literal
      int i = lit_location[subsumed_stack[subsumer_index]];
      //j is the location of the subsumer
      int j = lit_location[subsumer_index+subsumer_start];
      //if subsumed literal is "reversed" ...
      if (subsumed_stack[subsumer_index] == reversed_literal) {
	bool nt1 = (pool[i]->ftype == _negation);
	bool nt2 = (pool[j]->ftype == _negation);
	if (nt1 == nt2) {
	  subsumed_stack[subsumer_index]++;
	  continue;
	}
	i += (int) nt1;
	j += (int) nt2;
      }
      push_state();
      if (mapper(j, i, true))
	success = true;
      else
	pop_state();
      subsumed_stack[subsumer_index]++;
    }

    if (success) {
      subsumer_index++;
    } else {
      if (subsumer_index == 0) {
	restore_state(store_state);
	return false;
      } 
      //set subsumed_stack[subsumer_index] = 0 ???  cause need to start from
      //scratch next time around.
      subsumed_stack[subsumer_index] = subsumed_start;
      //keep going backward, until hit a not subsumed literal
      do { 
	subsumer_index--;  
      } while (subsumer_index > -1 && !pattern[subsumer_index+subsumer_start]);
      //there are still states that have been pushed
      if (subsumer_index > -1)
	pop_state();
    }
  }
  restore_state(store_state);
  return success;
}

/*
  generating next combination of n numbers
  1. create array a[n]
  2. populate a[n] by n; i.e.  a[0] = 0, a[1] = 1 ...
  3. to get next combination, increament the last location in a (a[n-1])
     if the last location is the max number, increment prev a loc by 1
     and set each after it to 1 + prev
*/
bool Unifier::nextCombination(){
  //scroll the clock forward
  for(int i = combination_size - 1; i > -1; i--) {
    int step = combination_size - i - 1;   //0, 1, ....
    //max_size, max_size - 1, ...
    int max_value_at_i = full_set.size() - step - 1; 
    //current combination is not the last one
    if (combination[i] < max_value_at_i) {
      combination[i]++;
      for (unsigned j = i + 1; j < combination_size; j++)
	combination[j] = combination[j-1] + 1;
      return true;
    }
  }
  //there are no more combinations with size elements, and size is the max.
  //  cout << "s f " << combination_size << " " << full_set.size() << endl;
  if (combination_size >= full_set.size())
    return false;
  //if size is 0 or there are no more combinations of size elements,
  //increase size and reset array
  combination_size++;
  for (unsigned i = 0; i < combination_size; i++)
    combination[i] = i;
  return true;
}  

void Unifier::dropLiteral(int l) { 
  if (sm[com_state].com_state_size > 1)
    create_state();
  pattern[l] = false;  
  //  if (sm[com_state].act_lits <= 0)
  //    cout << "down! " << sm[com_state].act_lits << endl;
  sm[com_state].act_lits--;
}

/* initial assumption:: 
 * in the very first call state == -1
 * state == 0: latest binary resolution result
 * state == 1: latest factorization AND/OR simplification
 */
bool Unifier::getNextResolvant() {
  /*
  cout << endl;
  cout << "lits: " << sm[com_state].act_lits << endl;
  cout << "state: " << state << endl;
  if (state == -1)
    cout << "." << endl;
  printPOOL();
  */
  assert(!subsumption_on);
  assert(tclause1->lits > 0);
  assert(tclause2->lits > 0);
  //this is the first call to this function
  if (full_set.size() == 0) {
    //this method has been visited before, but no combinations created - leave
    if (state > -1)
      return false;
    assert(!combination);
    //first produce new resolution [will also change res1 and res2 to resolved
    do {
      if (!makeBinaryStep())
	return false;    
      push_state();
      if (simplify()) {
	pop_state();
	pop_state();
      }
      else
	break;
    } while (true);
    //if there is nothing to factor, just return. next time it will fail on
    //call to getNextBinaryResolvant (<3 literals means 1 lit. per clause)
    if (ttotal->lits < 3)
      return true;
    //collect indexes of literals that are not part of the resolution
    full_set.resize(ttotal->lits);
    for (int i = 0; i < ttotal->lits; i++)
      full_set.set(i, i);
    full_set.erase(res2 + tclause1->lits);
    full_set.erase(res1);
    combination = new int[ttotal->lits - 2];
    combination_size = 0;    
    return true;
  }
  bool success = false;
  //if the prev step was factoring (state = 1), pop that state
  if (state == 1)
    pop_state();
  //push state for factorization (resolution state was pushed by "binaryRes.")
  int prev_com_state = com_state;
  push_state();
  //go over all the literals selected for factoring, and unify them
  //with the resolved literal in their respective clauses
  while (nextCombination()) {
    //    cout << "--- lits: " << sm[com_state].act_lits << endl;
    int originalLits = sm[com_state].act_lits;
    success = true;
    for (unsigned i = 0; i < combination_size; i++) {
      int chosen = full_set[combination[i]];
      if (chosen < tclause1->lits) {
	//literal 1
	if (!mapper(lit_location[chosen], lit_location[res1], false)) {
	  success = false;
	  break;
	}
      } else {
	//literal 2
	if (!mapper(lit_location[chosen], 
		    lit_location[res2 + tclause1->lits], false)) {
	  success = false;
	  break;
	}
      }
      dropLiteral(chosen);
    }
    //the state changed - something factored in
    if (success && !simplify())
      return true;
    else 
      sm[com_state].act_lits = originalLits;
    //if it was just 1 literal and it couldn't unify, don't try it again
    //(more proper way to do it is to remember any unsuccessful combination)
    if (!success && combination_size == 1) {
      //must be 0 here because combination_size=1
      full_set.erase(combination[0]);
      //decrement the cointer the make sure this spot is tried again
      combination[0]--;   
    }
    assert(state > 0);
    //refresh pattern information (instead of state poping) from prev. state
    //(but only if com_state changed)
    if (prev_com_state < com_state)
      memcpy(pattern, sm[prev_com_state].pattern.list,sizeof(bool)*ttotal->lits);
  }
  //pop the state used for factorization
  pop_state();
  //pop the state used for previouse binary resolution
  pop_state();
  do {
    if (!makeBinaryStep())
      return false;    
    push_state();
    if (simplify()) {
      pop_state();
      pop_state();
    }
    else
      break;
  } while (true);

  return true;
}

bool Unifier::makeBinaryStep() {
  assert(!subsumption_on);
  assert(tclause2->lits > 0);
  res2++;
  while(res1 < tclause1->lits) {
    while (res2 < tclause2->lits) {
      if (unify(res1, res2, true))
	return true;
      res2++;
    }
    res2 = 0;
    res1++;
  }
  return false;  
}

//assumes state starts from -1
bool Unifier::getNextBinaryResolvant() {
  if (state == 0)
    pop_state();
  push_state();
  if (!makeBinaryStep())
    return false;    
  simplify();
  return true;  
}

bool Unifier::getBinaryResolvant(Formula* f,int i1, int i2) {
  assert(!subsumption_on);
  setPeer(f);
  return (unify(i1, i2, true));
}

void Unifier::getSignature(FormulaSignature &fsig){
  TermList tempsymbols;
  createResolvant(tempsymbols, NULL);
  //copy everything into fsig but only ones
  for (int i = 0; i < tempsymbols.size(); i++)
    fsig.add(tempsymbols[i]->parent);
}

//return the remaining literals after mapping them, and reordering
//var is the new naming sequence for variables
//make sure variables in varset are given consistent names too - or
//it will not match itself when trying to insert into Store
//void Unifier::createResolvant(TermList &flatout, VarList_List &varset){
void Unifier::createResolvant(TermList &flatout, VARLIST_LIST varset){
  assert(!subsumption_on);
  push_state();
  //prepare stack full of stuff
  int LOCAL_STACK_SIZE = 60;
  int comeback[LOCAL_STACK_SIZE];	
  int stopat[LOCAL_STACK_SIZE];	        
  int vname[ttotal->vnum];
  if (varset != NULL) {
    //renaming variables (prevent name inflation)
    for (int i = 0; i < ttotal->vnum; i++)
      vname[i] = -1;
    int j = 0;
    //make new names only for vars that don't map to anything
    for (int i = 0; i < ttotal->vnum; i++)
      if (vname[vgroup[i]] == -1 && vmap[vgroup[i]] == -1)
	vname[vgroup[i]] = j++;
  }
  // current location is comback[i]  
  //  cout << "*** creating resolvant ***" << endl;
  //  printPOOL();
  for (int k = 0, l = 0; k < ttotal->size; k += pool[k]->weight, l++) {
    if (!pattern[l]) continue;
    //    cout << "literal: " << l << " location: " << k 
    //	 << " size: " << pool[k]->weight << endl;
    comeback[0] = k;
    stopat[0] = k + pool[k]->weight;
    VARLIST varlit;
    if (varset != NULL) {
      varlit = new VarList;
      varset->push(varlit);		//same order as the literals   
      //refresh mappings of vars to new names for each new literal
      for (int m = 0; m < ttotal->vnum; m++)
	if (vmap[m] < 0) vmap[m] = -1;
    }
    //until the literal is over...
    for (int i = 0; i > -1; ) {     
      //      cout << "Stack depth: " << i <<" bounds: " << comeback[i] 
      //	   << " ... " << stopat[i] <<  endl;
      //stack overflow check
      assert(i < LOCAL_STACK_SIZE);
      if (comeback[i] == stopat[i]) { 
	i--;
	continue;
      }      
      int val = (int) (pool[comeback[i]]);
      //      cout << "value: " << val << endl;
      if (term(val)) {
	//a term	
	if (term_mapping[comeback[i]] == comeback[i]) {
	  flatout.push(pool[comeback[i]]);
	  comeback[i]++;
	}
	//a term with a mapping (paramodulated)
	else {
	  int iloc = comeback[i];
	  comeback[i] += pool[iloc]->weight;	  
	  comeback[++i] = term_mapping[iloc];
	  stopat[i] = comeback[i] + pool[comeback[i]]->weight;	  
	}
      } else if (vmap[vgroup[val]] >= 0) {
	//a mapping to a term
	comeback[i]++;
        comeback[++i] = vmap[vgroup[val]];
        stopat[i] = comeback[i] + pool[comeback[i]]->weight;
      } else if (varset != NULL) {
	//a newly encountered variable
        if (vmap[vgroup[val]] == -1) {
	  //	  vmap[vgroup[val]] = - varlit->size() - 2;
	  setMap(val, - varlit->size() - 2);
	  varlit->push(vname[vgroup[val]]);
        }
	//previously encountered variable
        flatout.push((Formula*) (-vmap[vgroup[val]] - 2));
	comeback[i]++;
      } else {
	comeback[i]++;	
      }
    }
  }
  pop_state();
}

bool Unifier::same(int i1, int i2) {
  assert(state > -1);
  //they are virtually the same...
  if(i1 == i2)
    return true;
  //if we are comparing two equalities, compare their terms pairwise
  if (term((int)pool[i1]) && term((int)pool[i2]) && pool[i1]->symmetric && pool[i2]->symmetric) {
    //1+1 && 2+2 OR 1+2 && 2+1    
    bool res = ((same(i1+1, i2+1) && same(endpos(i1+1), endpos(i2+1))) ||
		(same(i1+1, endpos(i2+1)) && same(endpos(i1+1), i2+1)));
    //    cout << res << endl;
    return res;
  }
  //loop until the end of this term
  int end1 = endpos(i1);
  int end2 = endpos(i2);
  while (i1 < end1) {
    int val1 = (int) pool[i1];
    int val2 = (int) pool[i2];
    int map1 = (!term(val1) && vmap[vgroup[val1]] != -1) ? vmap[vgroup[val1]] : i1;
    int map2 = (!term(val2) && vmap[vgroup[val2]] != -1) ? vmap[vgroup[val2]] : i2;
    //virtually the same (at least map to the same location) -- jump over terms
    if (map1 == map2) {
      i1 += tweight(i1);
      i2 += tweight(i2);
      continue;
    }
    //one or both map to remote location
    if ((map1 != i1 || map2 != i2)) {
      if (!same(map1, map2))
	return false;      
      i1 += tweight(i1);
      i2 += tweight(i2);
      continue;
    }
    //have to be either terms with same parent or vars from same group
    if (term(val1) && term(val2)) {
      if (i1 != term_mapping[i1] || i2 != term_mapping[i2]) {
		if (!same(term_mapping[i1], term_mapping[i2]))
	  		return false;
		i1 += pool[i1]->weight;
		i2 += pool[i2]->weight;
		continue;
      } else if (pool[i1]->parent != pool[i2]->parent) 
	  	return false;      
    } else if (term(val1) || term(val2) || vgroup[val1] != vgroup[val2]) {
      return false;
    }
    i1++; i2++;
  }
  if (i2 != end2)
    return false;
  return true;
}

int Unifier::active_literals() {
  if (state == -1)
    return ttotal->lits;
  return sm[com_state].act_lits;
}

//returns true in case of tautology
//take 1st literal. if it's s!=s discard it. if it's s==s, return true.
//go over all literals after it. discard all literals identical to it.
//if there is opposite literal, return true.
bool Unifier::simplify() {
  assert(state > -1);
  assert(!subsumption_on);
  //contradiction, not tautology
  if (sm[com_state].act_lits == 0)
    return false;
  for (int k = 0, l = 0; k < ttotal->size; k += pool[k]->weight, l++) {
    if (!pattern[l])
      continue;
    // 	- trivial equality resolution
    if (isNegEq(k) && same(k + 2, endpos(k+2))) {
      dropLiteral(l);
      continue;
    }
    // 	- equality tautology deletion
    if (isEq(k) && same(k+1, endpos(k+1)))
      return true;   
    //checking literals deeper
    bool kneg = (pool[k]->ftype == _negation);
    int l2 = l+1;
    for (int s = endpos(k); s < ttotal->size; s = endpos(s), l2++){
      if (!pattern[l2])
	continue;
      bool sneg = (pool[s]->ftype == _negation);
      int kl2 = endpos(k + kneg + 1);
      if (!same(k + kneg, s + sneg))
	continue;	
      //  - factoring identical literals
      if (kneg == sneg)
	dropLiteral(l2);
      //  - propositional tautology deletion
      else
	return true;
    }     
  }
  return (sm[com_state].act_lits == 0);
}

bool Unifier::getNextEqualityResolvant() {
  //if this is not the first call, go back to original state
  if (nek > 0) 
    pop_state();
  //no point if there are no more literals left
  if (sm[com_state].act_lits == 0)
    return false;
  //look for the next ~equals literal
  if (state == -1)
    push_state();
  for (int k = nek, l = nel; k < ttotal->size; k += pool[k]->weight, l++) {
    if (!pattern[l])
      continue;
    //if the literal is negative equality and the two terms map to each
    //other, the can do the resolution
    push_state();
    if (isNegEq(k) && mapper(k + 2, endpos(k+2), false)) {      
      //here need to check that (T1 != T2)θ is not smaller than any literal in (T1 != T2 | C| )θ
      dropLiteral(l);
      if (!simplify())
	return true;      
    }
    pop_state();
  }  
  return false;
}

/*
bool Unifier::demodulate() {
  for (int k = 0, l = 0; k < ttotal->size; k += pool[k]->weight, l++) {
    if (!pattern[l])
      continue;
    // 	- trivial equality resolution
    if (isEq(k)) {
    }    
  }
}
*/

/* 0. if state is not 0, get it to zero (remove prev. paramodulation).
   1. choose equality literal eL in C1
   2. choose one of the sides of eL s.t. it is not a variable
   3. from C2 choose any term that has same principal symbol as in #2
   4. push state
   5. map. if can't, pop and go to #2 (or #1)
   6. if imposeKBOorder() then paramodulate
   7. call simplify
   8. return
 */
bool Unifier::getNextSuperposition() {
  assert(!subsumption_on);
  assert(tclause1->lits > 0);
  assert(tclause2->lits > 0);
  //if the prev step was factoring (state = 1), pop that state
  if (state > 0)
    restore_state(0);
  do {
    if (state == 0) {
      //      cout << "TT: " << TT << " T1: " << T1 << " L " << L << endl;
      int TTend;
      if (eL < tclause1->size)
	TTend = ttotal->size;
      else 
	TTend = tclause1->size;
      for ( ; TT < TTend; TT++) {
	//skip variables - do not paramodulate into them
	if (!term((int)pool[TT]))
	  continue;
	switch (pool[TT]->ftype) {
	case _negation: L = TT; TT++; continue;
	case _predicate: L = TT; continue;
	default:	 
	  if (!imposeKBOorder())          //make sure order is fine
	    continue;
	  paramodulate(TT, T2);           //do the paramodulation
	  dropLiteral(eLi);               //drop the equality literal
	  if (simplify())                 //because there was mapping
	    pop_state();
	  else {
	    TT++;                         //incriment for next visit
	    return true;
	  }
	}
      }
    } else {
      L = TT = tclause1->size;
      eL = 0;
      T1 = T2 = 0;
      eLi = 0;
      push_state();
    }
    if (!nextEqualityTerm())
      return false;
    if (eL < tclause1->size)
      L = TT = tclause1->size;      
    else 
      L = TT = 0;
  } while (true);
}

bool Unifier::nextEqualityTerm() {
  //this is the second call to the same equality, swap terms
  if (T2 > T1) {
    T1 = T2;
    T2 = eL + 1;
    return true;
  //this is the third call to the same equality make it 1st
  } else if (T2 < T1) {
    eL += pool[eL]->weight;
    eLi++;
  }
  //this is the first call, find me (another) equality
  while (eL < ttotal->size && (!isEq(eL) || !pattern[eLi])) {
    eL += pool[eL]->weight;
    eLi++;
  }    
  //no more equality terms
  if (eL == ttotal->size)
    return false;
  //return the left as T1 and right as T2
  T1 = eL + 1;
  T2 = endpos(T1);
  return true;    
}

/*given equality literal eL, grab one of the terms, and unify it with 
  some term in the second clause. now, check that:
  1. the after-unification equality literal is !< then any literal in 
     the 1st clause
  2. the unified term T1 is !< then it's peer in literal, T2
  3. the literal in second clause from which the sec. term came, !< any
     literal in the second clause
  4. if the second literal is also equality or in-equality, check that
     the chosen (second) term is !< then it's peer in the literal
*/
//fix --- skip litterals with pattern[i]=false   
/*
  e1 = e2 | C   	 +  	 L[t1]   | D
                   OR
  e1 = e2 | C   	 +  	 t1 = t2 | D
                   =>
  L[e2] | C | D
*/
bool Unifier::imposeKBOorder() {
  assert(!subsumption_on);

  //should not be a predicate or neg, or a variable.
  if (!term((int)pool[TT]) || !term((int)pool[T1]) || !term((int)pool[T2]))
    return false;
  
  if (pool[TT]->ftype != _functor)
    return false;

  //T1' = TT'
  push_state();
  if (!mapper(T1, TT, false) || !same(T1, TT)) {
    pop_state();
    return false;
  }

  //(T1 = T2)'  is not smaller than any literal in (T1 = T2 | C|)'
  int clEnd = clauseEnd(eL);
  for (int i = clauseStart(eL); i < clEnd; i += pool[i]->weight)
    if (kboGreater(i, eL)) {
      pop_state();
      return false;
    }

  //T1'  is not smaller than T2' in the term ordering
  if (kboGreater(T2, T1)) {
    pop_state();
    return false;
  }
  
  //L'  is not smaller than any literal in (L | D|)'
  int l = (L < tclause1->size ? 0 : tclause1->lits);
  clEnd = clauseEnd(L);
  for (int i = clauseStart(L); i < clEnd; i += pool[i]->weight, l++) {
    if (!pattern[l] || i == L)
      continue;   
    if (kboGreater(i, L)) {
      pop_state();
      return false;
    }
  }

  //additional check -- in case L is also an equality
  int t2 = -1; 
  if (isEq(L))
    TT == L+1 ? t2 = pool[TT]->weight + TT : t2 = L+1;
  if (isNegEq(L))
    TT == L+2 ? t2 = pool[TT]->weight + TT : t2 = L+2;
  if (t2 > -1 && kboGreater(t2, TT)) {
    pop_state();
    return false;
  }  
  return true;
}

void Unifier::countVars(int* var_occurence, int from, int to, int diff) {
  for (int i = from; i < to; i++) {
    int val = (int) pool[i];
    if (!term(val)) {
      int map = vmap[vgroup[val]];
      if (map > -1) 
	countVars(var_occurence, map, pool[map]->weight, diff);
      else
	var_occurence[val] += diff;
    }
  }
}

int Unifier::maxVarName(int from, int shift){
  int max_val = -1;
  int spool = pool.size();
  for (int i = from; i < spool; i++) {
    int val = (int) pool[i];
    if (!term(val)) {
      if (val > max_val)
	max_val = val;
      pool.set(i, (Formula *) (val + shift));
    }
  }
  return max_val + 1 + shift;
}

int Unifier::getKBO(int loc) {
  int kbo = 0;
  int end = endpos(loc);
  for (int i = loc; i < end; i++) {
    int val = (int) (pool[i]);
    if (!term(val)) {
      //a variable
      int map = vmap[vgroup[val]];
      if (map == -1)
	//with no mapping
	kbo++;
      else
	//mapped to something
	kbo += getKBO(map);
    } else {
      //a term
      if (term_mapping[i] == i)
	//not mapped to anything
	kbo += pool[i]->pKBO;
      else {
	//mapped to something
	kbo += getKBO(term_mapping[i]);
	//sump to next term - this one was replaced
	i = endpos(i);
      }
    }
    //    cout << "kbo " << kbo << endl;
  }
  return kbo;
}

//U. Martin. How to choose weights in the Knuth-Bendix ordering. In Proc. 2nd RTA, LNCS 256, Bordeaux, France, 1987.
bool Unifier::kboGreater(int i1, int i2) {
  //w(T1) > w(T2)  
  int kbo1 = getKBO(i1);
  int kbo2 = getKBO(i2);
  if (kbo1 > kbo2) {
    int var_occurence[ttotal->vnum];
    for (int i = 0; i < ttotal->vnum; i++)
      var_occurence[i] = 0;
    //every variable in T2 occurs at least the same number of times in T1
    countVars(var_occurence, i1, endpos(i1),  1);
    countVars(var_occurence, i2, endpos(i1), -1);
    for (int i = 0; i < ttotal->vnum; i++)
      if (var_occurence[i] < 0)
	return false;
  } else if (kbo1 == kbo2) {
    //w(T1) = w(T2) and p(symbol_of(T1)) > p(symbol_of(T2)) 
    //(the sign is inverted on purpose. I want predicates have bigger p)
    int val1 = (int) pool[i1];
    int val2 = (int) pool[i2];
    if (term(val1) && term(val2)) {
      if (pool[i1]->parent < pool[i2]->parent)
	return true;
      if (pool[i1]->parent > pool[i2]->parent)
	return false;
      if (pool[i1]->weight == 0)
	return false;
    }
    //for each pair of matching terms - if they are not the same, return
    for (int i = i1 + 1, j = i2 + 1; i < endpos(i1) && j < endpos(i2); 
	 i = endpos(i), j = endpos(j)) {
      if (!same(i, j)) {
	val1 = (int) pool[i];
	val2 = (int) pool[j];
	int map1 = (!term(val1) && vmap[vgroup[val1]] != -1) ? vmap[vgroup[val1]] : i;
	int map2 = (!term(val2) && vmap[vgroup[val2]] != -1) ? vmap[vgroup[val2]] : j;
	return kboGreater(i, j);
      }
    }
  }
  return false;
}

//----------------------------------------------------------------------
void Unifier::flush(){
  res1 = 0;
  res2 = -1;
  full_set.clear();
  if (combination) 
    delete [] combination;
  combination = NULL;
}

void Unifier::printPOOL() {
  for (unsigned i = 0 ; i < pool.size(); i++) {
    if (term((int) (pool[i])))
      cout << "(" << i <<") "<< symbols->at(pool[i]->parent).NameSymbol << " ";
    else {
      if (vgroup && vmap[vgroup[(int) (pool[i])]] != -1)
	cout << "[" << vmap[vgroup[(int) (pool[i])]] << "] ";
      else if (vgroup)
	cout << vgroup[(int) (pool[i])] << " ";
      else 
	cout << (int) pool[i] << " ";
    }
      //      cout << (int) (pool[i]) << " ";
  }
  cout << endl;
  if (vgroup && state > -1) {
    cout << "group: ";
    for (int i = 0; i < ttotal->vnum; i++)
      cout << vgroup[i] << " ";
    cout << endl;
  }
  if (gsize && state > -1) {
    cout << " size: ";
    for (int i = 0; i < ttotal->vnum; i++)
      cout << gsize[i] << " ";
    cout << endl;
  }
  if (vmap && state > -1) {
    cout << "  map: ";
    for (int i = 0; i < ttotal->vnum; i++)
    cout << vmap[i] << " ";
    cout << endl;
  }
  if (pattern && state > -1) {
    cout << "pattern: ";
    for (int i = 0; i < ttotal->lits; i++)
      cout << pattern[i] << " ";
    cout << endl;
  }
  if (term_mapping && state > -1) {      
    cout << "term map: ";
    for (int i = 0; i < ttotal->size; i++)
      cout << term_mapping[i] << " ";
    cout << endl;
  }
}

void Unifier::pop_state() {
  
  assert(state > -1);	  //can't have more pops then pushe

  state--;
  sm[com_state].com_state_size--;
  
  if (state == -1) {
    com_state = -1;
  } else if (sm[com_state].com_state_size == 0) {
    com_state--;
    setStatePointers();
  }
}

void Unifier::push_state() {

  state++;
  assert(state < MAX_STATES);

  if (state == 0)
    create_state();
  else
    sm[com_state].com_state_size++;
}

void Unifier::create_state() {

  if (com_state > -1)
    sm[com_state].com_state_size--;

  com_state++;
  sm[com_state].com_state_size = 1;

  OneState &c_state = sm[com_state];       //new state
  OneState &p_state = sm[com_state - 1];   //old state

  //making them bigger, in case allocated space is not enough
  if (ttotal->vnum > c_state.vgroup.size()) {
    c_state.vgroup.resize(ttotal->vnum);
    c_state.vmap.resize(ttotal->vnum);
    c_state.gsize.resize(ttotal->vnum);
  }
  if (ttotal->lits > c_state.pattern.size())
    c_state.pattern.resize(ttotal->lits);
  if (ttotal->size > c_state.term_mapping.size())
    c_state.term_mapping.resize(ttotal->size);    
  
  c_state.vnum = ttotal->vnum;	
  c_state.total_size = ttotal->size;
  c_state.total_lits = ttotal->lits;

  if (state > 0)
    c_state.act_lits = p_state.act_lits;
  else 
    c_state.act_lits = ttotal->lits;
  
  int svnum = 0;
  int stotal_lits = 0;
  int stotal_size = 0;
  if (state > 0) {
    svnum = c_state.vnum;	
    stotal_lits = c_state.total_lits;
    stotal_size = c_state.total_size;
    //take the smallest of the this and prev states sizes
    if (p_state.vnum < ttotal->vnum)
      svnum = p_state.vnum;
    if (p_state.total_lits < ttotal->lits)
      stotal_lits = p_state.total_lits;
    if (p_state.total_size < ttotal->size)
      stotal_size = p_state.total_size;
    //copy from prev state	
    memcpy(c_state.vgroup.list, p_state.vgroup.list, sizeof(int) *svnum);
    memcpy(c_state.vmap.list, p_state.vmap.list, sizeof(int) * svnum);
    memcpy(c_state.gsize.list, p_state.gsize.list, sizeof(int) * svnum);
    memcpy(c_state.pattern.list,p_state.pattern.list,sizeof(bool)*stotal_lits);
    memcpy(c_state.term_mapping.list, 
	   p_state.term_mapping.list, sizeof(int) * stotal_size);
  }

  //create new data in the region not covered by previouse state 
  //(or if prev. state is -1)
  for (int i = svnum; i < c_state.vnum; i++) {
    c_state.vgroup.set(i, i);
    c_state.vmap.set(i, -1);
    c_state.gsize.set(i, 1);
  }
  for (int i = stotal_lits; i < c_state.total_lits; i++)
    c_state.pattern.set(i, true);
  for (int i = stotal_size; i < c_state.total_size; i++)
    c_state.term_mapping.set(i, i);  
  setStatePointers();    
}

void Unifier::setStatePointers() {
  vgroup = NULL;
  vmap = NULL;
  gsize = NULL;
  pattern = NULL;
  term_mapping = NULL;
  if (state > -1) {
    //    if (ttotal->vnum > 0) {
      vgroup = sm[com_state].vgroup.list;
      vmap =  sm[com_state].vmap.list;
      gsize =  sm[com_state].gsize.list;
      //    }
      //    if (ttotal->lits > 0)
      pattern =  sm[com_state].pattern.list;
      //    if (ttotal->size > 0)
      term_mapping =  sm[com_state].term_mapping.list;
  }
}

void Unifier::restore_state(int old_state) {
  if (old_state == state)
    return;
  if (old_state == -1) {
    state = -1;
    com_state = -1;
    return;
  }
  assert (old_state < state && old_state > -1);
  while (state > old_state) {    
    state--;
    if (--sm[com_state].com_state_size == 0)
      com_state--;
  }  
  setStatePointers();
}

//fix varmap comparison
