/*
 * possible uses:
 * 	1. quicky binary resolution
 * 		- constractor
 * 		- getBinaryResolvant(Formula, 1, 2)
 * 	2. lazy binary resolution
 * 		- constructor
 * 		- setPeer 		
 * 		- repeatedly
 * 			- getNextBinaryResolvant
 *                      - createResolvant
 *                      - pop_state
 * 	3. lazy full resolution
 * 		- constructor
 * 		- setPeer 		
 * 		- repeatedly
 * 			- getNextResolvant
 *                      - createResolvant
 * 	4. subsumption
 *              - subsumptionON
 *              - repeat
 * 		        - subsume(Formula)
 *              - subsumptionOFF
 *      5. one-directional resolution (for compulsory resolution)
 *              
 *         call unify(opposite = true, one dir = true)
 *
 * simplification rules stages
 * 	look for ~eq(x,s), and remember to make substitution and skip it
 * 	(important to track down sim. lits that will appear as a consequence)
 * 	- destractive equality resolution
 * 		- if (x!=s) - replace all x by s, remove the (x!=s)
 * 			+ replace x by var, in it's group set it to s
 * 			+ keep replacing x by a var
 * 		
 * 	now start generating the clause
 * 	- propositional tautology deletion
 * 		- if one literal is negated - discard this clause
 * 	- factoring identical literals
 * 		- if not, discard the later literal [factoring]
 * 	- equality tautology deletion
 * 		- do same on content of $equal - if (s==s) - tautology
 * 	- trivial equality resolution
 * 		- if (s!=s) - discard this clause
 *
 * - subsumption resolution
 * 	vampire - for every literal, flip it's polarity and and try to subsume
 *      the modified clause. if it is subsumed, discard that literal
 * - demodulation - apply while looking for subsumption (?)
 * 	maintain equivalence sets for things that are or($equal(a,b))
 * 	choose one [first] symbol to represent all
 * 	where to apply it?
 *  		
 */

#ifndef UNIFIER__H_
#define UNIFIER__H_
 
#include "formula.h"

#define INITIAL_BUFFER_SIZE 200
#define MAX_STATES 60
#define term(i)  (i >= MAX_VARS_IN_FORMULA || i <= -MAX_VARS_IN_FORMULA)
//#define isLiteral(i) (term(i) && (pool[i]->ftype == _predicate || pool[i]->ftype == _negation))
#define tweight(i) ((term((int)pool[i])) ? pool[i]->weight : 1)
#define endpos(i) (i + tweight(i))
#define isEq(i) (symbols->at(pool[i]->parent)._equality)
#define isNegEq(i) (pool[i]->ftype == _negation && isEq(i+1))
#define clauseStart(i) ( i < tclause1->size ? 0 : tclause1->size)
#define clauseEnd(i) ( i < tclause1->size ? tclause1->size : ttotal->size)


struct OneState {  
  //                            Normal size       Size here: 200
  VarList  vgroup;         //   vnum
  VarList  gsize;          //   vnum
  VarList  vmap;           //   vnum
  FlagList pattern;        //   total_lits
  VarList  term_mapping;   //   total_size  
  int vnum, total_size, total_lits;
  int com_state_size;
  int act_lits;             //number of active literals (according to pattern)


  OneState() {
    vnum = total_lits = total_size = 0;
    com_state_size = 0;
    vgroup.resize(INITIAL_BUFFER_SIZE);   
    gsize.resize(INITIAL_BUFFER_SIZE);   
    vmap.resize(INITIAL_BUFFER_SIZE);        
    pattern.resize(INITIAL_BUFFER_SIZE);    
    term_mapping.resize(INITIAL_BUFFER_SIZE);    
  }

};

class ThinClause {
 public:
  int size;		//number of symbols
  int lits;		//number of literals
  int ires;		//index of latest resolved clause
  int vnum;		//size of both group and mapping and var num

  void set(ThinClause *c) {
    size = c->size;		//number of symbols
    lits = c->lits;		//number of literals
    vnum = c->vnum;		//number of variables
  }
};

class Unifier {
public:
  Unifier (Formula* f, vector<HB> *symbols, OneState* sm);
  ~Unifier();

  // -------------------------- RESOLUTION --------------------------
  // unless they return false, the state is pushed, waiting for createResolvant
  // to be called, and then state has to be poped

  //binary resolution on selected literals (i1 in clause 1, i2 in clause 2)
  bool getBinaryResolvant(Formula* f,int i1, int i2);

  //sets new second resolvant (for lazy resolution)
  void setPeer(Formula* f);   

  //lazy binary resolution (keep choosing different literals)
  bool getNextBinaryResolvant();

  //lazy full resolution (do binary, then do factorization)
  bool getNextResolvant();

  //lazy paramodulation
  bool getNextParamodulation();

  //sets resolution back to initial stage
  void flush();

  // ------------------------- SIMPLIFICATION -------------------------

  //returns true if this clause is a tautology / has to be discarded
  bool simplify();

  //returns number of not factored/resolved/simplified literals remaining
  int active_literals();

  // -------------------------- SUBSUMPTION ---------------------------

  //call before subsuming anything
  void subsumptionON();               

  //true if f got subsumed
  bool backSubsume(Formula* f);

  //true if f subsumed original clause
  bool forwSubsume(Formula* f, TERMLIST ancestors);

  //call after all subsumptions done
  void subsumptionOFF();               

  // ----------------------- PARAMODULATION -------------------------

  //returns true if just produced new paramodulation, false if unable
  bool getNextSuperposition();

  //equality resolution
  bool getNextEqualityResolvant();

  // ---------------------------- FINAL -----------------------------

  void createResolvant(TermList &flatout, VARLIST_LIST varset);

  //returns the signature of the current formula and the arity
  void getSignature(FormulaSignature &fsig);

  // ---------------------- state manipulation -----------------------
  void push_state();
  void pop_state();
  void restore_state(int old_state);   //restore state to a smaller value
  void create_state();

  // ----------------------------------------------------------------

  void printPOOL();  

  int com_state;
  int state;

private:
  //both clauses's literals crammed together into continuous vector
  TermList pool;
  //array of int[# unique vars in both clauses (not literals!)]
  int* vgroup;
  //size of each group
  int* gsize;
  //array of indexes to term locations [never to variables]
  int* vmap;
  //indicates which literals were dropped
  bool* pattern;
  //paremodulation information for each location (size = pool.size())
  int* term_mapping;
  VarList lit_location;
  //indexes of the latest resolved literals
  int res1, res2;
  //used for factoring
  VarList full_set;
  int* combination;
  unsigned combination_size;

  bool subsumption_on;

  vector<HB> *symbols;
  //the STATE MACHINE
  OneState *sm;
  //used in superposition
  int T1, T2, eL, eLi, TT, L;
  //used in equality resolution
  int nek, nel;

  ThinClause *tclause1, *tclause2, *ttotal, *ttemp;
  int act_lits, act_lits_back;

  /*
     RULE: the only f' that will not automatically return to original state is
     unify (on success)  --- because there is a need to createResolvant
     will extend by exactly one state.
  */

  void dropLiteral(int l);

  void setMap(int var, int value) {     
    if (sm[com_state].com_state_size > 1)
      create_state();
    vmap[vgroup[var]] = value; 
  }

  void paramodulate(int from, int to) {
    if (sm[com_state].com_state_size > 1)
      create_state();
    term_mapping[from] = to;
  }

  bool unify(int l1, int l2, bool opposite);
  bool mapper(int i1, int i2, bool one_to_two);
  bool same(int l1, int l2);
  bool mapper_or_same(int i1, int original_i1, 
		      int i2, int original_i2, bool one_to_two);

  bool occures(int group, int term_index);
  void reassign(int from_group, int to_group);
  bool nextCombination();
  void setStatePointers();
  void countVars(int* var_occurence, int from, int to, int diff);
  int maxVarName(int from, int shift);

  //the last parameter is location "reversed" literal. whenever something tries
  //to subsume it 1) if "reversed" is negative, ignore the negation
  //              2) else, ignore the negation of the subsumer
  bool subsume(int subsumer_start, int subsumer_end,
	       int subsumed_start, int subsumed_end,
	       int reversed_literal);

  bool makeBinaryStep();

  //try superpositioning term e2 from equality eL into term t1 of literal L
  //(all values correspond to locations in pool[])
  bool imposeKBOorder();

  //true if term/literal i1 > i2 in KBO ordering
  bool kboGreater(int i1, int i2);

  //return the KB order of the term
  int getKBO(int loc);

  //find next != or == term in any of the clauses
  bool nextEqualityTerm();

};

#endif
