#include <stdio.h>
#include <assert.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <unistd.h>
#include <sys/param.h>
#include <sys/types.h>
#include <pwd.h>
#include "DataTypes.h"
#include "Utilities.h"
#include "Tokenizer.h"
#include "FileUtilities.h"
//------------------------------------------------------------------------------
READFILE NewReadFile(void) {

    READFILE Stream;

    Stream = (READFILE)(Malloc(sizeof(ReadFile)));

    Stream->FileName = NULL;
    Stream->FileHandle = NULL;
    Stream->Line = 1;
    Stream->Character = 0;
    Stream->StringFileContent = NULL;
    Stream->Overshot = 0;
    Stream->StringOffset = 0;
    Stream->CurrentCharacter = 0;
    Stream->CurrentToken = NULL;
    Stream->Auxilliary = NULL;
    Stream->NeedNonLogicTokens = GetNeedForNonLogicTokens();

    return(Stream);
}
//------------------------------------------------------------------------------
void PathBasename(char * Path,String Basename,char * Suffix) {

    char * Slash;

//----Take from after the last slash
    if ((Slash = strrchr(Path,'/')) == NULL) {
        strcpy(Basename,Path);
    } else {
        strcpy(Basename,Slash+1);
    }
//----Chop off the suffix
    if (Suffix != NULL && (Slash = strstr(Basename,Suffix)) != NULL &&
!strcmp(Slash,Suffix)) {
        *Slash = '\0';
    }
}
//-----------------------------------------------------------------------------
char * ExpandFileName(char * FileName,char * CurrentFileName,
String ExpandedFileName) {

    char * Paths[] = {
        "",
        "Problems",
        "Axioms"
    };
    int NumberOfPaths = sizeof(Paths)/sizeof(char *);
    char * Extensions[] = {
        "",
        ".p",
        ".tptp",
        ".ax",
        ".eq" };
    int NumberOfExtensions = sizeof(Extensions)/sizeof(char *);

    int PathIndex;
    int ExtensionIndex;
    String TPTPDirectory;
    String Domain;
    int DomainOnOff;
    char * BaseName;
    char * Home;
    struct passwd *PasswdEntry;
    char * UserName;
    char * Slash;

//----If first character of filename is ~, expand to $HOME
    if (FileName[0] == '~' && FileName[1] == '/') {
        if ((Home = getenv("HOME")) != NULL) {
            strcpy(ExpandedFileName,Home);
            strcat(ExpandedFileName,&(FileName[1]));
            if (access(ExpandedFileName,R_OK) == 0) {
                return(ExpandedFileName);
            }
        }
        return(NULL);
    } 

//----Another user's home
    if (FileName[0] == '~' && isalpha(FileName[1])) {
        if ((Slash = strchr(FileName,'/')) != NULL && strlen(Slash+1) > 0) {
            UserName = strtok(FileName+1,"/");
            if ((PasswdEntry = getpwnam(UserName)) != NULL) {
                strcpy(ExpandedFileName,PasswdEntry->pw_dir);
                strcat(ExpandedFileName,Slash);
                if (access(ExpandedFileName,R_OK) == 0) {
                    return(ExpandedFileName);
                } 
            }
        }
        return(NULL);
    } 

//----Absolute
    if ((FileName[0] == '.' || FileName[0] =='/')) {
        strcpy(ExpandedFileName,FileName);
        if (access(ExpandedFileName,R_OK) == 0) {
            return(ExpandedFileName);
        } 
        return(NULL);
    }

//----Otherwise we have to search
//----Try the base directory of the including file 
    if (CurrentFileName != NULL) {
        strcpy(ExpandedFileName,CurrentFileName);
        if ((Slash = strrchr(ExpandedFileName,'/')) != NULL) {
            *(++Slash) = '\0';
            strcat(ExpandedFileName,FileName);
            if (access(ExpandedFileName,R_OK) == 0) {
                return(ExpandedFileName);
            }
        }
    }

//----If the including file has no name, use CWD
    if (CurrentFileName == NULL) {
        getcwd(ExpandedFileName,MAXPATHLEN);
        strcat(ExpandedFileName,"/");
        strcat(ExpandedFileName,FileName);
//----This short cut does not provide a full file name
// strcpy(ExpandedFileName,FileName);
        if (access(ExpandedFileName,R_OK) == 0) {
            return(ExpandedFileName);
        }
    }


//----Try the TPTP directory, and TPTP variations
    Home = getenv("TPTP");
    if (Home == NULL) {
        Home = getenv("TPTP_HOME");
//----If no TPTP_HOME, try the tptp user (aaargh)
        if (Home == NULL && (PasswdEntry = getpwnam("tptp")) != NULL) {
            Home = PasswdEntry->pw_dir;
        }
//----Now look in the TPTP directory from there
        if (Home != NULL) {
            strcpy(TPTPDirectory,Home);
            strcat(TPTPDirectory,"/TPTP");
        }
    } else {
        strcpy(TPTPDirectory,Home);
    }
        
    if (Home != NULL) {
//----Extract domain and make the path
        if ((Slash = strrchr(FileName,'/')) == NULL) {
            BaseName = FileName;
        } else {
            BaseName = Slash+1;
        }
        strncpy(Domain,BaseName,3);
        Domain[3] = '\0';
//----Check various extensions and paths
        for (PathIndex=0;PathIndex < NumberOfPaths;PathIndex++) {
            for (DomainOnOff=0; DomainOnOff <= 1; DomainOnOff++) {
                for (ExtensionIndex=0;ExtensionIndex < NumberOfExtensions;
ExtensionIndex++) {
                    strcpy(ExpandedFileName,TPTPDirectory);
                    if (strcmp(Paths[PathIndex],"")) {
                        strcat(ExpandedFileName,"/");
                        strcat(ExpandedFileName,Paths[PathIndex]);
                    }
                    if (DomainOnOff) {
                        strcat(ExpandedFileName,"/");
                        strcat(ExpandedFileName,Domain);
                    }
                    strcat(ExpandedFileName,"/");
                    strcat(ExpandedFileName,FileName);
                    strcat(ExpandedFileName,Extensions[ExtensionIndex]);
//DEBUG printf("look for ==%s===\n",ExpandedFileName);
                    if (access(ExpandedFileName,R_OK) == 0) {
                        return(ExpandedFileName);
                    }
                }
            }
        }
    }


    return(NULL);
}
//-----------------------------------------------------------------------------
FILE * OpenFileInMode(String FileName,char * Mode) {
    
    FILE * Stream;
    
    if ((Stream = fopen(FileName,Mode)) == NULL) {
        perror("Opening file");
        printf("ERROR: Could not open file %s in %s mode\n",FileName,Mode);
        return(NULL);
    } else {
        return(Stream);
    }
}   
//-----------------------------------------------------------------------------
READFILE OpenFile(char * OriginalFileName,char * CurrentFileName) {

    String FinalFileName;
    String FileName;
    READFILE Stream;

//----Removed quotes around file name
    strcpy(FileName,OriginalFileName);
    if ((FileName[0] == '\'' || FileName[0] == '\"') && 
FileName[0] == FileName[strlen(FileName)-1]) {
        strcpy(FileName,&FileName[1]);
        FileName[strlen(FileName)-1] = '\0';
    }

//----If the filename is "--" use stdin
    if (!strcmp(FileName,"--")) {
        return(OpenFILEFile(stdin));
    }

    if (ExpandFileName(FileName,CurrentFileName,FinalFileName) == NULL) {
        return(NULL);
    }

    Stream = NewReadFile();

    if ((Stream->FileHandle = OpenFileInMode(FinalFileName,"r")) == NULL) {
        Free((void **)&Stream);
        return(NULL);
    }

//----Record the file name
    Stream->FileName = CopyHeapString(FinalFileName);

////----Copy the base directory across if none there
//    if (strlen(BaseDirectory) == 0) {
////----If current file is relative, find out current
//        if (FinalFileName[0] != '/') {
//            getcwd(BaseDirectory,MAXPATHLEN);
//            strcat(BaseDirectory,"/");
//            strcat(BaseDirectory,FinalFileName);
//        } else {
//            strcpy(BaseDirectory,FinalFileName);
//        }
////----Lop off file name
//        if ((LastSlash = strrchr(BaseDirectory,'/')) != NULL) {
//            *LastSlash = '\0';
//        } else {
//            CodingError("Setting the base directory, no slash");
//        }
//    }

    return(Stream);
}
//-----------------------------------------------------------------------------
READFILE OpenFILEFile(FILE * OpenStream) {

    READFILE Stream;

    Stream = NewReadFile();

    if ((Stream->FileHandle = OpenStream) == NULL) {
        Free((void **)&Stream);
        return(NULL);
    }

    return(Stream);
}
//-----------------------------------------------------------------------------
READFILE OpenStringFile(char * Content) {

    READFILE Stream;

    if (Content == NULL) {
        return(NULL);
    }

    Stream = NewReadFile();
    Stream->StringFileContent = CopyHeapString(Content);
    Stream->StringOffset = 0;

    return(Stream);
}
//-----------------------------------------------------------------------------
void CloseFile(READFILE Stream) {

    assert((Stream->FileHandle != NULL || Stream->StringFileContent != NULL) &&
(Stream->FileHandle == NULL || Stream->StringFileContent == NULL));

    if (Stream->FileHandle != NULL) {
//----If a file name, then close the file
        if (Stream->FileName != NULL) {
            Free((void **)&(Stream->FileName));
            fclose(Stream->FileHandle);
        }
//----Else a string file
    } else if (Stream->StringFileContent != NULL) {
        Free((void **)&(Stream->StringFileContent));
    } 
    FreeToken(&(Stream->CurrentToken));
    FreeToken(&(Stream->Auxilliary));
    Free((void **)&Stream);
}
//-----------------------------------------------------------------------------
void RemoveFile(String FileName) {

    if (unlink(FileName) == -1) {
        perror("Deleting file");
        printf("ERROR: Could not delete file %s\n",FileName);
    }
}
//-----------------------------------------------------------------------------
