#include <stdio.h>
#include <assert.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include "Tokenizer.h"
#include "DataTypes.h"
#include "Utilities.h"
#include "Signature.h"
#include "Examine.h"
#include "ListStatistics.h"
#include "Modify.h"
#include "PrintTSTP.h"
#include "PrintKIF.h"
//-----------------------------------------------------------------------------
char * KIFConnectiveToString(ConnectiveType Connective) {

    switch (Connective) {
        case disjunction:
            return("or");
            break;
        case conjunction:
            return("and");
            break;
        case equivalence:
            return("<=>");
            break;
        case implication:
            return("=>");
            break;
        case reverseimplication:
            return("<=");
            break;
        case negation:
            return("not");
            break;
        case universal:
            return("forall");
            break;
        case existential:
            return("exists");
            break;
        default:
            CodingError("Not a KIF connective\n");
            return(NULL);
            break;
    }       
}
//-----------------------------------------------------------------------------
int KIFAssociative(ConnectiveType Connective) {

    return(Connective == disjunction || Connective == conjunction);
}
//-----------------------------------------------------------------------------
void KIFPrintFormula(FILE * Stream,FORMULA Formula,int Indent,
int AlreadyIndented,int Pretty,ConnectiveType LastConnective,
ConnectiveType NextConnective) {

    switch (Formula->Type) {
        case quantified:
            PrintIndent(Stream,Indent,AlreadyIndented,Pretty);
            fprintf(Stream,"( %s ",KIFConnectiveToString(
Formula->FormulaUnion.QuantifiedFormula.Quantifier));
            fprintf(Stream,"?%s",
GetSymbol(Formula->FormulaUnion.QuantifiedFormula.Variable));
//----Here's where types will be printed, in a future TSTP
//----List variables for same quantifiers
            while (Pretty && 
Formula->FormulaUnion.QuantifiedFormula.Formula->Type == quantified &&
Formula->FormulaUnion.QuantifiedFormula.Quantifier ==
Formula->FormulaUnion.QuantifiedFormula.Formula->FormulaUnion.QuantifiedFormula.Quantifier) {
                Formula = Formula->FormulaUnion.QuantifiedFormula.Formula;
                fprintf(Stream," ?%s",
GetSymbol(Formula->FormulaUnion.QuantifiedFormula.Variable));
            }
            fprintf(Stream," ");
//----If unary or atom, do on same line
            if ((Formula->FormulaUnion.QuantifiedFormula.Formula->Type == 
unary) || (Formula->FormulaUnion.QuantifiedFormula.Formula->Type == atom)) {
                KIFPrintFormula(Stream,Formula->
FormulaUnion.QuantifiedFormula.Formula,Indent,Indent,Pretty,
Formula->FormulaUnion.QuantifiedFormula.Quantifier,none);
            } else {
//----Otherwise on the next line
                fprintf(Stream,"%s",(Pretty?"\n":""));
//----If another quantified, no extra indent
                if (Formula->FormulaUnion.QuantifiedFormula.Formula->Type !=
quantified) {
                    Indent += 2;
                }
                KIFPrintFormula(Stream,Formula->
FormulaUnion.QuantifiedFormula.Formula,Indent,0,Pretty,
Formula->FormulaUnion.QuantifiedFormula.Quantifier,none);
            }
            fprintf(Stream," )");
            break;

        case binary:
            PrintIndent(Stream,Indent,AlreadyIndented,1);
            fprintf(Stream,"(%s ",KIFConnectiveToString(
Formula->FormulaUnion.BinaryFormula.Connective));
            KIFPrintFormula(Stream,Formula->FormulaUnion.BinaryFormula.LHS,
Indent+4,Indent+4,Pretty,none,Formula->FormulaUnion.BinaryFormula.Connective);
            fprintf(Stream,",\n");
            KIFPrintFormula(Stream,Formula->FormulaUnion.BinaryFormula.RHS,
Indent+4,0,Pretty,Formula->FormulaUnion.BinaryFormula.Connective,none);
            fprintf(Stream,")");
            break;

        case unary:
            PrintIndent(Stream,Indent,AlreadyIndented,1);
//----Special for infix negated equality
            if (Formula->FormulaUnion.UnaryFormula.Connective == negation && 
Formula->FormulaUnion.UnaryFormula.Formula->Type == atom &&
!strcmp(GetSymbol(Formula->FormulaUnion.UnaryFormula.Formula->
FormulaUnion.Atom),"$equal")) {
                PrintTSTPTerm(Stream,Formula->FormulaUnion.UnaryFormula.
Formula->FormulaUnion.Atom->Arguments[0],0);
                fprintf(Stream," \\= ");
                PrintTSTPTerm(Stream,Formula->FormulaUnion.UnaryFormula.
Formula->FormulaUnion.Atom->Arguments[1],0);
            } else {
                fprintf(Stream,"(%s ",KIFConnectiveToString(
Formula->FormulaUnion.UnaryFormula.Connective));
                KIFPrintFormula(Stream,Formula->FormulaUnion.UnaryFormula.
Formula,Indent+3,Indent+3,Pretty,Formula->FormulaUnion.UnaryFormula.Connective,
none);
                fprintf(Stream,")");
            }
            break;

        case atom:
            PrintIndent(Stream,Indent,AlreadyIndented,Pretty);
            PrintTSTPTerm(Stream,Formula->FormulaUnion.Atom,1);
            break;

        default:
            CodingError("Formula type unknown for printing KIF\n");
            break;
    }

}
//-----------------------------------------------------------------------------
void KIFPrintAnnotatedTSTPNode(FILE * Stream,ANNOTATEDFORMULA 
AnnotatedFormula) {

    String Name;
    String Comment;
    StatusType SubStatus;

    switch (AnnotatedFormula->Syntax) {
        case comment:
            strcpy(Comment,AnnotatedFormula->AnnotatedFormulaUnion.Comment);
            Comment[0] = ';';
            fprintf(Stream,"%s\n",Comment);
            break;
        case blank_line:
            fprintf(Stream,"\n");
            break;
        case tptp_cnf:
            fprintf(Stream,"; %s, %s\n",GetName(AnnotatedFormula,Name),
StatusToString(GetStatus(AnnotatedFormula,&SubStatus)));
            fprintf(Stream,"( ");
            KIFPrintFormula(Stream,AnnotatedFormula->AnnotatedFormulaUnion.
AnnotatedTSTPFormula.FormulaWithVariables->Formula,0,0,1,outermost,outermost);
            fprintf(Stream," ).\n");
            break;
        case tptp_fof:
            fprintf(Stream,"; %s, %s\n",GetName(AnnotatedFormula,Name),
StatusToString(GetStatus(AnnotatedFormula,&SubStatus)));
            KIFPrintFormula(Stream,AnnotatedFormula->AnnotatedFormulaUnion.
AnnotatedTSTPFormula.FormulaWithVariables->Formula,0,0,1,outermost,outermost);
            fprintf(Stream,".\n");
            break;
        default:
            CodingError("Syntax unknown for printing KIF\n");
            break;
    }

}
//-----------------------------------------------------------------------------
void KIFPrintHeader(FILE * Stream,LISTNODE Head,SIGNATURE Signature) {

    LISTNODE Node;

    Node = Head;
    while (Node != NULL && !LogicalAnnotatedFormula(Node->AnnotatedFormula)) {
        KIFPrintAnnotatedTSTPNode(Stream,Node->AnnotatedFormula);
        Node = Node->Next;
    }

}
//-----------------------------------------------------------------------------
void KIFPrintTailer(FILE * Stream) {

    fprintf(Stream,";------------------------------------------------------------------------------\n");
}
//-----------------------------------------------------------------------------
void KIFPrintListOfAnnotatedTSTPNodes(FILE * Stream,LISTNODE Head) {

    LISTNODE Node;
    int StartedThisPart;

    Node = Head;
//----Skip header
    while (Node != NULL && !LogicalAnnotatedFormula(Node->AnnotatedFormula)) {
        Node = Node->Next;
    }
    StartedThisPart = 0;
    while (Node != NULL) {
        if ((StartedThisPart && !LogicalAnnotatedFormula(
Node->AnnotatedFormula) && (Node->AnnotatedFormula->Syntax != comment ||
strstr(Node->AnnotatedFormula->AnnotatedFormulaUnion.Comment,"%--------------")
== NULL)) || LogicalAnnotatedFormula(Node->AnnotatedFormula)) {
            if (GetStatus(Node->AnnotatedFormula,NULL) == conjecture) {
                fprintf(Stream,
";----FOF proof by contradiction with negated conjecture\n");
                Negate(Node->AnnotatedFormula,0);
            }
            KIFPrintAnnotatedTSTPNode(Stream,Node->AnnotatedFormula);
            StartedThisPart = 1;
        }
        Node = Node->Next;
    }
    fprintf(Stream,"\n");

}
//-----------------------------------------------------------------------------
