#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>
#include "Signature.h"
#include "Utilities.h"
#include "DataTypes.h"
//-----------------------------------------------------------------------------
SYMBOLNODE NewSignatureNode(char * Name,int Arity) {

    SYMBOLNODE Symbol;

    Symbol = (SYMBOLNODE)Malloc(sizeof(SymbolNodeType));
    Symbol->NameSymbol = CopyHeapString(Name);
    Symbol->NumberOfUses = 1;
    Symbol->Arity = Arity;
    Symbol->LastSymbol = NULL;
    Symbol->NextSymbol = NULL;

    return(Symbol);
}
//-----------------------------------------------------------------------------
SIGNATURE NewSignature(void) {

    SIGNATURE Signature;

    Signature = (SIGNATURE)Malloc(sizeof(SignatureType));
    Signature->Variables = NULL;
    Signature->Functions = NULL;
    Signature->Predicates = NULL;
    Signature->NonLogicals = NULL;

    return(Signature);
}
//-----------------------------------------------------------------------------
void FreeSignatureList(SYMBOLNODE * Symbols) {

    if ((*Symbols) != NULL) {
        assert((*Symbols)->NumberOfUses == 0);
        FreeSignatureList(&((*Symbols)->LastSymbol));
        FreeSignatureList(&((*Symbols)->NextSymbol));
        Free((void **)(&((*Symbols)->NameSymbol)));
        Free((void **)Symbols);
    }
}
//-----------------------------------------------------------------------------
void FreeSignature(SIGNATURE * Signature) {

    FreeSignatureList(&((*Signature)->Variables));
    assert((*Signature)->Variables == NULL);
    FreeSignatureList(&((*Signature)->Functions));
    assert((*Signature)->Functions == NULL);
    FreeSignatureList(&((*Signature)->Predicates));
    assert((*Signature)->Predicates == NULL);
    FreeSignatureList(&((*Signature)->NonLogicals));
    assert((*Signature)->NonLogicals == NULL);

    Free((void **)Signature);
}
//-----------------------------------------------------------------------------
SYMBOLNODE RemoveSignatureNodeFromTree(SYMBOLNODE * OneToDeletePtr) {

    SYMBOLNODE * LargestPointer;
    SYMBOLNODE Deleted;

    Deleted = *OneToDeletePtr;

    if (Deleted->LastSymbol == NULL) {
        if (Deleted->NextSymbol != NULL) {
            *OneToDeletePtr = Deleted->NextSymbol;
        } else {
            *OneToDeletePtr = NULL;
        }
    } else if (Deleted->NextSymbol == NULL) {
        *OneToDeletePtr = Deleted->LastSymbol;
    } else {
        LargestPointer = &(Deleted->LastSymbol);
        while ((*LargestPointer)->NextSymbol != NULL) {
            LargestPointer = &((*LargestPointer)->NextSymbol);
        }
        *OneToDeletePtr = RemoveSignatureNodeFromTree(LargestPointer);
        (*OneToDeletePtr)->LastSymbol = Deleted->LastSymbol;
        (*OneToDeletePtr)->NextSymbol = Deleted->NextSymbol;
    }
    return(Deleted);
}
//-----------------------------------------------------------------------------
int RemovedUnusedSymbolsFromList(SYMBOLNODE * Symbols) {

    int NumberRemoved;
    SYMBOLNODE ToFree;

    NumberRemoved = 0;
    if (*Symbols != NULL) {
        NumberRemoved += RemovedUnusedSymbolsFromList(&((*Symbols)->
LastSymbol));
        NumberRemoved += RemovedUnusedSymbolsFromList(&((*Symbols)->
NextSymbol));
        if ((*Symbols)->NumberOfUses  == 0) {
            ToFree = RemoveSignatureNodeFromTree(Symbols);
            Free((void **)&ToFree);
            NumberRemoved++;
        }
    }

    return(NumberRemoved);
}
//-----------------------------------------------------------------------------
int RemovedUnusedSymbols(SIGNATURE Signature) {

    int TotalRemoved;

    TotalRemoved = 0;
    TotalRemoved += RemovedUnusedSymbolsFromList(&(Signature->Variables));
    TotalRemoved += RemovedUnusedSymbolsFromList(&(Signature->Functions));
    TotalRemoved += RemovedUnusedSymbolsFromList(&(Signature->Predicates));
    TotalRemoved += RemovedUnusedSymbolsFromList(&(Signature->NonLogicals));

    return(TotalRemoved);
}
//-----------------------------------------------------------------------------
char * GetSignatureSymbol(SYMBOLNODE SymbolNode) {

    return(SymbolNode->NameSymbol);
}
//-----------------------------------------------------------------------------
int GetSignatureArity(SYMBOLNODE SymbolNode) {

    return(SymbolNode->Arity);
}
//-----------------------------------------------------------------------------
int GetSignatureUses(SYMBOLNODE SymbolNode) {

    return(SymbolNode->NumberOfUses);
}
//-----------------------------------------------------------------------------
void IncreaseSymbolUseCount(SYMBOLNODE Symbol,int HowMuch) {

    Symbol->NumberOfUses += HowMuch;
}
//-----------------------------------------------------------------------------
SYMBOLNODE IsSymbolInSignatureList(SYMBOLNODE List,char * Name,int Arity) {

    if (List == NULL || !(!strcmp(GetSignatureSymbol(List),Name) &&
GetSignatureArity(List) == Arity)) {
        return(List);
    } else if (strcmp(Name,GetSignatureSymbol(List)) < 0) {
        return(IsSymbolInSignatureList(List->LastSymbol,Name,Arity));
    } else {
        return(IsSymbolInSignatureList(List->NextSymbol,Name,Arity));
    }
}
//-----------------------------------------------------------------------------
SYMBOLNODE InsertIntoSignatureList(SYMBOLNODE * List,char * Name,int Arity) {

    if (*List == NULL) {
        *List = NewSignatureNode(Name,Arity);
        return(*List);
    } else if (!strcmp(GetSignatureSymbol(*List),Name) &&
GetSignatureArity(*List) == Arity) {
        IncreaseSymbolUseCount(*List,1);
        return(*List);
    } else if (strcmp(Name,GetSignatureSymbol(*List)) < 0) {
        return(InsertIntoSignatureList(&((*List)->LastSymbol),Name,Arity));
    } else {
        return(InsertIntoSignatureList(&((*List)->NextSymbol),Name,Arity));
    }
}
//-----------------------------------------------------------------------------
SYMBOLNODE InsertIntoSignature(SIGNATURE Signature,TermType Type,
char * Name,int Arity) {

    switch (Type) {
        case variable:
            return(InsertIntoSignatureList(&(Signature->Variables),Name,Arity));
            break;
        case function:
            return(InsertIntoSignatureList(&(Signature->Functions),Name,Arity));
            break;
        case predicate:
            return(InsertIntoSignatureList(&(Signature->Predicates),Name,
Arity));
            break;
        case non_logical_data:
            return(InsertIntoSignatureList(&(Signature->NonLogicals),Name,
Arity));
            break;
        default:
            printf("ERROR: Unknown type for signature\n");
            exit(-1);
    }
}
//-----------------------------------------------------------------------------
void DoNextInSignature(SYMBOLNODE SignatureTree,SYMBOLNODE Current,
int * CurrentFound,int * NextFound,SYMBOLNODE * Next) {

//----If the tree exists and the next (after current) has not been found
    if (SignatureTree != NULL && !*NextFound) {
//----If the current has not even been found, search on
        if (!*CurrentFound) {
//----If there's a current one (not the first time) and this is it, woohoo
            if (Current != NULL && !strcmp(GetSignatureSymbol(SignatureTree),
GetSignatureSymbol(Current))) {
//----If there's a right subtree, get the left most of that
                if (SignatureTree->NextSymbol != NULL) {
                    DoNextInSignature(SignatureTree->NextSymbol,NULL,
CurrentFound,NextFound,Next);
//----Otherwise catch the next on the way out of the recursion
                } else {
                    *CurrentFound = 1;
                }
//----If there's no current one (first time) or the current is small, go left
            } else if (Current == NULL || strcmp(GetSignatureSymbol(Current),
GetSignatureSymbol(SignatureTree)) < 0) {
                DoNextInSignature(SignatureTree->LastSymbol,Current,
CurrentFound,NextFound,Next);
//----If current found to the (immediate ... next not found) left, catch 
//----the next one here
                if (*CurrentFound && !*NextFound) {
                    *NextFound = 1;
                    *Next = SignatureTree;
                }
//----Currennt exists and is larger ... head right
            } else {
                DoNextInSignature(SignatureTree->NextSymbol,Current,
CurrentFound,NextFound,Next);
            }
//----If the current has been found, then we're one level on the way out
//----of the recursion, and this is the next. Catch it.
        } else {
            *Next = SignatureTree;
            *NextFound = 1;
        }
//----If at the end of the tree, then the current does not exist, i.e., must
//----have been NULL, so we've reached the left corner. Mark current as found
//----so left corner node is caught on way out of recursion.
    } else {
        *CurrentFound = 1;
    }
}
//-----------------------------------------------------------------------------
SYMBOLNODE NextInSignatureTree(SYMBOLNODE SignatureTree,SYMBOLNODE Current) {

    SYMBOLNODE Next;
    int CurrentFound;
    int NextFound;

    Next = NULL;
    CurrentFound = 0;
    NextFound = 0;
    DoNextInSignature(SignatureTree,Current,&CurrentFound,&NextFound,&Next);
    return(Next);
}
//-----------------------------------------------------------------------------
void ListSignatureBySearch(SYMBOLNODE SignatureTree) {

    SYMBOLNODE Current;

    Current = NextInSignatureTree(SignatureTree,NULL);
    while (Current != NULL) {
        printf("%s/%d(%d)\n",GetSignatureSymbol(Current),
GetSignatureArity(Current),GetSignatureUses(Current));
        fflush(stdout);
        Current = NextInSignatureTree(SignatureTree,Current);
    }
}
//-----------------------------------------------------------------------------
void PrintSignatureTree(SYMBOLNODE SignatureTree) {

    if (SignatureTree != NULL) {
        PrintSignatureTree(SignatureTree->LastSymbol);
        printf("%s/%d(%d)\n",GetSignatureSymbol(SignatureTree),
GetSignatureArity(SignatureTree),GetSignatureUses(SignatureTree));
        fflush(stdout);
        PrintSignatureTree(SignatureTree->NextSymbol);
    }
}
//-----------------------------------------------------------------------------
void PrintSignature(SIGNATURE Signature) {

    printf("Variables:  ");
    PrintSignatureTree(Signature->Variables);
    printf("\n");
    printf("Functions:  ");
    PrintSignatureTree(Signature->Functions);
    printf("\n");
    printf("Predicates: ");
    PrintSignatureTree(Signature->Predicates);
    printf("\n");
    printf("NonLogical: ");
    PrintSignatureTree(Signature->NonLogicals);
    printf("\n");
}
//-----------------------------------------------------------------------------
//----PutUsageHere must be address of a malloced String
char * GetSignatureSymbolUsage(SIGNATURE Signature,char ** PutUsageHere,
char ** FunctorUsageStartsHere) {

    SYMBOLNODE Current;
    String OneSymbolUsage;
    int UsageLength = STRINGLENGTH;
    int PredicateCollectorLength;

    strcpy(*PutUsageHere,"");
    Current = NextInSignatureTree(Signature->Predicates,NULL);
    while (Current != NULL) {
        sprintf(OneSymbolUsage,"%s/%d/%d\n",GetSignatureSymbol(Current),GetSignatureArity(Current),GetSignatureUses(Current));
        ExtendString(PutUsageHere,OneSymbolUsage,&UsageLength);
        Current = NextInSignatureTree(Signature->Predicates,Current);
    }
    PredicateCollectorLength = strlen(*PutUsageHere);
    Current = NextInSignatureTree(Signature->Functions,NULL);
    while (Current != NULL) {
        sprintf(OneSymbolUsage,"%s/%d/%d\n",GetSignatureSymbol(Current),
GetSignatureArity(Current),GetSignatureUses(Current));
        ExtendString(PutUsageHere,OneSymbolUsage,&UsageLength);
        Current = NextInSignatureTree(Signature->Functions,Current);
    }
    *FunctorUsageStartsHere = (*PutUsageHere) + PredicateCollectorLength;
    return(*PutUsageHere);
}
//-----------------------------------------------------------------------------
