#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <ctype.h>
#include <stdlib.h>
#include "DataTypes.h"
#include "Utilities.h"
#include "FileUtilities.h"
#include "SystemOnTPTP.h"
//-----------------------------------------------------------------------------
void SystemOnTPTPFileName(char * Directory,char * BaseName,char * Extension,
String FileName) {

    String InternalFileName;

    if (Directory != NULL) {
        strcpy(InternalFileName,Directory);
        strcat(InternalFileName,"/");
    } else {
        strcpy(InternalFileName,"");
    }

    if (BaseName != NULL) {
        strcat(InternalFileName,BaseName);
    } else {
        strcat(InternalFileName,"SoTXXXXXX");
        mktemp(InternalFileName);
    }
    if (Extension != NULL) {
        strcat(InternalFileName,Extension);
    } 

//----Copy across at end to avoid clashes with same variable in call
    strcpy(FileName,InternalFileName);
}
//-----------------------------------------------------------------------------
int SystemOnTPTPAvailable(void) {

    String UNIXCommand;
    char * TPTPHome;

//----First look if user has a TPTP_HOME environment variable
    if ((TPTPHome = getenv("TPTP_HOME")) != NULL) {
        sprintf(UNIXCommand,"%s/%s",TPTPHome,SYSTEM_ON_TPTP);
//----If not, use the macro from compile time
    } else {
        sprintf(UNIXCommand,"%s/%s",TPTP_HOME,SYSTEM_ON_TPTP);
    }
    if (access(UNIXCommand,X_OK) == 0) {
        return(1);
    } else {
        perror(UNIXCommand);
        return(0);
    }
}
//-----------------------------------------------------------------------------
int SystemOnTPTPGetResult(char * ProblemFileName,char * ATPSystem,int TimeLimit,
char * SystemOutputPrefix,char * PutResultHere,int KeepOutputFiles,
char * FilesDirectory,char * UsersOutputFileName,char * OutputFileName) {

    String UNIXCommand;
    char * TPTPHome;
    String InternalOutputFileName;
    FILE * OutputFileHandle;
    FILE * SystemPipe;
    int GotResult;
    String SystemOutputLine;
    char * SaysPart;
    char * CPUPart;
    char * WCPart;

//----First look if user has a TPTP_HOME environment variable
    if ((TPTPHome = getenv("TPTP_HOME")) != NULL) {
        sprintf(UNIXCommand,"%s/%s -q0 %s %d %s",TPTPHome,SYSTEM_ON_TPTP,
ATPSystem,TimeLimit,ProblemFileName);
//----If not, use the macro from compile time
    } else {
        sprintf(UNIXCommand,"%s/%s -q0 %s %d %s",TPTP_HOME,SYSTEM_ON_TPTP,
ATPSystem,TimeLimit,ProblemFileName);
    }
    if ((SystemPipe = popen(UNIXCommand,"r")) == NULL) {
        perror("Running SystemOnTPTP");
        printf("ERROR: Could not start %s\n",UNIXCommand);
        return(0);
    }

    if (KeepOutputFiles) {
        SystemOnTPTPFileName(FilesDirectory,UsersOutputFileName,NULL,
InternalOutputFileName);
        if ((OutputFileHandle = OpenFileInMode(InternalOutputFileName,
"w")) == NULL) {
            pclose(SystemPipe);
            return(0);
        } else {
            if (OutputFileName != NULL) {
                strcpy(OutputFileName,InternalOutputFileName);
            }
        }
    }

//----Read SystemOnTPTP output echoing to file and looking for RESULT
    GotResult = 0;
    while (fgets(SystemOutputLine,STRINGLENGTH,SystemPipe) != NULL) {
        if (KeepOutputFiles) {
            fputs(SystemOutputLine,OutputFileHandle);
        }
//DEBUG printf("Line is %s",SystemOutputLine);
        if (!GotResult && 
strstr(SystemOutputLine,"RESULT: ") == SystemOutputLine &&
(SaysPart = strstr(SystemOutputLine," says ")) != NULL &&
(CPUPart = strstr(SystemOutputLine," - CPU =")) != NULL &&
(WCPart = strstr(SystemOutputLine," WC ="))) {
            *CPUPart = '\0';
            strcpy(PutResultHere,SaysPart+6);
            *CPUPart = ' ';
            *WCPart = '\0';
            if (SystemOutputPrefix != NULL) {
                printf("%s%s\n",SystemOutputPrefix,SystemOutputLine);
            }
            *WCPart = ' ';
            GotResult = 1;
//DEBUG printf("Got the result %s\n",PutResultHere);
        }
    }

    pclose(SystemPipe);
    if (KeepOutputFiles) {
        fclose(OutputFileHandle);
    }

    return(GotResult);
}
//-----------------------------------------------------------------------------
int SystemOnTPTP(LISTNODE Axioms,ANNOTATEDFORMULA Conjecture,
char * PositiveChecker,char * PositiveResult,int TestNegative,
char * NegativeChecker,char * NegativeResult,int TimeLimit,
char * SystemOutputPrefix,
int KeepOutputFiles,char * FilesDirectory,char * UsersOutputFileName,
String OutputFileName) {

    String ProblemFileName;
    String CopyUsersOutputFileName;
    FILE * ProblemFileHandle;
    String SystemResult;
    int Correct;

    SystemOnTPTPFileName(FilesDirectory,NULL,NULL,ProblemFileName);

    if ((ProblemFileHandle = OpenFileInMode(ProblemFileName,"w")) == NULL) {
        return(0);
    }
    if (Axioms != NULL) {
        PrintListOfAnnotatedTSTPNodesWithStatus(ProblemFileHandle,NULL,Axioms,
tptp,1,axiom);
    }
    if (Conjecture != NULL) {
        PrintAnnotatedTSTPNodeWithStatus(ProblemFileHandle,Conjecture,
tptp,1,conjecture);
    }
    fclose(ProblemFileHandle);

//----0 means no result yet
    Correct = 0;

//----Need to make a copy in case the same variable is used
    strcpy(CopyUsersOutputFileName,UsersOutputFileName);
    if (Correct == 0) {
        if (SystemOnTPTPGetResult(ProblemFileName,PositiveChecker,TimeLimit,
SystemOutputPrefix,SystemResult,KeepOutputFiles,FilesDirectory,
CopyUsersOutputFileName,OutputFileName)) {
            if (!strcmp(SystemResult,PositiveResult)) {
                Correct = 1;
//----Should not trust prover's disproofs
//            } else if (!strcmp(SystemResult,NegativeResult)) {
//                Correct = -1;
            }
        }
    }

//----Check if really not provable
    if (Correct == 0 && TestNegative) {
        strcat(OutputFileName,"_not");
        if (SystemOnTPTPGetResult(ProblemFileName,NegativeChecker,TimeLimit,
SystemOutputPrefix,SystemResult,KeepOutputFiles,FilesDirectory,
CopyUsersOutputFileName,OutputFileName)) {
            if (!strcmp(SystemResult,NegativeResult)) {
                Correct = -1;
//----Should not trust disprover's proofs
//            } else if (!strcmp(SystemResult,PositiveResult)) {
//                Correct = 1;
            }
        }
    }

    RemoveFile(ProblemFileName);

    return(Correct);
}
//-----------------------------------------------------------------------------
