#include <stdio.h>
#include <assert.h>
#include <string.h>
#include <stdlib.h>
#include "DataTypes.h"
#include "Utilities.h"
#include "Examine.h"
#include "Parsing.h"
#include "List.h"
#include "Tree.h"
//-----------------------------------------------------------------------------
void ResetTreeVisited(TREENODE Tree) {

    int Index;

    if (Tree != NULL && Tree->Visited) {
        Tree->Visited = 0;
        for (Index = 0; Index < Tree->NumberOfParents; Index++) {
            ResetTreeVisited(Tree->Parents[Index]);
        }
    }
}
//-----------------------------------------------------------------------------
void ResetRootListVisited(ROOTLIST RootListHead) {

    while (RootListHead != NULL) {
        if (RootListHead->TheTree != NULL) {
            ResetTreeVisited(RootListHead->TheTree);
        }
        RootListHead = RootListHead->Next;
    }
}
//-----------------------------------------------------------------------------
LISTNODE GetRootList(LISTNODE Head) {

    LISTNODE RootList;
    String AllParentNames;
    int NumberOfParents;
    StringParts ParentNames;
    int ParentNumber;
    LISTNODE * Remover;
    int ParentFound;
    LISTNODE ToRemove;
    String ParentName;

    RootList = DuplicateListOfNodes(Head,0);

    while ((Head = GetLogicNode(Head)) != NULL) {
        GetParentNames(Head->AnnotatedFormula,AllParentNames);
        NumberOfParents = Tokenize(AllParentNames,ParentNames,"\n");
        for (ParentNumber = 0;ParentNumber< NumberOfParents;ParentNumber++) {
            Remover = &RootList;
            ParentFound = 0;
            while (!ParentFound && *Remover != NULL) {
                if (!strcmp(ParentNames[ParentNumber],
GetName((*Remover)->AnnotatedFormula,ParentName))) {
                    ToRemove = *Remover;
                    *Remover = (*Remover)->Next;
                    FreeListNode(&ToRemove);
                    ParentFound = 1;
                } else {
                    Remover = &((*Remover)->Next);
                }
            }
        }
        Head = Head->Next;
    }

    return(RootList);
}
//-----------------------------------------------------------------------------
TREENODE * NewParentsList(int NumberOfParents) {

    TREENODE * ParentsList;
    int Index;

    ParentsList = (TREENODE *)Malloc(NumberOfParents * sizeof(TREENODE));
    for (Index = 0; Index < NumberOfParents; Index++) {
        ParentsList[Index] = NULL;
    }

    return(ParentsList);
}
//-----------------------------------------------------------------------------
TREENODE NewTreeNode(ANNOTATEDFORMULA AnnotatedFormula) {

    TREENODE TreeNode;

    TreeNode = (TREENODE)Malloc(sizeof(TreeNodeType));
    TreeNode->NumberOfUses = 1;
    TreeNode->AnnotatedFormula = AnnotatedFormula;
    TreeNode->AnnotatedFormula->NumberOfUses++;
    TreeNode->NumberOfParents = 0;
    TreeNode->Parents = NULL;
    TreeNode->Visited = 0;
    TreeNode->StatisticsCache = 0;
    return(TreeNode);
}
//-----------------------------------------------------------------------------
ROOTLIST NewRootNode(TREENODE TheTree) {

    ROOTLIST RootList;

    RootList = (ROOTLIST)Malloc(sizeof(RootListType));
    RootList->Next = NULL;
    RootList->TheTree = TheTree;
    if (RootList->TheTree != NULL) {
        RootList->TheTree->NumberOfUses++;
    }

    return(RootList);
}
//-----------------------------------------------------------------------------
void AddRootNode(ROOTLIST * From,ROOTLIST Next,TREENODE TheTree) {

    ROOTLIST NewNode;

    NewNode = NewRootNode(TheTree);
    NewNode->Next = Next;
    *From = NewNode;
}
//-----------------------------------------------------------------------------
void FreeTree(TREENODE * Tree) {

    int Index;
//DEBUG String FormulaName;

    if (*Tree != NULL) {
//DEBUG printf("enter %s with %d uses\n",GetName((*Tree)->AnnotatedFormula,FormulaName),(*Tree)->NumberOfUses);

//----Recursively do parents first
        if (!(*Tree)->Visited) {
            (*Tree)->Visited = 1;
            if ((*Tree)->NumberOfParents > 0) {
//DEBUG printf("do the %d parents of %s\n",(*Tree)->NumberOfParents,FormulaName);
                for (Index = 0; Index < (*Tree)->NumberOfParents; Index++) {
                    FreeTree(&((*Tree)->Parents[Index]));
                }
                (*Tree)->NumberOfParents = 0;;
                Free((void **)&((*Tree)->Parents));
            }
        }

        --((*Tree)->NumberOfUses);
//DEBUG printf("%d more uses for %s\n",(*Tree)->NumberOfUses,FormulaName);

//----If no more uses then free it up
        if ((*Tree)->NumberOfUses == 0) {
//DEBUG printf("and now free memory for %s\n",FormulaName);
            FreeAnnotatedFormula(&((*Tree)->AnnotatedFormula));
            Free((void **)Tree);
        }
    }
}
//-----------------------------------------------------------------------------
void DoFreeRootList(ROOTLIST * Head,int MustFreeTree) {

    if (*Head != NULL) {
//----Recursively do rest first
        if ((*Head)->Next != NULL) {
            DoFreeRootList(&((*Head)->Next),MustFreeTree);
        }
        if (MustFreeTree) {
            FreeTree(&((*Head)->TheTree));
        } else {
//----Simply decrement the direct node counter
            (*Head)->TheTree->NumberOfUses--;
        }
        Free((void **)Head);
    }
}
//-----------------------------------------------------------------------------
void FreeRootList(ROOTLIST * Head,int MustFreeTree) {

    if (MustFreeTree) {
        ResetRootListVisited(*Head);
    }
    DoFreeRootList(Head,MustFreeTree);
}
//-----------------------------------------------------------------------------
TREENODE DoAnnotatedFormulaInTree(TREENODE ATree,ANNOTATEDFORMULA 
LookingForThis) {

    int ParentIndex;
    TREENODE AncestorNode;

    if (ATree != NULL) {
        if (!ATree->Visited) {
//----Is this the one?
            if (ATree->AnnotatedFormula == LookingForThis) {
                return(ATree);
            } else {
//----Look in subtrees
                for (ParentIndex = 0; ParentIndex < ATree->NumberOfParents;
ParentIndex++) {
                    if ((AncestorNode = DoAnnotatedFormulaInTree(ATree->
Parents[ParentIndex],LookingForThis)) != NULL) {
                        return(AncestorNode);
                    }
                }
            }
            ATree->Visited = 1;
        } else {
            return(NULL);
        }
    } 

    return(NULL);
}
//-----------------------------------------------------------------------------
TREENODE AnnotatedFormulaInTree(TREENODE ATree,ANNOTATEDFORMULA 
LookingForThis) {

    TREENODE TreeNodeFound;

    ResetTreeVisited(ATree);
    TreeNodeFound = DoAnnotatedFormulaInTree(ATree,LookingForThis);
    ResetTreeVisited(ATree);
    return(TreeNodeFound);
}
//-----------------------------------------------------------------------------
TREENODE AnnotatedFormulaInTrees(ROOTLIST RootListHead,ANNOTATEDFORMULA 
LookingForThis) {

    TREENODE TreeNode;

    while (RootListHead != NULL) {
        if ((TreeNode = AnnotatedFormulaInTree(RootListHead->TheTree,
LookingForThis)) != NULL) {
            return(TreeNode);
        } else {
            RootListHead = RootListHead->Next;
        }
    }
    return(NULL);
}
//-----------------------------------------------------------------------------
int CountParents(char * ParentNames) {

    int NumberOfParents;
    int Index;

    NumberOfParents = 0;
    for (Index = 0; Index < (int)strlen(ParentNames); Index++) {
        if (ParentNames[Index]  == '\n') {
            NumberOfParents++;
        }
    }

    return(NumberOfParents);
}
//-----------------------------------------------------------------------------
TREENODE InListOfTreeNodes(ROOTLIST Head,ANNOTATEDFORMULA LookingForThis) {

    while (Head != NULL) {
        if (Head->TheTree != NULL && Head->TheTree->AnnotatedFormula ==
LookingForThis) {
            return(Head->TheTree);
        }
        Head = Head->Next;
    }

    return(NULL);
}
//-----------------------------------------------------------------------------
int RemoveTheoryParents(char * ParentNames) {

    int NumberOfTheoryParents;
    char * TheoryParent;
    char * EndOfTheoryParent;

    NumberOfTheoryParents = 0;
    while ((TheoryParent = strstr(ParentNames,"theory(")) != NULL) {
        NumberOfTheoryParents++;
        if ((EndOfTheoryParent = strchr(TheoryParent,'\n')) == NULL) {
            CodingError("Broken parent name list in RemoveTheoryParents\n");
        }
        strcpy(TheoryParent,EndOfTheoryParent+1);
    }
    return(NumberOfTheoryParents);
}
//-----------------------------------------------------------------------------
//----Using a reference parameter rather than returning so that the linking
//----is done immediately to allow search what we have done already
TREENODE BuildTree(LISTNODE Head,ROOTLIST RootListHead,ANNOTATEDFORMULA 
RootAnnotatedFormula,TREENODE * TheTree,ROOTLIST * ListOfTreeNodes) {

    String AllParentNames;
    StringParts ParentNames;
    int ParentIndex;
    LISTNODE ParentNode;

//DEBUG printf("Building tree for root %s \n",GetName(RootAnnotatedFormula,AllParentNames));
//----Try link to an existing node in the tree
    if (((*TheTree) = InListOfTreeNodes(*ListOfTreeNodes,
RootAnnotatedFormula)) != NULL) {
//DEBUG printf("already in tree\n");
        (*TheTree)->NumberOfUses++;
    } else {

//DEBUG printf("making the tree node\n");
//----Make new tree node
        (*TheTree) = NewTreeNode(RootAnnotatedFormula);

//----Add to linear list
        AddRootNode(ListOfTreeNodes,*ListOfTreeNodes,*TheTree);
//        LinearNode = NewRootNode(*TheTree);
//        LinearNode->Next = *ListOfTreeNodes;
//        *ListOfTreeNodes = LinearNode;
        
//----Do parents if derived 
        if (DerivedAnnotatedFormula((*TheTree)->AnnotatedFormula)) {
            GetParentNames((*TheTree)->AnnotatedFormula,AllParentNames);
//DEBUG printf("All parents are %s\n",AllParentNames);
            RemoveTheoryParents(AllParentNames);
//DEBUG printf("Non-theory parents are %s\n",AllParentNames);
            if (((*TheTree)->NumberOfParents = Tokenize(AllParentNames,
ParentNames,"\n")) == 0) {
//DEBUG printf("There are no parents\n");
                (*TheTree)->Parents = NULL;
            } else {
//DEBUG printf("Parents are %s\n",AllParentNames);
                (*TheTree)->Parents = NewParentsList((*TheTree)->NumberOfParents);
                for (ParentIndex = 0;ParentIndex < (*TheTree)->NumberOfParents;
ParentIndex++) {
                    if ((ParentNode = *GetNodeFromListByAnnotatedFormulaName(
&Head,ParentNames[ParentIndex])) == NULL || BuildTree(Head,RootListHead,
ParentNode->AnnotatedFormula,&((*TheTree)->Parents[ParentIndex]),
ListOfTreeNodes) == NULL) {
//DEBUG printf("ERROR: Could not build parent trees\n");
                        FreeTree(TheTree);
                        return(NULL);
                    }
                }
            }
        } else {
            (*TheTree)->NumberOfParents = 0;
            (*TheTree)->Parents = NULL;
        }
    }
    return(*TheTree);
}
//-----------------------------------------------------------------------------
ROOTLIST BuildRootList(LISTNODE Head) {

    ROOTLIST RootListHead;
    ROOTLIST * NextRootList;
    ROOTLIST ListOfTreeNodes;
    LISTNODE RootAnnotatedFormulae;
    LISTNODE RootAnnotatedFormulaNode;
    String RootName;

//----This is the linear list of all tree nodes
    ListOfTreeNodes = NULL;
//----This is the list of tree roots
    RootListHead = NULL;

//----Keep a pointer to last link field in the list of tree roots
    NextRootList = &RootListHead;
//----Find all the root annotated formulae - each will start a tree
    RootAnnotatedFormulae = GetRootList(Head);
    RootAnnotatedFormulaNode = RootAnnotatedFormulae;
    while (RootAnnotatedFormulaNode != NULL) {
        *NextRootList = NewRootNode(NULL);
        if (BuildTree(Head,RootListHead,RootAnnotatedFormulaNode->
AnnotatedFormula,&((*NextRootList)->TheTree),&ListOfTreeNodes) == NULL) {
            printf("ERROR: Could not build tree for root %s\n",
GetName(RootAnnotatedFormulaNode->AnnotatedFormula,RootName));
            FreeRootList(NextRootList,1);
        } else {
            NextRootList = &((*NextRootList)->Next);
        }
        RootAnnotatedFormulaNode = RootAnnotatedFormulaNode->Next;
    }

//----Free list of root annotated formulae
    FreeListOfAnnotatedFormulae(&RootAnnotatedFormulae);
//----Free linear list
    FreeRootList(&ListOfTreeNodes,0);
//    while (ListOfTreeNodes != NULL) {
//        LinearNode = ListOfTreeNodes;
//        ListOfTreeNodes = ListOfTreeNodes->Next;
//        Free((void **)&LinearNode);
//    }

    return(RootListHead);
}
//-----------------------------------------------------------------------------
void DFSWithIndent(FILE * Stream,TREENODE Root,char * Indent) {

    String Name;
    String NewIndent;
    int Index;

    fprintf(Stream,"%s%s (%d uses, %d visited)\n",Indent,
GetName(Root->AnnotatedFormula,Name),Root->NumberOfUses,Root->Visited);
    strcpy(NewIndent,Indent);
    strcat(NewIndent,"  ");
    if (Root->Parents != NULL) {
        for (Index=0;Index < Root->NumberOfParents; Index++) {
            DFSWithIndent(Stream,Root->Parents[Index],NewIndent);
        }
    }

}
//-----------------------------------------------------------------------------
void DFS(FILE * Stream,TREENODE Root) {

    DFSWithIndent(Stream,Root,"");
}
//-----------------------------------------------------------------------------
void PrintRootList(FILE * Stream,ROOTLIST RootListHead) {

    while (RootListHead != NULL) {
        if (RootListHead->TheTree == NULL) {
            fprintf(Stream,"Empty tree\n");
        } else {
           DFS(Stream,RootListHead->TheTree);
        }
        RootListHead = RootListHead->Next;
    }
}
//-----------------------------------------------------------------------------
