

// Code written by Hans de Nivelle, February 2006. 


#include "application.h"


#include "swll/searchstate.h"





int application::comparerulesnrpremisses( const application& other )
{
   unsigned int p1 = f. r. nr_premisses( ); 
   unsigned int p2 = other. f. r. nr_premisses( ); 

   if( p1 < p2 ) 
      return -1;
   if( p2 > p1 ) 
      return 1;

   return 0;
}




namespace
{


   // Supremum of a vector of unsigneds. 

   unsigned int supremum( const std::vector< unsigned int > & set )
   {
      unsigned int sup = 0;

      for( std::vector< unsigned int > :: const_iterator
              p = set. begin( );
	      p != set. end( );
              ++ p )
      {
         if( *p + 1 > sup )
	    sup = *p + 1;
      }

      return sup;
   }



   // Returns the supremum of the subset of elements of set
   // that are strictly smaller than border. 


   unsigned int bordersupremum( unsigned int border,
                                const std::vector< unsigned int > & set )
   {
      unsigned int sup = 0;
      for( std::vector< unsigned int > :: const_iterator 
	      p = set. begin( );
	      p != set. end( );
	      ++ p )
      {
         if( *p < border )
         {
            if( *p + 1 > sup )
               sup = *p + 1;
         }
      }
      return sup;
   }


   unsigned int nroccurrences( unsigned int el,
		               const std::vector< unsigned int > & set )
   {
      unsigned int nr = 0;
      for( std::vector< unsigned int > :: const_iterator 
              p = set. begin( );
	      p != set. end( );
	      ++ p )
      {
         if( *p == el )
            ++ nr;
      }
      return nr;
   }


   // Compare set1, set2 in a multiset fashion.
   // upperbound should be an upperbound of both sets. 


   int multisetcompare( const std::vector< unsigned int > & set1,
		        const std::vector< unsigned int > & set2,
		        unsigned int upperbound ) 
   {


      while( upperbound != 0 )
      {
         unsigned int sup1 = bordersupremum( upperbound, set1 );
	 unsigned int sup2 = bordersupremum( upperbound, set2 );

	 if( sup1 < sup2 )
            return -1;
	 if( sup1 > sup2 )
            return 1;

	 if( sup1 == 0 )
            return 0;
         
// If you put 1 here, there will be true multiset comparison. 
#if 0
	 unsigned int cnt1 = nroccurrences( sup1 - 1, set1 );
	 unsigned int cnt2 = nroccurrences( sup2 - 1, set2 );

	 if( cnt1 < cnt2 )
            return -1;
	 if( cnt1 > cnt2 )
            return 1;
#endif

	 upperbound = sup1 - 1; 
      }
      return 0;
   }


   // Returns number of distinct elements in set. 

   unsigned int nr_elements( const std::vector< unsigned int > & set )
   {
      unsigned int sup = supremum( set );
      unsigned int nr = 0;

      while( sup != 0 )
      {
	 sup = bordersupremum( sup, set );
	 if( sup == 0 )
            return nr;

         ++ nr;
	 sup = sup - 1;
      }
      return nr; 
   }

}



int application::comparematchingsmultiset( const application& other )
{

   unsigned int sup1 = supremum( f. prem_match ); 
   unsigned int sup2 = supremum( other. f. prem_match ); 

   if( sup1 < sup2 )
      return -1; 

   if( sup1 > sup2 )
      return 1; 

   return multisetcompare( f. prem_match, other. f. prem_match, sup1 );
}


int application::comparematchingscardinality( const application& other )
{
   unsigned int card1 = nr_elements( f. prem_match );
   unsigned int card2 = nr_elements( other. f. prem_match );

   if( card1 < card2 )
      return -1;
   if( card1 > card2 )
      return 1;
   return 0;
}


int application::comparematchingssupremum( const application& other ) 
{
   unsigned int sup1 = supremum( f. prem_match );
   unsigned int sup2 = supremum( other. f. prem_match ); 

   if( sup1 < sup2 )
      return -1;
   if( sup1 > sup2 )
      return 1;
   return 0;
}



int application::comparerulenrs( const application& other )
{
   if( rule_nr < other. rule_nr )
      return -1;
   if( rule_nr > other. rule_nr )
      return 1;
   return 0;
}


bool application::conclusionistrue( const model& m, const rule& r,
		                    groundsubst& s )  
{


#if 0
   std::cout << "checking conclusion of " << r << " under ";
   std::cout << s << "\n\n";
#endif 

   if( r. isexistential( ))
   {
      unsigned int subst_size = s. size( );
      varatom va = r. getbody( );

      for( model::atom_iterator atom = m. atoms_begin( );
              atom != m. atoms_end( );
	      ++ atom )
      {
         groundatom ga = m. getatom( atom );
	 if( ga.p == va.p )
         {
            if( s. match( va, ga ))
            {
               s. backtrack( subst_size );
	       return true;
            }
	    s. backtrack( subst_size );
         }
      }
      return false;
   }
   else
   {
      // Disjunctive rule. We look for a true disjunct:

      for( rule::conc_iterator c = r. concs_begin( );
              c != r. concs_end( );
	      ++ c )
      {
         groundatom ga = s. apply_on( *c );
	 for( model::atom_iterator atom = m. atoms_begin( );
                 atom != m. atoms_end( );
	         ++ atom )
         {
            if( m. getatom( atom ) == ga )
               return true;
         }
      }
      return false;
   }
}




partial< application >
application::bestforward( const rulesystem& rs, const model& m ) 
{

   partial< application > best;

   unsigned int rule_nr = 0; 
   for( std::vector< rulesystem::rule_in_system > :: const_iterator
	   rule = rs. rules. begin( );
           rule != rs. rules. end( );
	   ++ rule, ++ rule_nr )
   {
      if( ! rule -> deleted && 
          ( rule -> r. isexistential( ) || 
	    rule -> r. nr_conclusions( ) != 0 )) 
      {

#if 0
         std::cout << "Considering forward rule: " << rule_nr << ": "; 
	 std::cout << rule -> r << "\n";
#endif

         application appl( false_instance::false_instance( rule -> r ), 
			   rule_nr, m ); 

         if( best. isundefined( ) ||
             ( best. contents( ). branching_degree >= 
	       appl. branching_degree ))
         {
            appl. improve( m, best );
         }
      }
   }
#if 0
   std::cout << "returning: "; 
   std::cout << best << "\n\n"; 
#endif

   return best; 
}




void application::improve( const model& m, partial< application > & best )
{
   unsigned int subst_size = f. s. size( );
   unsigned int prem_match_size = f. prem_match. size( );

   improve( m, f. r. prems_begin( ), f. r. prems_end( ), best );

   ASSERT( f. s. size( ) == subst_size );
   ASSERT( f. prem_match. size( ) == prem_match_size );
}



void application::improve( const model& m,
		           rule::prem_iterator p1,
			   rule::prem_iterator p2,
			   partial< application > & best )
{


#if 0
   std::cout << "Entering improve for premisse " << *p1 << ":\n"; 
   std::cout << *this << "\n";
#endif


   if( p1 == p2 )
   {
      if( ! conclusionistrue( m, f. r, f. s ))
      {
         if( best. isdefined( ))
         {
            if( branching_degree < best. contents( ). branching_degree )
               best = *this;
	    else
            {
               if( branching_degree == best. contents( ). branching_degree )
               {
                  int c = comparematchingssupremum( best. contents( ) );
		  if( c == -1 )
                  {
                     best = *this;
                  }

		  int p = comparerulesnrpremisses( best. contents( )); 
		  if( c == 0 && p == -1 )
                  {
                     best = *this;
                  }

		  if( c == 0 && p == 0 && 
				comparerulenrs( best. contents( ) ) == -1 )
                  {
                     best = *this;
                  }
               }
            }
         }
         else
            best = *this; 
      }
   }
   else
   {
      unsigned int subst_size = f. s. size( );
      varatom va = * ( p1 ++ );

      if( va. p != predicate::diseq )
      { 
         unsigned int atom_nr = 0;  
         for( model::atom_iterator atom = m. atoms_begin( );
                 atom != m. atoms_end( );
	         ++ atom, ++ atom_nr )
         {
            groundatom ga = m. getatom( atom );
            if( va. p == ga. p && f. s. match( va, ga ))
            {
               f. prem_match. push_back( atom_nr ); 
               improve( m, p1, p2, best ); 
	       f. prem_match. pop_back( ); 
            }
            f. s. backtrack( subst_size );
         }
      }
      else
      {
         std::vector< variable > :: const_iterator a = va. a1;
	 variable v1 = *(a++);
	 variable v2 = *a;

	 const element* e1 = f. s. lookup(v1);
	 const element* e2 = f. s. lookup(v2);

	 ASSERT( e1 );
	 ASSERT( e2 );

         if( *e1 != *e2 )
         {
            element e = *e1;
	    if( *e2 > *e1 )
               e = *e2;
	    
            unsigned int atom_nr = 0; 
	    for( model::atom_iterator atom = m. atoms_begin( );
                    atom != m. atoms_end( );
	            ++ atom, ++ atom_nr ) 
            {
               groundatom ga = m. getatom( atom );
	       if( ga. p == predicate::create && *( ga. a1 ) == e )
               {
                  f. prem_match. push_back( atom_nr ); 
                  improve( m, p1, p2, best ); 
                  f. prem_match. pop_back( );
               }
            }
         }
      }
   }
}





partial< application >
application::findclosinglemma( const rulesystem& rs, const model& m, 
		               bool best ) 
{

   partial< application > lemma; 

   static swll::statistics_collector stat;

   unsigned int rule_nr = 0; 
   for( std::vector< rulesystem::rule_in_system > :: const_iterator
	   rule = rs. rules. begin( );
           rule != rs. rules. end( );
	   ++ rule, ++ rule_nr )
   {
      if( ! rule -> deleted && 
          rule -> r. isdisjunctive( ) && 
	  rule -> r. nr_conclusions( ) == 0 ) 
      {

#if 0
         std::cout << "Considering lemma: " << rule_nr << ": "; 
	 std::cout << rule -> r << "\n";
#endif


	 swll::searchstate state;
	 bool b = state. setproblem( rule -> r, m );
	 stat. nr_calls ++ ; 
	 if(b)
         {
            stat. nr_passedsetproblem ++ ; 
            state. interreduce( );

	    b = state. searchsolution( );

	    if(b)
               stat. nr_solution ++ ;

	    if( state. update_stats( stat ))
            {
               std::cout << "Application Statistics:\n"; 
	       std::cout << stat << "\n\n";
	       std::cout << rule -> r << "\n";
	       std::cout << m << "\n";
	       std::cout << "======================\n\n"; 
            }

	    if(b)
            {
               application appl( 
		        false_instance::false_instance( rule -> r ), 
			rule_nr, m ); 

               appl. f. s = state. makegroundsubst( );
	       bool acc = appl. f. computematching(m);

               ASSERT( acc );

               if( lemma. isdefined( )) 
               {
                  int c = appl. comparematchingssupremum( 
				            lemma. contents( ) );

                  if( c == -1 )
		  {
                     lemma = appl; 
                  }

		  int p = appl. comparerulesnrpremisses( lemma. contents( )); 

		  if( c == 0 && p == -1 )
                        lemma = appl;

		  int m = appl. comparerulenrs( lemma. contents( ) ); 
		  if( c == 0 && p == 0 && m == -1 )
                  {
                     lemma = appl;
                  }
               }
	       else
               {
                  lemma = appl; 
		  if( ! best )
                        return lemma;
               }
            }
         }
      }
   }
#if 0
   std::cout << stat << "\n";
   std::cout << "returning: "; 
   std::cout << lemma << "\n\n"; 
#endif

   return lemma; 
}





partial< application >
application::makeclosurewith( const model& m, const rule& lem, 
		              unsigned int nr ) 
{
   
   ASSERT( lem. isdisjunctive( ) && lem. nr_conclusions( ) == 0 );

   swll::searchstate state;
   bool b = state. setproblem( lem, m );

   if(b)
   {
      state. interreduce( );

      b = state. searchsolution( );

      if(b)
      {
         application appl = application::application( 
			       false_instance::false_instance( lem ), nr, m ); 

         appl. f. s = state. makegroundsubst( );
	 bool acc = appl. f. computematching(m);

         ASSERT( acc );
         return appl;
      }
   }
   return partial< application > :: partial( ); 
}





void application::test_swll_match( const model& m, const rulesystem& rs )
{
   ASSERT( rs. size( ) == 1 );
   rule r = rs. rules [0]. r;

   std::cout << m << "\n";
   std::cout << r << "\n";

   ASSERT( r. isdisjunctive( ));
   ASSERT( r. nr_conclusions( ) == 0 );

   swll::searchstate state;

   bool b = false;
   for( unsigned int i = 0; i < 1; ++ i )
      b = state. setproblem( r, m );

   if(b)
   {
      for( unsigned int i = 0; i < 1; ++ i )
         state. interreduce( );

      std::cout << state << "\n";

      for( unsigned int i = 0; i < 1; ++ i )
         b = state. searchsolution( );
      if(b)
      {
         std::cout << "found solution: ";
	 std::cout << state. makegroundsubst( ) << "\n"; 
      }
      else
         std::cout << "found no solution\n"; 
   }
   else
      std::cout << "failed in setproblem\n"; 

}


std::ostream& operator << ( std::ostream& stream,
		            const application& appl )
{
   stream << "application( ";
   stream << appl. f << ", ";
   stream << "rule number = " << appl. rule_nr << ", "; 
   stream << "branching degree = " << appl. branching_degree << " )";
   return stream;
}


