

// Written by Hans de Nivelle, this version is from June 2006. 

// An application is a false_instance decorated with
// some additional statistics.



#ifndef APPLICATION_INCLUDED
#define APPLICATION_INCLUDED   1


#include "false_instance.h"
#include "partial.h"
#include "rulesystem.h"
#include "model.h" 



struct application
{

   false_instance f;

   unsigned int rule_nr;
   unsigned int branching_degree;
      // For a disjunctive rule, the branching degree is
      // the number of conclusions.
      // For an existential rule, the branching degree is
      // the number of elements in the current model + 1.
   
   application( const false_instance& f,
		unsigned int rule_nr, 
		const model& m ) 
   : f(f),
     rule_nr( rule_nr )
   {
      if( f. r. isexistential( ))
         branching_degree = m. nrelements( ) + 1;
      else
         branching_degree = f. r. nr_conclusions( ); 
   }


   // Each of the following comparisons returns -1 if
   // this application is smaller, 0 if this application is equivalent to the 
   // other application, and 1 if the other application is smaller.

   int comparerulesnrpremisses( const application& other );
   int comparematchingsmultiset( const application& other ); 
   int comparematchingscardinality( const application& other );
   int comparematchingssupremum( const application& other ); 
   int comparerulenrs( const application& other );
   

   static bool conclusionistrue( const model& m, const rule& r,
		                 groundsubst& s );
      // true if s makes the conclusion of r true in model m.



   static partial< application > 
      bestforward( const rulesystem& rs, const model& m );
         // Finds the best possible application of a forward reasoning
	 // rule. 


   static partial< application > 
      findclosinglemma( const rulesystem& rs, const model& m, bool best );
         // Returns a closing lemma in case there exists one. 
         // In case best = true, findclosinglemma will try to look
	 // for the best closing lemma. This means: 
         // Smallest, or with smallest number 
	 // of disequalities, smallest number of variables, etc. 

   static partial< application > 
      makeclosurewith( const model& m, const rule& lem, unsigned int nr ); 
         // Construct a closing application from lem if this is possible.
         // The instantiation needs not be the best possible. 


private: 
   void improve( const model& m, partial< application > & best );
      // Tries to extend this matching in the best possible way,
      // and to improve best with it.
 
   void improve( const model& m, 
                 rule::prem_iterator p1,
		 rule::prem_iterator p2,
		 partial< application > & best ); 

public: 
   static void test_swll_match( const model& m, const rulesystem& rs );

};


std::ostream& operator << ( std::ostream& , const application& );


#endif

