
// Written by Hans de Nivelle, November 2005.


#ifndef CALCULUS_INCLUDED
#define CALCULUS_INCLUDED  1



#include <vector>
#include <map>


#include "variable.h"
#include "false_instance.h"
#include "rulesystem.h"


namespace calculus
{
 
   bool ext_res_possible( const false_instance& existential,
			  const false_instance& lemma,
			  unsigned int branch_border );
      // In the lemma, the rule and the 
      // matching prem_match must be filled in. 

   false_instance existential_resolvent( 
           std::vector< predicate > & predicates_base,
	   std::vector< variable > & arguments_base,
           const false_instance& existential,
           const false_instance& lemma,
	   const model& m, 
	   unsigned int branch_border );  

      // In the lemma and the rule prem_match must
      // be filled in. prem_match is used for determining which
      // premisses should resolve, and for computing the substitution. 
      // In the result, prem_match and the ground substitution
      // will be filled in.

   // We have only binary at this moment:

   false_instance disjunction_resolvent(
           std::vector< predicate > & predicates_base,
	   std::vector< variable > & arguments_base,
	   const false_instance& disjunction,
	   unsigned int alternative, 
	   const false_instance& lemma, 
	   const model& m, 
	   unsigned int branch_border );

      // The lemma and the disjunction should have
      // their prem_match filled in.
      // In the result, both prem_match and the ground substitution
      // will be filled in.


   false_instance sortlemma(
      std::vector< predicate > & predicates_base,
      std::vector< variable > & arguments_base,
      const false_instance& orig,
      const model& m,
      unsigned int lookahead );

      // sortlemma first deletes create atoms #X for which X occurs
      // in another non-disequality atom.
      // After that it sorts the lemma in such a way
      // that at each point the relevant variables are minimized.
      // This is done heuristically.
      // lookahead determines how far sortlemma looks ahead
      // when deciding how to sort the literals. 

   false_instance functional_reduction( 
      std::vector< predicate > & predicates_base,
      std::vector< variable > & arguments_base,
      const false_instance& orig,
      const model& m );

      // Functional reduction is the following rule:
      // If the lemma contains two atoms of form
      // f(X1,...,Xn, Y1,..., Ym) and f(X1,...,Xn,Z1,...,Zm),
      // and the functionality of f is m, then the atoms are unified.
      // This is repeated until no more unifications are possible.

   false_instance condense(
      std::vector< predicate > & predicates_base,
      std::vector< variable > & arguments_base,
      const false_instance& orig,
      const model& m );

      // Return smallest subset of orig that is subsumed by
      // orig. 


   
   bool decreasefunctionality( const rule& r,
		               varatom conc,
			       unsigned int func ); 
      
   bool decreasefunctionality( const rule& r,
		    std::map< predicate, unsigned int > & occurrences ); 

   void setfunctionalities( const rulesystem& rs );
      // This algorithm needs to be improved. 

   void testing( ); 
}


#endif


