
#include <fstream>


#include "firstorder/nnf.h"
#include "firstorder/simplifications.h"
#include "firstorder/replacement.h"
#include "firstorder/antiskolemization.h"
#include "firstorder/factor.h"
#include "firstorder/geometric.h"


#include "firstorder/testing.h"


#include "swll/searchstate.h"
#include "swll/lemmasystem.h"


#include "parser/parser.h"
#include "parser/token.h"
#include "parser/errors.h"


#include "subsumption.h"

#include "varmatching.h"

#include <iostream>

#include "varset.h"

#include "search.h" 
#include "false_instance.h"

#include "rulesystem.h"
#include "model.h"

#include "normalizer.h"


#include "varsubst.h"
#include "varatom.h" 
#include "element.h" 
#include "predicate.h"
#include "groundatom.h" 
#include "calculus.h"
#include "partial.h"
#include "application.h"


#include "predicate.stat"
#include "firstorder/function.stat"



int main( unsigned int nrargs, char* ( args [] ) )
{


#if 0
   variable v0;
   variable v1 = v0; v1 ++ ;
   variable v2 = v1; v2 ++ ;
   variable v3 = v2; v3 ++ ;
   variable v4 = v3; v4 ++ ;
   variable v5 = v4; v5 ++ ;
   variable v6 = v5; v6 ++ ;
   variable v7 = v6; v7 ++ ;
   variable v8 = v7; v8 ++ ;
   variable v9 = v8; v9 ++ ;
   variable v10 = v9; v10 ++ ; 
     
   element e0;
   element e1 = e0; e1 ++ ;
   element e2 = e1; e2 ++ ;
   element e3 = e2; e3 ++ ;
   element e4 = e3; e4 ++ ;
   element e5 = e4; e5 ++ ;
   element e6 = e5; e6 ++ ;
   element e7 = e6; e7 ++ ;
   element e8 = e7; e8 ++ ;
   element e9 = e8; e9 ++ ; 

   return 0;
#endif 

   std::cout << "hallo\n\n";


#if 1 

   if( nrargs != 2 )
   {
      std::cout << "ERROR: "; 
      std::cout << "wrong number of parameters\n\n"; 
      return 0; 
   }

   std::ifstream inputfile( args [1] ); 
   if( ! inputfile )
   {
      std::cout << "ERROR ";
      std::cout << "unable to open file " << args [1] << "\n\n";
      return 0;
   }

   parser::inputstream is = 
	      parser::inputstream::inputstream( inputfile,
			                        args [1] );
#else
   parser::inputstream is = parser::inputstream::inputstream( );
#endif


   model m;
   rulesystem rs( "input" );
   firstorder::sequent fol;


   try
   {
      readproblem( is, m, rs, fol );
   }
   catch( parser::syntaxerror s )
   {
      std::cout << "ERROR: "; 
      std::cout << s << "\n\n";
      return 0; 
   }

   fol. collectusedidentifiers( );

   std::cout << m << "\n";
   std::cout << rs << "\n"; 
   firstorder::simplifications::flatten( fol ); 

   if( fol. formulas. size( ))
   {

      firstorder::simplifications::flatten( fol );
      std::cout << "-----------------------------------------------------\n";
      std::cout << "Input:\n\n";
      std::cout << fol << "\n";


      firstorder::nnf::convert_to_nnf( fol );
      firstorder::simplifications::flatten( fol ); 
      std::cout << "-----------------------------------------------------\n"; 
      std::cout << "NNF:\n\n";
      std::cout << fol << "\n";


      firstorder::simplifications::fully_simplify( fol );
      std::cout << "-----------------------------------------------------\n";
      std::cout << "Full simplification:\n\n";
      std::cout << fol << "\n";


      firstorder::antiskolemization::antiskolemize( fol );
      firstorder::simplifications::flatten( fol );
      std::cout << "-----------------------------------------------------\n";
      std::cout << "Anti-Skolemization:\n\n";
      std::cout << fol << "\n"; 


      firstorder::factor::convert_to_cuden( fol );
      firstorder::simplifications::flatten( fol );
      std::cout << "-----------------------------------------------------\n";
      std::cout << "Factored normal form (CUDEN):\n\n";
      std::cout << fol << "\n";


      firstorder::geometric::addtherules( fol, rs );

      std::cout << "-----------------------------------------------------\n";
      std::cout << "Resulting rule system:\n\n";
      std::cout << rs << "\n";

   }

   calculus::setfunctionalities( rs );
   std::cout << "\n";
   std::cout << "predicates / arities , functionalities\n\n";

   predicate::printtable( std::cout );

#if 0
   application::test_swll_match( m, rs );
   return 0;
#endif
#if 0
   subsumption::test_swll_subsumption( rs );
   return 0;
#endif

   // CASC seems to dislike empty models. For this
   // reason, we start the search with a one element model.

   if( m. size( ) != 0 )
   {
      std::cout << "ERROR: expected model to be empty at this point\n\n";
      exit(0);
   }

   {
      element e0;
      std::vector< element > args;
      args. push_back( e0 );

      std::vector< unsigned int > parents;

      m. addatom( groundatom::groundatom( predicate::create, args ),
                  9999999, parents );
   }
	
   partial< application > res = search::extendstomodel( m, rs );

   if( m. size( ) < 1 )
   {
      std::cout << "ERROR: interpretations must be nonempty\n";
      exit(0);
   }

   if( res. isdefined( )) 
   {
      std::cout << "\n"; 
      std::cout << "START-OF-PROOF\n"; 
      std::cout << rs << "\n\n";
      std::cout << "END-OF-PROOF\n\n"; 
      std::cout << "could not extend to model in ";
      std::cout << m. closure_counter << " attempts\n\n";
   }

   if( res. isundefined( ))
   {
      std::cout << "\n"; 
      std::cout << "START-OF-MODEL\n\n";
      std::cout << m << "\n";
      std::cout << "END-OF-MODEL\n\n"; 
   }

   return 0; 
}




