
// Code written by Hans de Nivelle, November 2005. 


#include "model.h"



bool model::addatom( groundatom at, 
		     unsigned int rule_nr, 
		     const std::vector< unsigned int > & ps,
		     unsigned int alternative,
		     unsigned int nr_alternatives )
{

   for( atom_iterator p = atoms_begin( );
        p != atoms_end( );
	++ p )
   {
      if( getatom(p) == at )
         return false;
   }

   atoms. push_back( 
      atom::atom( at. p, 
	          arguments. size( ), arguments. size( ) + at.p.getarity( ),
		  rule_nr, 
		  parents. size( ), parents. size( ) + ps. size( ),
                  closure_counter,
		  alternative, nr_alternatives ));

   for( std::vector< element > :: const_iterator a = at. a1;
        a != at. a2;
	++ a )
   {
      arguments. push_back( *a );
   }

   for( std::vector< unsigned int > :: const_iterator p = ps. begin( );
        p != ps. end( );
	++ p )
   {
      parents. push_back(*p);
      ASSERT( *p < atoms. size( ) - 1 ); 
         // Parents should come before their children. 
   }

   return true; 
}


bool model::canbeadded( groundatom at ) const
{
   if( at. p != predicate::create )
   {
      for( std::vector< element > :: const_iterator 
	      a = at. a1;
	      a != at. a2;
	      ++ a )
      {
         for( atom_iterator 
	         i = atoms_begin( );
                 i != atoms_end( );
	         ++ i )
         {
            if( getatom(i). p == predicate::create &&
                *( getatom(i). a1 ) == *a )
            {
               goto found;
            }
         }
         return false;
      found:
         ;
      }
   }
   return true;
}


void model::contradiction( const std::vector< unsigned int > & contr )
{
   for( std::vector< unsigned int > :: const_iterator i = contr. begin( );
        i != contr. end( );
        ++ i )
   {
      ASSERT( *i < atoms. size( ) ); 
      atoms [*i]. usedinproof = true;
   }

   ++ closure_counter; 
}


bool model::backtrack( unsigned int s0 )
{
   ASSERT( atoms. size( ) >= s0 );

   if( s0 == atoms. size( ))
      return false;

   bool one_is_used = false;

   unsigned int s = atoms. size( );
   do 
   {
      -- s;

      if( atoms [s]. usedinproof ) 
      {
         one_is_used = true; 
         for( unsigned int i = atoms [s]. pars1;
              i != atoms [s]. pars2;
	      ++ i )
         {
            atoms [ parents [i] ]. usedinproof = true;
         }
      }
   }
   while( s != s0 );

   arguments. erase( arguments. begin( ) + atoms [ s0 ]. args1, 
		     arguments. end( ) ); 
   parents. erase( parents. begin( ) + atoms [ s0 ]. pars1, 
		   parents. end( ) ); 

   atoms. erase( atoms. begin( ) + s0, atoms. end( ) );

   return one_is_used; 
}


unsigned int model::nrelements( ) const
{
   unsigned nr = 0;

   for( model::atom_iterator a = atoms_begin( );
        a != atoms_end( );
	++ a )
   {
      if( getatom(a). p == predicate::create ) 
         ++ nr;
   }
   return nr;
}


std::ostream& operator << ( std::ostream& stream, const model& m )
{
   stream << "Interpretation:\n\n";

   unsigned int atomcntr = 0;

   for( model::atom_iterator a = m. atoms_begin( );
        a != m. atoms_end( );
	++ a )
   {
      stream << atomcntr << " : ";
      stream << m. getrule_nr(a); 
      stream << "(";

      for( model::parent_iterator p = m. parents_begin(a);
           p != m. parents_end(a);
	   ++ p )
      {
         if( p != m. parents_begin(a) )
            stream << ","; 
         stream << *p;
      }
      stream << ") ";
      stream << "(" << m. getalternative(a) << "/";
      stream << m. getnralternatives(a) << ")    "; 

      stream << "[" << m. gettimestamp(a) << "]   ";

      if( m. isused(a) )
         stream << "* ";

      ++ atomcntr;

      stream << m. getatom(a) << "\n"; 
   }
   stream << "\n"; 
   stream << "The model contains " << m. nrelements( ) << " elements\n"; 
   stream << "It has been closed " << m. closure_counter << " times\n";
   return stream; 
}


