

// Code written by Hans de Nivelle, April 2006.


#ifndef PREDICATE_INCLUDED
#define PREDICATE_INCLUDED   1


#include <vector> 
#include <map>
#include <string>
#include <list>
#include <iostream>


#include "assert.h"


class predicate
{

   static std::map < std::string, std::list< unsigned int > > lookuptable;
      // Assigns to each string the list of its representations. 
      // The distinct representations correspond to distinct arities.

   struct prop 
   { 
      std::string stringrepr;
      unsigned int arity;
      unsigned int functionality;
         // It should be the case that functionality <= arity.
	 // It specifies how many of the arguments are functional
	 // in the other arguments. The functional arguments
	 // are always at the end.

      prop( const std::string& stringrepr,
            unsigned int arity,
	    unsigned int functionality )
         : stringrepr( stringrepr ),
	   arity( arity ),
	   functionality( functionality )
      {
      }

   };  

   static std::vector< prop > properties;

   unsigned int repr; 

   predicate( );
      // Has no definition. There is no default predicate. 

public: 
   predicate( const std::string& stringrepr, unsigned int arity );
       
   static bool isdefined( const std::string& stringrepr, unsigned int arity );
      // True if the string is known with the given arity.

   const std::string& getstring( ) const;
   unsigned int getfunctionality( ) const; 
   unsigned int getarity( ) const; 

   void setfunctionality( unsigned int f );
      // Functionality = f means that the last f arguments
      // are functional in the first arity - f arguments.

   friend std::ostream& operator << ( std::ostream& ,
		                      const predicate p ); 

   static void printtable( std::ostream& stream );
      // Prints the table of all predicates. 

   friend bool operator == ( predicate, predicate );
   friend bool operator != ( predicate, predicate );
   friend bool operator < ( predicate, predicate ); 

   static predicate diseq; 
      // Predicate for disequality. 
   static predicate create;
      // Predicate for creating a new object in model. 


   friend std::ostream& operator << ( std::ostream& , const prop& ); 

};


inline const std::string& predicate::getstring( ) const
{
   return properties [ repr ]. stringrepr; 
}

inline unsigned int predicate::getfunctionality( ) const
{
   return properties [ repr ]. functionality;
}

inline unsigned int predicate::getarity( ) const
{
   return properties [ repr ]. arity; 
}

inline void predicate::setfunctionality( unsigned int f ) 
{
   ASSERT( f <= getarity( ));
   properties [ repr ]. functionality = f;
}


inline std::ostream& operator << ( std::ostream& stream,
		                   const predicate p )
{
   stream << p. getstring( );
   return stream;
}


inline bool operator == ( predicate p1, predicate p2 )
{
   return p1. repr == p2. repr;
}

inline bool operator != ( predicate p1, predicate p2 )
{
   return p1. repr != p2. repr;
}

inline bool operator < ( predicate p1, predicate p2 )
{
   return p1. repr < p2. repr;
}


#endif



