

// Code written by Hans de Nivelle, June 2006. 
// A rulestack is simply a stack of integers, where
// the integers are supposed to represent rules. 
// 
// Its intended use if for storing the essential rules during
// the searchprocess. 


#ifndef RULESTACK_INCLUDED
#define RULESTACK_INCLUDED    1



#include <iostream>
#include <vector>

#include "assert.h"


class rulestack
{


private: 
   std::vector< unsigned int > repr; 

public:

   void clear( );

   void reserve( unsigned int x );
      // Reserve space for x rules. Note that existing 
      // assignments are counted.  

   unsigned int size( ) const;
      // Returns number of rules in the rulestack. 
      // The result can be used for backtracking. 

   void backtrack( unsigned int x ); 
    
   void append( unsigned int r ); 
      // Append the rule r at the end. 

   bool contains( unsigned int r ) const; 
      // True if the rulestack contains r.

   friend std::ostream& operator << ( std::ostream&, const rulestack& );

};


inline void rulestack::clear( )
{
   repr. clear( ); 
}
   

inline void rulestack::reserve( unsigned int x )
{
   repr. reserve( x ); 
}   
    

inline unsigned int rulestack::size( ) const
{
   return repr. size( ); 
}


inline void rulestack::backtrack( unsigned int x ) 
{
   ASSERT( x <= repr. size( )); 

   repr. erase( repr. begin( ) + x, repr. end( )); 
}


inline void rulestack::append( unsigned int x )
{
   repr. push_back(x);
}


#endif


