

// Code written by Hans de Nivelle, November 2005.


#include "rulesystem.h"
#include "subsumption.h"



void rulesystem::addrule( rule r, type_of_derivation derivedwith,
		          const std::vector< unsigned int > & pars ) 
{
   unsigned int pars1 = parents. size( );
   for( std::vector< unsigned int > :: const_iterator p = pars. begin( );
        p != pars. end( );
	++ p )
   {
      parents. push_back( *p );
   }
   unsigned int pars2 = parents. size( );

   rule r_copy = r. clone_into( predicates_base, arguments_base );

   rules. push_back( rule_in_system::rule_in_system(
             r_copy,
	     derivedwith, 
	     pars1, pars2 ));

}


void rulesystem::deletesubsumed( const rule& r ) 
{
   for( unsigned int i = 0; i < size( ); ++ i ) 
   {
      if( ! rules [i]. deleted && 
	  subsumption::findsubsumption( r, rules [i]. r ). isdefined( ))
      {
         std::cout << "Deleting " << i << ":   "; 
	 std::cout << rules [i]. r << "\n"; 
         rules [i]. deleted = true;
      }
   }
}


std::ostream& operator << ( std::ostream& stream,
		            const rulesystem& rs )
{
   stream << "Rulesystem \"" << rs. name << "\":\n\n"; 

   unsigned int cntr = 0; 
   unsigned int active_cntr = 0;

   for( std::vector< rulesystem::rule_in_system > :: const_iterator 
	   r = rs. rules. begin( );
           r != rs. rules. end( );
	   ++ r ) 
   {
      {
         stream << cntr << ":   "; 
         switch( r -> derivedwith )
         {
         case rulesystem::rule_initial:
            stream << "I"; 
	    break; 
         case rulesystem::rule_factor:
            stream << "F"; 
	    break; 
         case rulesystem::rule_disjunction:
            stream << "D"; 
	    break; 
         case rulesystem::rule_existential:
	    stream << "E"; 
	    break; 
         default:
            ASSERT( false );
         }

         if( rs. parents_begin(r) != rs. parents_end(r) )
         {
            stream << "(";
	    for( rulesystem::parent_iterator p = rs. parents_begin(r);
                 p != rs. parents_end(r);
	         ++ p )
            {
               if( p != rs. parents_begin(r) )
                  stream << ",";
	       stream << *p; 
	    }
            stream << ")";
         }
         stream << "   "; 

         stream << "closed(" << ( r -> nr_closures ) << ")    ";

         if( ! r -> deleted )
         {
            stream << " (ACTIVE)   ";
            ++ active_cntr ;
         }

	 stream << "\n"; 
         stream << ( r -> r ) << "\n"; 
      }
      cntr ++ ; 
   }

   std::cout << "\n"; 
   std::cout << "There are " << active_cntr << " active rules\n"; 

   return stream;
}

