
// Code written by Hans de Nivelle, November 2005. 


#ifndef RULESYSTEM_INCLUDED
#define RULESYSTEM_INCLUDED   1


#include <vector>
#include <iostream>


#include "variable.h"
#include "varatom.h"
#include "rule.h"


class rulesystem
{

   std::vector< predicate > predicates_base; 
   std::vector< variable > arguments_base; 
   std::vector< unsigned int > parents; 

public: 
   const std::string name;

   enum type_of_derivation 
   {
      rule_initial,
      rule_factor,
      rule_disjunction,
      rule_existential 
   };


   struct rule_in_system 
   {
      rule r;

      type_of_derivation derivedwith; 
      unsigned int pars1; 
      unsigned int pars2; 

      mutable unsigned int nr_closures;
         // How often the rule was used in closing a branch.

      bool deleted;
         // True if rule is deleted. This allows for
	 // subsumption check. 

      mutable bool usedinproof;
         // True if rule was used in the proof.

      rule_in_system( const rule& r, 
		      type_of_derivation derivedwith,
		      unsigned int pars1, unsigned int pars2 ) 
         : r(r), 
	   derivedwith( derivedwith ),
	   pars1( pars1 ), pars2( pars2 ),
	   nr_closures(0),
	   deleted(false),
	   usedinproof(false) 
      {
      }

   };


   std::vector< rule_in_system > rules;

public: 
   rulesystem( const std::string& name ) 
      : name(name) { }


   void addrule( rule r, type_of_derivation derivedwith, 
		 const std::vector< unsigned int > & pars );
                    // Add a rule to the system. 

   unsigned int size( ) const
   {
      return rules. size( );
   }


   void deletesubsumed( const rule& r );
      // Mark rules that are subsumed by r as deleted. 


   // A parent_iterator iterates through the parents of a rule. 

   struct parent_iterator
   {
      std::vector< unsigned int > :: const_iterator p;

      explicit parent_iterator( 
		      std::vector< unsigned int > :: const_iterator p ) 
         : p(p)
      {
      }


      parent_iterator operator ++ ( int ) 
      {
         return parent_iterator::parent_iterator( p ++  );
      }

      parent_iterator operator ++ ( ) 
      {
         return parent_iterator::parent_iterator( ++ p );
      }

      bool operator == ( parent_iterator other )
      {
         return p == other. p;
      }

      bool operator != ( parent_iterator other )
      {
         return p != other. p;
      }

      unsigned int operator * ( ) const
      {
         return *p;
      }
   }; 

   parent_iterator parents_begin( 
	  std::vector< rule_in_system > :: const_iterator r ) const
   {
      return parent_iterator::parent_iterator( 
		  parents. begin( ) + ( r -> pars1 ));
   }

   parent_iterator parents_end( 
	  std::vector< rule_in_system > :: const_iterator r ) const
   {
      return parent_iterator::parent_iterator( 
		  parents. begin( ) + ( r -> pars2 ));
   }

   friend std::ostream& operator << ( std::ostream& , const rulesystem& ); 
};


std::ostream& operator << ( std::ostream& , const rulesystem & ); 

#endif


