

// Code written by Hans de Nivelle, February 2006. 



#include "subsumption.h"
#include "swll/searchstate.h"



partial< varmatching >
subsumption::findselfsubsumption( const rule& self_rule )
{
   partial< varmatching > res;
   return res;

   if( self_rule. isdisjunctive( ) &&
       self_rule. nr_conclusions( ) == 0 )
   {
      varset rulevars = self_rule. getvariables( self_rule. prems_begin( ),
		                                 self_rule. prems_end( ));

      while( rulevars. size( ) != 0 )
      {
         variable v = rulevars. select( );
         rulevars. remove( v );

         static swll::statistics_collector sc;

	 swll::searchstate state;
	 bool b = state. setproblem( self_rule, self_rule, v );
	 sc. nr_calls ++ ;
         
	 if(b)
         {
            sc. nr_passedsetproblem ++ ;
	    state. interreduce( );
	    b = state. searchsolution( );

	    if(b) 
               sc. nr_solution ++ ;

            if( state. update_stats( sc ))
            {
               std::cout << "Condensation Statistics:\n";
	       std::cout << sc << "\n";
	       std::cout << "self_rule: " << self_rule << "\n"; 
               std::cout << "===================\n\n";
            }

	    if(b)
            {
               res = state. makevarmatching( );
	       ASSERT( check_subsumption( res. contents( ), 
				          self_rule, self_rule ));

	       return res; 
	    }
         }
      }
   }
   return res; 
}


partial< varmatching >
subsumption::findsubsumption( const rule& from_rule, const rule& into_rule )
{
   partial< varmatching > res;
   return res; 

#if 0 
   std::cout << "\n"; 
   std::cout << "=================================\n"; 
   std::cout << "Entering subsumes\n";
   std::cout << "From: " << from_rule << "\n";
   std::cout << "Into: " << into_rule << "\n";
   std::cout << "Base: " << base << "\n"; 
   std::cout << "\n"; 
#endif


   if( from_rule. isdisjunctive( ) &&
       from_rule. nr_conclusions( ) == 0 )
   {
      static swll::statistics_collector sc;

      swll::searchstate state;

      bool b = state. setproblem( from_rule, into_rule,
		                  partial< variable > :: partial( ));
      sc. nr_calls ++ ;

      if(b)
      {
         sc. nr_passedsetproblem ++ ; 
         state. interreduce( );
	 b = state. searchsolution( );
     
	 if(b)
            sc. nr_solution ++ ;

         if( state. update_stats( sc ))
         {
            std::cout << "Subsumption Statistics:\n";
            std::cout << sc << "\n\n";
	    std::cout << "from: " << from_rule << "\n";
	    std::cout << "into: " << into_rule << "\n";
	    std::cout << "====================\n\n"; 
         }

	 if(b) 
         {
	    res = state. makevarmatching( ); 
	    ASSERT( check_subsumption( res.contents( ), 
				       from_rule, into_rule )); 
         }
      }
   }
   return res; 
}



bool subsumption::check_subsumption( const varmatching& subst, 
	                             const rule& from_rule,
	                             const rule& into_rule )
{
   if( ! from_rule. isdisjunctive( ))
      return false;
   if( ! from_rule. nr_conclusions( ) == 0 )
      return false;

   for( rule::prem_iterator p = from_rule. prems_begin( );
                            p != from_rule. prems_end( );
			    ++ p )
   {
      varatom at = subst. apply_on( *p );

#if 0 
      std::cout << "( " << *p << ", " << at << " ) " << "\n";
#endif

      if( at. p == predicate::create )
      {
         variable v = * ( at. a1 );

	 // We only demand that v is a variable of into_rule:

	 for( rule::prem_iterator q = into_rule. prems_begin( );
			          q != into_rule. prems_end( );
				  ++ q )
         {
            for( std::vector< variable > :: const_iterator b = (*q). a1;
                    b != (*q). a2;
		    ++ b )
             {
                if( *b == v ) 
                   goto found;
             }
         }
	 return false; 
      }

      if( at. p == predicate::diseq )
      { 

         // Disequalities can match in two orientations. 

         std::vector< variable > :: const_iterator a = at. a1;
	 variable v1 = * ( a ++ );
	 variable v2 = * a;

         for( rule::prem_iterator q = into_rule. prems_begin( );
                                  q != into_rule. prems_end( );
				  ++ q )
         {
            if( (*q). p == predicate::diseq )
            {
               std::vector< variable > :: const_iterator b = (*q). a1;
	       variable z1 = * ( b ++ );
	       variable z2 = * b;

	       if( v1 == z1 && v2 == z2 ) goto found;
	       if( v1 == z2 && v2 == z1 ) goto found;

	    }
         }
	 return false; 
      }

      if( at. p != predicate::create && at. p != predicate::diseq ) 
      {
         for( rule::prem_iterator q = into_rule. prems_begin( );
                                  q != into_rule. prems_end( );
			          ++ q )
         {
            if( *q == at )
               goto found;
         }
         return false;
      }

   found:
      ;
   }
   return true;
}




void subsumption::test_swll_subsumption( const rulesystem& rs )
{
   std::cout << "Testing subsumption: \n";

   ASSERT( rs. size( ) == 2 );
   
   rule from = rs. rules [0]. r;
   rule into = rs. rules [1]. r;

   std::cout << "from: " << from << "\n";
   std::cout << "into: " << into << "\n";

   ASSERT( from. isdisjunctive( ));
   ASSERT( from. nr_conclusions( ) == 0 );

   swll::searchstate state;

   varmatching vm;

   bool b = false;
   for( unsigned int i = 0; i < 1000; ++ i )
      b = state. setproblem( from, into, partial< variable > :: partial( ));

   if(b)
   {
      for( unsigned int i = 0; i < 1; ++ i )
         state. interreduce( );

      std::cout << state << "\n";

      for( unsigned int i = 0; i < 1; ++ i )
         b = state. searchsolution( );

      if(b)
      {
	 varmatching res = state. makevarmatching( ); 
         std::cout << "found solution: ";
	 std::cout << res << "\n";
	 ASSERT( check_subsumption( res, from, into ));
      }
      else
         std::cout << "found no solution\n";
   }
   else
      std::cout << "failed in setproblem\n";
}

  



