<?xml version="1.0"?>
<!-- An XSL style sheet for creating MATHsAiD prefix syntax from OMDoc
     Initial version 20051004 by Juergen Zimmer, 
     send bug-reports, patches, suggestions to jzimmer@ags.uni-sb.de

     Copyright (c) 2005 Juergen Zimmer, 

     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
     License as published by the Free Software Foundation; either
     version 2.1 of the License, or (at your option) any later version.

     This library is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     Lesser General Public License for more details.

     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
-->

<!-- java org.apache.xalan.xslt.Process -TT -TS -IN FOF.xml -XSL xtstp2post.xsl -OUT out.post -->
<!-- java org.apache.xalan.xslt.Process -IN FOF.xml -XSL xtstp2post.xsl  -OUT out -->

<xsl:stylesheet xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
                xmlns:om="http://www.openmath.org/OpenMath"
                xmlns:dc="http://purl.org/dc/elements/1.1/"
                xmlns:xtstp="http://xff.sf.net/xff/ext/tstp/xff-v0.3-ext-tstp.xsd"
		xmlns:omdoc="http://www.mathweb.org/omdoc"
        	xmlns="http://www.mathweb.org/omdoc"
                version="1.0">

<xsl:output method="xml" indent="yes" version="1.0" standalone="yes"/>

<!-- the default behavior is that everything (for instance OMOBJects)
     are copied verbatim. This is overridden later by the other rules
  -->

<!--   <xsl:template match="*"> -->
<!--     <xsl:copy-of select="."/> -->
<!--   </xsl:template> -->

  <!-- xsl:import href="copyover.xsl"/ -->

  <!-- list processer -->
  <xsl:template name="clist">
    <xsl:param name="separator"/>
    <xsl:param name="sublist"/>
    <xsl:for-each select="$sublist">
      <xsl:value-of select="$separator" />
      <xsl:apply-templates select="." />
    </xsl:for-each>
  </xsl:template>

  <!-- list processer -->
  <xsl:template name="list">
    <xsl:param name="separator"/>
    <xsl:param name="sublist"/>

    <xsl:apply-templates select="$sublist[1]" />
    <xsl:call-template name="clist">
      <xsl:with-param name="separator" select="$separator"/>
      <xsl:with-param name="sublist" select="$sublist[position()>1]"/>
    </xsl:call-template>
  </xsl:template>

  <xsl:key name="funcKey" match="xtstp:func" use="@name"/>
  <xsl:key name="predKey" match="xtstp:pred" use="@name"/>

 
  <xsl:template match="xtstp:xff">
    <omdoc xml:id="XTSTP-auto-generated-omdoc">
<!--     <metadata> -->
<!--       <Title>XTSTP-auto-omdoc </Title> -->
<!--       <Creator role="trl"><xsl:text>The xtstp2omdoc style sheet</xsl:text></Creator> -->
<!--     </metadata> -->
    <xsl:text>&#xA;</xsl:text>
    <theory xml:id="local-theory">
      <xsl:for-each select=".//xtstp:func[count( . | key('funcKey', @name)[1] ) = 1]">
        <xsl:variable name="n" select="@name"/>
	<symbol name="{$n}" kind="object" scope="global">
          <type system="simpletypes"> 
            <om:OMOBJ>
              <xsl:choose>
               <xsl:when test="*">
               <om:OMA>
                <om:OMS cd="simpletypes" name="funtype"/>
                <om:OMS cd="ind" name="ind"/>
                <om:OMS cd="ind" name="ind"/>
               </om:OMA> 
             </xsl:when>
             <xsl:otherwise>
               <om:OMS cd="ind" name="ind"/>
             </xsl:otherwise>
             </xsl:choose>
            </om:OMOBJ>
          </type>
        </symbol>
      </xsl:for-each>

      <xsl:for-each select=".//xtstp:pred[count( . | key('predKey', @name)[1] ) = 1]">
        <xsl:variable name="n" select="@name"/>
 	<symbol name="{$n}" kind="object" scope="global">
          <type system="simpletypes"> 
            <om:OMOBJ>
              <xsl:choose>
               <xsl:when test="*">
              <om:OMA>
                <om:OMS cd="simpletypes" name="funtype"/>
                <om:OMS cd="truthval" name="bool"/>
                <om:OMS cd="truthval" name="bool"/>
              </om:OMA> 
             </xsl:when>
             <xsl:otherwise>
               <om:OMS cd="truthval" name="bool"/>
             </xsl:otherwise> 
             </xsl:choose>
           </om:OMOBJ>
          </type>
        </symbol>
      </xsl:for-each>
      <xsl:apply-templates/>
    </theory>
  </omdoc>
  </xsl:template>


  <xsl:template match="xtstp:include">
    <CMP>
    <xsl:text><!-- include skipped --></xsl:text>
    <xsl:text>&#xa;</xsl:text>
    </CMP>
  </xsl:template>
  
  <xsl:template match="xtstp:formula">
      <xsl:variable name="n" select="@name"/>        
      <xsl:choose>
        <xsl:when test="@type='axiom'"> 
	   <axiom xml:id="{$n}">
             <FMP>
	       <om:OMOBJ>
                 <xsl:apply-templates/>
	       </om:OMOBJ>
             </FMP>
	   </axiom>
        </xsl:when>
        <xsl:when test="@type='conjecture'"> 
	   <assertion xml:id="{$n}" type="conjecture">
             <FMP>
	       <om:OMOBJ>
                 <xsl:apply-templates/>
	       </om:OMOBJ>
             </FMP>
	   </assertion>
        </xsl:when>
      </xsl:choose>
  </xsl:template>

  <xsl:template name="lambda">
    <xsl:param name="quantor"/>
    <xsl:param name="varlist"/>
    <xsl:param name="body"/>
    <om:OMBIND>
      <om:OMS cd="pl1" name="{$quantor}"/>
      <om:OMBVAR> 
        <xsl:for-each select="$varlist">
           <xsl:variable name="n" select="string(@name)" />
           <om:OMV name="{$n}"/>
        </xsl:for-each>
      </om:OMBVAR>
      <xsl:apply-templates select="$body" />
    </om:OMBIND>
  </xsl:template>

  <xsl:template match="xtstp:quantify">
        <xsl:call-template name="lambda">
          <xsl:with-param name="quantor" select="@type"/>
          <xsl:with-param name="varlist" select="xtstp:qvar"/>
          <xsl:with-param name="body" select="*[node() != 'xtstp:var']"/>
        </xsl:call-template>
  </xsl:template>

  <!-- processess connective subtree
       $symbol  paramter specifies the connective symbol (it is converted
                to a TSTP symbol
       $sublist gives a list of subtrees to process; for each subtree, TSTP
                symbol glyph is emited and then the subtree is processed
		recursively
  -->
  <xsl:template name="connector">
    <xsl:param name="symbol"/>
      <!-- OPTIMIZE: is there a way to store the result for repetive use? -->
      <xsl:choose>
        <xsl:when test="$symbol='not'">
	  <om:OMS cd="pl0" name="not"/>
	</xsl:when>
        <xsl:when test="$symbol='or'">  
	  <om:OMS cd="pl0" name="or"/>
	</xsl:when>
        <xsl:when test="$symbol='and'">
	  <om:OMS cd="pl0" name="and"/>
<!-- 	          <xsl:text disable-output-escaping="yes">and </xsl:text> -->
        </xsl:when>
	<xsl:when test="$symbol='impl'">
	  <om:OMS cd="pl0" name="implies"/>
	</xsl:when>
        <xsl:when test="$symbol='equiv'">
	  <om:OMS cd="pl0" name="iff"/>
        </xsl:when>
        <xsl:otherwise>
          <xsl:message terminate="yes">
            Unknown connective '<xsl:value-of select="$symbol" />'
          </xsl:message>
        </xsl:otherwise>
      </xsl:choose>
  </xsl:template>

  <!-- processess connective subtree
       $symbol  paramter specifies the connective symbol (it is converted
                to an Openmath symbol
       $sublist gives a list of subtrees to process; for each subtree, 
                symbol glyph is emited and then the subtree is processed
		recursively
  -->

  <xsl:template match="xtstp:pred|xtstp:func">
    <xsl:variable name="n" select="string(@name)" />
    <xsl:if test="count(*)>0">
      <om:OMA>
        <om:OMS cd="local-theory" name="{$n}"/>
        <xsl:call-template name="list">
          <xsl:with-param name="separator" select="' '"/>
          <xsl:with-param name="sublist" select="*"/>
        </xsl:call-template>
      </om:OMA>
    </xsl:if>
    <xsl:if test="count(*)=0">
      <om:OMS cd="foobar" name="{$n}"/>
    </xsl:if>
  </xsl:template>

  <xsl:template match="xtstp:equal">
    <om:OMA>
      <om:OMS cd="relation1" name="eq"/>
      <xsl:apply-templates select="*[1]" />
      <xsl:apply-templates select="*[2]" />
    </om:OMA>
  </xsl:template>
  
 <xsl:template match="xtstp:not">
    <om:OMA>
      <om:OMS cd="pl0" name="not"/>
      <xsl:apply-templates select="*[1]" />
    </om:OMA>
 </xsl:template>

 <xsl:template match="xtstp:and">
    <om:OMA>
      <om:OMS cd="pl0" name="and"/>
      <xsl:apply-templates select="*[1]" />
      <xsl:apply-templates select="*[2]" />
    </om:OMA>
 </xsl:template>

 <xsl:template match="xtstp:or">
    <om:OMA>
      <om:OMS cd="pl0" name="or"/>
      <xsl:apply-templates select="*[1]" />
      <xsl:apply-templates select="*[2]" />
    </om:OMA>
 </xsl:template>

 <xsl:template match="xtstp:impl">
    <om:OMA>
      <om:OMS cd="pl0" name="implies"/>
      <xsl:apply-templates select="*[1]" />
      <xsl:apply-templates select="*[2]" />
    </om:OMA>
 </xsl:template>

 <xsl:template match="xtstp:equiv">
    <om:OMA>
      <om:OMS cd="pl0" name="equivalent"/>
      <xsl:apply-templates select="*[1]" />
      <xsl:apply-templates select="*[2]" />
    </om:OMA>
 </xsl:template>

  <xsl:template match="xtstp:true">
    <om:OMS cd="pl0" name="true"/>
  </xsl:template>

  <xsl:template match="xtstp:false">
    <om:OMS cd="pl0" name="false"/> 
  </xsl:template>

  <xsl:template match="xtstp:var">
    <xsl:variable name="n" select="string(@name)" />
    <om:OMV name="{$n}"/>
  </xsl:template>
  
  <xsl:template match="xtstp:number">
    <om:OMI><xsl:value-of select="." /></om:OMI>
  </xsl:template>

  <xsl:template match="xtstp:qstring">
    <CMP>
      <xsl:value-of select="."/>
    </CMP>
  </xsl:template>
  

  <xsl:template match="xtstp:comment">
  </xsl:template>


  <xsl:template match="xtstp:head">
<!--     <xsl:text>%</xsl:text> -->
<!--     <xsl:if test="@field!=''"> -->
<!--       <xsl:text> </xsl:text> -->
<!--       <xsl:value-of select="@field"/> -->
<!--       <xsl:value-of select="substring('            ',string-length(@field))"/> -->
<!--       <xsl:text> : </xsl:text> -->
<!--     </xsl:if> -->
<!--     <xsl:call-template name="replace-string"> -->
<!--      <xsl:with-param name="text" select="text()"/> -->
<!--      <xsl:with-param name="replace" select="'&#xa;'"/> -->
<!--      <xsl:with-param name="with" select="'&#xa;%                 '"/> -->
<!--     </xsl:call-template> -->
<!--     <xsl:text>&#xa;</xsl:text> -->
<!--     <xsl:if test="following-sibling::comment"> -->
<!--       <xsl:text>% Comment         </xsl:text> -->
<!--     </xsl:if> -->
  </xsl:template>

</xsl:stylesheet>
