<?xml version="1.0"?>
<!-- An XSL style sheet for creating MATHsAiD prefix syntax from OMDoc
     Initial version 20051004 by Juergen Zimmer, 
     send bug-reports, patches, suggestions to jzimmer@ags.uni-sb.de

     Copyright (c) 2005 Juergen Zimmer, 

     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
     License as published by the Free Software Foundation; either
     version 2.1 of the License, or (at your option) any later version.

     This library is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     Lesser General Public License for more details.

     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
-->

<!-- java org.apache.xalan.xslt.Process -TT -TS -IN FOF.xml -XSL xtstp2post.xsl -OUT out.post -->
<!-- java org.apache.xalan.xslt.Process -IN FOF.xml -XSL xtstp2post.xsl  -OUT out -->

<xsl:stylesheet xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
                xmlns:om="http://www.openmath.org/OpenMath"
                xmlns:dc="http://purl.org/dc/elements/1.1/"
                xmlns:xtstp="http://xff.sf.net/xff/ext/tstp/xff-v0.3-ext-tstp.xsd"
		xmlns="http://www.mathweb.org/omdoc"
                version="1.0">

<xsl:output method="xml" indent="yes" version="1.0" standalone="yes"/>

<!-- the default behavior is that everything (for instance OMOBJects)
     are copied verbatim. This is overridden later by the other rules
  -->

<!--   <xsl:template match="*"> -->
<!--     <xsl:copy-of select="."/> -->
<!--   </xsl:template> -->

  <!-- xsl:import href="copyover.xsl"/ -->

  <!-- list processer -->
  <xsl:template name="clist">
    <xsl:param name="separator"/>
    <xsl:param name="sublist"/>
    <xsl:for-each select="$sublist">
      <xsl:value-of select="$separator" />
      <xsl:apply-templates select="." />
    </xsl:for-each>
  </xsl:template>

  <!-- list processer -->
  <xsl:template name="list">
    <xsl:param name="separator"/>
    <xsl:param name="sublist"/>

    <xsl:apply-templates select="$sublist[1]" />
    <xsl:call-template name="clist">
      <xsl:with-param name="separator" select="$separator"/>
      <xsl:with-param name="sublist" select="$sublist[position()>1]"/>
    </xsl:call-template>
  </xsl:template>

  <xsl:key name="funcKey" match="xtstp:function" use="@name"/>
  <xsl:key name="predKey" match="xtstp:predicate" use="@name"/>

 
  <xsl:template match="xtstp:tstp">
    <omdoc id="XTSTP-auto-generated-omdoc">
    <metadata>
      <Title>XTSTP-auto-omdoc </Title>
      <Creator role="trl"><xsl:text>The xtstp2omdoc style sheet</xsl:text></Creator>
    </metadata>
    <xsl:text>&#xA;</xsl:text>
    <theory id="local-theory">
      <xsl:for-each select=".//xtstp:function[count( . | key('funcKey', @name)[1] ) = 1]">
        <xsl:variable name="n" select="@name"/>
	<symbol id="{$n}" kind="object" scope="global"/>
      </xsl:for-each>

      <xsl:for-each select=".//xtstp:predicate[count( . | key('predKey', @name)[1] ) = 1]">
        <xsl:variable name="n" select="@name"/>
 	<symbol id="{$n}" kind="object" scope="global"/>
      </xsl:for-each>
      <xsl:apply-templates/>
    </theory>
  </omdoc>
  </xsl:template>


  <xsl:template match="xtstp:include">
    <CMP>
    <xsl:text><!-- include skipped --></xsl:text>
    <xsl:text>&#xa;</xsl:text>
    </CMP>
  </xsl:template>
  
  <xsl:template match="xtstp:formula">
      <xsl:variable name="n" select="@name"/>        
      <xsl:choose>
        <xsl:when test="@status='axiom'"> 
 	   <axiom id="{$n}">
             <FMP>
	       <om:OMOBJ>
                 <xsl:apply-templates/>
	       </om:OMOBJ>
             </FMP>
	   </axiom>
        </xsl:when>
        <xsl:when test="@status='conjecture'"> 
	   <assertion id="{$n}" type="conjecture">
             <FMP>
	       <om:OMOBJ>
                 <xsl:apply-templates/>
	       </om:OMOBJ>
             </FMP>
	   </assertion>
        </xsl:when>
      </xsl:choose>
  </xsl:template>

  <xsl:template name="lambda">
    <xsl:param name="quantor"/>
    <xsl:param name="varlist"/>
    <xsl:param name="body"/>
    <om:OMBIND>
      <!--      <om:OMS cd="sthol" name="{$quantor}"/> -->
      <xsl:if test="@type='existential'"> 
        <om:OMS cd="quant1" name="exists"/>
      </xsl:if>
      <xsl:if test="@type='universal'"> 
        <om:OMS cd="quant1" name="forall"/>
      </xsl:if>
      <om:OMBVAR> 
        <xsl:for-each select="$varlist">
           <xsl:variable name="n" select="string(@name)" />
             <om:OMATTR> 
               <om:OMATP>
                <om:OMS name="type" cd="simpletypes"/>
                <om:OMS name="i"/>
              </om:OMATP>
              <om:OMV name="{$n}"/>
            </om:OMATTR>
        </xsl:for-each>
      </om:OMBVAR>
      <xsl:apply-templates select="$body" />
    </om:OMBIND>
  </xsl:template>

  <xsl:template match="xtstp:quantifier">
        <xsl:call-template name="lambda">
          <xsl:with-param name="quantor" select="@type"/>
          <xsl:with-param name="varlist" select="xtstp:variable"/>
          <xsl:with-param name="body" select="*[node() != 'xtstp:variable']"/>
        </xsl:call-template>
  </xsl:template>

  <!-- processess connective subtree
       $symbol  paramter specifies the connective symbol (it is converted
                to a TSTP symbol
       $sublist gives a list of subtrees to process; for each subtree, TSTP
                symbol glyph is emited and then the subtree is processed
		recursively
  -->
  <!-- processess connective subtree
       $symbol  paramter specifies the connective symbol (it is converted
                to an Openmath symbol
       $sublist gives a list of subtrees to process; for each subtree, 
                symbol glyph is emited and then the subtree is processed
		recursively
  -->

  <xsl:template match="xtstp:predicate|xtstp:function">
    <xsl:variable name="n" select="string(@name)" />
    <xsl:if test="count(*)>0">
      <om:OMA>
        <om:OMS cd="local-theory" name="{$n}"/>
        <xsl:call-template name="list">
          <xsl:with-param name="separator" select="' '"/>
          <xsl:with-param name="sublist" select="*"/>
        </xsl:call-template>
      </om:OMA>
    </xsl:if>
    <xsl:if test="count(*)=0">
      <om:OMS cd="local-theory" name="{$n}"/>
    </xsl:if>
  </xsl:template>

  <xsl:template match="xtstp:defined-predicate">
    <xsl:if test="@name='equal'"> 
    <om:OMA>
      <om:OMS cd="relation1" name="eq"/>
      <xsl:apply-templates select="*[1]" />
      <xsl:apply-templates select="*[2]" />
    </om:OMA>
    </xsl:if>
  </xsl:template>
  
 <xsl:template match="xtstp:negation">
    <om:OMA>
      <om:OMS cd="pl0" name="not"/>
      <xsl:apply-templates select="*[1]" />
    </om:OMA>
 </xsl:template>

 <xsl:template match="xtstp:conjunction">
    <om:OMA>
      <om:OMS cd="pl0" name="and"/>
      <xsl:apply-templates select="*[1]" />
      <xsl:apply-templates select="*[2]" />
    </om:OMA>
 </xsl:template>

 <xsl:template match="xtstp:disjunction">
    <om:OMA>
      <om:OMS cd="pl0" name="or"/>
      <xsl:apply-templates select="*[1]" />
      <xsl:apply-templates select="*[2]" />
    </om:OMA>
 </xsl:template>

 <xsl:template match="xtstp:implication">
    <om:OMA>
      <om:OMS cd="pl0" name="implies"/>
      <xsl:apply-templates select="*[1]" />
      <xsl:apply-templates select="*[2]" />
    </om:OMA>
 </xsl:template>

  <xsl:template match="xtstp:true">
    <om:OMS cd="pl0" name="true"/>
  </xsl:template>

  <xsl:template match="xtstp:false">
    <om:OMS cd="pl0" name="false"/> 
  </xsl:template>

  <xsl:template match="xtstp:variable">
    <xsl:variable name="n" select="string(@name)" />
    <om:OMV name="{$n}"/>
  </xsl:template>
  
  <xsl:template match="xtstp:number">
    <om:OMI><xsl:value-of select="." /></om:OMI>
  </xsl:template>

  <xsl:template match="xtstp:qstring">
    <CMP>
      <xsl:value-of select="."/>
    </CMP>
  </xsl:template>
  

  <xsl:template match="xtstp:comment">
  </xsl:template>


  <xsl:template match="xtstp:head">
<!--     <xsl:text>%</xsl:text> -->
<!--     <xsl:if test="@field!=''"> -->
<!--       <xsl:text> </xsl:text> -->
<!--       <xsl:value-of select="@field"/> -->
<!--       <xsl:value-of select="substring('            ',string-length(@field))"/> -->
<!--       <xsl:text> : </xsl:text> -->
<!--     </xsl:if> -->
<!--     <xsl:call-template name="replace-string"> -->
<!--      <xsl:with-param name="text" select="text()"/> -->
<!--      <xsl:with-param name="replace" select="'&#xa;'"/> -->
<!--      <xsl:with-param name="with" select="'&#xa;%                 '"/> -->
<!--     </xsl:call-template> -->
<!--     <xsl:text>&#xa;</xsl:text> -->
<!--     <xsl:if test="following-sibling::comment"> -->
<!--       <xsl:text>% Comment         </xsl:text> -->
<!--     </xsl:if> -->
  </xsl:template>

</xsl:stylesheet>
