// -*- indent-tabs-mode: nil -*-
/*****************************************************************
This file is part of the MathServe system

Authors: Juergen Zimmer

email: jzimmer@ags.uni-sb.de
   
For information about this program, write to:                          
MathWeb-Project                                                        
AG Siekmann/FB Informatik                                            
Universitaet des Saarlandes                                          
Gebaeude E 1.1                                                     
D-66041 Saarbruecken                                                 
Germany    

For information about the newest version of MathServe, see 
http://www.ags.uni-sb.de/~jzimmer/mathserve.html

See the file "LICENSE.txt" or
http://www.ags.uni-sb.de/~jzimmer/LICENSE.html
for information on usage and redistribution
of this file, and for a DISCLAIMER OF ALL WARRANTIES.
*****************************************************************/

/*  
 * Class for shutting down MathServe services properly 
 */

import java.util.List;

import javax.xml.namespace.QName;

import org.jdom.*;
import org.jdom.Namespace;
import org.jdom.xpath.XPath;

import org.apache.axis.client.Call;
import org.apache.axis.client.Service;

import org.mathweb.core.BasicService;
import org.mathweb.onto.OWLJDom;

/**
@author Juergen Zimmer, Saarland University, Germany
@version $Date: 2006/05/11 12:48:54 $ $Revision: 1.9 $
*/

public class ShutdownMathServ extends BasicService
{
  
  private String serviceConfigFile = getSystemProperty("mathserv.services.config", "MATHSERV_SERVICES_CONFIG");
  private static OWLJDom owlJDom = OWLJDom.getInstance();
  private static String endpoint = "http://localhost:8080/axis/services/";
  public static final Namespace WSDD_NS = Namespace.getNamespace("wsdd","http://xml.apache.org/axis/wsdd/");
  public static final XPath servicex = newXPath("./wsdd:service");
  public static final XPath classNamex = newXPath("./wsdd:parameter[@name='className']/@value");
  public static final XPath methodsx = newXPath("./wsdd:parameter[@name='allowedMethods']/@value");
  public static final XPath msdlx = newXPath("./wsdd:parameter[@name='msdlDescription']/@value");
  
  public static XPath newXPath(String x)
  {
    try{
      XPath xpath =  XPath.newInstance(x);
      xpath.addNamespace(WSDD_NS);
      return xpath;
    } catch(Exception e){
      return null;
    }
  }

  public void shutdown()
  {
    try{
      String servXML = null;
      Element root = null;
      List servs = null;
      servXML = readFile(serviceConfigFile);
      if (servXML != null) {
	try{
	  root = owlJDom.getRootElement(servXML);    
	  servs = servicex.selectNodes(root);
	} catch(Exception e) {
	}
      }
      if (root == null) return;
      for(int i=0 ; i<servs.size(); i++){
	try { 
	  Element s = (Element) servs.get(i);
	  String servName = s.getAttributeValue("name");
	  String className = classNamex.valueOf(s);
          String methods = methodsx.valueOf(s);
          String[] mNames = methods.split(",");
          boolean shutdownMethod = false; 
          for(int k = 0; k< mNames.length; k++)
            if (mNames[k].equals("shutdown")){
              shutdownMethod = true;
              k = mNames.length;
            }
          
          if (shutdownMethod) {
            //System.out.println(" methods:  " + ms);
            
            String[] path = className.trim().split("[.]");
            String qname = "http://";
            for (int k = (path.length - 2); k > -1 ; k--){
              qname = qname + path[k];
              if (k>0) 
                qname = qname + ".";
            }
            try {
              Service  service = new Service();
              Call     call    = (Call) service.createCall();
              call.setTargetEndpointAddress( new java.net.URL(endpoint + servName) );
              call.setOperationName(new QName(qname, "shutdown") );	
              call.setTimeout(new Integer(4000));
              System.out.print("shutting down " +endpoint + servName + "...");
              call.invoke( new Object[0] );
              System.out.println(" done.");
              //  	  logger.info(" parsed!!!");
              // 	  put(service);
            } catch (Exception e) {
              System.out.println(" failed.");
              e.printStackTrace();
	    //	    System.err.println(e.toString());
            }
          }
	} catch(Exception e){
	  e.printStackTrace();
	}
      }
    } catch(Exception e){
      e.printStackTrace();
    }
  }

  public static void main(String [] args) {
    ShutdownMathServ shut = new ShutdownMathServ();
    shut.shutdown();
    System.exit(0);
  }
}



