// -*- indent-tabs-mode: nil -*-
/*****************************************************************
This file is part of the MathServe system

Authors: Juergen Zimmer

email: jzimmer@ags.uni-sb.de
   
For information about this program, write to:                          
MathWeb-Project                                                        
AG Siekmann/FB Informatik                                            
Universitaet des Saarlandes                                          
Gebaeude E 1.1                                                     
D-66041 Saarbruecken                                                 
Germany    

For information about the newest version of MathServe, see 
http://www.ags.uni-sb.de/~jzimmer/mathserve.html

See the file "LICENSE.txt" or
http://www.ags.uni-sb.de/~jzimmer/LICENSE.html
for information on usage and redistribution
of this file, and for a DISCLAIMER OF ALL WARRANTIES.
*****************************************************************/

package org.mathweb.atp;

import java.io.*;
import org.apache.log4j.Level;
import org.apache.log4j.Logger; 
import org.apache.log4j.FileAppender; 
import org.apache.log4j.SimpleLayout;

import org.mathweb.util.TSTPCreator;
import org.mathweb.core.BasicService;
import org.mathweb.onto.OWLJDom;


/**
@author Juergen Zimmer University of Edinburgh
@version $Date: 2005/09/26 14:59:53 $ $Revision: 1.39 $
*/
public class KRHyper extends ATP {
  
  public String KRHYPER_HOME = getSystemProperty("krhyper.home", "KRHYPER_HOME");
  public String TPTP_HOME = getSystemProperty("tptp.home.dir", "TPTP.HOME.DIR");
  public String SWI_HOME =  getSystemProperty("tptp.swi.prolog", "TPTP_SWI_PROLOG");
  public String swiExecutable = findSystemExecutable(SWI_HOME+sep+"bin", "pl", "pl.exe");
  
  public String[] krhyper_envs = new String[] {"KRHYPER_HOME="+KRHYPER_HOME, "TPTP_HOME="+TPTP_HOME, "TPTPDirectory="+TPTP_HOME, "SWI_HOME="+SWI_HOME};
  public String krhyperExecutable = findSystemExecutable(KRHYPER_HOME, "krh", "KRH.exe");

  private static Logger logger = Logger.getLogger("KRHyper");
  private static String logFileName = getSystemProperty("krhyper.log.file", "KRHYPER_LOG_FILE");
  private static FileAppender appender = newAppender(logFileName);
        
  private static final int defaultNice = 15;
  
  public KRHyper()
  {
    super();
    name = "KRHyper_2005_04_22";
    acceptedConcept = "TstpProblem";
    resultConcept = "FOBrFPResult";
    proofConcept = "TstpCnfBrFPRefutation";
    calculusConcept = "krhyper_calc";
    
    logger.setLevel(Level.ALL);
    //    logger.setLevel(Level.OFF);
    if (!logger.getAllAppenders().hasMoreElements()) logger.addAppender(appender);
  }
  
  public String[] callProver(File problemFile, String language, String format, int time, String options) throws Exception
  {
    File inputFile = null;
    File outputFile = null;
    try{
      outputFile = File.createTempFile("krhyper", ".out");
      
      String krhyperOptions = " ";
      //      System.out.println(inputFile + " " + outputFile);
      
      if (language.equals(OWLJDom.MW_NS_STR+"TSTP")) {
        inputFile = tstp2X.tstp2X(problemFile, 
                                  "protein",
                                  "none");

        String command = swiExecutable + " -s " + KRHYPER_HOME + "/scripts/pre.pl " + inputFile.getAbsolutePath() + 
          " | " + krhyperExecutable + " " + KRHYPER_HOME+"/tptp_before.tme - " + KRHYPER_HOME + "/tptp_after.tme" +
          " > " + outputFile.getAbsolutePath() + " 2>&1";
        logger.debug("[KRHYPER] command: " + command);
        String[] commands = {"sh", "-c", command};
        logger.debug("[KRHYPER] starting prover...");
        this.start(command, krhyper_envs, defaultNice, time);

        logger.debug("reading result..." + outputFile.getAbsolutePath());
        String result = this.readFile(outputFile);
        if ((result != null) && (result.length() > 100000))
          return  new String[] {"Unknown", ""};

        logger.debug("[KRHYPER] output..." + result);
        
//         logger.debug("[KRHYPER] transforming to TSTP...");
//         String tstp = x2tstp.x2tstp(result, "KRHyper---3.3");
//         //        System.out.println("[KRHYPER] TSTP:\n" + tstp);
        String status = "foo";
        return new String[] {status, null};
      } 
      else{
        throw new Exception("Input format "+language+" not known or not supported.");
      }
    } catch(Exception e){
      logger.debug("[KRHyper] Exception " + e);
      throw e;
    } finally{
      try{ 
//         if (inputFile != null) inputFile.delete();
//         if (outputFile != null) outputFile.delete();
      } catch(Exception e){ logger.debug("[KRHyper] Exception " + e); }
    }
  }
  
  
  public static void main(String argv[]) throws Exception {
    String msHome = System.getProperty("mathserv.home");
      KRHyper e = new KRHyper();
      OWLJDom owlJDom = OWLJDom.getInstance();
      TSTPCreator creator = new TSTPCreator();
      String timeInd = owlJDom.createTimeIndi(20,20);
      //      timeInd = e.readFile(msHome + "/examples/owl/fault-time.owl");
      String servF1;
      System.out.println("path: " +System.getProperty("java.library.path"));
                         //       String problem = e.readFile(servF1);
      servF1 = msHome + "/examples/owl/fof2.owl";
      //servF1 = msHome + "/examples/owl/fault.owl";
      servF1 = msHome + "/examples/owl/cnf2.owl";
      //        servF1 = msHome + "/examples/owl/cnf-exec.owl";
      //      servF1 = msHome + "/examples/owl/fof3.owl";
      //        String servF1 = msHome + "/src/b";
      //        logger.info("result1: " + result);

      String ind = e.readFile(servF1);
      //ind = creator.tstp2OWLProblem(msHome + "/thirdParty/TPTPWorld/TPTP/Problems/SYN/SYN880-1.p");
      ind = creator.tstp2OWLProblem(msHome + "/thirdParty/TPTPWorld/TPTP/Problems/ALG/ALG001-2.p");
      //ind = creator.tstp2OWLProblem(msHome + "/thirdParty/TPTPWorld/TPTP/Problems/GRA/GRA001-1.p");
//       ind = creator.tstp2OWLProblem(msHome + "/thirdParty/TPTPWorld/TPTP/Problems/SET/SET062+3.p");
      //ind = creator.tstp2OWLProblem(msHome + "/thirdParty/TPTPWorld/TPTP/Problems/GRP/GRP012-2.p");
      
      System.out.println(ind);
      
      try {
        System.out.println(timeInd);
        String result = null;
//         for (int i=0; i< 700; i++) 
//         {
//             System.out.println(" call " + i);            
//             e = new KRHyper();
        result = e.proveProblem(ind, timeInd);
        //result = e.proveProblemXml(ind, timeInd);
        //System.out.println(result);
        
        //        logger.info("result1: " + result);
        
        //         result = e.proveProblem(ind);
        
        //        logger.info("result2: " + result);
        System.out.println(result);
      }
      catch(Exception ex){
        ex.printStackTrace();
        System.out.println(ex);
      }
  }
  
  
} // end KRHyper.java

/* Settings for Emacs.
 * Local variables:
 * c-basic-offset: 2
 * tab-width: 2
 * indent-tabs-mode: nil
 * End:
 */
