// -*- indent-tabs-mode: nil -*-
/*****************************************************************
This file is part of the MathServe system

Authors: Juergen Zimmer

email: jzimmer@ags.uni-sb.de
   
For information about this program, write to:                          
MathWeb-Project                                                        
AG Siekmann/FB Informatik                                            
Universitaet des Saarlandes                                          
Gebaeude E 1.1                                                     
D-66041 Saarbruecken                                                 
Germany    

For information about the newest version of MathServe, see 
http://www.ags.uni-sb.de/~jzimmer/mathserve.html

See the file "LICENSE.txt" or
http://www.ags.uni-sb.de/~jzimmer/LICENSE.html
for information on usage and redistribution
of this file, and for a DISCLAIMER OF ALL WARRANTIES.
*****************************************************************/
package org.mathweb.atp;

import java.io.*;
import java.util.*;
import java.rmi.Remote;
import java.rmi.RemoteException;
import javax.xml.namespace.QName;
import javax.xml.rpc.ParameterMode;

import org.apache.axis.client.Call;
import org.apache.axis.client.Service;
import org.apache.axis.encoding.XMLType;

import org.mindswap.query.ValueMap;

import org.mathweb.owls.OWLSParser;
import org.mathweb.onto.OWLJDom;
import org.mathweb.serv.ReasoningService;
import org.mathweb.util.TSTPCreator;
import org.mathweb.util.Java2WSDL;

public class VampireClient
{

  public String readFile(String fileName){
    try {
      FileInputStream isFile = new FileInputStream(fileName);
      int nChar;
      StringBuffer sbRead = new StringBuffer();
      nChar = isFile.read();
      while (nChar > -1) {
	sbRead.append((char) nChar);
	nChar = isFile.read();
      }
      isFile.close();
      return sbRead.toString();
    } catch (IOException ioe) {
      return null;
    }
  }

  public static void main(String [] args) {
    String msHome = System.getProperty("mathserv.home");
    OWLJDom owlJDom = OWLJDom.getInstance();
    TSTPCreator creator = new TSTPCreator();
    
    int time = 10;
    String timeInd = owlJDom.createTimeIndi(time, time);
    Integer timeInt = new Integer(time);
    
    try {
      String hostname;
      if ((args.length > 0) && (!args[0].trim().equals(""))) 
	hostname = args[0];
      else 
	hostname = "localhost";
      
      Service service = new Service();
      OWLSParser owlsParser = new OWLSParser();
      
      VampireClient c = new VampireClient();
      
      String fofProb = c.readFile(msHome + "/examples/owl/fof2.owl");
      //fofProb = creator.tstp2OWLProblem(msHome + "/thirdParty/TPTPWorld/TPTP/Problems/SYN/SYN880-1.p");
      //      fofProb = creator.tstp2OWLProblem(msHome + "/thirdParty/TPTPWorld/TPTP/Problems/GRP/GRP001-1.p");
      fofProb = creator.tstp2OWLProblem("/home/jzimmer/fuer_Juergen.tptp");
      String tptpProb = c.readFile("/home/jzimmer/fuer_Juergen.tptp");

      String endpoint = "http://" + hostname + ":8080/axis/services/VampireService";
      
      Call call = (Call) service.createCall();
      call.setTargetEndpointAddress(new java.net.URL(endpoint));
      call.addParameter("in0", XMLType.SOAP_STRING, ParameterMode.IN);
      //      call.addParameter("in1", XMLType.SOAP_STRING, ParameterMode.IN);
      call.addParameter("in1", XMLType.SOAP_INT, ParameterMode.IN);
      call.setReturnType(XMLType.SOAP_STRING);
      call.setTimeout(new Integer((time + 120)*1000));

//       call.setOperationName("proveProblem");
//       String result = (String) call.invoke( new Object [] {fofProb, timeInd});
      call.setOperationName("proveTPTPProblem");
      String result = (String) call.invoke( new Object [] {tptpProb, timeInt});
      System.out.println(result);

//       String wsdl = null;
//       Vector ms = new Vector();
//       ms.add("proveProblem");
//       ms.add("proveProblemXml");
      
//       wsdl = (new Java2WSDL()).java2WSDL(hostname, "VampireService", "org.mathweb.atp.Vampire" , ms );
//       //System.out.println(" wsdl:  " + wsdl);
      
//       ReasoningService service = owlsParser.parseServiceFile(msHome + "/etc/services/owls/1.1/VampireService.owl", wsdl);
      //ReasoningService service = owlsParser.parseServiceFile(msHome + "/etc/services/owls/1.1/VampireXmlService.owl", wsdl);
//       System.out.println("Service " + service);
//       System.out.println("time " + timeInd);
//       Hashtable map = new Hashtable();
//       map.put("tstp_problem", fofProb);
//       map.put("time_res", timeInd);
//       ValueMap resTab = service.invokeOWLS(map);
//       System.out.println("Sent problem and got '" + resTab + "'");

      
    } catch (Exception e) {
      e.printStackTrace();
      System.err.println(e.toString());
    }
  }
}
