// -*- indent-tabs-mode: nil -*-
/*****************************************************************
This file is part of the MathServe system

Authors: Juergen Zimmer

email: jzimmer@ags.uni-sb.de
   
For information about this program, write to:                          
MathWeb-Project                                                        
AG Siekmann/FB Informatik                                            
Universitaet des Saarlandes                                          
Gebaeude E 1.1                                                     
D-66041 Saarbruecken                                                 
Germany    

For information about the newest version of MathServe, see 
http://www.ags.uni-sb.de/~jzimmer/mathserve.html

See the file "LICENSE.txt" or
http://www.ags.uni-sb.de/~jzimmer/LICENSE.html
for information on usage and redistribution
of this file, and for a DISCLAIMER OF ALL WARRANTIES.
*****************************************************************/

package org.mathweb.broker;

import java.io.*;
import java.util.Locale;
import javax.swing.*;
import javax.swing.text.*;
import javax.imageio.ImageIO;

import java.awt.*;
import java.awt.image.BufferedImage;

import org.mathweb.broker.GuiEvent;
import org.mathweb.broker.Broker;

public class BrokerGUI extends JFrame // ModalDialog
{
  
  Broker myAgent; // pointer to AppointmentAgent
  private BufferedImage image;    // the rasterized image
  
  public class MyHighlightPainter extends DefaultHighlighter.DefaultHighlightPainter {
    public MyHighlightPainter(Color color) {
      super(color);
    }
  }
  
  public BrokerGUI(Broker a, String title) {
    this((JFrame)null,title);
    myAgent=a;
  }

  //  JLabel directiveLabel;
  JLabel imageLabel;
  ImageIcon imageIcon;
  
  JTextField directiveTextField;
  JTextField servicesTextField;
  JTextArea queryTextArea;
  JTextArea resultTextArea; 
  JTextArea gologTextArea; 
  JTextArea policyTextArea; 
  JButton  query1Button;
  JButton  query2Button;
  JButton  sendButton;
  JButton  prevButton;
  JButton  nextButton;
  JButton  execButton;
  JButton  policyButton;

  
  Highlighter.HighlightPainter myHighlightPainter = new MyHighlightPainter(Color.cyan);
  class Query1Mouse extends java.awt.event.MouseAdapter
  {
    String msHome = System.getProperty("mathserv.home");
    public void mouseClicked(java.awt.event.MouseEvent event)    {
      Object object = event.getSource();
      if (object == query1Button)
	try {
          //          String query1 = myAgent.readFile(msHome + "/examples/queries/query1.xml");
          String query1 = myAgent.readFile(msHome + "/examples/queries/owls/1.1/anyResult.owl");
          removeHighlights(queryTextArea);
          queryTextArea.setText(query1);
          gologTextArea.setText("");
          policyTextArea.setText("");
          resultTextArea.setText("");
          try { 
            image = ImageIO.read(new File(msHome + "/etc/images/BrokerEmpty.jpg")); }
          catch(IOException e) {
            e.printStackTrace();
            image = null;
          }
          if (image != null){
            imageIcon = new ImageIcon(image);
            imageLabel.setIcon(imageIcon);
          }
          
	  highlight(queryTextArea,"<mw:TstpCNFProblem",25);
	  setResult("");
	}
	catch(Exception e) { e.printStackTrace();}
    }}

  class Query2Mouse extends java.awt.event.MouseAdapter
  {
    String msHome = System.getProperty("mathserv.home");
    public void mouseClicked(java.awt.event.MouseEvent event)
    {
      Object object = event.getSource();
      if (object == query2Button)
	try {
          //          String query2 = myAgent.readFile(msHome + "/examples/queries/query2.xml");
          String query2 = myAgent.readFile(msHome + "/examples/queries/owls/1.1/nd-query1.owl");
          removeHighlights(queryTextArea);
	  queryTextArea.setText(query2);
          gologTextArea.setText("");
          policyTextArea.setText("");
          resultTextArea.setText("");
          try { 
            image = ImageIO.read(new File(msHome + "/etc/images/BrokerEmpty.jpg")); }
          catch(IOException e) {
            e.printStackTrace();
            image = null;
          }
          if (image != null){
            imageIcon = new ImageIcon(image);
            imageLabel.setIcon(imageIcon);
          }
	  highlight(queryTextArea,"<mw:TstpFOFProblem",21);
	  setResult("");
	} catch(Exception e) { e.printStackTrace();}
    }}
  
  class SendMouse extends java.awt.event.MouseAdapter {
    public void mouseClicked(java.awt.event.MouseEvent event) {
      Object object = event.getSource(); 
      if (object == sendButton){ GuiEvent ev = new GuiEvent(this, myAgent.COMP_PLANS); ev.addParameter(getQuery());  myAgent.onGuiEvent(ev);} } }

  class PrevMouse extends java.awt.event.MouseAdapter {
    public void mouseClicked(java.awt.event.MouseEvent event){
      Object object = event.getSource();
      if (object == prevButton) { GuiEvent ev = new GuiEvent(this, myAgent.PREV_PLAN); myAgent.onGuiEvent(ev);}}}
  class NextMouse extends java.awt.event.MouseAdapter {
    public void mouseClicked(java.awt.event.MouseEvent event){
      Object object = event.getSource();
      if (object == nextButton) { GuiEvent ev = new GuiEvent(this, myAgent.NEXT_PLAN); myAgent.onGuiEvent(ev);}}}

  class ExecMouse extends java.awt.event.MouseAdapter {
    public void mouseClicked(java.awt.event.MouseEvent event) {
      Object object = event.getSource();
      if (object == execButton){ GuiEvent ev = new GuiEvent(this, myAgent.EXEC_PLAN);  myAgent.onGuiEvent(ev);} } }
  

  class PolicyMouse extends java.awt.event.MouseAdapter {
    public void mouseClicked(java.awt.event.MouseEvent event) {
      Object object = event.getSource();
      if (object == policyButton){ GuiEvent ev = new GuiEvent(this, myAgent.COMP_POLICY); ev.addParameter(getQuery()); myAgent.onGuiEvent(ev);} } }
  
  
  public BrokerGUI(JFrame parent, String title) 
  {
    String msHome = System.getProperty("mathserv.home");
    setTitle(title);
    setSize(1350, 1000);
    setBackground( Color.gray );
    Rectangle bounds = this.getBounds();
    setLocation(20 + bounds.x, 15 + bounds.y);
    
    getContentPane().setLayout(new BorderLayout());
//     directiveLabel = new JLabel("Directive:");
//     directiveTextField = new JTextField(200);
//     directiveTextField.setPreferredSize(new Dimension(100,1));

    JLabel servicesLabel = new JLabel("Services:");
    JLabel queryLabel = new JLabel("Query:");	
    JLabel resultLabel = new JLabel("Dataflow Graphs:");	
    JLabel gologLabel = new JLabel("Golog Domain:");	
    JLabel policyLabel = new JLabel("Optimal Policy:");	
    
    servicesTextField= new JTextField(1);
    servicesTextField.setText("");
    servicesTextField.setPreferredSize(new Dimension(100,10));
    
    queryTextArea= new JTextArea();
    queryTextArea.setPreferredSize(new Dimension(80,200));
    queryTextArea.setMaximumSize(new Dimension(80,200));
    queryTextArea.setColumns(50);
    queryTextArea.setRows(400);

    JScrollPane queryPane = new JScrollPane();
    queryPane.getViewport().add(queryTextArea);
    queryPane.setVerticalScrollBarPolicy(JScrollPane.VERTICAL_SCROLLBAR_ALWAYS);
    queryPane.setPreferredSize(new Dimension(600, 3000));
    //    queryPane.setBounds(1,1,280,300);

     Font f = new Font("SansSerif", Font.PLAIN, 14);
     queryTextArea.setFont(f);

    try { 
      image = ImageIO.read(new File(msHome + "/etc/images/BrokerEmpty.jpg")); }
    catch(IOException e) {
      e.printStackTrace();
      image = null;
    }
    if (image != null){
      imageIcon = new ImageIcon(image);
      imageLabel = new JLabel(imageIcon);
      imageLabel.setText(null);
      imageLabel.setIcon(imageIcon);
    } else {
      imageLabel = new JLabel("No execution plan yet.                                 ");
      imageLabel.setPreferredSize(new Dimension(100,300));
      imageLabel.setMinimumSize(new Dimension(100,300));
      imageLabel.setMaximumSize(new Dimension(100,400));
      
    }

    JScrollPane planPane = new JScrollPane();
    planPane.getViewport().add(imageLabel);


    gologTextArea= new JTextArea(new DefaultStyledDocument());
    gologTextArea.setPreferredSize(new Dimension(80,100));
    gologTextArea.setMinimumSize(new Dimension(80,100));
    gologTextArea.setMaximumSize(new Dimension(80,100));
    gologTextArea.setColumns(80);
    gologTextArea.setRows(1200);
    gologTextArea.setText("Golog Domain");
    //    gologTextArea.setLineWrap(true);

    JScrollPane gologPane = new JScrollPane();
    gologPane.setVerticalScrollBarPolicy(JScrollPane.VERTICAL_SCROLLBAR_ALWAYS); 
    gologPane.setHorizontalScrollBarPolicy(JScrollPane.HORIZONTAL_SCROLLBAR_ALWAYS);
    gologPane.setPreferredSize(new Dimension(250, 1800));
    gologPane.getViewport().add(gologTextArea);

    policyTextArea= new JTextArea(new DefaultStyledDocument());
    policyTextArea.setPreferredSize(new Dimension(80,60));
    policyTextArea.setMaximumSize(new Dimension(80,60));
    policyTextArea.setColumns(50);
    policyTextArea.setRows(300);
    policyTextArea.setText("Optimal Policy");

    JScrollPane policyPane = new JScrollPane();
    policyPane.getViewport().add(policyTextArea);
    policyPane.setVerticalScrollBarPolicy(JScrollPane.VERTICAL_SCROLLBAR_ALWAYS);
    policyPane.setHorizontalScrollBarPolicy(JScrollPane.HORIZONTAL_SCROLLBAR_ALWAYS);
    policyPane.setPreferredSize(new Dimension(250, 1800));

    resultTextArea= new JTextArea(new DefaultStyledDocument());
    resultTextArea.setPreferredSize(new Dimension(80,60));
    resultTextArea.setMaximumSize(new Dimension(80,60));
    resultTextArea.setColumns(50);
    resultTextArea.setRows(300);
    resultTextArea.setText("Result");
    
    JScrollPane resultPane = new JScrollPane();
    resultPane.getViewport().add(resultTextArea);
    resultPane.setVerticalScrollBarPolicy(JScrollPane.VERTICAL_SCROLLBAR_ALWAYS);
    resultPane.setPreferredSize(new Dimension(250, 1800));
    //    resultPane.setBounds(1,1,280,40);

    
    //     int hpolicy = pane.getHorizontalScrollBarPolicy();
//     int vpolicy = pane.getVerticalScrollBarPolicy();

    // Make the scrollbars always appear
    //pane.setHorizontalScrollBarPolicy(JScrollPane.HORIZONTAL_SCROLLBAR_NEVER);
    //    pane.setHorizontalScrollBarPolicy(JScrollPane.HORIZONTAL_SCROLLBAR_ALWAYS);
    //pane.setVerticalScrollBarPolicy(JScrollPane.VERTICAL_SCROLLBAR_ALWAYS);
    
//     // Make the scrollbars never appear
//     pane.setVerticalScrollBarPolicy(JScrollPane.VERTICAL_SCROLLBAR_NEVER);


    query1Button = new JButton("Query 1");
    query2Button = new JButton("Query 2");
    sendButton = new JButton("Generate Plans");
    prevButton = new JButton("Previous Plan");
    nextButton = new JButton("Next Plan");
    execButton = new JButton("Execute Plan");
    policyButton = new JButton("Compute Policy");
    //    queryTextArea.setText("fooo \n hallo");
    
    //     JScrollBar scrollBar = new JScrollBar();
    //     queryTextArea.add(scrollBar);
    

    JPanel p1 = new JPanel();
    p1.setLayout(new BoxLayout(p1 , BoxLayout.Y_AXIS));
    
    JPanel spanel = new JPanel();
    spanel.setLayout(new BoxLayout(spanel , BoxLayout.X_AXIS));
    spanel.add(Box.createRigidArea(new Dimension(10,0)));
    servicesLabel.setPreferredSize(new Dimension(75,15));
    spanel.add(servicesLabel);
    spanel.add(Box.createHorizontalGlue());
    spanel.add(servicesTextField);
    spanel.add(Box.createRigidArea(new Dimension(10,0)));
      
    JPanel p2 = new JPanel();
    p2.setLayout(new BoxLayout(p2 , BoxLayout.X_AXIS));
//     directiveLabel.setPreferredSize(new Dimension(75,15));
    p2.add(Box.createRigidArea(new Dimension(10,0)));
    p2.add(queryLabel);
    p2.add(Box.createRigidArea(new Dimension(500,0)));
    p2.add(resultLabel);
    p2.add(Box.createRigidArea(new Dimension(10,0)));

//     JPanel gLabels = new JPanel();
//     gLabels.setLayout(new BoxLayout(gLabels , BoxLayout.X_AXIS));
// //     directiveLabel.setPreferredSize(new Dimension(75,15));
//     gLabels.add(Box.createRigidArea(new Dimension(10,0)));
//     gLabels.add(gologLabel);
//     gLabels.add(Box.createRigidArea(new Dimension(400,0)));
//     gLabels.add(policyLabel);
//     gLabels.add(Box.createRigidArea(new Dimension(10,0)));
    
    JPanel p3 = new JPanel();
    p3.setLayout(new BoxLayout(p3, BoxLayout.X_AXIS));
    //    queryLabel.setPreferredSize(new Dimension(75,15));
    p3.add(Box.createRigidArea(new Dimension(10,0)));
    p3.add(queryPane);
    p3.add(Box.createRigidArea(new Dimension(10,0)));
    //    p3.add(Box.createHorizontalGlue());
    p3.add(planPane);
    p3.add(Box.createRigidArea(new Dimension(10,0)));
    
    JPanel p4 = new JPanel(); // 
    p4.setLayout(new BoxLayout(p4, BoxLayout.X_AXIS));
    p4.add(Box.createHorizontalGlue());
    p4.add(query1Button);
    p4.add(query2Button);
    p4.add(sendButton);
    p4.add(policyButton);
    p4.add(Box.createHorizontalGlue());
    
    p4.add(prevButton);
    p4.add(nextButton);
    p4.add(execButton);
    p4.add(Box.createHorizontalGlue());
    
    JPanel p5 = new JPanel();
    p5.setLayout(new BoxLayout(p5, BoxLayout.X_AXIS));
    p5.add(resultPane);


    JPanel p6 = new JPanel();
    p6.setLayout(new BoxLayout(p6, BoxLayout.X_AXIS));
    p6.add(gologPane);
    
    JPanel p7 = new JPanel();
    p7.setLayout(new BoxLayout(p7, BoxLayout.X_AXIS));
    p7.add(policyPane);

    JPanel gologPanel = new JPanel();
    gologPanel.setLayout(new BoxLayout(gologPanel, BoxLayout.Y_AXIS));
    gologPanel.add(p6);
    gologPanel.add(Box.createRigidArea(new Dimension(0,5)));
    gologPanel.add(p7);

    /**
       p1: main area
       p2: directive box
       p3: query box & execution plan
       p4: buttons 
       p5: result of execution
       p6: golog program
       p7: golog optimal policy
    */
    p1.add(Box.createRigidArea(new Dimension(0,15)));
    p1.add(spanel);
    p1.add(Box.createRigidArea(new Dimension(0,15)));
    p1.add(p2);
    p1.add(Box.createRigidArea(new Dimension(0,5)));
    p1.add(p3);
    p1.add(Box.createRigidArea(new Dimension(0,5)));
    p1.add(p4);
    p1.add(Box.createRigidArea(new Dimension(0,5)));
    //    p1.add(gLabels);
    p1.add(Box.createRigidArea(new Dimension(0,5)));
    p1.add(gologPanel);
    p1.add(Box.createRigidArea(new Dimension(0,5)));
    p1.add(p5);
    p1.add(Box.createRigidArea(new Dimension(0,5)));
    getContentPane().add(p1);		

    //{{REGISTER_LISTENERS
    Query1Mouse aSymMouse1 = new Query1Mouse();
    Query2Mouse aSymMouse2 = new Query2Mouse();
    SendMouse aSymMouse3 = new SendMouse();
    PrevMouse aSymMouse4 = new PrevMouse();
    NextMouse aSymMouse5 = new NextMouse();
    ExecMouse aSymMouse6 = new ExecMouse();
    PolicyMouse aSymMouse7 = new PolicyMouse();
    query1Button.addMouseListener(aSymMouse1);
    query2Button.addMouseListener(aSymMouse2);
    sendButton.addMouseListener(aSymMouse3);
    prevButton.addMouseListener(aSymMouse4);
    nextButton.addMouseListener(aSymMouse5);
    execButton.addMouseListener(aSymMouse6);
    policyButton.addMouseListener(aSymMouse7);
    //     SymAction lSymAction = new SymAction();
//     directiveTextField.addActionListener(lSymAction);
//     queryTextArea.addActionListener(lSymAction);
    //}}
  }

  public BrokerGUI(JFrame parent)
  {
    this(parent, "MathWeb Proving Service Broker");
  }

  // Add a constructor for Interactions (ignoring modal)
  public BrokerGUI(JFrame parent, boolean modal)
  {
    this(parent);
  }

  // Add a constructor for Interactions (ignoring modal)
  public BrokerGUI(JFrame parent, String title, boolean modal)
  {
    this(parent, title);
  }
  
  // to embed in frames
  public JLabel getJLabel() {
    if (image == null) return null;         // no image available
    ImageIcon icon = new ImageIcon(image);
    return new JLabel(icon);
  }
  
  public void readImage(String filename) {
    try { 
      image = ImageIO.read(new File(filename)); }
    catch(IOException e) {
      e.printStackTrace();
      image = null;
    }
  }
  
  public void  showImage(String filename){
    // create the GUI for viewing the image if needed
    if (filename != null){
      readImage(filename);
    }
    if (image != null){
      imageIcon = new ImageIcon(image);
      imageLabel.setText(null);
      imageLabel.setIcon(imageIcon);
    }
    
    this.repaint();
  }
  public void  showImage(BufferedImage img){
    // create the GUI for viewing the image if needed
    if (img != null){
      image = img;
      imageIcon = new ImageIcon(image);
      imageLabel.setText(null);
      imageLabel.setIcon(imageIcon);
    }
    this.repaint();
  }
  
  // accessor methods
    public int height() { return image.getHeight(null); }
    public int width()  { return image.getWidth(null);  }

    // return Color of pixel (x, y)
    public Color getColor(int x, int y) {
        return new Color(image.getRGB(x, y));
    }

    // change color of pixel (x, y) to c
    public void setColor(int x, int y, Color c) {
        image.setRGB(x, y, c.getRGB());
    }

  public String getQuery()
  {
    return queryTextArea.getText();
  }
  
  public void setQuery(String pass)
  {
    //queryTextArea.insert(pass,0);
    queryTextArea.setText(pass);
    highlight(queryTextArea,"<problem",33);
  }
  
  public void setServices(String pass)
  {
    servicesTextField.setText(pass);
  }

  public void setGolog(String pass)
  {
    gologTextArea.setText(pass);
  }
  
  public void setPolicy(String pass)
  {
    policyTextArea.setText(pass);
  }
  
  public void setResult(String pass)
  {
    removeHighlights(resultTextArea);
    resultTextArea.setText(pass);
    highlight(resultTextArea,"TstpCnfBrFPRefutation", 22);
    highlight(resultTextArea,"% File       :",12);
  }
  

  //{{DECLARE_CONTROLS
  //}}
  
  
  public void highlight(JTextComponent textComp, String pattern, int offset) {
    // First remove all old highlights
    try {
      Highlighter hilite = textComp.getHighlighter();
      Document doc = textComp.getDocument();
      String text = doc.getText(0, doc.getLength());
      int pos = 0;
      
      // Search for pattern
      while ((pos = text.indexOf(pattern, pos)) >= 0) {
	// Create highlighter using private painter and apply around pattern
	hilite.addHighlight(pos, pos+pattern.length()+offset, myHighlightPainter);
	pos += pattern.length()+offset;
      }
    } catch (BadLocationException e) {
    }
  }
  
  // Removes only our private highlights
  public void removeHighlights(JTextComponent textComp) {
    Highlighter hilite = textComp.getHighlighter();
    Highlighter.Highlight[] hilites = hilite.getHighlights();
    
    for (int i=0; i<hilites.length; i++) {
      if (hilites[i].getPainter() instanceof MyHighlightPainter) {
	hilite.removeHighlight(hilites[i]);
      }
    }
  }
}
