// -*- indent-tabs-mode: nil -*-
/*****************************************************************
This file is part of the MathServe system

Authors: Juergen Zimmer

email: jzimmer@ags.uni-sb.de
   
For information about this program, write to:                          
MathWeb-Project                                                        
AG Siekmann/FB Informatik                                            
Universitaet des Saarlandes                                          
Gebaeude E 1.1                                                     
D-66041 Saarbruecken                                                 
Germany    

For information about the newest version of MathServe, see 
http://www.ags.uni-sb.de/~jzimmer/mathserve.html

See the file "LICENSE.txt" or
http://www.ags.uni-sb.de/~jzimmer/LICENSE.html
for information on usage and redistribution
of this file, and for a DISCLAIMER OF ALL WARRANTIES.
*****************************************************************/

/** For testing:

runt org.mathweb.eval.SoapBroker

*/
package org.mathweb.broker;

import java.io.File;
import java.util.*;

import org.apache.log4j.Level;
import org.apache.log4j.Logger; 
import org.apache.log4j.FileAppender; 

import org.jdom.*;

import org.mathweb.atp.X2TSTP;
import org.mathweb.serv.Query;
import org.mathweb.serv.ReasoningProblem;
import org.mathweb.util.*;
import org.mathweb.owls.*;
import org.mathweb.onto.*;
import org.mathweb.core.BasicService;


/**
   @author Juergen Zimmer, Saarland University, Germany
   @version $Date: 2006/05/26 13:39:25 $ $Revision: 1.4 $
*/
public class SoapBroker extends BasicService {
  
  private static Logger logger = Logger.getLogger("SoapBroker");
  private static String logFileName = getSystemProperty("xmlrpcbroker.log.file", "XMLRPCBROKER_LOG_FILE");
  private static FileAppender appender = newAppender(logFileName);

  private Random rand = new Random();
  private List concepts;
  
  private static OWLSParser owlsParser = new OWLSParser();
  private ServiceRegistry serviceRegistry = new ServiceRegistry();
  private TSTPCreator creator = new TSTPCreator();

  protected static final String sep = File.separator;
  
  private String choiceProgram = readFile(msHome + sep + "thirdParty" + sep + "golog" + sep + "dtgolog" + sep + "atpChoice");
  private String splitProgram = readFile(msHome + sep + "thirdParty" + sep + "golog" + sep + "dtgolog" + sep + "atpSplit");
  private String splitJoinProgram = readFile(msHome + sep + "thirdParty" + sep + "golog" + sep + "dtgolog" + sep + "atpSplitJoin");
  private String optProgram = readFile(msHome + sep + "thirdParty" + sep + "golog" + sep + "dtgolog" + sep + "atpOpt-300");
  private String zenonProgram = readFile(msHome + sep + "thirdParty" + sep + "golog" + sep + "dtgolog" + sep + "zenonOpt");
  
  private int serverPort = 0;
  
  /* The transformation engine for the provers' output */
  protected X2TSTP x2tstp = new X2TSTP();
  
  public SoapBroker()  {
    super();
    
    serviceRegistry.loadDefaultOwlsServices(owlsParser);
    
    logger.setLevel(Level.ALL);
    logger.setLevel(Level.OFF);
    if (!logger.getAllAppenders().hasMoreElements()) logger.addAppender(appender);
    logger.info("new broker object!");
  }
  
  public String ping()
  {
    System.out.println("I'm being pinged!!!");
    return new String("pong");
  }
  
  public String proveProblemXml(String tstpProblem, Integer timeInt)
  {
    String owlsQuery = creator.tstp2OWLSQuery(tstpProblem);
    //    System.out.println(owlsQuery);
    int time = timeInt.intValue();
    Query query = owlsParser.parseQuery(owlsQuery);
    String msHome = System.getProperty("mathserv.home");
    String program = readFile(msHome + sep + "thirdParty" + sep + "golog" + sep + "dtgolog" + sep + "atpOptXml");
    GologExecutor executor = new GologExecutor();
    String result = executor.executeProgram(program, query, serviceRegistry, time, false, null, false);
    //    System.out.println("result: "+ result);
    return result;
  }

  public String proveProblemChoice(String tstpProblem, Integer timeInt)
  {
      //    System.out.println(owlsQuery);
    int time = timeInt.intValue();
    String owlsQuery = creator.tstp2OWLSQuery(tstpProblem);
    GologExecutor executor = new GologExecutor();
    Query query = owlsParser.parseQuery(owlsQuery);
    String result = executor.executeProgram(choiceProgram, query, serviceRegistry, time, false, null, false);
    //    System.out.println("result: "+ result);
    return result;
  }

  public String proveProblemSplit(String tstpProblem, Integer timeInt)
  {
    //    System.out.println(owlsQuery);
    int time = timeInt.intValue();
    String owlsQuery = creator.tstp2OWLSQuery(tstpProblem);
    GologExecutor executor = new GologExecutor();
    Query query = owlsParser.parseQuery(owlsQuery);
    String result = executor.executeProgram(splitProgram, query, serviceRegistry, time, false, null, false);
    //    System.out.println("result: "+ result);
    return result;
  }

  public String proveProblemSplitJoin(String tstpProblem, Integer timeInt)
  {
    //    System.out.println(owlsQuery);
    int time = timeInt.intValue();
    String owlsQuery = creator.tstp2OWLSQuery(tstpProblem);
    GologExecutor executor = new GologExecutor();
    Query query = owlsParser.parseQuery(owlsQuery);
    String result = executor.executeProgram(splitJoinProgram, query, serviceRegistry, time, false, null, false);
    //    System.out.println("result: "+ result);
    return result;
  }

  public String proveProblemOpt(String tstpProblem, Integer timeInt)
  {
    //    System.out.println(owlsQuery);
    int time = timeInt.intValue();
    String owlsQuery = creator.tstp2OWLSQuery(tstpProblem);
    Query query = owlsParser.parseQuery(owlsQuery);
    GologExecutor executor = new GologExecutor();
    String result = executor.executeProgram(optProgram, query, serviceRegistry, time, false, null, false);
    //    System.out.println("result: "+ result);
    return result;
  }

  public String proveOmdocProblem(String omdocProblem, Integer timeInt)
  {
    //    System.out.println(owlsQuery);
    int time = timeInt.intValue();
    String tstpProblem = x2tstp.omdoc2TSTP(omdocProblem);
    System.out.println("tstp: "+ tstpProblem);
    String owlsQuery = creator.tstp2OWLSQuery(tstpProblem);
    GologExecutor executor = new GologExecutor();
    Query query = owlsParser.parseQuery(owlsQuery);
    String result = executor.executeProgram(optProgram, query, serviceRegistry, time, false, null, false);
    //    System.out.println("result: "+ result);
    return result;
  }


  public String testProve()
  {
    //    System.out.println(owlsQuery);
    String omdocProblem= readFile("/home/jzimmer/test.omdoc");
    int time = 20;
    String tstpProblem = x2tstp.omdoc2TSTP(omdocProblem);
    System.out.println("tstp: "+ tstpProblem);
    String owlsQuery = creator.tstp2OWLSQuery(tstpProblem);
    GologExecutor executor = new GologExecutor();
    Query query = owlsParser.parseQuery(owlsQuery);
    String result = executor.executeProgram(optProgram, query, serviceRegistry, time, false, null, false);
    //    System.out.println("result: "+ result);
    return result;
  }


  public String proveProblemFocal(String tstpProblem, Integer timeInt)
  {
    int time = timeInt.intValue();
    String owlsQuery = creator.tstp2OWLSQuery(tstpProblem);
    Query query = owlsParser.parseQuery(owlsQuery);
    GologExecutor executor = new GologExecutor();
    String result = executor.executeProgram(zenonProgram, query, serviceRegistry, time, false, null, false);
    System.out.println("result: "+ result);
    return result;
  }

    
  /*  the name of a golog program  and of a TSTP SPC file and time in seconds */
  public static void main(String argv[]) throws Exception {
    SoapBroker broker = new SoapBroker();
    //    String queryXML = broker.readFile(msHome + "/examples/queries/result-query2.xml");
    String queryXML = broker.readFile(msHome + "/examples/queries/result-query1.xml");
    System.out.println("XML: " + queryXML);
    //    broker.brokerService(queryXML, 20);
  }
    
//     SoapBroker broker = new SoapBroker();
//     //    String problem = broker.readFile("/home/jzimmer/puz.omdoc");
//     String msHome = System.getProperty("mathserv.home");
//     String problem = broker.readFile(msHome + "/examples/omdoc/error.xml");
//     broker.proveOmdocProblem(problem, 10);
  
} // end SoapBroker.java

      
/* Settings for Emacs.
 * Local variables:
 * c-basic-offset: 2
 * tab-width: 2
 * indent-tabs-mode: nil
 * End:
 */

