// -*- indent-tabs-mode: nil -*-
/*****************************************************************
This file is part of the MathServe system

Authors: Juergen Zimmer

email: jzimmer@ags.uni-sb.de
   
For information about this program, write to:                          
MathWeb-Project                                                        
AG Siekmann/FB Informatik                                            
Universitaet des Saarlandes                                          
Gebaeude E 1.1                                                     
D-66041 Saarbruecken                                                 
Germany    

For information about the newest version of MathServe, see 
http://www.ags.uni-sb.de/~jzimmer/mathserve.html

See the file "LICENSE.txt" or
http://www.ags.uni-sb.de/~jzimmer/LICENSE.html
for information on usage and redistribution
of this file, and for a DISCLAIMER OF ALL WARRANTIES.
*****************************************************************/

/** For testing:

runt org.mathweb.eval.XmlRpcBroker atpChoice Test1 3

*/
package org.mathweb.broker;

import java.io.File;
import java.io.IOException;
import java.util.*;

import org.apache.xmlrpc.*;

import org.apache.log4j.Level;
import org.apache.log4j.Logger; 
import org.apache.log4j.FileAppender; 

import org.jdom.*;

import org.mathweb.atp.X2TSTP;
import org.mathweb.serv.Query;
import org.mathweb.util.*;
import org.mathweb.owls.*;
import org.mathweb.core.BasicService;


/**
   @author Juergen Zimmer, Saarland University, Germany
   @version $Date: 2006/07/28 08:35:19 $ $Revision: 1.28.2.2 $
*/
public class XmlRpcBroker extends BasicService {
  
  //  private static Logger logger = Logger.getLogger("XmlRpcBroker");
  private static String logFileName = getSystemProperty("xmlrpcbroker.log.file", "XMLRPCBROKER_LOG_FILE");
  private static FileAppender appender = newAppender(logFileName);

  private Random rand = new Random();
  private List concepts;
  
  private static OWLSParser owlsParser = new OWLSParser();
  private ServiceRegistry serviceRegistry = new ServiceRegistry();
  private TSTPCreator creator = new TSTPCreator();

  protected static final String sep = File.separator;
  private String msHome = System.getProperty("mathserv.home");
  
  private String choiceProgram = readFile(msHome + sep + "thirdParty" + sep + "golog" + sep + "dtgolog" + sep + "atpChoice");
  private String splitProgram = readFile(msHome + sep + "thirdParty" + sep + "golog" + sep + "dtgolog" + sep + "atpSplit");
  private String splitJoinProgram = readFile(msHome + sep + "thirdParty" + sep + "golog" + sep + "dtgolog" + sep + "atpSplitJoin");
  private String optProgram = readFile(msHome + sep + "thirdParty" + sep + "golog" + sep + "dtgolog" + sep + "atpOpt-600");
  private String zenonProgram = readFile(msHome + sep + "thirdParty" + sep + "golog" + sep + "dtgolog" + sep + "zenonOpt");
  
  private WebServer xmlRpcServer;
  private int serverPort = 0;

  /* The transformation engine for the provers' output */
  protected X2TSTP x2tstp = new X2TSTP();
  
  public XmlRpcBroker()  {
    super();
    
    serviceRegistry.loadDefaultOwlsServices(owlsParser);
    
//     logger.setLevel(Level.ALL);
//     logger.setLevel(Level.OFF);
//     if (!logger.getAllAppenders().hasMoreElements()) logger.addAppender(appender);
//     logger.info("new broker object!");
  }
  
  
  private void startXmlRpcServer(int givenPort) throws ClassNotFoundException {
    //    XmlRpc.setDriver(Class.forName(".SAXDriver"));
    boolean success = false;
    serverPort = givenPort;
    int defaultPort = givenPort;
    do{
      try{
        xmlRpcServer = new WebServer(serverPort);
        xmlRpcServer.start();
        System.out.println("Started Java XMLRPC server on port "+serverPort);
        xmlRpcServer.addHandler ("$default", this);
        success = true;
      }catch(IOException e){
        serverPort++;
      }
      if ((serverPort - defaultPort) > 40) {
        System.out.println("More than 40 ports over "+ defaultPort + " are bound. Giving up!");
        serverPort = 0;
        break;
      }
    } while (!success);
  }
  
  
  public String ping()
  {
    System.out.println("I'm being pinged!!!");
    return new String("pong");
  }
  
  public String proveProblemXml(String tstpProblem, int time)
  {
    String owlsQuery = creator.tstp2OWLSQuery(tstpProblem);
    //    System.out.println(owlsQuery);
    Query query = owlsParser.parseQuery(owlsQuery);
    String msHome = System.getProperty("mathserv.home");
    String program = readFile(msHome + sep + "thirdParty" + sep + "golog" + sep + "dtgolog" + sep + "atpOptXml");
    GologExecutor executor = new GologExecutor();
    String result = executor.executeProgram(program, query, serviceRegistry, time, false, null, false);
    //    System.out.println("result: "+ result);
    return result;
  }

  public String proveProblemChoice(String tstpProblem, int time)
  {
      //    System.out.println(owlsQuery);
    String owlsQuery = creator.tstp2OWLSQuery(tstpProblem);
    GologExecutor executor = new GologExecutor();
    Query query = owlsParser.parseQuery(owlsQuery);
    String result = executor.executeProgram(choiceProgram, query, serviceRegistry, time, false, null, false);
    //    System.out.println("result: "+ result);
    return result;
  }

  public String proveProblemSplit(String tstpProblem, int time)
  {
    //    System.out.println(owlsQuery);
    String owlsQuery = creator.tstp2OWLSQuery(tstpProblem);
    GologExecutor executor = new GologExecutor();
    Query query = owlsParser.parseQuery(owlsQuery);
    String result = executor.executeProgram(splitProgram, query, serviceRegistry, time, false, null, false);
    //    System.out.println("result: "+ result);
    return result;
  }

  public String proveProblemSplitJoin(String tstpProblem, int time)
  {
    //    System.out.println(owlsQuery);
    String owlsQuery = creator.tstp2OWLSQuery(tstpProblem);
    GologExecutor executor = new GologExecutor();
    Query query = owlsParser.parseQuery(owlsQuery);
    String result = executor.executeProgram(splitJoinProgram, query, serviceRegistry, time, false, null, false);
    //    System.out.println("result: "+ result);
    return result;
  }

  public String proveProblemOpt(String tstpProblem, int time)
  {
    //    System.gc();
    String owlsQuery = creator.tstp2OWLSQuery(tstpProblem);
    //System.out.println(owlsQuery);
    Query query = owlsParser.parseQuery(owlsQuery);
    //    System.out.println(query);
    GologExecutor executor = new GologExecutor();
    String result = executor.executeProgram(optProgram, query, serviceRegistry, time, false, null, false);
    //    System.out.println("result: "+ result);
    return result;
  }

  public String proveOmdocProblem(String omdocProblem, int time)
  {
    System.out.println(omdocProblem);
    String tstpProblem = x2tstp.omdoc2TSTP(omdocProblem);
    System.out.println("tstp: "+ tstpProblem);
    String owlsQuery = creator.tstp2OWLSQuery(tstpProblem);
    GologExecutor executor = new GologExecutor();
    Query query = owlsParser.parseQuery(owlsQuery);
    String result = executor.executeProgram(optProgram, query, serviceRegistry, time, false, null, false);
    //    System.out.println("result: "+ result);
    return result;
  }

  public String testProve()
  {
    //    System.out.println(owlsQuery);
    String omdocProblem= readFile("/home/jzimmer/test.omdoc");
    int time = 20;
    
    String tstpProblem = x2tstp.omdoc2TSTP(omdocProblem);
    System.out.println("tstp: "+ tstpProblem);
    String owlsQuery = creator.tstp2OWLSQuery(tstpProblem);
    GologExecutor executor = new GologExecutor();
    Query query = owlsParser.parseQuery(owlsQuery);
    String result = executor.executeProgram(optProgram, query, serviceRegistry, time, false, null, false);
    //    System.out.println("result: "+ result);
    return result;
  }


  public String proveProblemFocal(String tstpProblem, int time)
  {
    String owlsQuery = creator.tstp2OWLSQuery(tstpProblem);
    Query query = owlsParser.parseQuery(owlsQuery);
    GologExecutor executor = new GologExecutor();
    String result = executor.executeProgram(zenonProgram, query, serviceRegistry, time, false, null, false);
    System.out.println("result: "+ result);
    return result;
  }

  /*  the name of a golog program  and of a TSTP SPC file and time in seconds */
  public static void main(String argv[]) throws Exception {
    int portNumber = 0;
    try{ if (argv.length > 0) {
      portNumber = Integer.parseInt(argv[0]);
    }
    if (portNumber > 0) {
      XmlRpcBroker broker = new XmlRpcBroker();
      broker.startXmlRpcServer(portNumber);
    }
    }catch(Exception e){
    }
    
//     XmlRpcBroker broker = new XmlRpcBroker();
//     //    String problem = broker.readFile("/home/jzimmer/puz.omdoc");
//     String msHome = System.getProperty("mathserv.home");
//     String problem = broker.readFile(msHome + "/examples/omdoc/error.xml");
//     broker.proveOmdocProblem(problem, 10);
  }
  
} // end XmlRpcBroker.java

      
/* Settings for Emacs.
 * Local variables:
 * c-basic-offset: 2
 * tab-width: 2
 * indent-tabs-mode: nil
 * End:
 */

