// -*- indent-tabs-mode: nil -*-
/*****************************************************************
This file is part of the MathServe system

Authors: Juergen Zimmer

email: jzimmer@ags.uni-sb.de
   
For information about this program, write to:                          
MathWeb-Project                                                        
AG Siekmann/FB Informatik                                            
Universitaet des Saarlandes                                          
Gebaeude E 1.1                                                     
D-66041 Saarbruecken                                                 
Germany    

For information about the newest version of MathServe, see 
http://www.ags.uni-sb.de/~jzimmer/mathserve.html

See the file "LICENSE.txt" or
http://www.ags.uni-sb.de/~jzimmer/LICENSE.html
for information on usage and redistribution
of this file, and for a DISCLAIMER OF ALL WARRANTIES.
*****************************************************************/
package org.mathweb.core;

import java.io.*;

public class CommandThread extends Thread
{

  private String command;
  private String[] envs;
  private Process myProcess;
  private int nice;
  private int maxCPU;
  private boolean wait;
  private boolean finished;
  
  static boolean verbose=true;
  
  public CommandThread(String cmd, String[] es)
  {
    command = cmd;
    envs = es;
    nice = -1;
    maxCPU = -1;
    wait = true;
  }
    
  public CommandThread(String cmd, String[] es, int niceVal, int cpuLimit, boolean doWait)
  {
    command = cmd;
    envs = es;
    nice = niceVal;
    maxCPU = cpuLimit;
    wait = doWait;
  }
    
  
  public void run()    {
    //System.out.println("executing threaded: " + commands);
    finished = false;
    String new_command;
        
    if (maxCPU < 1) {
      new_command = command;
    }
    else {
      new_command = BasicService.cpuLimitedRun + " " + maxCPU + " " + maxCPU + " " + command;
    }
    String[] commands = {"/bin/sh", "-c", new_command};
    //    System.out.println("executing: "+ new_command);
    this.exec(commands, envs, false, wait);
  }
    
    
  public boolean exec(String[] commands, String[] envs, boolean printResults, boolean wait) 
  {
    try {
      // Start running command, returning immediately.
      myProcess = java.lang.Runtime.getRuntime().exec(commands,envs);
      
      // Print the output. Since we read until
      // there is no more input, this causes us
      // to wait until the process is completed

      System.out.println("printing output?: " + printResults);
      if(printResults) {
	BufferedInputStream buffer = new BufferedInputStream(myProcess.getInputStream());
	BufferedReader  commandResult = new BufferedReader(new InputStreamReader(buffer));
	String s = null;
	try {
	  while ((s = commandResult.readLine()) != null)
	    System.out.println("Output: " + s);
	  commandResult.close();
	  if (myProcess.exitValue() != 0) {
	    if (verbose)
	      printError(commands + " -- myProcess.exitValue() != 0");
	    return(false);
	  }
	  // Ignore read errors; they mean process is done
	} catch (Exception e) {}
                
	// If you don't print the results, then you
	// need to call waitFor to stop until the process
	// is completed
      } else if (wait) {
	try {
	  System.out.println(" reading return value!");
	  int returnVal = myProcess.waitFor();
	  if (returnVal != 0) {
	    if (verbose)
	      printError(commands);
	    return(false);
	  }
	} catch (Exception e) {
	  if (verbose)
	    printError(commands, e);
	  return(false);
	}
      }
    } catch (Exception e) {
      if (verbose)
	printError(commands, e);
      return(false);
    } finally{
      finished = true;
        
      //       if (myProcess != null) {
      //         try{ InputStream is = myProcess.getErrorStream();
      //         System.out.println(" exec() error stream " + is);
      //         is.close();
      //         } catch (Exception e){      e.printStackTrace();
      //         }
      //         try{ InputStream is = myProcess.getInputStream();
      //         System.out.println(" exec() input stream " + is);
      //         is.close();
      //         } catch (Exception e){      e.printStackTrace();
      //         }
      //         try{ OutputStream os = myProcess.getOutputStream();
      //         System.out.println(" exec() output stream " + os);
      //         os.close();
      //         } catch (Exception e){      e.printStackTrace();
      //         }
      //         myProcess.destroy();
      //       }
    }
    
    return(true);
  }
    
  public boolean isFinished()
  {
    return finished;
  }
    
  protected static void printError(String command,
                                   Exception e) {
    System.out.println("Error doing exec(" +
                       command + "): " + e.getMessage());
    System.out.println("Did you specify the full " +
                       "pathname?");
  }
  
  protected static void printError(String[] commands,
                                   Exception e) {
    String mess = "";
    for(int i=0; i< commands.length; i++) 
      mess = mess + " " + commands[i];
    System.out.println("Error doing exec(" +
                       mess + "): " + e.getMessage());
    System.out.println("Did you specify the full " +
                       "pathname?");
  }
  
  protected static void printError(String[] commands) {
    String mess = "";
    for(int i=0; i< commands.length; i++) 
      mess = mess + " " + commands[i];
    System.out.println("Error doing exec(" +
                       mess + ").");
    System.out.println("Return value not 0 ");
  }

  protected static void printError(String command) {
    System.out.println("Error doing exec(" +
                       command + ").");
    System.out.println("Did you specify the full " +
                       "pathname?");
  }

  protected static void printError(String command, int value) {
    System.out.println("Error executing '" +
                       command + " " + value + "'.");
  }

  public void shutdown()
  {
    myProcess.destroy();
  }
  
}
