// -*- indent-tabs-mode: nil -*-
/*****************************************************************
This file is part of the MathServe system

Authors: Juergen Zimmer

email: jzimmer@ags.uni-sb.de
   
For information about this program, write to:                          
MathWeb-Project                                                        
AG Siekmann/FB Informatik                                            
Universitaet des Saarlandes                                          
Gebaeude E 1.1                                                     
D-66041 Saarbruecken                                                 
Germany    

For information about the newest version of MathServe, see 
http://www.ags.uni-sb.de/~jzimmer/mathserve.html

See the file "LICENSE.txt" or
http://www.ags.uni-sb.de/~jzimmer/LICENSE.html
for information on usage and redistribution
of this file, and for a DISCLAIMER OF ALL WARRANTIES.
*****************************************************************/
package org.mathweb.gen;

import java.io.*;
import java.util.*;
import java.net.*;

import org.jdom.*;
import org.jdom.input.*;
import org.jdom.output.*;
import org.jdom.Namespace;
import org.jdom.JDOMException;
import org.jdom.xpath.XPath;
import org.xml.sax.*;
import org.xml.sax.helpers.XMLFilterImpl;
import org.xml.sax.helpers.AttributesImpl;

public class OMDocJDom
{
  //******************** Namespaces ******************************
  public static final String MW_NS_STR = "http://www.mathweb.org/owl/mathweb.owl#";
  public static final String OM_NS_STR = "http://www.openmath.org/OpenMath";
  public static final String OMDOC_NS_STR = "http://www.mathweb.org/omdoc";
  public static final String STAT_NS_STR = "http://www.mathweb.org/owl/status.owl#";
  public static final String RDF_NS_STR = "http://www.w3.org/1999/02/22-rdf-syntax-ns#";
  public static final String RDFS_NS_STR = "http://www.w3.org/2001/XMLSchema#";
  
  public static final Namespace MW_NS = Namespace.getNamespace("mw", MW_NS_STR);
  public static final Namespace OMDOC_NS = Namespace.getNamespace("omdoc", OMDOC_NS_STR);
  public static final Namespace OM_NS = Namespace.getNamespace("om", OM_NS_STR);
  public static final Namespace STAT_NS = Namespace.getNamespace("stat", STAT_NS_STR);
  public static final Namespace RDF_NS = Namespace.getNamespace("rdf", RDF_NS_STR );

 //***********  XPaths  ****************************************
  public static final XPath RDF_ID = newXPath("./@rdf:ID");
//   public static final XPath RDF_ABOUT = newXPath("./@rdf:about");
//   public static final XPath XML_BASE = newXPath("./@xml:base");
//   public static final XPath FAILURE_MESSAGE = newXPath("./mw:Failure/mw:message");
//   public static final XPath TSTP_FOF_PROBLEM = newXPath("./mw:TstpFOFProblem");
//   public static final XPath TSTP_CNF_PROBLEM = newXPath("./mw:TstpCNFProblem"); 
//   public static final XPath OMDOC_PROBLEM = newXPath("./mw:OmdocProvingProblem");
//   public static final XPath PROVING_PROBLEM_TIME_RESOURCE = newXPath("./mw:time/mw:TimeResource/mw:wallClockTime");
  public static final XPath TIME_RESOURCE_WC = newXPath("./mw:TimeResource/mw:wallClockTime");
  public static final XPath TIME_RESOURCE_CPU = newXPath("./mw:TimeResource/mw:cpuTime");

  public static final XPath OMDOC_THEORY_FORMAL_DESCRIPTIONS = newXPath(".//mw:formalDescription");

//   public static final XPath OMDOC_AXIOMS = newXPath(".//omdoc:axiom");
//   public static final XPath OMDOC_AXIOMS = newXPath(".//omdoc:axiom");
//   public static final XPath OM_SYMBOLS = newXPath(".//om:OMS");
//   public static final XPath PROVING_PROBLEM_LOGIC = newXPath("./mw:inLogic");
//   public static final XPath PROVING_PROBLEM_LANGUAGE =  newXPath("/mw:language");
//   public static final XPath FO_ATP_RESULT = newXPath("./mw:FoAtpResult");
//   public static final XPath FO_ATP_RESULT_PROOF = newXPath("./mw:proof");
//   public static final XPath PROOF_FORMAL_PROOF = newXPath("./mw:formalProof");
//   public static final XPath PROOF_FORMAL_PROOFS = newXPath(".//mw:formalProof");
//   public static final XPath PROOF_PROOF_OF = newXPath("./mw:proofOf/@rdf:resource");
//   public static final XPath BRFP_REFUTATION = newXPath("./mw:TstpCnfBrFPRefutation");
//   public static final XPath STATUS = newXPath(".//mw:status/@rdf:resource");
//   public static final XPath POTENTIAL_ID = newXPath(".//@rdf:ID");

  private SAXBuilder builder;
  private XMLOutputter outputter;

  private static OMDocJDom theInstance = (new OMDocJDom()).initialize();
  
  public static OMDocJDom getInstance() {
    return theInstance;
  }
  
  public static XPath newXPath(String x)
  {
    try{
      XPath xpath =  XPath.newInstance(x);
      xpath.addNamespace(RDF_NS);
      xpath.addNamespace(MW_NS);
      xpath.addNamespace(OM_NS);
      xpath.addNamespace(OMDOC_NS);
      return xpath;
    } catch(Exception e){
      return null;
    }
  }
  
  public void OMDocJDom()
  { return;
  }
  
  public OMDocJDom initialize(){
    builder = new SAXBuilder();
    builder.setExpandEntities(false);
    Format format = Format.getPrettyFormat();
    format.setLineSeparator("\n");
    outputter = new XMLOutputter(format);
    //outputter = new XMLOutputter(" ", false);
    return this;
  }
  
  public Element getRootElement(String xml)  {
    Document doc = null;
    try {
    StringReader reader = new StringReader(xml);
    System.out.println(reader);
    doc = builder.build(reader);
    //    System.out.println(doc);
    }catch (Exception ex) {
      System.out.println(ex);   ex.printStackTrace();
     }
    return doc.getRootElement();
  }

  public Document parseLargeDocument(String xml) throws JDOMException, IOException
  {
    StringReader reader = new StringReader(xml);
    SAXBuilder oldbuilder = builder;
    builder = new SAXBuilder();
    builder.setXMLFilter(new XMLFilterImpl() {
        boolean isReadingIntoFile = false;
        File file = null;
        Writer out = null;

        public void startElement(String uri, String localName, String qName, Attributes atts) throws SAXException {
          if("formalDescription".equals(localName)) {
            isReadingIntoFile = true;
            try {
              file  = File.createTempFile("formalDescription","");
              System.out.println("externalised in " + file.getAbsolutePath());
              out = new OutputStreamWriter(new FileOutputStream(file),"utf-8");
              AttributesImpl attsNew = new AttributesImpl(atts);
              for(int i= 0; i< attsNew.getLength(); i++)
                if (attsNew.getLocalName(i).equals("datatype"))
                  attsNew.removeAttribute(i);
              attsNew.addAttribute("http://www.w3.org/1999/02/22-rdf-syntax-ns#","resource","rdf:resource",null,file.toURI().toString());
              atts = attsNew;
            } catch(IOException ex) {ex.printStackTrace();}
          }
          super.startElement(uri,localName,qName,atts);
        }
        
        public void endElement(String uri, String localName, String qName) throws SAXException {
          if(isReadingIntoFile && "formalDescription".equals(localName)) {
            try {
              out.flush();
              out.close();
              isReadingIntoFile = false;
              out = null; file = null;
            } catch(IOException ex) {ex.printStackTrace(); }
          }
          super.endElement(uri, localName, qName);    //To change body of overridden methods use File | Settings | File Templates.
            }
        
        public void characters(char ch[], int start, int length) throws SAXException {
          if(isReadingIntoFile) {
            try {
              out.write(ch,start,length);
            } catch(IOException ex) {ex.printStackTrace();}
          }
                else
                  super.characters(ch, start, length);    //To change body of overridden methods use File | Settings | File Templates.
        }
        
        public void ignorableWhitespace(char ch[], int start, int length) throws SAXException {
          if(isReadingIntoFile) {
            try {
              out.write(ch,start,length);
            } catch(IOException ex) {ex.printStackTrace();}
          }
          else
            super.ignorableWhitespace(ch, start, length);    //To change body of overridden methods use File | Settings | File Templates.
        }
      });
    
    Document doc = builder.build(reader);
    builder = oldbuilder;
    return doc;
  }
  
  public String newXMLDocument(Element root) throws JDOMException, IOException {
    StringWriter writer = new StringWriter();
    Document doc = new Document(root);
    //doc.setDocType(new DocType("RDF", RDF_NS.getURI()));
    outputter.output(doc, writer);
    return writer.toString();
  }
  
  public String valueOf(XPath xpath, Element root) {
    try {
      return xpath.valueOf(root);
    } catch(Exception e){
      e.printStackTrace();
    }
    return "";
  } 

  public Object getSingleNode(XPath xpath, Element root) {
    try {
      return xpath.selectSingleNode(root);
    } catch(Exception e){
      e.printStackTrace();
    }
    return null;
  }

  public boolean hasChild(Element node, String name, Namespace ns) {
    //    System.out.println("node:" + node);
    List cs = node.getChildren(name, ns);
    //    System.out.println("list:" + cs);
    
    return (!cs.isEmpty());
  }

  public String createTimeIndi(long cpuTimeVal,  long wcTimeVal){
    int newCpu = (new Long(cpuTimeVal)).intValue();
    int newWc = (new Long(wcTimeVal)).intValue();
    return createTimeIndi(newCpu, newWc);
  }

  public String createTimeIndi(int cpuTimeVal,  int wcTimeVal){
    try{
      Element timeRes = OMDocJDom.createTimeResource(cpuTimeVal, wcTimeVal);
      Element rdf = new Element("RDF", OMDocJDom.RDF_NS);
      rdf.addNamespaceDeclaration(OMDocJDom.MW_NS);
      rdf.setAttribute("base", OMDocJDom.MW_NS.getURI(), Namespace.XML_NAMESPACE);
      rdf.addContent(timeRes);
      return newXMLDocument(rdf);
    } catch(Exception e) {
      e.printStackTrace();
      return "Some problem occured while creating an OWL Failure: " + e.getMessage();
    }
  }
  
  public Element createTimeIndiElement(long cpuTimeVal,  long wcTimeVal){
    int newCpu = (new Long(cpuTimeVal)).intValue();
    int newWc = (new Long(wcTimeVal)).intValue();
    return OMDocJDom.createTimeResource(newCpu, newWc);
  }
  public static Element createTimeResource(int cpuTimeVal,  int wcTimeVal){
    Integer intCpuTime = new Integer(cpuTimeVal);
    Integer intWCTime = new Integer(wcTimeVal);
    //    Element time = new Element("time",OMDocJDom.MW_NS);
    Element timeres = new  Element("TimeResource",OMDocJDom.MW_NS);
    Element wcTime = new  Element("wallClockTime",OMDocJDom.MW_NS);
    Element cpuTime = new  Element("cpuTime",OMDocJDom.MW_NS);
    
    wcTime.addContent(intWCTime.toString());
    wcTime.setAttribute("datatype", OMDocJDom.RDFS_NS_STR+"integer", OMDocJDom.RDF_NS);
    
    cpuTime.addContent(intCpuTime.toString());
    cpuTime.setAttribute("datatype", OMDocJDom.RDFS_NS_STR+"integer", OMDocJDom.RDF_NS);
    
    timeres.addContent(cpuTime);
    timeres.addContent(wcTime);
    return timeres;
  }
  
  public static Element createTimeResource(long startTime, long cpuStart){
    Integer intWCTime = new Integer((new Long((System.currentTimeMillis()-startTime))).intValue());
    Integer intCpuTime = new Integer(0); //new Integer((new Long((JProf.getCurrentThreadCpuMillis() - cpuStart))).intValue());
    return createTimeResource(intCpuTime.intValue(), intWCTime.intValue());
  }
  
  public String createTheory(String omdoc) {
    try{
      Element rdf = new Element("RDF", OMDocJDom.RDF_NS);
      rdf.addNamespaceDeclaration(OMDocJDom.MW_NS);
      rdf.setAttribute("base", OMDocJDom.MW_NS.getURI(), Namespace.XML_NAMESPACE);

      Element omdocTheory = new Element("OmdocTheory", OMDocJDom.MW_NS);
      omdocTheory.setAttribute("ID", "OMDoc_Theory_" + System.currentTimeMillis(), RDF_NS);
      Element formal = new Element("formalDescription", OMDocJDom.MW_NS);
      formal.addContent(omdoc);
      omdocTheory.addContent(formal);
      rdf.addContent(omdocTheory);
      
      return newXMLDocument(rdf);
    } catch(Exception e) {
      e.printStackTrace();
      return "Some problem occured while creating an OWL OmdocTheory: " + e.getMessage();
    }
  }
  
  public String createFailure(String probID, long startTime, long cpuStart, String str){
    try{
      Element rdf = new Element("RDF",  OMDocJDom.RDF_NS);
      Element failure = new Element("Failure",OMDocJDom.MW_NS);
      Element message = new Element("message",OMDocJDom.MW_NS);
      Element time = createTimeResource(startTime, cpuStart);
      //      System.out.println("string:\n" + str);
      message.addContent(str);
      failure.addContent(message);
      rdf.addContent(failure);
      rdf.addContent(time);
      return newXMLDocument(rdf);
    } catch(Exception e) {
      e.printStackTrace();
      return "Some problem occured while creating an OWL Failure: " + e.getMessage();
    }
  }
 
  /** 
   * Reads the contens of a file in a java.lang.String.
   * @param fileName the name of the file to read.
   * @return The contents of the file.
   */
  public String readFile(String fileName){
    FileInputStream isFile  = null;
    try {
      isFile = new FileInputStream(fileName);
      int nChar;
      StringBuffer sbRead = new StringBuffer();
      nChar = isFile.read();
      while (nChar > -1) {
        sbRead.append((char) nChar);
        nChar = isFile.read();
      }
      isFile.close();
      return sbRead.toString();
    } catch (IOException ioe) {
      return null;
    } finally {
      if (isFile != null)
        try{ isFile.close(); } catch (Exception e) {}
    }
  }

  /** 
   * Writes a string into a file
   * @param fileName the name of the file to write into.
   * @param content the string to be written.
   * @return True if the file as written successfully, false otherwise.
   */
  protected boolean writeFile(String fileName, String content){
    FileOutputStream outFile = null;
    try {
      outFile = new FileOutputStream(fileName) ;
      outFile.write(content.getBytes());
      outFile.close();
      return true;
    } catch (IOException ioe) {
      return false;
    } finally {
      if (outFile != null)
        try{ outFile.close(); } catch (Exception e) {}
    }
  }
  
  /** 
   * Writes a string into a file
   * @param file the java.io.File to write into.
   * @param content the string to be written.
   * @return True if the file as written successfully, false otherwise.
   */
  protected boolean writeFile(File file, String content){
    FileOutputStream outFile = null;
    try {
      outFile = new FileOutputStream(file) ;
      outFile.write(content.getBytes());
      outFile.close();
      return true;
    } catch (IOException ioe) {
      return false;
    } finally {
      if (outFile != null)
        try{ outFile.close(); } catch (Exception e) {}
    }
  }

  public String createConfirmation(String str){
    try{
      Element rdf = new Element("RDF",  OMDocJDom.RDF_NS);
      Element conf = new Element("Confirmation",OMDocJDom.MW_NS);
      Element message = new Element("message",OMDocJDom.MW_NS);
      conf.addContent(str);
      conf.addContent(message);
      rdf.addContent(conf);
      return newXMLDocument(rdf);
      
    } catch(Exception e) {
      e.printStackTrace();
      return "Some problem occured while creating an OWL Confirmation: " + e.getMessage();
    }
  }
  
  public static void main(String argv[]) throws Exception {
   OMDocJDom omd = new OMDocJDom();
   omd.initialize();
   String omdoc = omd.readFile("C:/Temp/puz.omdoc");
   System.out.println(omdoc);
   Element root = omd.getRootElement(omdoc);
   System.out.println(root);  
   System.out.println("hi");
  }

}
      
/* Settings for Emacs.
 * Local variables:
 * c-basic-offset: 2
 * tab-width: 2
 * indent-tabs-mode: nil
 * End:
 */
