// -*- indent-tabs-mode: nil -*-
/*****************************************************************
This file is part of the MathServe system

Authors: Juergen Zimmer

email: jzimmer@ags.uni-sb.de
   
For information about this program, write to:                          
MathWeb-Project                                                        
AG Siekmann/FB Informatik                                            
Universitaet des Saarlandes                                          
Gebaeude E 1.1                                                     
D-66041 Saarbruecken                                                 
Germany    

For information about the newest version of MathServe, see 
http://www.ags.uni-sb.de/~jzimmer/mathserve.html

See the file "LICENSE.txt" or
http://www.ags.uni-sb.de/~jzimmer/LICENSE.html
for information on usage and redistribution
of this file, and for a DISCLAIMER OF ALL WARRANTIES.
*****************************************************************/
package org.mathweb.serv;

import java.util.*;

import org.mathweb.serv.*;
/** 
 * Example of a useful base class for handing agent arguments.
 * Classes which formerly extended jade.Agent may now extend this class.
 */
public class ReasoningProblem {

  public String name;
  public String header;
  
  public List inputs;
  public List outputs;
  public HashMap locals;
  public List preconds;
  public List staticPreconds;
  public List dynamicPreconds;
  public List effects;
  public List stochEffects;
  
   public ReasoningProblem()
   {
     name = "";
     inputs = new ArrayList();
     outputs = new ArrayList();
     preconds = new ArrayList();
     staticPreconds = null;
     dynamicPreconds = null; 
     effects = new ArrayList();
     stochEffects = new ArrayList();
   }
  
  public void addInput(Input i)
  {
    inputs.add(i);
  }
  
  public void addOutput(Output o){
    outputs.add(o);
  }

  public String getName(){return name;}

  public List getInputs()
  {
    return inputs;
  }

  public Input getInput(int i)
  {
    return (Input) inputs.get(i);
  }

  public Input getInput(String name)
  {
    for(int i=0; i<inputs.size(); i++){
      Input in = (Input) ((ArrayList) inputs).get(i);
      if (in.getName().equals(name))
        return in;
    }
    return null;
  }
  public List getOutputs()
  {
    return outputs;
  }

  public Output getOutput(int i)
  {
    return (Output) outputs.get(i);
  }

  public Output getOutput(String name)
  {
    for(int i=0; i<outputs.size(); i++){
      Output out = (Output) ((ArrayList) outputs).get(i);
      if (out.getName().equals(name))
        return out;
    }
    return null;
  }
  
  public List getPreconds()
  {
    return preconds;
  }
  public HashMap getLocals()
  {
    return locals;
  }

  public List getStaticPreconds(){
    return preconds;
  }
  
  public void setPreconds(ArrayList pre)
  {
    preconds = pre;
  }

  public List getEffects()
  {
    return effects;
  }

  public void setEffects(List effs )
  {
    effects = effs;
  }
  

  public List getStochasticEffects()
  {
    return stochEffects;
  }
  
  public void setStochEffects(List stochEffs)
  {
    stochEffects = stochEffs;
  }
  
  public void trim()
  {
    name = name.trim();
    header = header.trim();
//     for(int i=0; i<inputs.size(); i++){
//       ((Input)  ((ArrayList)inputs).get(i)).trim();
//     }
//     for(int i=0; i<outputs.size(); i++){
//       ((Output) ((ArrayList)outputs).get(i)).trim();
//     }  
}
  
  
  public String toString()
  {
    StringBuffer result = new StringBuffer();
    result.append(" problem `" + name + "'\n    with inputs: \n");
    for(int i=0; i<inputs.size(); i++){
      result.append("  " + inputs.get(i).toString() + "\n");
    }
    result.append("\n    outputs:\n");
    for(int i=0; i<outputs.size(); i++){
      result.append("  " + outputs.get(i).toString() + " ");
    }
    result.append("\n    preconditions: ");
    for(int i=0; i<preconds.size(); i++){
      result.append("  " + preconds.get(i).toString() + " ");
    }
    
    result.append("\n    effects: ");
    for(int i=0; i<effects.size(); i++){
      result.append("  " + effects.get(i).toString() + " ");
    }
    result.append("\n");

    return result.toString();
     
  }
  
}
