// -*- indent-tabs-mode: nil -*-
/*****************************************************************
This file is part of the MathServe system

Authors: Juergen Zimmer

email: jzimmer@ags.uni-sb.de
   
For information about this program, write to:                          
MathWeb-Project                                                        
AG Siekmann/FB Informatik                                            
Universitaet des Saarlandes                                          
Gebaeude E 1.1                                                     
D-66041 Saarbruecken                                                 
Germany    

For information about the newest version of MathServe, see 
http://www.ags.uni-sb.de/~jzimmer/mathserve.html

See the file "LICENSE.txt" or
http://www.ags.uni-sb.de/~jzimmer/LICENSE.html
for information on usage and redistribution
of this file, and for a DISCLAIMER OF ALL WARRANTIES.
*****************************************************************/
package org.mathweb.serv;

import java.net.URI;
import java.net.URL;
import java.util.List;
import java.util.ArrayList;
import java.util.Hashtable;
import java.util.HashMap;
import java.util.Vector;
import java.util.Enumeration;

import org.mathweb.serv.*;
import org.mathweb.reasoning.logic.*;
import org.mathweb.reasoning.planning.*;
//import org.apache.axis.client.Service;
import javax.xml.namespace.QName;

import org.mindswap.owls.OWLSFactory;
import org.mindswap.owls.grounding.Grounding;
import org.mindswap.owls.process.execution.ProcessExecutionListener;
import org.mindswap.owls.process.Process;
import org.mindswap.owls.process.AtomicProcess;
//import org.mindswap.owls.process.Input;
import org.mindswap.owls.process.InputList;
import org.mindswap.owls.process.OutputList;
import org.mindswap.owls.process.ConditionList;
import org.mindswap.owls.process.Condition;
import org.mindswap.query.ValueMap;
import org.mindswap.owls.profile.Profile;
import org.mindswap.owls.profile.ServiceParameter;
import org.mindswap.owls.profile.ServiceCategory;
import org.mindswap.owls.service.Service;
import org.mindswap.owls.grounding.Grounding;
import org.mindswap.owls.grounding.AtomicGrounding;
import org.mindswap.owls.grounding.WSDLAtomicGrounding;
import org.mindswap.owls.process.execution.ProcessExecutionEngine;
import org.mindswap.wsdl.WSDLOperation;
import org.mindswap.wsdl.WSDLParameter;
import org.mindswap.wsdl.WSDLService;

/**

 */
public class ReasoningService {
  
  public String name;
  public String uri;
  
  public ReasoningProblem problem;
  public TaxonomyEntry taxonomy ;
  public String directive;
  private WSDLService wsdlService;
  private String wsdlXML;
  private String operation;
  private Service owlsService = null;

  private ArrayList paramNames;
  private ArrayList newVars;
  public static int maxObjs = 1;
  
  private static URI serviceParameter = URI.create("http://www.daml.org/services/owl-s/1.0DL/Profile.owl#serviceParameter");

  private static URI onEffect = URI.create("http://www.mathweb.org/owl/mathweb.owl#onEffect");
  private static URI whenCondURI = URI.create("http://www.mathweb.org/owl/mathweb.owl#whenCondition");
  private static URI propPred = URI.create("http://www.w3.org/2003/11/swrl#propertyPredicate");
  private static URI arg1 = URI.create("http://www.w3.org/2003/11/swrl#argument1");
  private static URI arg2 = URI.create("http://www.w3.org/2003/11/swrl#argument2");
  private static Hashtable fluents = new Hashtable();
  private static int defaultReward = 0;
  private static int defaultCosts = 3000;

  private static boolean  forProbabilisticReasoning = true; 
  //  private static boolean fake = true;
  private static boolean fake = false;

  //  public Service wsdlOperation;
      
  public String getName()  { return name; }

  public String getURI(){
    return uri;
  }
  
  public void setURI(String newuri){
    uri = newuri;
  }


  public ReasoningProblem getProblem()
  {
    return problem;
  }
  public String getDirective()
  {
    return directive;
  }

  public WSDLService getWsdlService()
  {
    return wsdlService;
  }

  public String getWsdlXML()
  {
    return wsdlXML;
  }
  
  public void trim()
  {
    name = name.trim();
    directive = directive.trim();
    problem.trim();
    taxonomy.trim();
  }
    
  public TaxonomyEntry getTaxonomyEntry()
  {
    return taxonomy;
  }
  
  public void setName(String name)
  {
    this.name = name;
  }

  public void setOwlsService(Service serv)
  {
    this.owlsService = serv;
  }

  public Service getOwlsService()
  {
    return this.owlsService;
  }

  public void setWsdlGrounding(String wsdlFileName, String operation)
  {
    try{
      this.wsdlService = WSDLService.createService("file:" + wsdlFileName);
      this.operation = operation;
    } catch(Exception e){
      System.out.println(e);
      e.printStackTrace();
      return;
    }
    
  }

  public void setWsdlXML(String wsdl)
  {
    this.wsdlXML = wsdl;
  }
  
  
  public ValueMap invokeOWLS(Hashtable map) throws Exception
  {
    ProcessExecutionEngine exec = OWLSFactory.createExecutionEngine();
//     exec.addExecutionListener(new ProcessExecutionListener() {
//         String TAB = "   ";
//         String indent = "   ";
         
//         public void setCurrentExecuteService(Process p) {
//           System.out.println(indent + "Start executing process " + p);
//           indent += TAB;
//         }
        
//         public void printMessage(String message) {
//           if(message.equals("[DONE]"))
//             indent = indent.substring(0, indent.length()-TAB.length());
//           System.out.println(indent + message);
//         }

//         public void finishExecution(int retCode) {
//           System.out.println("Finished execution " + 
//                              ((retCode == ProcessExecutionListener.EXEC_ERROR)
//                               ? "with errors" : "successfully"));				
//         }			
//       });		

    AtomicProcess process = (AtomicProcess) owlsService.getProcess();
    //    System.out.println("process: " + process);
    WSDLAtomicGrounding atomic = (WSDLAtomicGrounding) process.getGrounding();
//     System.out.println("grounding: " + atomic);
//     System.out.println("grounding: " +atomic.getClass().getName());
//     System.out.println("op: " + atomic.getOperation());
    Hashtable results = new Hashtable();
//     List groundings = ground.getAtomicGroundings();
//     if (groundings.isEmpty())
//       return results;
//     AtomicGrounding atomic = (AtomicGrounding) groundings.get(0);
//     Process process = owlsService.getProcess();
// new for 1.1.0 
    ValueMap vMap = new ValueMap();
    //    ValueMap vMap = OWLSFactory.createValueMap();
    
    for(Enumeration e= map.keys(); e.hasMoreElements(); ){
      String iName = (String) e.nextElement();
      
      //      System.out.println("input: " +process.getInputs().getParameter(iName) +" --> " + map.get(iName));
      
      //    URI uri = owlsService.getURI();
//       URI paramUri = new URI(uri.getScheme(),
//                              uri.getSchemeSpecificPart(),
//                              iName);
      //      System.out.println("param: "+  paramUri);
      //       Parameter in = process.getParameter(paramUri);
      //       System.out.println("mapping: " + in + " -> " + map.get(iName));
      //       vMap.setValue(in, map.get(iName));
      
// new for 1.1.0 
      vMap.setDataValue(process.getInputs().getParameter(iName), ((String) map.get(iName)));		
//      vMap.setValue(process.getInputs().getParameter(iName), map.get(iName));		
    }
    ValueMap resMap= null;
    try {
       resMap = exec.execute(process, vMap);	
    } catch(Exception e){
      System.out.println(e);
      e.printStackTrace();
    }
    //    System.out.println("resultMap: " + resMap);
    return resMap;
  }
  
  public ValueMap invoke(Hashtable map) throws Exception
  {
    if (owlsService != null)
      return invokeOWLS(map);
// new for 1.1.0 
    ValueMap vMap = new ValueMap();
//    ValueMap vMap = OWLSFactory.createValueMap();
    
    List ops = wsdlService.getOperations();
    //    System.out.println("ops: " + ops);
    
    Hashtable results = new Hashtable();
    if (fake){
      List outputs = problem.getOutputs();
      for(int i = 0; i < outputs.size(); i++) {		
        Output servOut = (Output) outputs.get(i);
        results.put(servOut.getType(), "fake");		
      }
    }
    else{
      WSDLOperation op = wsdlService.getOperation(operation);
      if(op == null) op = wsdlService.getOperation(getLocalName(operation));
      
      for(int i = 0; i < op.getInputs().size(); i++) {		
        WSDLParameter opIn = op.getInput(i);
        Input servIn = problem.getInput(i);
        Object value = map.get(servIn.getName());
        opIn.setValue(value);
      }
      op.invoke();
      
      for(int i = 0; i < op.getOutputs().size(); i++) {		
        WSDLParameter opOut = op.getOutput(i);
        String outputValue = opOut.getValue().toString();
        Output servOut = problem.getOutput(i);
        results.put(servOut.getType(), outputValue);		
      }
    }
    return vMap ;
  }

  public static String getLocalName(URI uri) {
    return getLocalName(uri.toString());
  }
  public static String getLocalName(String uri) {
    int index = splitPos(uri);
    
    if(index == -1) return null;
    
    return uri.substring(index + 1);
  }
  private static int splitPos(String uri) {
    int pos = uri.indexOf("#");
    
    if(pos == -1) pos = uri.lastIndexOf("/");
    
    return pos; 
  }
  
  public void setProblem(ReasoningProblem prob)
  {
    problem = prob;
  }

  public void shutdown()
  {
    try{
      List ops = wsdlService.getOperations();
      //      System.out.println("ops: " + ops);
      WSDLOperation op = wsdlService.getOperation("shutdown");
      op.invoke();
    }catch(Exception e){
      System.out.println("shutdown error: " + e);
      e.printStackTrace();
    }
  }
  
  
  /* ==============================================================*/
  /*      Translation methods for GOLOG domain descriptions        */
  /* ==============================================================*/

  /** 
   * @param services A list of ReasoningServices
   * @return A DTGolog domain representing all the given services, including senseCondition and successor-state axioms. 
   * Stochastic effects are translated into probability statements of nondeterministic DTGolog actions. 
   */
  public static String computeDTGologDomain(List services)
  {
    StringBuffer result = new StringBuffer();
    fluents = new Hashtable();
    List allEffects = new ArrayList();
    for(int k=0; k < services.size(); k++){
      ReasoningService serv = (ReasoningService) services.get(k);
      ReasoningProblem problem = serv.getProblem();
      List inputs = problem.getInputs();
      List outputs = problem.getOutputs();
      ArrayList params = new ArrayList();
      params.addAll(inputs);
      params.addAll(outputs);
      List effects = problem.getEffects();
      
      for(int e=0; e < effects.size(); e++){
        Effect eff = (Effect) effects.get(e);
        CachedHashCodeObject condition = eff.getCondition();
        if (condition instanceof Literal) {
          Literal lit = (Literal) condition;
          String predicate = lit.getPredicate().getName();
          String litStr = getGologLitString(lit);
          String abstrLitStr = getAbstractGologLitString(lit);
          String detActStr = actionString(serv.getURI(), params);
          result.append(litStr.substring(0, (litStr.length() - 1))  + ", do(A, S)) :-\n   A = " + detActStr + ".\n\n");
          if (!predicate.equals("equal"))
            if (!fluents.containsKey(abstrLitStr)){
              List servs = new ArrayList();
              servs.add(serv);
              fluents.put(abstrLitStr, servs);
            } else {
              List servs = (List) fluents.get(abstrLitStr);
              servs.add(serv);
            }
        } else if (condition instanceof Conjunct){
          Conjunct conj = (Conjunct) condition;
          List lits = conj.getLiterals(); 
          for(int i=0; i<lits.size(); i++){
            Literal lit = (Literal) lits.get(i);
            String predicate = lit.getPredicate().getName();
            String abstrLitStr = getAbstractGologLitString(lit);
            if (!predicate.equals("equal"))
              if (!fluents.containsKey(abstrLitStr)){
                List servs = new ArrayList();
                servs.add(serv);
                fluents.put(abstrLitStr, servs);
              } else {
                List servs = (List) fluents.get(abstrLitStr);
                servs.add(serv);
              }
          }
        } else if (condition instanceof Disjunct){
          Disjunct disj = (Disjunct) condition;
          List literals = disj.getLiterals();
          for(int l=0; l<literals.size(); l++){
            Literal lit = (Literal) literals.get(l);
            String predicate = lit.getPredicate().getName();
            String abstrLitStr = getAbstractGologLitString(lit);
            if (!predicate.equals("equal"))
              if (!fluents.containsKey(abstrLitStr)){
                List servs = new ArrayList();
                servs.add(serv);
                fluents.put(abstrLitStr, servs);
              } else {
                List servs = (List) fluents.get(abstrLitStr);
                servs.add(serv);
              }
          }
        }
      }
    }
    
    //    System.out.println(fluents);

    for(Enumeration en = fluents.keys(); en.hasMoreElements();){
      String fluent = (String) en.nextElement();
      //      System.out.println("fluent: " + fluent);
      result.append("restoreSitArg(" + fluent + ", S, " + fluent.substring(0, (fluent.length() - 1)) + ", S)).\n");
    }
    
    for(int k=0; k < services.size(); k++){
      ReasoningService serv = (ReasoningService) services.get(k);
      result.append(serv.toDTGolog());
    }

    for(Enumeration en = fluents.keys(); en.hasMoreElements();){
      String fluent = (String) en.nextElement();
      result.append(fluent.substring(0, (fluent.length() - 1)) + ", do(A, S)) :- " + fluent.substring(0, (fluent.length() - 1)) + ", S).\n");
    }
    result.append("myReward(A, " + defaultReward + ").\n\n");
    
    return result.toString();
    }
    
  
  /** 
   * @return A DTGolog representation of this reasoning service including senseCondition and successor-state axioms. 
   * Stochastic effects are translated into probability statements of nondeterministic DTGolog actions. 
   */
  public String toDTGolog(){
    List inputs = problem.getInputs();
    List outputs = problem.getOutputs();
    ArrayList params = new ArrayList();
    params.addAll(inputs);
    params.addAll(outputs);
    
    HashMap locals = problem.getLocals();
    List preconds = problem.getPreconds();
    List effects = problem.getEffects(); 
    List stochEffects = problem.getStochasticEffects();

    /* maps string representation of literals of classical effects to the literals themselves  */
    Hashtable detLiterals = new Hashtable();

    /* maps string representation of literals of disjunctive or stochastic effects to the literals themselves  */
    Hashtable nonDetLiterals = new Hashtable();
    
    String result = "";
    String detActStr = actionString(getURI(), params);

    
    try{
    String senseConditions = "";
    String succStateAxioms = "";

    /* First look for effects of the service that are disjunctions which identify the service as a nondeterministic action. 
       Collect all literals of all disjunctions in the nonDetLiterals list. */
    for(int k=0; k < effects.size(); k++){
      //      System.out.println("post: " + postconds.get(i).getClass().getName());
      Effect eff = (Effect) effects.get(k);
      CachedHashCodeObject condition = eff.getCondition();
      if (condition instanceof Disjunct){
        Disjunct disj = (Disjunct) condition;
        List literals = disj.getLiterals();
        for(int l=0; l<literals.size(); l++){
          Literal lit = (Literal) literals.get(l);
          String litString = getGologLitString(lit);          
          //          System.out.println("disj literal: " + litString);
          nonDetLiterals.put(litString,  lit);
        }
      } else if (condition instanceof Literal){
        Literal lit = (Literal) condition;
        String litStr = getGologLitString(lit);          
        detLiterals.put(litStr, lit);
        senseConditions += "senseCondition(" + detActStr + ", Phi) :- Phi = " + litStr + ".\n";
      } else if (condition instanceof Conjunct){
        Conjunct conj = (Conjunct) condition;
        List literals = conj.getLiterals();
        for(int l=0; l<literals.size(); l++){
          Literal lit = (Literal) literals.get(l);
          String litStr = getGologLitString(lit);          
          detLiterals.put(litStr, lit);
          //          System.out.println("disj literal: " + litString);
          senseConditions += "senseCondition(" + detActStr + ", Phi) :- Phi = " + litStr + ".\n";
        }
      }
    } 
    result += "\n";
       
    /* maps literals (string representation) to action terms */
    Hashtable nonDetActions = new Hashtable();
    
    /* First we introduce this service as a new agent action */
    result += "agentAction(" + actionString(getURI(), params) + ").\n\n";

    /* Then we look for all the preconditions that 'match' any fluent. 
       These are the only preconditions that might become true 
       during the execution of the DTGolog program, we produce 'poss(A,S)' statements for them. 
       All other preconditions are ignored. They should have been dealt 
       with during classical planning. */
    String possStr = "poss(" + detActStr + ", S) :-\n";
    for(int m=0; m < preconds.size(); m++){
      PreCondition pre = (PreCondition) preconds.get(m);
      CachedHashCodeObject condition = pre.getCondition();
      if (condition instanceof Literal) {
        Literal lit = (Literal) condition;
        String abstrLitStr = getAbstractGologLitString(lit);
        if (fluents.containsKey(abstrLitStr))
          possStr += "   holds(" + getGologLitString(lit) + ",  S),\n";
      } else if (condition instanceof Conjunct){
        Conjunct conj = (Conjunct) condition;
        List lits = conj.getLiterals(); 
        for(int i=0; i<lits.size(); i++){
          Literal lit = (Literal) lits.get(i);
          String abstrLitStr = getAbstractGologLitString(lit);
          if (fluents.containsKey(abstrLitStr))
            possStr += "   holds(" + getGologLitString(lit) + ",  S),\n";
        }
      }
    }
    possStr += "   true.\n\n";
    result += possStr;
    
    /* For every literal we define a new nondeterministic DTGolog action. We also add 'senseCondition' statements for
       those nondet. actions.*/
    int nonDetCount = 0;
    String nonDetActionsString = "";

    for(Enumeration en = nonDetLiterals.keys(); en.hasMoreElements();){
      String litStr = (String) en.nextElement();
      String sensedStr = null;
      
      Literal lit = (Literal) nonDetLiterals.get(litStr);
      String nonDetPredicate = lit.getPredicate().getName();
      String actName = "" + getURI()+nonDetCount;
      String nonDetActStr = actionString(actName, params);
      List argVars = new ArrayList();
      
      nonDetActions.put(litStr, nonDetActStr);
      
      /* translate service IO parameters into Variables or Constants of our logic language */
      for(int i=0; i < params.size(); i++){
        Parameter par = (Parameter) params.get(i);
        /* Service IO parameters are always variables */
        argVars.add(new Variable(par.getName()));
      }

      /* For every nondeterministic DTGolog action we create successor state axioms. */
      for(Enumeration detEn = detLiterals.keys(); detEn.hasMoreElements();){
        String detLitStr = (String) detEn.nextElement();
        Literal detLit = (Literal) detLiterals.get(detLitStr);
        //        System.out.println(detLitO + " " + detLitO.getClass().getName());
        
        String succStateAx = detLitStr.substring(0, (detLitStr.length() - 1))  + ", do(A, S)) :-\n   A = " + nonDetActStr + ".\n\n";

        /* if we deal with equality between a variable and a constant, check for a determinstic effect that contains that 
           variable as a second argument. Then use that effect as the conditions sensed together with the equality */
        if (nonDetPredicate.equals("equal")) {
          //          System.out.println("equality case " + detLitStr + " " +litStr);
          
          Argument[] detArgs =  { detLit.getArgument(0), detLit.getArgument(1)};
          Argument[] nonDetArgs = { lit.getArgument(0), lit.getArgument(1)};
          
          if ((nonDetArgs[0] instanceof Variable) && (detArgs[1] instanceof Variable))
            if(nonDetArgs[0].getName().equals(detArgs[1].getName())){
              succStateAx = detLitStr.substring(0, (detLitStr.length() - 1))  + ", do(A, S)) :-\n   A = " + nonDetActStr + 
                ", " + litStr + ".\n\n";
              if (sensedStr == null) 
                sensedStr = detLitStr; // + ", " + litStr;
              for(int g=0; g < argVars.size(); g++){
                Argument arg = (Argument) argVars.get(g);
                if ((arg instanceof Variable) && (arg.getName().equals(nonDetArgs[0].getName()))){
                  argVars.set(g, nonDetArgs[1]);
                  g = argVars.size();
                }
              }
            }
        }
        succStateAxioms += succStateAx;
      }
      
      if (sensedStr == null)
        sensedStr = litStr;
      
      senseConditions += "senseCondition(" + nonDetActStr + ", Phi) :- Phi = " + sensedStr + ".\n";

      String newActStr = actionStringWithArguments(actName, argVars);
      nonDetActionsString += newActStr;
      
      //      System.out.println("action " + newActStr);
      if (en.hasMoreElements())
        nonDetActionsString += ",";
      nonDetActionsString += "\n";
      
      nonDetCount++;
    }
    if (nonDetCount > 0)
      result += "nondetActions(" + actionString(getURI(), params) + ", S,\n [" + nonDetActionsString + 
        " ]).\n\n" + senseConditions + "\n"  + succStateAxioms + "\n";;
    
    
    /* A hashtable which maps literals to a list of all the stochastic effects belonging to that literal */
    Hashtable stochEffsForLit = new Hashtable();
    
    /* Define the default cost and reward for actions */
    String costsString = "cost(" + detActStr + ", Cost, S) :-\n";
    String rewardsString = "";

    /* For every nondeterministic DTGolog action we try to find all stochastic effects whose effect literal is 
       'sensed' by the nondeterministic action */
    for(Enumeration en = nonDetLiterals.keys(); en.hasMoreElements();){
      String litStr = (String) en.nextElement();
      
      for(int k=0; k < stochEffects.size() ; k++){
        StochasticEffect se = (StochasticEffect) stochEffects.get(k);
                
        Effect eff = se.getEffect();
        Literal effLit = (Literal) eff.getCondition();        
        String effLitStr = getGologLitString(effLit);
        if (effLitStr.equals(litStr)) {
          if(stochEffsForLit.containsKey(litStr)){
            List oldLits = (List) stochEffsForLit.get(litStr);
            oldLits.add(se);
          } else {
            List newLits = new ArrayList();
            newLits.add(se);
            stochEffsForLit.put(litStr, newLits);
          }
        }
      }
      
      List stochEffs = (List) stochEffsForLit.get(litStr);
      //      System.out.println(stochEffs);
      
      /* There are stochastic effects for this literal */
      if(stochEffs != null && stochEffs.size() > 0) {
        StochasticEffect se = (StochasticEffect) stochEffs.get(0);
        rewardsString += "myReward(" + nonDetActions.get(litStr) + ", " + se.getReward() + ").\n";
      }else{
        if (stochEffects.isEmpty())
          rewardsString += "myReward(" + nonDetActions.get(litStr) + ", " + defaultReward + ").\n";
        else
          rewardsString += "myReward(" + nonDetActions.get(litStr) + ", 0).\n";
      }
      

      /* For every nondeterministic DTGolog action we translate the corresponding stochastic effects into 
         probability and cost statements for DTGolog  */
      /*        
        prob('http://www.mathweb.org/owl/atp/OtterService.owl3'(X, Time,Y),Pr3,S) :- 
	prob('http://www.mathweb.org/owl/atp/OtterService.owl1'(X, Time,Y),Pr1, S), 
	prob('http://www.mathweb.org/owl/atp/OtterService.owl2'(X, Time,Y),Pr2, S), 
          (Pr3 is 1.0 - (Pr1 + Pr2)).
      */
      
      if(en.hasMoreElements()){
        result += "prob(" + nonDetActions.get(litStr) + ", Pr, S) :-\n";

        /* Create probability clauses for all stochastic effects for this literal */
        if(stochEffs != null && stochEffs.size() > 0) {
          for(int i=0; i< stochEffs.size(); i++){
            StochasticEffect se = (StochasticEffect) stochEffs.get(i);
            PreCondition pre = se.getCondition();
            Literal condLit = (Literal) pre.getCondition();
            String condLitStr = getGologLitString(condLit);
            result += "   holds(" + condLitStr + ", S), ( Pr is " + se.getProbability() + ") ;\n";
            costsString += "   holds(" + condLitStr + ", S), ( Cost is " + se.getCost() + ") ;\n";
          }
          result += "   ( Pr is 0.0 ).\n\n";
          /* There are NO stochastic effects for this literal */
        } else {
          result += "   ( Pr is " + (1.0/nonDetCount) + " ).\n\n";
        }
        //        costsString += "cost(" + detActStr + ", Cost, S) :-\n";
      } else {
        result += "prob(" + nonDetActions.get(litStr) + ", RestPr, S) :-\n";
        int count = 0;
        String appendix = "   (RestPr is 1.0 - (";
        for(Enumeration actEn = nonDetActions.keys(); actEn.hasMoreElements();){
          String oldAct = (String) actEn.nextElement();
          if(!oldAct.equals(litStr)){
            result += "   prob(" + nonDetActions.get(oldAct) + ", Pr" + count + ", S),\n";
            appendix += "Pr" + count++ + " + ";
          }
        }
        result += appendix + "0.0)).\n\n";
      } // if
    } //  for(Enumeration en = nonDetLiterals.keys(); en.hasMoreElements();)

    if (stochEffects.isEmpty())
      costsString += "   (Cost is " + defaultCosts + ").\n\n";
    else 
      costsString += "   (Cost is 0 ).\n\n";
    
    result += costsString;
    result += rewardsString;

    /* For every nondeterministic DTGolog action we try to find all stochastic effects whose effect literal is 
       'sensed' by the nondeterministic action */
    for(Enumeration en = nonDetLiterals.keys(); en.hasMoreElements();){
      String litStr = (String) en.nextElement();
      Literal lit = (Literal) nonDetLiterals.get(litStr);
      
      String predicate = lit.getPredicate().getName();      
      if (!predicate.equals("equal")){
        String sitLitStr = getGologSituatedLitString(lit, "do(A,S)");
        result += sitLitStr + " :-\n   A = " + nonDetActions.get(litStr) + " ;\n";
        result += "   " + getGologSituatedLitString(lit, "S") + ".\n";
      } //for     
    }
    
    //    System.out.println(stochEffsForLit);
    } catch(Exception e){
      e.printStackTrace();
    }
    
    // nondetActions('http://www.mathweb.org/owl/atp/OtterService.owl'(TstpProblem, Time, ATPResult), S, 
//                   [
// 'http://www.mathweb.org/owl/atp/OtterService.owl1'(TstpProblem, Time, ATPResult), 
// 'http://www.mathweb.org/owl/atp/OtterService.owl2'(TstpProblem, Time, ATPResult), 
// 'http://www.mathweb.org/owl/atp/OtterService.owl3'(TstpProblem, Time, ATPResult)]).
    

    return result.toString();
  }
  
  public static String actionString(String actName, List params)
  {
    String result = "'" + actName + "'(";
    for(int i=0; i < params.size(); i++){
      Parameter par = (Parameter) params.get(i);
      result += par.getName().toUpperCase();
      if (i < params.size() -1)
        result += ", ";
    }
    result += ")";
    return result ;
  }  

  /*  Computes a Prolog term representing an service as an agent action with parameters given as logical Variables or Constants
   * @param actName The name of the action
   * @param argNames A list of argument names as instances of class Argument.
   * @return The Prolog term representing the action
   */
  public static String actionStringWithArguments(String actName, List args)
  {
    String result = "'" + actName + "'(";
    for(int i=0; i < args.size(); i++){
      Argument arg = (Argument) args.get(i);
      if (arg instanceof Variable)
        result += arg.getName().toUpperCase();
      else 
        result += "'" + arg.getName() + "'";
      if (i < args.size() -1)
        result += ", ";
    }
    result += ")";
    return result ;
  }  
  
  /*  Computes a Prolog term representing an service as an agent action with parameters given as strings
   * Note: this method is used by the ServicePlanner class. 
   * @param actName The name of the action
   * @param argNames A list of argument names as Strings.
   * @return The Prolog term representing the action
   */
  public static String actionStringWithArgnames(String actName, List argNames)
  {
    String result = "'" + actName + "'(";
    for(int i=0; i < argNames.size(); i++){
      String arg = (String) argNames.get(i);
      result += arg.toUpperCase();
      if (i < argNames.size() -1)
        result += ", ";
    }
    result += ")";
    return result ;
  }  
  
  private static String getGologLitString(Literal lit){
    Argument[] args = { lit.getArgument(0), lit.getArgument(1)};
    String an0, an1;
    String parName0, parName1;
    if (args[0] instanceof Variable)
      parName0 = args[0].getName().toUpperCase();
    else parName0 = "'" + args[0].getName() + "'";
    if (args[1] instanceof Variable)
      parName1 = args[1].getName().toUpperCase();
    else parName1 = "'" + args[1].getName() + "'";
    String predicate = lit.getPredicate().getName();
    if (predicate.equals("equal"))
      return "(" + parName0 + " = "+parName1+")";
    else
      return predicate + "(" + parName0 + ", "+parName1+")";
  }
  
  /* needed for restoreSitArg(status(X,Y),S,status(X,Y, S)).  */
  private static String getAbstractGologLitString(Literal lit){
    return lit.getPredicate().getName() + "(X, Y)";
  }

  private static String getAbstractSituatedGologLitString(Literal lit, String sitTerm){
    return lit.getPredicate().getName() + "(X, Y, " + sitTerm + ")";
  }

  private String getGologSituatedLitString(Literal lit, String sitTerm){
    Argument[] args = { lit.getArgument(0), lit.getArgument(1)};
    String an0, an1;
    String parName0, parName1;
    if (args[0] instanceof Variable)
      parName0 = args[0].getName().toUpperCase();
    else parName0 = "'" + args[0].getName() + "'";
    if (args[1] instanceof Variable)
      parName1 = args[1].getName().toUpperCase();
    else parName1 = "'" + args[1].getName() + "'";

    return lit.getPredicate().getName() + "(" + parName0 + ", "+parName1+", " + sitTerm + ")";
  }
  

  /* ==============================================================*/
  /*      Translation methods for PRODIGY Planning domains         */
  /* ==============================================================*/
  public String toProdigyOperator()
  {
    return toProdigyOperator(1);
  }
  

  public String toProdigyOperator(int objectCounter)
  {
    List inputs = problem.getInputs();
    List outputs = problem.getOutputs();
    HashMap locals = problem.getLocals();
    List preconds = problem.getPreconds();
    List postconds= problem.getEffects(); 
    List stochEffects = problem.getStochasticEffects();
    String params = "";
    String precondsDecl = "";
    String precondsStr = "";
    String effectsDecl = "";
    String effects = "";
    String rejectRule = "";
    String rejectRuleAppend = "";
    String objects = "";
    String opName = getURI().substring(7);
    
    String newName = ((name.replace('(', '_')).replace(')', '_')).replace(' ', '_');
    
    this.paramNames = new ArrayList();
    rejectRule += "(Control-Rule reject-wrong-binding-for-";
    rejectRule += newName;
    rejectRule += "\n  (if (and\n        (applicable-operator (";
    rejectRule += opName;
    rejectRuleAppend += "        (or\n ";
    
    /*
      all inputs have to be bound.
     */
    for(int i=0; i<inputs.size(); i++){
      //      System.out.println("in: " + inputs.get(i).getClass().getName());
      
      Input e = (Input) inputs.get(i);
      params = params + " <"+e.getName() + ">";
      precondsDecl = precondsDecl + "  (<" +e.getName() + "> OWL-" +e.getType() + ")\n";
      precondsStr += "  (bound <" +e.getName() + ">)\n";
      rejectRule += " <"+e.getName() + "> ";
      paramNames.add(e.getName().toLowerCase());
    }

    if (forProbabilisticReasoning) {
      /* HACK: TODO: all antecedents of stochastic effects are put into the precondition list of the action */
      Hashtable newPreconds = new Hashtable();
      int preCounter = 1;
    
      for(int k=0; k < stochEffects.size() ; k++){
        StochasticEffect se = (StochasticEffect) stochEffects.get(k);
      
        PreCondition pre = se.getCondition();
        Literal preLit = (Literal) pre.getCondition();        
        Argument[] args = { preLit.getArgument(0), preLit.getArgument(1)};
      
        String predicate = preLit.getPredicate().getName();
      
        //        System.out.println(preLit);
        //        System.out.println(newPreconds);
        if (!newPreconds.containsKey(predicate)){
          newPreconds.put(predicate, "foo");
          String argName0, argName1;
          if (args[0] instanceof Constant) {
            argName0 = "x" + preCounter;
            precondsDecl = precondsDecl + "  (<" + argName0 + "> OWL-Thing)\n";
            preCounter ++;
          } else {
            argName0 = args[0].getName();
          }
        
          if (args[1] instanceof Constant) {
            argName1 = "x" + preCounter;
            precondsDecl = precondsDecl + "  (<" + argName1 + "> OWL-Thing)\n";
            preCounter ++;
          } else {
            argName1 = args[1].getName();
          }
          precondsStr += "  (" + predicate + " <"+ argName0 + "> <" + argName1 + ">)\n";
        }
      }
    }
    
    /*
      all output parameters have to be NOT bound before the operator application.
    */
    for(int i=0; i<outputs.size(); i++){
      //      System.out.println("out: " + outputs.get(i).getClass().getName());
      
      Output e = (Output) outputs.get(i);
      String owlType = " OWL-" +e.getType();
      params = params + " <"+e.getName() + ">";
      precondsDecl = precondsDecl + "  (<" +e.getName() + ">" + owlType + ")\n";
      precondsStr += "  (~ (bound <" +e.getName() + ">))\n";
      rejectRule += " <"+e.getName() + ">";
      rejectRuleAppend += "        (~ (exact-type-of-object <" + e.getName() + ">" + owlType + "))\n";

      //      effects += "(del (unbound <" +e.getName() + ">))\n";
      for(int x=0; x< maxObjs; x++)
        objects += "(pinstance-of object_"+ e.getName() + "_" + objectCounter++ + owlType + ")\n";
      effects += "   (add (bound <" +e.getName() + ">))\n";
      paramNames.add(e.getName().toLowerCase());
    }
    rejectRule += "))\n";
    rejectRuleAppend += "         )))\n  (then sub-goal))\n";
    
    newVars= new ArrayList();
    
    for(int k=0; k<preconds.size(); k++){
      //      System.out.println("pre: " + preconds.get(i).getClass().getName());
      PreCondition pre = (PreCondition) preconds.get(k);
      CachedHashCodeObject condition = pre.getCondition();
      //      System.out.println("pre: " + condition);
      ArrayList strs = getCondStrings(condition, paramNames);
      for (int s=0; s<strs.size(); s++) {
        if (strs.get(s) != null) {
          String str = (String) strs.get(s);
          precondsStr += "  " + str + "\n";
        }
      }
    }

    /*
      free variables (not bound in precondDecls) have to be introduced. Since the type is not known, we choose OWL-Thing
    */
    for(int l=0; l<newVars.size(); l++){
      String type = "OWL-Thing";
      String name = (String) newVars.get(l);
      if (locals.containsKey(name)){
        type = "OWL-" + ((String) locals.get(name));
      }
      precondsDecl += "  (<" +name + "> " + type + ")\n";
    }
    ArrayList oldVars = new ArrayList(newVars);
    
    newVars= new ArrayList();
    for(int k=0; k<postconds.size(); k++){
      //      System.out.println("post: " + postconds.get(i).getClass().getName());
      Effect po = (Effect) postconds.get(k);
      CachedHashCodeObject condition = po.getCondition();
      //      System.out.println("eff: " + condition);
      ArrayList strs = getCondStrings(condition, paramNames);
      
      for (int s=0; s<strs.size(); s++) { 
        if (strs.get(s) != null) {
          String str = (String) strs.get(s);
          
          if (condition instanceof Implication){
            effects += "   "+ str + "\n";
          } else{
            effects += "   (add "+ str + ")\n";          
          }
        }
      }
    }

    for(int l=0; l<newVars.size(); l++){
      String name = (String) newVars.get(l);
      if (!oldVars.contains(name))
        effectsDecl += "(<" +name + "> OWL-Thing)\n";
    }
    
    StringBuffer buf = new StringBuffer();
    buf.append("(OPERATOR ");
    buf.append(opName + "\n");
    buf.append("(params " + params + ")\n");
    buf.append("(preconds \n (\n"+precondsDecl+ " )\n (and\n" + precondsStr + " ))\n");
    buf.append("(effects \n ("+effectsDecl+ " )\n (\n" + effects + ")))\n\n");

    buf.append(objects);
    buf.append("\n\n");
    
    buf.append(rejectRule);
    buf.append(rejectRuleAppend);
    
    
    // (Control-Rule reject-wrong-binding-otter
    // 	      (if (and
    // 		   (applicable-operator (OtterATP <x> <result>))
    // 		   (~ (exact-type-of-object <result> OWL-FOBrFPResult))
    // 		   ))
    //               (then sub-goal))

    return buf.toString();
  }

  private String getLitString(Literal lit, List paramNames){
    // check whether the both arguments of the literal are variables and whether they 
    // are the inputs of the service. 
    Argument[] args = { lit.getArgument(0), lit.getArgument(1)};
    String an0, an1;
    if ((args[0] instanceof Variable) && (args[1] instanceof Variable)){
      String varName0 = args[0].getName().toLowerCase();
      String varName1 = args[1].getName().toLowerCase();
      String type = null; 
      //      System.out.println("params: " + varName0 + " " + varName1 + " " + paramNames);
     if (!paramNames.contains(varName0)) {
        Variable v0 = (Variable) args[0];
        type = v0.getType();
        if (!newVars.contains(varName0))
          newVars.add(varName0);
      }
      if (!paramNames.contains(varName1)){
        Variable v1 = (Variable) args[1];
        type = v1.getType();
        if (!newVars.contains(varName1))
          newVars.add(varName1);
      }
      return "(" + lit.getPredicate().getName() + " <"+ varName0 + "> <"+varName1+">)";
    }
    return null;
  }

  private ArrayList getCondStrings(CachedHashCodeObject condition, List paramNames){
    ArrayList condStrs = new ArrayList();
    
    if (condition instanceof Literal) {
      Literal lit = (Literal) condition;
      condStrs.add(getLitString(lit, paramNames));
    }
    else if (condition instanceof Conjunct){
      Conjunct conj = (Conjunct) condition;
      List lits = conj.getLiterals(); 
      for(int i=0; i<lits.size(); i++){
        Literal lit = (Literal) lits.get(i);
        // check whether the both arguments of the literal are variables and whether they 
        // are the inputs of the service. If so, then add it to the list of
        // literals. Otherwise omit it.
        condStrs.add(getLitString(lit, paramNames));
      }
    }
    else if (condition instanceof Implication){
      Implication imp = (Implication) condition;
      List lits = imp.getBody();
      Literal head = imp.getHead();
      String impStr = "(if (and\n";
      for(int i=0; i<lits.size(); i++){
        Literal lit = (Literal) lits.get(i);
        String litStr = getLitString(lit, paramNames);
        impStr += "        " + litStr + "\n";
      }
      impStr += "       )\n       ((add " + getLitString(head, paramNames) + ")))";
      condStrs.add(impStr);
    }
    return condStrs;
  }
  
  
  public String toString()
  {
    String result = "service `" + name + "' with directive "+ directive;
    if (problem != null) { result = result + " solves: \n" + problem.toString(); }
    return result;
  }

  public Action toPlanningOperator()
  {
    List inputs = problem.getInputs();
    List outputs = problem.getOutputs();
    List preconds= problem.getPreconds();
    List postconds= problem.getEffects(); 

    List inputNames = new ArrayList();
    
    ArrayList ls = new ArrayList(); // list of literals
    for(int i=0; i<inputs.size(); i++){
      //      System.out.println("in: " + inputs.get(i).getClass().getName());
      
      Input e = (Input) inputs.get(i);
      Predicate p = new Predicate("type");
      Argument[] args = { new org.mathweb.reasoning.logic.Variable(e.getName()),
                          new org.mathweb.reasoning.logic.Constant(e.getType())};
      ls.add(new org.mathweb.reasoning.logic.Literal(p, args, true));
      inputNames.add(e.getName().toLowerCase());
    }

    for(int i=0; i<preconds.size(); i++){
      //      System.out.println("pre: " + preconds.get(i).getClass().getName());
      PreCondition pre = (PreCondition) preconds.get(i);
      List l = pre.toPlanningLiterals(inputNames);
      ls.addAll(l);
    }
    org.mathweb.reasoning.logic.Literal[] pre= new org.mathweb.reasoning.logic.Literal[ls.size()];
    Object[] ps = ls.toArray();
    for(int i=0; i<ps.length; i++){
      //      System.out.println("lit: " + ps[i].getClass().getName());
      pre[i] = (org.mathweb.reasoning.logic.Literal) ps[i];
    }
    
    ls = new ArrayList();
    for(int i=0; i<outputs.size(); i++){
      //      System.out.println("out: " + outputs.get(i).getClass().getName());
      
      Output e = (Output) outputs.get(i);
      Predicate p = new Predicate("type");
      Argument[] args = { new org.mathweb.reasoning.logic.Variable(e.getName()),
                          new org.mathweb.reasoning.logic.Constant(e.getType())};
      ls.add(new org.mathweb.reasoning.logic.Literal(p, args, true));
    }
    for(int i=0; i<postconds.size(); i++){
      //      System.out.println("post: " + postconds.get(i).getClass().getName());
      Effect po = (Effect) postconds.get(i);
      List l = po.toPlanningLiterals(inputNames);
      ls.addAll(l);
    }
    org.mathweb.reasoning.logic.Literal[] post = new org.mathweb.reasoning.logic.Literal[ls.size()];
    ps = ls.toArray();
    for(int i=0; i<ps.length; i++){
      //      System.out.println("lit: " + ps[i].getClass().getName());
      post[i] = (org.mathweb.reasoning.logic.Literal) ps[i];
    }
    return new Action(new Task(name), pre, post, new BindingConstraints());
  }
  
}


//   public String getProdigyPreferenceRule()
//   {
//     if (owlsService == null) return "";
//     Profile prof = owlsService.getProfile();
//     ServiceParameter param = prof.getServiceParameter(serviceParameter);
//     OWLResource value = param.getParameter();
    
//     OWLResource whenCond = value.getProperty(whenCondURI);
//     System.out.println(whenCond);
//     String oneff = value.getProperty(onEffect).toString();
//     System.out.println(oneff);
//     String whenCondPred = whenCond.getProperty(propPred).toString();
//     String argName1 = whenCond.getProperty(arg1).toString();
//     String argName2 = whenCond.getProperty(arg2).toString();
//     Predicate pred = new Predicate(whenCondPred); 
//     Argument[] args = new Argument[2];
//     args[0] = new Variable(argName1);
//     args[1] = new Variable(argName2);
//     Literal whenCondLit =  new Literal(pred, args, true);
//     String whenCondLitStr = getCondString(whenCondLit, new ArrayList());
    
//     System.out.println("whenCond: " + whenCondLitStr);
    
//     CachedHashCodeObject condition = null;
//     List postconds= problem.getEffects(); 
//     for(int k=0; k<postconds.size(); k++){
//       //      System.out.println("post: " + postconds.get(i).getClass().getName());
//       Effect po = (Effect) postconds.get(k);
//       if (po.getName().equals(oneff))
//         condition = po.getCondition();
//     }
//     String condStr = getCondString(condition, new ArrayList());
//     System.out.println("cond: " + condStr);
    
//     StringBuffer rule = new StringBuffer();
//     rule.append("(Control-Rule prefer-");
//     rule.append(this.name);
//     rule.append("\n  (if (and\n    (current-goal ");
//     rule.append(condStr);
//     rule.append(")\n    (true-in-state ");
//     rule.append(whenCondLitStr);
//     rule.append(")\n    ))\n  (then prefer operator ");
//     rule.append(this.name);
//     rule.append(" <anything>))\n");
    
    
//     /*
//       (Control-Rule prefer-ep-1
// 	      (if (and
// 		   (current-goal (resultFor <x> <y>))
// 		   (true-in-state (problem-class <y> sat_epr_fof))
// 		   ))
// 	      (then prefer operator EpATP <anything>))
// */
//       return rule.toString();
//   }

  /** transforms services into planning operators, such as:  
action {
  tramp
  pre {
    type(prob1, FOFPROBLEM),
    type(cnf, CNFPROBLEM),
    type(proof1, FORESOLUTIONPROOF),
    constraint(EQUIVALENTTO, cnf, prob1),
    constraint(FOR, proof1, cnf)
  }
  post {
    type(proof2, NDPROOF),
    constraint(FOR, proof2, prob1) 
  }
  constraints {
    prob1 != cnf
  }
}
  */

  

