// -*- indent-tabs-mode: nil -*-
/*****************************************************************
This file is part of the MathServe system

Authors: Juergen Zimmer

email: jzimmer@ags.uni-sb.de
   
For information about this program, write to:                          
MathWeb-Project                                                        
AG Siekmann/FB Informatik                                            
Universitaet des Saarlandes                                          
Gebaeude E 1.1                                                     
D-66041 Saarbruecken                                                 
Germany    

For information about the newest version of MathServe, see 
http://www.ags.uni-sb.de/~jzimmer/mathserve.html

See the file "LICENSE.txt" or
http://www.ags.uni-sb.de/~jzimmer/LICENSE.html
for information on usage and redistribution
of this file, and for a DISCLAIMER OF ALL WARRANTIES.
*****************************************************************/

package org.mathweb.texmacs;

import java.io.*;
import java.net.URI;
import java.net.URL;
import java.util.*;
import javax.xml.rpc.ParameterMode;


import org.apache.log4j.Level;
import org.apache.log4j.Logger; 

import org.jdom.*;
import org.jdom.input.*;
import org.jdom.output.*;
import org.jdom.Namespace;
import org.jdom.JDOMException;
import org.jdom.xpath.XPath;

import org.apache.axis.client.Call;
import org.apache.axis.client.Service;
import org.apache.axis.encoding.XMLType;
import org.apache.axis.configuration.FileProvider;

import org.mathweb.core.*;
import org.mathweb.onto.*;
import org.mathweb.atp.ATP;

// import org.mathweb.util.Java2WSDL;
// import org.mathweb.util.TSTPCreator;
// import org.mathweb.reasoning.logic.*;
// import org.mathweb.serv.TaxonomyEntry;
// import org.mathweb.serv.ReasoningService;
// import org.mathweb.serv.PreCondition;
// import org.mathweb.serv.Effect;
// import org.mathweb.serv.StochasticEffect;
// import org.mathweb.serv.ReasoningProblem;
// import org.mathweb.serv.Query;

public class MathserveClient extends BasicService
{
  
  private OWLJDom owlJDom = OWLJDom.getInstance();
  
  private class MyException extends Exception
  {
    public MyException(String msg)
    {
      super(msg);
    }
    
  }
  
  public String callMathserve(String hostname, String inFileName, int timeRes, String resultClass)
  {
    String eProof = readFile(msHome + "/examples/owl/e-proof2.owl");
    Element eProofRoot= null;
    
    try{
      eProofRoot = owlJDom.getRootElement(eProof);
      String texmacsXML = readFile(inFileName);
      StringBuffer buffer = new StringBuffer();
      buffer.append(ATP.tstpFixString);
      buffer.append("%--- This file was automatically generated from a TeXmacs XML buffer dump.\n\n");
      
      Element root = owlJDom.getRootElement(texmacsXML);
      System.out.println("root: " + root.getName());
      List axiomEs = root.getChildren("axiom");
      List children = new ArrayList();
      System.out.println(axiomEs.size() + " axioms");
      
      for(int i=0; i<axiomEs.size(); i++){
        Element axiom = (Element) axiomEs.get(i);
        children = new ArrayList();
        children.addAll(axiom.getChildren("mexists"));
        children.addAll(axiom.getChildren("mforall"));
        children.addAll(axiom.getChildren("mappl"));
        children.addAll(axiom.getChildren("meq"));
        children.addAll(axiom.getChildren("mimpl"));
        children.addAll(axiom.getChildren("mand"));
        children.addAll(axiom.getChildren("mor"));
        children.addAll(axiom.getChildren("mneg"));
        if (children.size() > 0) {
          buffer.append("fof(ax_" + i + ",axiom,\n   ");
          buffer.append(translateFormula((Element) children.get(0)));
          buffer.append(").\n\n");
        }
      }
      List conjEs = root.getChildren("conj");
      if (conjEs.size() > 0){
        Element conj = (Element) conjEs.get(0);
        children = new ArrayList();
        children.addAll(conj.getChildren("mexists"));
        children.addAll(conj.getChildren("mforall"));
        children.addAll(conj.getChildren("mappl"));
        children.addAll(conj.getChildren("meq"));
        children.addAll(conj.getChildren("mimpl"));
        children.addAll(conj.getChildren("mand"));
        children.addAll(conj.getChildren("mor"));
        children.addAll(conj.getChildren("mneg"));            
        if (children.size() > 0) {
          buffer.append("fof(conj_1, conjecture,\n   ");
          buffer.append(translateFormula((Element) children.get(0)));
          buffer.append(").\n");
        }
      }
      
      System.out.println(buffer);
      writeFile("/home/jzimmer/example.tstp", buffer.toString());
      String endpoint = "http://" + hostname + ":8080/axis/services/Broker";
      String method = "proveProblemOpt";
      
      FileProvider config = new FileProvider(msHome + "/conf/client-config.wsdd");
      Service  service = new Service(config);
      Call     call    = (Call) service.createCall();
      
      call.setTargetEndpointAddress(new java.net.URL(endpoint));
      call.setOperationName( method );
      call.addParameter("in0", XMLType.SOAP_STRING, ParameterMode.IN);
      call.addParameter("in1", XMLType.SOAP_INT, ParameterMode.IN);
      call.setReturnType(XMLType.SOAP_STRING);
      call.setTimeout(new Integer(600*1000));
      
      String result = (String) call.invoke( new Object [] {buffer.toString(), new Integer(timeRes) });
      Element resultRoot = owlJDom.getRootElement(result);
      
      return xml2SExpr(resultRoot).toString();
      
    } catch(Exception e){
      e.printStackTrace();
      try {
        String failure = owlJDom.createFailure(e.getMessage());
        Element failRoot = owlJDom.getRootElement(failure);
        return xml2SExpr(failRoot).toString();
      } catch(Exception ex){
        return "#f";
      }
    }
    
  }
  
  private StringBuffer xml2SExpr(Element elem)
  {
    List children = elem.getChildren();
    List attribs = elem.getAttributes();
    StringBuffer result = new StringBuffer();
    
    result.append("(" + elem.getName() + " ");
    
    for(int i = 0 ; i< attribs.size(); i++){
      Attribute att = (Attribute) attribs.get(i);
      result.append("(= \"" + att.getName() + "\" \"" + att.getValue() + "\") ");
    }
    for(int i = 0 ; i< children.size(); i++){
      Element ch = (Element) children.get(i);
      result.append(xml2SExpr(ch));
      result.append(" ");
    }
    List content = elem.getContent();
    /* content elements are of classes Text, Element, Comment, ProcessingInstruction, CDATA, and EntityRef */
    String text = "";
    for(int i = 0; i< content.size(); i++){
      Object cont = content.get(i);
      if (cont instanceof Text) 
        text += ((Text) cont).getText().trim();
      else if (cont instanceof CDATA)
        text += cont.toString().trim();
    }
    if ((text != null) && (! text.equals(""))){
      result.append("\"");
      result.append(text);
      result.append("\"");
    }
    
    result.append(")");
    return result;
  }
  
  private StringBuffer translateFormula(Element formE) throws Exception
  {
    String name = "";
    List args = new ArrayList();
    try{
      //      System.out.println("translate formula  " + formE);
      StringBuffer result = new StringBuffer();
      name = formE.getName();
      if (name.equals("var")){
        String varName = formE.getAttributeValue("name");
        return new StringBuffer(varName.toUpperCase());
      }
      else if (name.equals("arg")){
        String argName = formE.getAttributeValue("name");
        String startLetter = argName.substring(0,1);
        //        System.out.println("constant " + argName + "? " + startLetter.matches("[A-Z]"));
        
        if (startLetter.matches("[A-Z]"))
          return new StringBuffer(argName.toLowerCase());
      else 
        return new StringBuffer(argName.toUpperCase());
      }
      else if (name.equals("mappl")){
        args = formE.getChildren();
        result.append(formE.getAttributeValue("name"));
        if (args.size() > 0) {
          result.append("(");
          for(int i=0; i < args.size(); i++){
            Element arg = (Element) args.get(i);
            result.append(translateFormula(arg));
            if (i < args.size() -1) 
            result.append(", ");
          }
          result.append(")");
        }
        return result;
      }
      else if (name.equals("mand")){
        args = formE.getChildren();
      if (args.size() == 2) {
        result.append("(");
        Element arg1 = (Element) args.get(0);
        Element arg2 = (Element) args.get(1);
        result.append(translateFormula(arg1));
        result.append(" & ");
        result.append(translateFormula(arg2));
        result.append(")");
      }
      return result;
    }
    else if (name.equals("mor")){
      args = formE.getChildren();
      if (args.size() == 2) {
        result.append("(");
        Element arg1 = (Element) args.get(0);
        Element arg2 = (Element) args.get(1);
        result.append(translateFormula(arg1));
        result.append(" | ");
        result.append(translateFormula(arg2));
        result.append(")");
      }
      return result;
    }
    else if (name.equals("meq")){
      args = formE.getChildren();
      if (args.size() == 2) {
        result.append("(");
        Element arg1 = (Element) args.get(0);
        Element arg2 = (Element) args.get(1);
        result.append(translateFormula(arg1));
        result.append(" = ");
        result.append(translateFormula(arg2));
        result.append(")");
      }
      return result;
    }
    else if (name.equals("mimpl")){
      args = formE.getChildren();
      if (args.size() == 2) {
        result.append("(");
        Element arg1 = (Element) args.get(0);
        Element arg2 = (Element) args.get(1);
        result.append(translateFormula(arg1));
        result.append(" => ");
        result.append(translateFormula(arg2));
        result.append(")");
      }
      return result;
    }
    else if (name.equals("mneg")){
       args = formE.getChildren();
      if (args.size() == 1) {
        result.append("~(");
        Element arg1 = (Element) args.get(0);
        result.append(translateFormula(arg1));
        result.append(")");
      }
      return result;
    }
    else if (name.equals("mforall")){
       args = formE.getChildren();
      if (args.size() == 2) {
        Element arg1 = (Element) args.get(0);
        Element arg2 = (Element) args.get(1);
        result.append("( ! [");
        result.append(translateFormula(arg1));
        result.append("] : \n"); 
        result.append(translateFormula(arg2));
        result.append(")");
      }
      return result;
    }
    else if (name.equals("mexists")){
       args = formE.getChildren();
      if (args.size() == 2) {
        Element arg1 = (Element) args.get(0);
        Element arg2 = (Element) args.get(1);
        result.append("( ? [");
        result.append(translateFormula(arg1));
        result.append("] : \n"); 
        result.append(translateFormula(arg2));
        result.append(")");
      }
      return result;
    }
      throw new MyException("Error: Don't know how to translate at application of " + name);
    } catch(Exception ex){
      if (ex instanceof MyException) 
        throw ex;
      else {
        String msg = "Error: Don't know how to translate near " + name + "(";
        for(int i = 0; i< args.size(); i++) {
          Element arg = (Element) args.get(i);
          String aName = arg.getAttributeValue("name");
          if (aName != null) 
            msg += aName;
          else 
            msg += "...";
          if (i < args.size() - 1) msg += ",";
        }
        throw new Exception(msg);
      }
    }
  }
  
    

  public static void main(String argv[]) throws Exception {
    if (argv.length > 3){
      MathserveClient c = new MathserveClient();
      String hostname = argv[0];
      String inFile = argv[1];
      int timeRes = Integer.parseInt(argv[2]);
      String resultClass = argv[3];
      String outFile = argv[4];
      
      String result = c.callMathserve(hostname, inFile, timeRes, resultClass);
      c.writeFile(outFile, result);
    }
  }
  
      

}

 
      
/* Settings for Emacs.
 * Local variables:
 * c-basic-offset: 2
 * tab-width: 2
 * indent-tabs-mode: nil
 * End:
 */

