package org.mathweb.trans;

import java.util.Vector;
import java.net.MalformedURLException;

import org.apache.xmlrpc.*;
import org.apache.log4j.Level;
import org.apache.log4j.Logger; 
import org.apache.log4j.FileAppender; 
import org.apache.log4j.SimpleLayout; 

/** The XML-RPC server for the lisp process */
public class TrampThreadXmlRpcHandler implements XmlRpcHandler {
  
  private static Logger logger = Logger.getLogger("TrampThreadXmlRpcHandler");
  private static String msHome = System.getProperty("mathserv.home");
  private static String logFileName = msHome + "/logs/trampxmlrpchandler.log";
  private static FileAppender appender = newAppender(logFileName);
  private TrampThread tramp;
  
  public TrampThreadXmlRpcHandler(TrampThread tr)
  {
    super();
    this.tramp = tr;
    logger.setLevel(Level.ALL);
    logger.addAppender(appender);
    logger.debug("TrampThread XmlRpc handler started.");
  }
  
  public static FileAppender newAppender(String logFileName){
    FileAppender appender = null;
    try {
      appender = new FileAppender(new SimpleLayout(), logFileName, false);
    } catch(Exception e) {}
    return appender;
  }

  public Object execute(String methodName, Vector params ) 
  {
    try{ 
      logger.debug("RPC call method: " + methodName);
      logger.debug("args: " + params);
      for (int i=0; i<params.size(); i++){
	logger.debug("args " + params.get(i) + " " + params.get(i).getClass());
      }
      if (methodName.equals("initLispServer")) {
	String lispServerURL = (String) params.get(0);
	logger.debug("lisp server: " + lispServerURL);
	tramp.setLispServerConnection(lispServerURL);
	return new String("T");
      }
      else if (methodName.equals("ready")) {
	logger.debug("lisp says that it is ready!");
	if (tramp.getLispServerConnection() != null){
	  tramp.setReady();
	  return new String("T");
	} else {
	  logger.debug("Lisp server has not yet been initialised!");
	  logger.debug("Returning NIL!");
	  return new String("NIL");
	}
      }
    } catch(Exception e){
      e.printStackTrace();
    }
    return null;
    }
}
