// -*- indent-tabs-mode: nil -*-
/*****************************************************************
This file is part of the MathServe system

Authors: Juergen Zimmer

email: jzimmer@ags.uni-sb.de
   
For information about this program, write to:                          
MathWeb-Project                                                        
AG Siekmann/FB Informatik                                            
Universitaet des Saarlandes                                          
Gebaeude E 1.1                                                     
D-66041 Saarbruecken                                                 
Germany    

For information about the newest version of MathServe, see 
http://www.ags.uni-sb.de/~jzimmer/mathserve.html

See the file "LICENSE.txt" or
http://www.ags.uni-sb.de/~jzimmer/LICENSE.html
for information on usage and redistribution
of this file, and for a DISCLAIMER OF ALL WARRANTIES.
*****************************************************************/
package org.mathweb.util;

import java.io.*;
import java.util.*;
import java.net.URI;

import org.jdom.*;
import org.jdom.input.*;
import org.jdom.output.*;
import org.jdom.Namespace;
import org.jdom.JDOMException;
import org.jdom.xpath.XPath;
 
import org.mathweb.core.BasicService;
import org.mathweb.onto.OWLJDom;

/**
@author Juergen Zimmer, Saarland University, Germany
@version $Date: 2006/05/26 13:47:08 $ $Revision: 1.7 $
*/
public class Performance extends BasicService {
  private Quoter quoter = new Quoter();
  private static String xmlBase = "http://www.tptp.org/Problems#";
  private static OWLJDom owlJDom = OWLJDom.getInstance();
  private static String sep = File.separator;
  
  private static Element probClass = new Element("propertyPredicate", OWLJDom.SWRLX_NS);

  private static Element status = new Element("propertyPredicate", OWLJDom.SWRLX_NS);
  
  
  public Performance(){
    probClass.setAttribute("resource", OWLJDom.MW_NS_STR + "problemClass", OWLJDom.RDF_NS);
    status.setAttribute("resource", OWLJDom.MW_NS_STR + "status", OWLJDom.RDF_NS);
  }
                                   
  public String performance(String fileName) { 
    try{
      //      String data = readFile(msHome + sep + "thirdParty" + sep + "TPTPWorld" + sep + "performance" + sep + "new" +
      String data = readFile(msHome + sep + "thirdParty" + sep + "TPTPWorld" + sep + "performance" + sep + "2006" +
                             sep + fileName);
      String[] lines = data.split("\n");

      if (lines.length > 0) {
        String servName = lines[0];
        Namespace SERV_NS =  Namespace.getNamespace("serv", servName + "#");
        Element root = new Element("Performance", OWLJDom.PERF_NS);
        root.addNamespaceDeclaration(OWLJDom.RDF_NS);
        root.addNamespaceDeclaration(OWLJDom.SWRLX_NS);
        root.addNamespaceDeclaration(OWLJDom.PERF_NS);
        root.addNamespaceDeclaration(SERV_NS);
        
        //        root.setAttribute("base", OWLJDom.(), Namespace.XML_NAMESPACE);
        
        for(int i=1 ; i< lines.length; i++){
          System.out.println("line " + lines[i]);
          String[] parts = lines[i].split(" ");
          if (parts.length > 2){
            String className = parts[0];
            Float probability = new Float(parts[1]);
            Float costFloat = new Float(parts[2]);
            int cost = costFloat.intValue();
            
            String atpStatus;
            if (className.indexOf("FOF")>-1)
              if (className.indexOf("_CSA") >-1)
                atpStatus = "CounterSatisfiable";
              else if(className.indexOf("_SAT") > -1)
                atpStatus = "Satisfiable";
              else atpStatus = "Theorem";
            else 
              if(className.indexOf("_SAT") > -1)
                atpStatus = "Satisfiable";
              else atpStatus = "Unsatisfiable";
            
            Integer rewardInt = new Integer(300000);
            
            Element perform = new Element("Perform", OWLJDom.PERF_NS);
            Element cond = new Element("Condition", OWLJDom.PERF_NS);
            Element atom = new Element("IndividualPropertyAtom", OWLJDom.SWRLX_NS);
            Element pred = (Element) probClass.clone();
            Element arg1 = new Element("argument1", OWLJDom.SWRLX_NS);
            Element arg2 = new Element("argument2", OWLJDom.SWRLX_NS);
            arg1.setAttribute("resource", servName + "#tstp_problem" , OWLJDom.RDF_NS);
            arg2.setAttribute("resource", OWLJDom.MW_NS_STR + className, OWLJDom.RDF_NS);
            
            atom.addContent(pred);
            atom.addContent(arg1);
            atom.addContent(arg2);
            cond.addContent(atom);

            perform.addContent(cond);
            

            Element eff = new Element("Effect", OWLJDom.PERF_NS);
            Element eatom = new Element("IndividualPropertyAtom", OWLJDom.SWRLX_NS);
            Element epred = (Element) status.clone();
            Element earg1 = new Element("argument1", OWLJDom.SWRLX_NS);
            Element earg2 = new Element("argument2", OWLJDom.SWRLX_NS);
            earg1.setAttribute("resource", servName + "#atp_result" , OWLJDom.RDF_NS);
            earg2.setAttribute("resource", OWLJDom.STAT_NS_STR + atpStatus, OWLJDom.RDF_NS);
            
            eatom.addContent(epred);
            eatom.addContent(earg1);
            eatom.addContent(earg2);
            eff.addContent(eatom);
            
            perform.addContent(eff);
            
            Element probab = new Element("Probability", OWLJDom.PERF_NS);
            probab.setAttribute("type", OWLJDom.XSD_NS_STR + "float" , OWLJDom.PERF_NS);
            probab.addContent(probability.toString());

            Element c = new Element("AverageCost", OWLJDom.PERF_NS);
            c.setAttribute("type", OWLJDom.XSD_NS_STR + "nonNegativeInteger" , OWLJDom.PERF_NS);
            c.addContent("" + cost);

            Element reward = new Element("Reward", OWLJDom.PERF_NS);
            reward.setAttribute("type", OWLJDom.XSD_NS_STR + "nonNegativeInteger" , OWLJDom.PERF_NS);
            reward.addContent(rewardInt.toString());
            
            perform.addContent(probab);
            perform.addContent(c);
            perform.addContent(reward);
            
            root.addContent(perform);
            
            //            System.out.println(className + ":" + prob + " " + cost);
          }
          
        }
        return owlJDom.newXMLDocument(root);
      }
      return null;
      
    } catch(Exception e) {
      e.printStackTrace();
      
      return null;
    }
  }
  
  public static void main(String[] args) {
    Performance creat  = new Performance();
    String result= creat.performance(args[0]);
    System.out.println(result);
    creat.writeFile("/home/jzimmer/public_html/atp/performance/" + args[1], result);
    
    System.exit(0);
  }
}

  
  
