;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; MODULE      : mathserve.scm
;; DESCRIPTION : Stub for interfacing to the MathServe system
;; AUTHOR      : Juergen Zimmer, Uni Saarbrucken, 2006.
;;
;; This software falls under the GNU general public license and comes WITHOUT
;; ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for details.
;; If you don't have this file, write to the Free Software Foundation, Inc.,
;; 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(load "xml-rpc.scm")

(texmacs-module (mathserve)
		(:use (utils plugins plugin-cmd)))

(if (not (defined? 'ms-menu-bound))
    (begin
      (menu-bind mathserve-menu
		 ("MathServe ATP Result" (ms-call-mathserve "FoAtpResult"))
		 )
      
      (menu-extend texmacs-menu
		   (=> "MathServe" (link mathserve-menu)))
      ))

(define-public ms-menu-bound #t)

;; Translating objects back & forth
(define (ms-texmacs2mathserve o)
; What it does : (markup (document (concat a b) c)) --> (markup a b c)
  (display "stripping tree!\n\n")
  (traverse-tree o
		 (lambda (c)
		   (cond ((and (pair? c)
			       (pair? (cdr c))
			       (pair? (cadr c))
			       (in? (caadr c) '(concat document)))
			  (set-cdr! c (cdadr c)))
			 (else #t))
		   (cond ((and (pair? c)
			       (pair? (cdr c))
			       (pair? (cadr c))
			       (in? (caadr c) '(concat document)))
			  (set-cdr! c (cdadr c)))
			 (else #t))
		   )))

(define (ms-mathserve2texmacs o)
  o)

(define (ms-find-element elem lst)
;;;  (display* "find-elem " elem " " lst "\n")
  (cond ((pair? lst)
	 (if (eq? (car lst) elem)
	     lst
	     (if (eq? (car lst) '=)
		 nil
		 (let ((res (ms-find-element elem (car lst))))
		   (if (eq? res nil)
		       (ms-find-element elem (cdr lst))
		       res)))))
	(else nil)))

; Extracting fragments of the markup
(tm-define (ms-extract-fragment path tags)
	   (let* ((p (tm-upper-path path tags))
		  (p0 p)
		  )
	     (if p0
		 (set! p0 (tm-upper-path p0 '(keyed))))
	     (if p0
		 (set! p p0))
	     (let*((o (if p (stree-at p) p))
		   (r (if p (get-self-keyed-path p) p))
		   )
	      (if r
	 	  (set! o (stree-at r)))
	      (display* "p=" p "\n")
	      (display "o=")
	      ;;(write (texmacs2scheme (copy-tree o)))
	      (write o)
	      (newline)
	      o)))

(define (ms-call-mathserve resultClass)
  (let* ((obj0 (ms-extract-fragment (cursor-path) '(mathserve)))
	 (obj (if (not (eq? obj0 #f))
		  (caddr obj0)
		  #f))
	 )
    (if (and (not (eq?  (pair? obj) #f))
	     (in? (car obj) '(concat document)))
	(set! obj (cadr obj)))
    (display* " obj : " obj "\n\n")
    (if (not (eq? obj #f))
	(let* ((newObj (ms-texmacs2mathserve obj))
	       (asdfasf  (display* newObj "\n\n"))
	       (key (caddr (cadr obj0)))
	       (asdfasfasfd  (display* key "\n\n"))
	       (inFileName (tmpnam))
	       (outFileName (tmpnam))
	       (inFile (open-file inFileName "w"))
	       
	       (hostname   "hades.ags.uni-sb.de")
	       (timeResource "30")
	       
	       )
	  (if (and newObj inFile)
	      (begin (with-output-to-port inFile
		       (lambda ()
			 (sexpr->xml-bis newObj)))
		     (flush-all-ports)
		     (close-port inFile)
		     ;;; execute the MathServe client shell command
		     (with command (string-concatenate
				    (list "/home/jzimmer/import/MathServe/bin/texmacsCallMS "
					  hostname " " inFileName " " timeResource " " resultClass " " outFileName))	
			   (display* "command: " command "\n\n")
			   (system command))
		     
		     (delete-file inFileName)
		     
		     (let ((outFile (open-file outFileName "r"))
			   (result "")
			   (new-line "")
			   (end-of-file #f))
		       (if outFile
			   (let* ((result-expr (read outFile))
				  (afadfjsf (display* result-expr "\n\n"))
				  (failure-expr (ms-find-element 'Failure result-expr))
				  (status  "Unknown")
				  (answer "Unknown"))
			     (delete-file outFileName)
			     (display* failure-expr "\n\n")
			     (if (not (eq? failure-expr nil))
				 (begin (set! status "Failure")
					(set! answer
					      (string-concatenate (list "MathServe reported failure: "
									(second (second failure-expr))))))
				 (let* ((statusExpr (ms-find-element 'status result-expr))
					(statusStr (if statusExpr (third (second  statusExpr))))
					(split  (if statusStr
						    (string-split statusStr #\#)))
					(systemExpr (ms-find-element 'system result-expr))
					(system (if (and systemExpr (pair? systemExpr))
						    (second systemExpr))))
				   (set! status (if split
						    (second split)
						    statusStr))
				   (set! answer (string-concatenate
						 (list "The system " system " says the status of the problem is '" status "'."))))
				 )
			     (close-port outFile)
			     (display* answer "\n")
			     (if status
				 (let* ((arity (tree-arity (path->tree (append (tree->path (index-find-keyed-root? "MathServe")) '(1 0 0)))))
					(tree-path (append (tree->path (index-find-keyed-root? "MathServe"))
							   '(1 0 0)))
					(status-path (append (tree->path (index-find-keyed-root? "MathServe"))
							     (list 1 0 0 (- arity 1))))
					(old-status (tree->stree (path->tree status-path)))
					)
				   (display* "old status: ")
				   (write old-status)
				   (if (and (pair? old-status) (eq? (car old-status) 'status))
				       (path-assign status-path
						    (o2t (stree->tree (list 'status answer))))
					 ;;; use path-insert instead
				       (path-assign tree-path
						    (tree-append (path->tree tree-path)
								 (o2t (stree->tree
								       (list 'foo (list 'status answer))))))
				       )
				   (go-to status-path)
				   )
				 )
			     (display* "\n")
			     )
			   (display "Error reading result file " outFile))
		       )
		     )
	      (begin (display "Error while opening file " inFile))))
	(display "You have to be inside a MathServe environment to call MathServe.\n")
	)
    )
  )
	
;; Shortcuts for easy editing
; Shortcut to toggle the Source tree edit mode
(kbd-map ("S-tab" (toggle-preamble)))

; Adding new leaves to a tree using Shift-Alt-(Right|Left Arrow)
(kbd-map ("C-S-left" (structured-insert #f))
         ("C-S-right" (structured-insert #t)))


; Keyboard shortcuts
(kbd-map ("C-S"
  (insert 
    '(keyed (tuple "_key" "MathServe")
     (mathserve "")))))
      

(kbd-map ("C-G" (ms-call-mathserve "FoAtpResult")))

(kbd-map ("C-L" (begin (load "/home/jzimmer/.TeXmacs/plugins/mathserve/progs/mathserve.scm")
		       (load "/home/jzimmer/.TeXmacs/plugins/mathserve/progs/xml-rpc.scm"))))

; Special tags
(plugin-descr! "mathserve" lisp-import identity ms-mathserve2texmacs ms-texmacs2mathserve)
(keyed-tag! 'mathserve "mathserve" #t 'mathserve #f #f)

; Initialization
(tm-define (mathserve-initialize)
  (display* "Loading MathServe plugin...\n"))


					; 		     (begin (while (not end-of-file)
					; 				   (set! new-line (read-line outFile))
					; 				   (display* "new " new-line)
					; 				   (if (not (eof-object? new-line))
					; 				       (set! result (string-concatenate (list result "\n" new-line)))
					; 				       (set! end-of-file #t)))
					; 			     (write  (index-find-keyed-value? "Status")) (newline)
					; 			     (write  (tree->path (index-find-keyed-value? "Status"))) (newline)
; ; 				   (display* "index keys: " index-keys "\n\n")
; ; 				   (display* "tree root: " (index-find-keyed-root? "MathServe") "\n")
; ; 				   (display* "tree arity: " (tree-arity (index-find-keyed-root? "MathServe")) "\n")
; ; 				   (display* "path: " (tree->path (index-find-keyed-root? "MathServe")) "\n")
; 				   (display* "subtree: " (path->tree (append (tree->path (index-find-keyed-root? "MathServe")) '(1 0))) "\n\n")
; 				   (with arity (tree-arity (path->tree (append (tree->path (index-find-keyed-root? "MathServe")) '(1 0 0))))
; 					 (display* "stree: "
; 						   (tree->stree (path->tree (append (tree->path (index-find-keyed-root? "MathServe"))
; 										    (list 1 0 0 (- arity 2))))))
; 					 )
