#include <stdio.h>
#include <assert.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include "DataTypes.h"
#include "Utilities.h"
#include "Tokenizer.h"
#include "Parsing.h"
#include "Signature.h"
#include "Examine.h"
#include "Modify.h"
#include "List.h"
#include "ListStatistics.h"
#include "PrintTSTP.h"
//-----------------------------------------------------------------------------
int main(int argc, char *argv[]) {

    SIGNATURE Signature;
    ANNOTATEDFORMULA AnnotatedFormula;
    READFILE Stream; 
    PrintFormatType Format;
    int ArgOffset;
    int KeepNonLogicals;
    int AllowFreeVariables;
    int Pretty;
    SyntaxType LastNodeType;

    ArgOffset = 0;
    LastNodeType = nontype;

    Format = tptp;
    Pretty = 1;
    KeepNonLogicals = 0;
    AllowFreeVariables = 0;

    while (argc > ArgOffset+2 && argv[ArgOffset+1][0] == '-') {
//----Change output length if specified
        if (argc > ArgOffset+3 && !strcmp(argv[ArgOffset+1],"-f")) {
            Format = StringToPrintFormat(argv[ArgOffset+2]);
            ArgOffset += 2;
        }
        if (argc > ArgOffset+3 && !strcmp(argv[ArgOffset+1],"-u")) {
            if (!strcmp("machine",argv[ArgOffset+2])) {
                Pretty = 0;
            }
            ArgOffset += 2;
        }
//----Decide whether or not to keep non-logicals
        if (!strcmp(argv[ArgOffset+1],"-k")) {
            KeepNonLogicals = 1;
            ArgOffset++;
        }
//----Decide whether or not to allow free variables
        if (!strcmp(argv[ArgOffset+1],"-V")) {
            AllowFreeVariables = 1;
            ArgOffset++;
        }
    }


//----One signature for all testing
    Signature = NewSignature();

//----Reading formulae one-by-one from stdin. 
    SetNeedForNonLogicTokens(KeepNonLogicals);
    SetAllowFreeVariables(AllowFreeVariables);
    if ((Stream = OpenFile(argv[ArgOffset+1],NULL)) != NULL) {
        NextToken(Stream);
        while (!CheckTokenType(Stream,endeof)) {
            AnnotatedFormula = ParseAnnotatedFormula(Stream,Signature);
//----Blank line after every formula
            if ((LastNodeType == tptp_fof || LastNodeType == tptp_cnf) &&
GetSyntax(AnnotatedFormula) != blank_line) {
                printf("\n");
            }
//----Skip double blank lines
            if (LastNodeType != blank_line || GetSyntax(AnnotatedFormula) !=
blank_line) {
                PrintAnnotatedTSTPNode(stdout,AnnotatedFormula,Format,Pretty);
            }
            fflush(stdout);
            LastNodeType = GetSyntax(AnnotatedFormula);
            FreeAnnotatedFormula(&AnnotatedFormula);
        }
        CloseFile(Stream);
    } else {
        printf("ERROR: Failed to open %s\n",argv[ArgOffset+1]);
    }

    return(EXIT_SUCCESS);
}
//-----------------------------------------------------------------------------
