#include <stdio.h>
#include <assert.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include "DataTypes.h"
#include "Utilities.h"
#include "Tokenizer.h"
#include "Parsing.h"
#include "Signature.h"
#include "Examine.h"
#include "List.h"
#include "Tree.h"
#include "TreeStatistics.h"
#include "ListStatistics.h"
#include "PrintTSTP.h"
//-----------------------------------------------------------------------------
void DetermineFOFSPC(ListStatisticsRecordType Statistics,String SPC) {

    if (Statistics.NumberOfVariables == 0 || Statistics.MaxFunctorArity < 1) {
        strcat(SPC,"EPR");
    } else {
        strcat(SPC,"RFO_");
        if (Statistics.NumberOfEqualityAtoms == 0) {
            strcat(SPC,"NEQ");
        } else {
            strcat(SPC,"EQU");
        }
    }
}
//-----------------------------------------------------------------------------
void DetermineCNFSPC(ListStatisticsRecordType Statistics,String SPC) {

    if (Statistics.NumberOfVariables == 0 || Statistics.MaxFunctorArity < 1) {
        strcat(SPC,"EPR");
    } else {
        strcat(SPC,"RFO_");
        if (Statistics.NumberOfEqualityAtoms == Statistics.NumberOfAtoms) {
            strcat(SPC,"PEQ_");
            if (Statistics.NumberOfFormulae == 
Statistics.NumberOfUnitFormulae) {
                strcat(SPC,"UEQ");
            } else {
                strcat(SPC,"NUE");
            }
        } else {
            if (Statistics.NumberOfEqualityAtoms == 0) {
                strcat(SPC,"NEQ_");
            } else {
                strcat(SPC,"SEQ_");
            }
            if (Statistics.NumberOfClauses == Statistics.NumberOfHornClauses) {
                strcat(SPC,"HRN");
            } else {
                strcat(SPC,"NHN");
            }
        }
    }

}
//-----------------------------------------------------------------------------
void DetermineSPC(SyntaxType Syntax,ListStatisticsRecordType Statistics,
int HasAConjecture,String SPC) {

//----Separate CNF and FOF
    if (Syntax == tptp_fof) {
//----Here it is done right, as opposed to SPCForTPTPProblem which puts
//----Unknowns without a conjecture into NKC.
        if (HasAConjecture) {
            strcpy(SPC,"FOF_NKC_");
        } else {
            strcpy(SPC,"FOF_NKS_NUN_");
        }
        DetermineFOFSPC(Statistics,SPC);
    } else if (Syntax == tptp_cnf) {
//----Assume an unsatisfiable set
        strcpy(SPC,"CNF_NKS_");
        DetermineCNFSPC(Statistics,SPC);
    } else {
        strcpy(SPC,"UNKNOWN_FORM");
    }
}
//-----------------------------------------------------------------------------
int main(int argc, char *argv[]) {

    LISTNODE Head;
    SIGNATURE Signature;
    ListStatisticsRecordType Statistics;
    String SPC;

    if (argc == 1) {
        printf("Usage: SPCForProblem <file name>\n");
        return(0);
    }

    Signature = NewSignature();
    SetNeedForNonLogicTokens(0);
    if ((Head = ParseFileOfFormulae(argv[1],NULL,Signature,1,NULL)) != NULL) {
        if (GetListStatistics(Head,&Statistics) != NULL) {
            DetermineSPC(GetListSyntax(Head),Statistics,
ThereIsAConjecture(Head),SPC);
            printf("%s\n",SPC);
        } else {
            printf("ERROR: Could not get statistics for %s\n",argv[1]);
        }
        FreeListOfAnnotatedFormulae(&Head);
        assert(Head == NULL);
    } else {
        printf("ERROR: Could not parse %s\n",argv[1]);
    }

    return(0);
}
//-----------------------------------------------------------------------------
