/*--------------------------------------------------------**
* modN.c (Version for Octopus)                             *     
* copyrighted by Monty Newborn 2005, 2006                  *
*                                                          *
* Added to Octopus December 2002                           *
*                                                          *
* Controls the modification of clauses                     *
*     determine_potential_changes().............#1         *
*     modify_clause()...........................#2         *
*     form_weakened_clause()....................#3         *
*     mark_questionable().......................#4         *
*     print_change_array()......................#5         *
*     reconsider_proof()........................#6         *
*     reinstate_base_clause()...................#7         *
*     bres_instantiated_proof().................#8         *
*     bfac_instantiated_proof().................#9         *
*     bsub_instantiated_proof().................#10        *
*     filter_ulist()............................#11        *
*     skip_equal_all_vars().....................#12        *
*     skip_equal_first_last()...................#13        *
*     print_xl_array()..........................#14        * 
*     det_dependencies()........................#15        *
*     void analyze_decomposeability()...........#16        *
*                                                          *
*---------------------------------------------------------*/

#include <stdio.h>
#include <string.h>
#include <math.h>

#include "const.h"
#include "clauseS.h"

/*----------------------------------------------------------*
* externally defined functions                              *
*----------------------------------------------------------*/

extern void print_clause();
extern int grab_number();
extern void standardize();
extern void clear_var_substitution_tble();
extern int unify();
extern clause *generate_iclause();
extern void calc_clause_hashcode();
extern int calc_lit_hc();
extern int time_now();
extern void clean_var_names();
extern void replace_vars();
extern void copy_clause_all();
extern void normalize_literal();
extern int simple_subsume2();
extern void final_clean();
extern void delete_literal();
extern int biggest_var_in_clause();
		   
/*----------------------------------------------------------*
* externally defined constants                              *
*----------------------------------------------------------*/

extern char *OutPtr;
extern int ply, nextu;
extern int base, org_base, nc;
extern int batch_tried;
extern int o_maxvar, o_maxlit, o_maxterm;
extern int inc_o_maxlit;
extern int o_batchtime, o_maxtime;
extern int s_starttime;
extern int s_nod;
extern int s_nod_last;
extern int learn_nod;
extern int o_debug;
extern int o_learn, weakened_ply, weakened_clause,o_weakened_index;
extern int length_of_proof;
extern int clauseToWeaken;
extern int weakened_proof_remember;
extern int weakened_clause_count;
extern char *weakProofBeginsHere, *weakProofEndsHere;
extern int good_clauses[];
extern int good_inferred_cls;
extern int satisfiable;
extern int equal_pred;
extern clause *clist[];
extern clause *ulist[];
extern int *u_space;

extern char input_buffer[];
extern int cl[PROOF_ARRAY_SIZE][14];
extern int xl[PROOF_ARRAY_SIZE][14];
extern int equal_xyyzxz;
extern dnode depth[];
extern int cade_format;
extern int spawned, TC, theorem_number;
extern int reinstate_parent[];
extern int questionableUsed;
extern int numb_trivial_proofs;
extern int split_proof_reported;

extern int pass_count;
extern int o_relevent;
extern int repeat_theorem;

/*----------------------------------------------------------*
* internally defined functions                              *
*----------------------------------------------------------*/

void determine_potential_changes();
void modify_clause();
void form_weakened_clause();
void mark_questionable();
void print_change_array();
void reconsider_proof();
void reinstate_base_clause();
int bres_instantiated_proof();
int bfac_instantiated_proof();
int bsub_instantiated_proof();
void filter_ulist();
int skip_equal_all_vars();
int skip_equal_first_last();
void print_xl_array();
void det_dependencies();
void analyze_decomposeability();

/*----------------------------------------------------------*
* internally defined constants                              *
*----------------------------------------------------------*/

char clause_number[5];
char literal_number[5];
char char_number[5];
char old_char[80];
char new_char[5];
int replaced_char;
int deleted_terms[200];
int questionable[2000];
int total_questionable;
int used, used_nc;
int c_split, l_split;

int excellent_weakened_proof;

/*-----------------------------------------------------------*
* change_array:                                              *
*         column 0:  batch_tried                             *
*         column 1:  clause number                           *
*         column 2:  literal number                          *
*         column 3:  position in literal                     * 
*         column 4:  substitute (variable)                   *
*         cols 5-9:  characters substituted for              *
*-----------------------------------------------------------*/

int change_array[50][10];
int change_level;

/*-----------------------------------------------------------*
* potential_change_array:                                    *
*          column 0: clause                                  *
*          column 1: literal                                 *
*          column 2: position                                *
*          column 3: char                                    *
*          column 4: #var in clause                          *
*          column 5: substituted char.                       *
*-----------------------------------------------------------*/

int weak[2000][7];
int p_change_number;
int randy, randy2, randy_last;
int repeat_weakening;

int decompose, decompose2, decomposible3, decomposible4;
int most_dc1, most_dc2, most_dc3, most_dc4;
int dc_parts1, dc_parts2, dc_parts3, dc_parts4;
int lits_1, lits_2, lits_3, lits_4;
int var_vector[32][17];
int split_temp[17];
int best_var_v1[32][17],best_var_v2[32][17];
int best_var_v3[32][17],best_var_v4[32][17];

int split_pbs, unsolved_split_pbs, solved_split_lits[32];
int this_split;
int unsolved_exists_pbs;
int solved_split_pbs[32];
int exists_split_pbs[32];
int exists_split_lits[32];
int pick_split_pbm1, pick_split_pbm2;
int pick_split_pbm3, pick_split_pbm4;
int decomposible_clauses[32], dc_ptr;
int final_split_vector[32][17];

/*------------------------------------------------------------*/
/* determine_potential_changes() ..................#1         */
/*                                                            */
/* called by:                                                 */
/*                                                            */
/* calls:   skip_equal_all_vars()                             */
/*          skip_equal_first_last()                           */
/*          tdump.c/print_clause()                            */
/*                                                            */
/* determines a list of potential changes that can be made to */
/* the base clauses to weaken them.  At most 2000 changes can */
/* be listed on potential_change_array[][].                   */
/*------------------------------------------------------------*/

void determine_potential_changes()
{
int i, j, jj, k, kk, a, *p, v_cnt, position, litl_index;
int jjtemp, var_cnt[64], p_value;
literal *lit;
clause *c;
int second_half;
int const_sub,var_sub,fxn_sub,lit_del, cls_split;
int meg, kil;
int unifies;

/*--------------------------------------------------*/
/* Heuristic function to weight weakeneings:        */
/*     1: Constant in set-of-support                */
/*     2: function in set of support                */
/*     3: literal in set-of-support                 */
/*     4: Constant in axioms                        */
/*     5: Function in axioms                        */
/*     6: Literal in axioms                         */
/*--------------------------------------------------*/

/*--------------------------------------------------*/
/* List positions where a constant can be replaced  */
/*--------------------------------------------------*/

for(i = 0; i < 2000; i++) {
   for(j = 0; j < 7; j++) {weak[i][j] = 0;}}

const_sub = var_sub = fxn_sub = lit_del = cls_split = 0;
meg = 100000000;
kil = 100000;
k = -1;
for(i = base; i >= 0; i--) {

/*--------------------------------------*/
/* for base clauses                     */
/*--------------------------------------*/

   for(j = 0; j < 64; j++) {var_cnt[j] = 0;}
   c = clist[i];
   if(c->c_id == most_dc1 || c->c_id == most_dc2) continue;
   if(c->c_id == most_dc3 || c->c_id == most_dc4) continue;
   if(c->c_bits & EQCLS) {continue;}
   if(c->c_nlit == 1 && skip_equal_all_vars(c)) {continue;}
   v_cnt = 0;
   kk = k;
   litl_index = 0;
   for(lit = c->c_lit; lit != 0; lit = lit->l_next) {
      p = &lit->l_id;
      litl_index++;position = second_half = 0;
      for(a = *p++ & ARGMASK; a > 0; a--) {
	 position++;
	 if(a == 1 && second_half == 0) {second_half = position;}
         if(*p & FCNMASK) {
            if((*p & ARGMASK) == 0) {
               if(!skip_equal_first_last(c,p)) {

/*----------------------------------------*/
/* the replacement of any constant by a   */
/* variable is a potential change.        */
/* for clauses in the set-of-support      */
/*----------------------------------------*/

                  if(k < 1500) {
		     k++;
		     const_sub++;
	             weak[k][0] = c->c_id;
	             weak[k][1] = litl_index;
	             weak[k][2] = position;
                     weak[k][3] = *p;
                     weak[k][5] = a;
		     p_value = (*p & IDMASK) >> 6;
		     if(c->c_id >= base - nc + 1) {
                        weak[k][6] = meg + (const_sub - a)*kil - 
			100*c->c_id - (p_value*position);}
                     else weak[k][6] = 3*meg + (const_sub - a)*kil - 
		        100*c->c_id - (p_value*position);}}}
	    else {

/*---------------------------------------*/
/* replace function with variable        */
/* but..... first ELIMINATE weakenings   */
/* for unit clauses                      */
/* where pred is ~equal(f....,g....)     */
/* and weakening replaces f with var     */
/* or g with a var, as in these cases    */
/* weakening won't help much!!.          */
/*---------------------------------------*/

               if(lit->l_id == (equal_pred | FCNMASK) && c->c_nlit==1) {
                  if(position == 1 || position == second_half) { 
		     goto Skip_saving;}}

	       if(k < 1500) {
                  k++;
		  fxn_sub++;
	          weak[k][0] = c->c_id;
	          weak[k][1] = litl_index;
	          weak[k][2] = position;
	          weak[k][3] = *p;
                  weak[k][5] = a;
		  if(c->c_id >= 3*base - nc + 1) {
                     weak[k][6] = meg + (fxn_sub - a)*kil - 
		     100*c->c_id - position;}
                  else weak[k][6] = 3*meg + (fxn_sub - a)*kil - 
		       100*c->c_id - position;}}
Skip_saving:
            a += *p++ & ARGMASK;}
         else {
	    if(var_cnt[*p + 1] == 0) {
	       var_cnt[*p + 1] = *p + 1;
	       v_cnt++;
	       if(v_cnt > o_maxvar) goto Skip_saving;}
            else {
               if(k < 1500) {
                  k++;
		  var_sub++;
                  weak[k][0] = c->c_id;
                  weak[k][1] = litl_index;
                  weak[k][2] = position;
		  weak[k][3] = *p;
                  weak[k][5] = a;
                  if(c->c_id >= base - nc) {
                     weak[k][6] = 3*meg + (var_sub - a)*kil - 2000*c->c_id
		            - position;}
                  else weak[k][6] = 3*meg + (var_sub - a)*kil - 2000*c->c_id
		            - position;}}
	   
            p++;}}}

   for(j = kk + 1; j <= k; j++) {weak[j][4] = v_cnt;}}

/*-------------------------------------------------*/
/* list literals that can be deleted from clauses  */
/* with more than 2 literals.                      */
/*-------------------------------------------------*/

for(i = base; i >= 0; i--) {
   c = clist[i];
   if(c->c_id == most_dc1 || c->c_id == most_dc2) continue;
   if(c->c_id == most_dc3 || c->c_id == most_dc4) continue;
   if(c->c_bits & EQCLS) {continue;}
   litl_index = 0;
   for(lit = c->c_lit; lit != 0; lit = lit->l_next) {
      litl_index++;
      if((litl_index == c->c_nlit && litl_index > 1) 
		      || (c->c_nlit > 1 && (i >= base - nc))) {
         if(k < 1999) {
            k++;
	    lit_del++;
            weak[k][0] = c->c_id;
            weak[k][1] = litl_index;
            weak[k][2] = -1;
            weak[k][3] = -1;
            weak[k][4] = c->c_nv;
            if(c->c_id >= base - nc + 1) {
	       weak[k][6] = meg + lit_del*kil - 100*c->c_id;}
            else weak[k][6] = meg + lit_del*kil - 100*c->c_id;
	    for(j = 0; j < dc_ptr; j++) {
	       if(c->c_id == decomposible_clauses[j]) {
	          cls_split++;
	          weak[k][6] = meg 
		         + (cls_split*kil/2) - 100*c->c_id;}}}}}} 

if(weak[k][0] == 0 && k >= 0) k--;

/*-----------------------------------------*/
/* sort based on heuristic score           */
/*-----------------------------------------*/

if( k != -1) {
   for(i = 0; i < k; i++) {
      for(j = i + 1; j <= k; j++) {
         if(weak[i][6] <= weak[j][6]) {continue;}
         for(jj = 0; jj < 7; jj++) {
            jjtemp = weak[i][jj];
            weak[i][jj] = weak[j][jj];
            weak[j][jj] = jjtemp;}}}}

/*-------------------------------------------------*/
/* print potential changes                         */
/*-------------------------------------------------*/

if(o_debug > 0 || TC == 0) {
   if(k == -1) {
      printf("\n\nPOTENTIAL CHANGE ARRAY: no modifications possible\n");
      p_change_number = 0;
      return;}
   else {
      printf(
      "\n\nPOTENTIAL CHANGE ARRAY: %d modifications possible\n",k+1);}
      printf(
      " # Clause  Literal  Position  Contents  Variables Heuristic\n");}

if(o_debug > 0 || TC == 0) {
for(i = 0;  i <= k && i < 60; i++) {
printf("%2d %5d    %3d       %2d      %8x     %2d    %1d %d\n",
	 i,
	 weak[i][0], weak[i][1], weak[i][2], weak[i][3],
	 weak[i][4], weak[i][6], weak[i][5]);}
   printf("\n\n");}

/*--------------------------------------------------*/
/* print clauses that can be modified               */
/*--------------------------------------------------*/

p_change_number = k + 1;
return;
}

/*-----------------------------------------------------------*
* modify_clause() .................................#2        *
*                                                            *
* modifies a clause as desired                               *
*                                                            *
* called from: main()             theo.c                     *
*                                                            *
* calls:       determine_potential_changes()                 *
*              print_clause()                                *
*              grab_number()                                 *
*              clean_var_names()                             *
*                                                            *
*-----------------------------------------------------------*/

void modify_clause()
{
int j, mm;
char ichar;
clause *c;
int timeTook, timeToGo;
int good1;
int tc;
literal *lit2;

/*--------------------------------------------*
* determine clause to modify                  *
*--------------------------------------------*/

if(spawned > 1) analyze_decomposeability(0);
determine_potential_changes();
if(TC == 0) return;

/*--------------------------------------------*
* if no weakenings possible, return           *
*--------------------------------------------*/

if(p_change_number < 2) {weakened_ply = 1;return;}

/*--------------------------------------------*
* pick a weakening                            *
*--------------------------------------------*/

timeTook = time_now() - s_starttime;
timeToGo = o_batchtime - timeTook;

sprintf(OutPtr++,"\n");
OutPtr += sprintf(OutPtr,
     "\nC%d: -------------------------------------------------\n",TC);
OutPtr += sprintf(OutPtr,"C%d:                MODIFY THEOREM\n",TC);

OutPtr += sprintf(OutPtr,
          "C%d: Spawned = %d   Number of weakenings = %d   Pass = %d\n",
	  TC, spawned, p_change_number, pass_count++);

/*----------------------------------------------------*
* if remaining time is less than .2 of total time,    *
* then do not weaken any clause.                      *
*----------------------------------------------------*/

if(timeToGo < o_batchtime/5) {
   OutPtr += sprintf(OutPtr,
            "C%d: Time left < .2 of total => No weakening.\n",TC);
   o_maxtime = timeToGo;
   weakened_clause_count = 0;
   used_nc = 0;
   inc_o_maxlit = 1;
   o_maxlit += inc_o_maxlit;
   return;}

o_maxtime = timeToGo;

OutPtr += sprintf(OutPtr,
       "C%d: Nodes searched last time = %d\n",
       TC,s_nod_last);
OutPtr += sprintf(OutPtr,
       "C%d: Time taken=%d  Time to go=%d  Time for this pass=%d\n",
       TC, timeTook,timeToGo,o_maxtime);

if(o_learn == 1 && clauseToWeaken != -1) {
   if(weakened_clause_count == 1) {
      if(weakened_proof_remember == 0) {
         o_maxlit++;
	 o_maxterm++;
         weakened_clause_count = 1;
         for(mm = 0; mm <= base; mm++) {good_clauses[mm] = 0;} 
	 randy = (randy + spawned - 1) % p_change_number;
         OutPtr += sprintf(OutPtr,
	    "C%d: No proof => Try  weakening %d.\n",TC,randy);}

      if(weakened_proof_remember == 1) {
         good1 = change_array[1][1];
         OutPtr += sprintf(OutPtr,
	        "C%d: Proof obtained by weakening clause %d.  ",
                TC,good1);
	 if(good_clauses[good1 - 1] == 1) {
	    OutPtr += sprintf(OutPtr,"(used).\n");}
	 else {OutPtr += sprintf(OutPtr,"(not used).\n");}
	 
         if(repeat_theorem || numb_trivial_proofs > 1 || length_of_proof > 6  ||
			 (length_of_proof > 1 && used_nc)) {
	    OutPtr += sprintf(OutPtr,
            "C%d: Set w_clause_count to 0, o_maxtime=timeToGo\n",TC);
	    weakened_clause_count = 0;
	    o_maxtime = timeToGo;
	    if(numb_trivial_proofs <= 1) {
               OutPtr += sprintf(OutPtr,
                  "C%d Single w-proof obtained => Try none.\n",TC);
               OutPtr += sprintf(OutPtr,
     "C%d: -------------------------------------------------\n\n",TC);}
            if(numb_trivial_proofs > 1) {
               OutPtr += sprintf(OutPtr,
                   "C%d: Too many trivial proofs => Try none.\n",TC);
               OutPtr += sprintf(OutPtr,
     "C%d: --------------------------------------------------\n\n",TC);}
	    used_nc = 0;
            inc_o_maxlit = 1; 
            o_maxlit += inc_o_maxlit;
	    return;}
         else {

/*----------------------------------------------*
* trivial proof found, try another.             *
*----------------------------------------------*/

            for(mm = 0; mm <= base; mm++) {good_clauses[mm] = 0;}
	    weakened_clause_count = 1; 
	    numb_trivial_proofs++;
            randy = (randy + spawned - 1) % p_change_number;
            OutPtr += sprintf(OutPtr,
            "C%d: #%d trivial %d-step w-proof obtained; new randy=%d\n",
	    TC,numb_trivial_proofs,length_of_proof,randy);}}}

   else if(weakened_clause_count == 0) {
      OutPtr += sprintf(OutPtr,
     "C%d: --------------------------------------------------\n\n",TC);
      used_nc = 0;
      return;}}

/*-----------------------------------------------*
* first pass....                                 *
*-----------------------------------------------*/

if(o_learn == 1 && clauseToWeaken == -1) {
   used_nc = 0;
   weakened_clause_count = 1;

   if(randy == -1) {
      if(TC <= (spawned - 1)/4) randy = (((TC + 1)/2) - 1) % p_change_number; 
      else randy = (TC - (spawned - 1)/8) % p_change_number;}}

/*---------------------------------------------*
* m is the integfer value of the new variable  *
* determine pointer to literal to modify       *
* and then modify literal: replace l with m    *
*---------------------------------------------*/

/*----------------------------------------------*
* form initial weakening                        *
*----------------------------------------------*/

weakened_proof_remember = 0;
form_weakened_clause(randy);
j = weak[randy][0];
questionable[j] = 1;
c = clist[j - 1];
c->c_bits |= NC;
c->c_nv = weak[randy][4] + 1;

if(most_dc4 != -1) { 
   OutPtr += 
   sprintf(OutPtr,
       "C%d: X%d V%d L%d W%d SPLIT PBS %d:%d.%d.%d.%d  R%d\n",
        TC,o_maxterm,o_maxvar,o_maxlit,randy,
	split_pbs,pick_split_pbm1,
	pick_split_pbm2,pick_split_pbm3,pick_split_pbm4,
        repeat_theorem);
   printf(
       "C%d: X%d V%d L%d W%d SPLIT PBS %d:%d.%d.%d.%d  R%d TH%d\n",
        TC,o_maxterm,o_maxvar,o_maxlit,randy,
	split_pbs,pick_split_pbm1,
	pick_split_pbm2,pick_split_pbm3,pick_split_pbm4,
        repeat_theorem,theorem_number);
	if(TC!=0) fflush(stdout);}
else if(most_dc3 != -1) { 
   OutPtr += 
   sprintf(OutPtr,"C%d: X%d V%d L%d W%d SPLIT PBS %d:%d.%d.%d  R%d\n",
        TC,o_maxterm,o_maxvar,o_maxlit,randy,
	split_pbs,pick_split_pbm1,pick_split_pbm2,pick_split_pbm3,
	repeat_theorem);
   printf("C%d: X%d V%d L%d W%d SPLIT PBS %d:%d.%d.%d  R%d TH%d\n",
        TC,o_maxterm,o_maxvar,o_maxlit,randy,
	split_pbs,pick_split_pbm1,pick_split_pbm2,pick_split_pbm3,
        repeat_theorem,theorem_number);
	if(TC!=0) fflush(stdout);}
else if(most_dc2 != -1) { 
   OutPtr += sprintf(OutPtr,
        "C%d: X%d V%d L%d W%d SPLIT PBS %d:%d.%d  R%d\n",
        TC,o_maxterm,o_maxvar,o_maxlit,randy,
	split_pbs,pick_split_pbm1,pick_split_pbm2,
	repeat_theorem);
   printf("C%d: X%d V%d L%d W%d SPLIT PBS %d:%d.%d  R%d TH%d\n",
        TC,o_maxterm,o_maxvar,o_maxlit,randy,
	split_pbs,pick_split_pbm1,pick_split_pbm2,
	repeat_theorem,theorem_number);
	if(TC!=0) fflush(stdout);}
else if(most_dc1 != -1) {
   printf("C%d: X%d V%d L%d W%d SPLIT PBM: %d.%d  R%d TH%d\n",
        TC,o_maxterm,o_maxvar,o_maxlit,randy,
	split_pbs,this_split,repeat_theorem,theorem_number);
	if(TC!=0) fflush(stdout);
   OutPtr += sprintf(OutPtr,
          "C%d: X%d V%d L%d W%d SPLIT PBM: %d.%d  R%d\n",
        TC,o_maxterm,o_maxvar,o_maxlit,randy,
	split_pbs,this_split,repeat_theorem);}
else {
   printf("C%d: X%d V%d L%d W%d TH%d\n",
   TC,o_maxterm,o_maxvar,o_maxlit,randy,theorem_number);
   OutPtr += sprintf(OutPtr,"C%d: X%d V%d L%d W%d TH%d\n",
   TC,o_maxterm,o_maxvar,o_maxlit,randy,theorem_number);
   if(TC!=0) fflush(stdout);}

if(TC == 0) print_change_array();

change_array[change_level][0] = 0;
change_array[change_level][1] = j;                 /* j = c->id */
change_array[change_level][2] = weak[randy][1];
change_array[change_level][3] = weak[randy][2];
change_array[change_level][4] = weak[randy][4];
change_array[change_level][5] = replaced_char;
change_array[change_level + 1][0] = 0;
change_array[change_level + 1][1] = 0;
clean_var_names(c);

if(c->c_nlit == 1) {
   for(lit2 = c->c_lit; ;lit2 = lit2->l_next) {
      if((lit2->l_id & ELIMLIT) == 0) break;}
   normalize_literal(lit2,c);}

prepare_exit:
used_nc = 0;
clauseToWeaken = j;
OutPtr += sprintf(OutPtr,
     "C%d: ---------------------------------------------------------------\n\n",TC);
return;
}

/*-----------------------------------------------------------*
* form_weakened_clause() ........................#3          *
*                                                            *
* mark clauses that were derived from modified clauses       *
* and their descendants                                      *
*                                                            *
* called by:                                                 *
*                                                            *
* calls:     Nobody                                          *
*                                                            *
*-----------------------------------------------------------*/

void form_weakened_clause(randy)
int randy;
{
int j, k, kk, l, f, m, nv, n, dt, aa, lit_end;
int position, a, *p, *pp, *ppp;
clause *c;
literal *lit;
int tc;

j = weak[randy][0];
k = weak[randy][1];
l = weak[randy][2];
f = weak[randy][3];
m = weak[randy][4];
nv = weak[randy][4];
c = clist[j - 1];

/*---------------------------------------------*
* now carry out the weakening of the clause    *
* first determine the literal to weaken.       *
*---------------------------------------------*/

kk = k;
for(lit = c->c_lit; lit != 0; lit = lit->l_next) {
   if(--kk == 0) break;}

if(l != 0 && l != -1) {
   p = &lit->l_id;
   position = 1;
   if(nv > o_maxvar - 3) {o_maxvar++;}
   for(a = *p++ & ARGMASK; a > 0; a--) {
      if(*p & FCNMASK) {
         if(position++ == l) {
            if((*p & ARGMASK) == 0) {

/*-----------------------------------------*
* replace constant with a variable!        *
*-----------------------------------------*/

               replaced_char = *p;*p = nv++;break;}
            else {

/*-----------------------------------------*
* replace function with a variable!        *
*-----------------------------------------*/

               pp = p;
	       dt = 1;

/*-----------------------------------*
* find end of function               *
*-----------------------------------*/

	       for(aa = *pp++ & ARGMASK; aa > 0; aa--) {
                  if(*pp & FCNMASK) {aa += *pp++ & ARGMASK;}
                  else {pp++;}}

/*-----------------------------------*
* p points to fxn to be replaced     *
* pp points to first term after fxn  *
*-----------------------------------*/

               if(lit->l_next != 0) lit_end = (int) lit->l_next;
	       else lit_end = (int) c->c_end;

	       for(ppp=p;(unsigned int) ppp<(unsigned int) lit_end; ppp++) {
		  deleted_terms[dt++] = *ppp;}

               deleted_terms[0] = dt - 1;

/*------------------------------------*
* find end of literal                 *
*------------------------------------*/

                replaced_char = *p;
                *p = nv++;
                p++;
                while((unsigned int) pp < (unsigned int) lit_end) {
                   *p = *pp;
                   p++;pp++;}
                break;}}
          a += *p++ & ARGMASK;}
       else {
          if(position++ == l) {
             replaced_char = *p;
             n = *p;*p = nv++;break;}
          p++;}}}

/*------------------------------------*
* delete literal                      *
*------------------------------------*/

else if(l == -1) {
   replaced_char = 0;
   c->c_nlit--;
   lit->l_id = lit->l_id | ELIMLIT;}

return;
}

/*------------------------------------------------------------*
* mark_questionable() ............................#4         *
*                                                            *
* mark clauses that were derived from modified clauses       *
* and their descendants as modified (or questionable).       *
*                                                            *
* called by:                                                 *
*                                                            *
* calls:     Nobody                                          *
*                                                            *
*-----------------------------------------------------------*/

void mark_questionable()
{
int i, lp, rp;

if(o_learn == 0 || weakened_clause_count == 0) return;

total_questionable = 0;

lp = rp = 0;
for(i = org_base + 1; i <= base; i++) {
   if(clist[i]->c_lpar != 0) lp = clist[i]->c_lpar->c_id;
   if(clist[i]->c_rpar != 0) rp = clist[i]->c_rpar->c_id;
   if(reinstate_parent[i] != 0) {
      rp = reinstate_parent[i];}

/*----------------------------------------------*
* if either parent is questionable, mark this   *
* clause as questionable and increase count     *
* of questionable clauses.                      *
*----------------------------------------------*/

   if(questionable[lp] == 1) {
      questionable[i+1] = 1;total_questionable++;}
   else if(questionable[rp] == 1) {
      questionable[i+1] = 1;total_questionable++;}}

for(i = 0;cl[i][0] != 0 ;i++) {
   lp = ((cl[i][0] >> 4) & PARENT_MASK);
   rp = ((cl[i][1] >> 4) & PARENT_MASK);
   if(questionable[lp] == 1) {
      questionable[base + 2 + i] = 1;total_questionable++;}
   if(questionable[rp] == 1) {
      questionable[base + 2 + i] = 1;total_questionable++;}}
return;
}

/*----------------------------------------------------------*
* print_change_array() ........................#5           *
*                                                           *
* called by:                                                *
*                                                           *
* calls:      Nobody                                        *
*                                                           *
*----------------------------------------------------------*/

void print_change_array()
{

printf("\nLEVEL  CLAUSE  LITERAL POSITION     NEW   OLD HEURISTIC\n");
OutPtr += sprintf(OutPtr,
       "\nLEVEL  CLAUSE  LITERAL POSITION     NEW   OLD HEURISTIC\n");

printf("%3d   %5d %7d %8x  %8x    %8x\n",
      change_array[1][0],change_array[1][1],change_array[1][2],
      change_array[1][3],change_array[1][4],change_array[1][5]);
OutPtr += sprintf(OutPtr,"%3d   %5d %7d %8x  %8x    %8x\n",
      change_array[1][0],change_array[1][1],change_array[1][2],
      change_array[1][3],change_array[1][4],change_array[1][5]);
if(weakened_clause_count == 2) {
   printf("%3d   %5d %7d %8x  %8x    %8x\n",
      change_array[2][0],change_array[2][1],change_array[2][2],
      change_array[2][3],change_array[2][4],change_array[2][5]);
   OutPtr += sprintf(OutPtr,"%3d   %5d %7d %8x  %8x    %8x\n",
      change_array[2][0],change_array[2][1],change_array[2][2],
      change_array[2][3],change_array[2][4],change_array[2][5]);}
printf("\n");
sprintf(OutPtr++,"\n");
}

/*-----------------------------------------------------------*
* reconsider_proof() ............................#6          *
*                                                            *
* called by:                                                 *
*                                                            *
* calls:      reinstate_base_clause()                        *
*             bres_instantiated_proof()                      *
*             bfac_instantiated_proof()                      *
*             bsub_instantiated_proof()                      *
*             filter_ulist()                                 *
*                                                            *
* examine proof to see what can be used for next proof       *
* (1) substitute constants, etc., back into original clauses.*
* (2) try proof with revised clauses.                        *
* (3) add good clauses to ulist!                             *
*-----------------------------------------------------------*/

void reconsider_proof(proof_fnd)
int proof_fnd;
{
int i, ii, j, jj, k, ui, clpar, crpar, clindex, crindex;
int c1, c2;                      /* index of clauses to resolve  */
int c1_index, c2_index;          /* index of literals to resolve */
int nextu0, xl_ptr;
clause *c, *c_nc;
literal *lit, *lit2, *lit3;
int TCt;

/*---------------------------------------*
* Undo changes to base clauses           *
*---------------------------------------*/

if(o_learn == 0 || weakened_ply == 0) return;

used_nc = 0;
for(used = org_base - nc + 1; used <= org_base; used++) {
   c_nc = clist[used];
   if(c_nc->c_bits & LIST) {
   used_nc++;}}

/*
OutPtr += sprintf(OutPtr,"\n----------------------------");
OutPtr += sprintf(OutPtr,"-----------------------------------\n");
OutPtr += sprintf(OutPtr,"                       RECONSIDER PROOF\n\n");
*/

nextu0 = nextu;
reinstate_base_clause();

/*--------------------------------------*
* clear xl array.....                   *
*--------------------------------------*/

for(i = 0; xl[i][0] != 0 ; i++) {
   for(ii = 0; ii < 14; ii++) {
      xl[i][ii] = 0;}}
xl_ptr = 0;

if(satisfiable == 1) {
   nextu = 0;
   ulist[nextu] = (clause *) u_space;
   return;}

if(proof_fnd!=1 || (proof_fnd=1 && 
(used_nc==0 || (ply+good_inferred_cls<2 && split_pbs==0)))) {
   OutPtr += sprintf(OutPtr,"Dont add clauses to ulist\n");
   nextu = 0;
   ulist[nextu] = (clause *) u_space;
   return;}

/*---------------------------------------*
* try proof with revised clauses         *
* examine first Phase 0 clauses          *
*---------------------------------------*/

for(i = org_base + 1; i <= base; i++) {
   if(!(clist[i]->c_bits & LIST)) {continue;}
   if(clist[i]->c_lpar == 0) {continue;}

/*---------------------------------------*
* determine clpar,crpar,clindex,crindex  *
* in order to form resolvent.            *
*---------------------------------------*/

   clpar = clist[i]->c_lpar->c_id;
   if(clist[i]->c_rpar != 0) crpar = clist[i]->c_rpar->c_id;
   else crpar = 0;
   clindex = clist[i]->c_indices >> 6;
   crindex = (clist[i]->c_indices & 0x0000003F);

   xl[xl_ptr][0] = clist[i]->c_id << 16;
   xl[xl_ptr][0] = xl[xl_ptr][0] + (clpar << 4) + clindex;
   xl[xl_ptr][1] = (crpar << 4) + crindex;
   xl_ptr++;

   if(questionable[clist[i]->c_id] == 0) {continue;}

   if(crpar == 0) {
      if(bfac_instantiated_proof(clpar,clindex,crindex,clist[i])) {
         ulist[nextu - 1]->c_id = clist[i]->c_id;
         continue;}}
 
   if(bres_instantiated_proof(i,clpar,crpar,clindex,crindex,clist[i])) {
      ulist[nextu - 1]->c_id = clist[i]->c_id;}}

/*---------------------------------------*
* save cl array in xl array...           *
*---------------------------------------*/

for(i = 0; cl[i][0] != 0; i++) {
   for(ii = 0; ii < 14; ii++) {
      xl[xl_ptr][ii] = cl[i][ii];}
   xl[xl_ptr][13] = 0;
   xl_ptr++;}

/*---------------------------------------*
* Now consider phase 1,2 clauses         *
*---------------------------------------*/

k = 0;
for(i = base + 1; i <= base + ply; i++) {
next_clause:

   if(excellent_weakened_proof == 1) {
      for(jj = 0; jj < nextu; jj++) {
         ulist[jj]->c_bits |= BASE;
         ulist[jj]->c_lpar = clist[i]->c_lpar;
         ulist[jj]->c_rpar = clist[i]->c_rpar;
         ulist[jj]->c_indices = clist[i]->c_indices;}
      return;}

   k++;
   if(questionable[i+1] == 0) {continue;}

/*-------------------------*
* substitute string??      *
*-------------------------*/

   c1       = (cl[k - 1][0] >> 4) & PARENT_MASK;
   c1_index = (cl[k - 1][0] & INDEX_MASK);
   c2       = (cl[k - 1][1] >> 4) & PARENT_MASK;
   c2_index = (cl[k - 1][1] & INDEX_MASK);

   if(cl[k - 1][11] > 0 && clist[c1 - 1]->c_nlit > 1) {
      goto det_clause;}

   if(c2 == 0 && c2_index == 0) {goto det_next_clause;}

   if(c1_index == 0) {

/*----------------------------------------------------*
* substitute string-- not programmed to handle        *
*----------------------------------------------------*/

      for(j = i + 1; j < ply + base; j++) {
         clpar = clist[j]->c_lpar->c_id;
         if(clpar <= base) {i = j; break;}}
      if(i != j) i = j + 1;
      if(i >= base + ply) {break;}}

det_clause:
   clpar = clist[i]->c_lpar->c_id;
   if(clist[i]->c_rpar != 0) {crpar = clist[i]->c_rpar->c_id;}
   else crpar = 0;
   clindex = clist[i]->c_indices >> 6;
   crindex = (clist[i]->c_indices & 0x0000003F);

    if(clpar*crpar*clindex*crindex == 0) {
det_next_clause:

/*------------------------------*
* try to instantiate            *
*------------------------------*/

       if(c2 == 0 & c2_index == 0) {
          if(bsub_instantiated_proof(c1)) {continue;}}

/*------------------------------*
* try to factor                 *
*------------------------------*/

       if(crpar == 0) {
          clpar = clist[i]->c_id;
          if(bfac_instantiated_proof(clpar,clindex,crindex,clist[i])) {
             ulist[nextu - 1]->c_id = clist[i]->c_id;
             for(ui = 0; ui < nextu - 2; ui++) {
                if(ulist[ui]->c_id == clpar) {
		   ulist[ui]->c_bits &= ~ACTIVE;}}
             continue;}}

       for(j = i + 1; j < base + ply; j++) {
          clpar = clist[j]->c_lpar->c_id;
	  if(clpar <= base) {i = j;break;}}
       if(i != j) i++;
       if(i < ply + base) {continue;}
       else {break;}}

   if(bres_instantiated_proof(i,clpar,crpar,clindex,crindex,clist[i])) {
      ulist[nextu - 1]->c_id = clist[i]->c_id;
      for(ui = 0; ui < nextu - 2; ui++) {
         if(ulist[ui]->c_id == clpar) {
	    ulist[ui]->c_bits &= ~ACTIVE;}}}}
	    
for(j = 0; j < nextu; j++) {
   c = ulist[j];
   c->c_nlit = 0;
   for(lit = c->c_lit; lit != 0; lit = lit->l_next) {
      c->c_nlit++;
      lit->l_id = lit->l_id & ~ELIMLIT;}
   calc_clause_hashcode(c);
   if(c->c_nlit == 0) excellent_weakened_proof = 1;}

for(i = 0; i < nextu; i++) {
factor_again:
   c = ulist[i];
   for(lit = c->c_lit; lit != 0; lit = lit->l_next) {
      j = 0;
      for(lit2 = lit->l_next; lit2 != 0; lit2 = lit2->l_next) {
         j++;
         if(lit->l_X == lit2->l_X && lit->l_Y == lit2->l_Y) { 
	    delete_literal(c,0,lit2,j);
	    for(lit3=c->c_lit; lit3!=0; lit3=lit3->l_next) {
	       calc_lit_hc(lit3,c);}
	    calc_clause_hashcode(c);
	    goto factor_again;}}}}

if(excellent_weakened_proof == 0) final_clean(2);

c = clist[change_array[1][1] - 1];
if(change_array[1][3] == -1) {
   j = change_array[1][2];
   for(lit = c->c_lit; lit != 0; lit = lit->l_next) {
      if(--j == 0) {break;}}
   c->c_nlit++;
   calc_clause_hashcode(c);
   c->c_bits = c->c_bits & ~UNIT;
   lit->l_id = lit->l_id & ~ELIMLIT;}

filter_ulist(nextu0);
return;
}

/*-----------------------------------------------------------*
* reinstate_base_clause()  ............................#7    *
*                                                            *
* called by: reconsider_proof()                              *
*                                                            *
* calls:     clean_var_names()                               *
*                                                            *
* substitute constants, etc., back into original clause.     *
*-----------------------------------------------------------*/


void reinstate_base_clause()
{
int i, *p, fix_lit, ca_ptr;
clause *c;
literal *lit, *lit2;

/*-----------------------------------------*
* Undo changes to base clauses             *
* determine modified clause and literal    *
*-----------------------------------------*/

for(ca_ptr = 1; change_array[ca_ptr][1] != 0; ca_ptr++) {
   c = clist[change_array[ca_ptr][1] - 1];
   c->c_nv = change_array[ca_ptr][4] + 1;
   fix_lit = change_array[ca_ptr][2];

   for(lit = c->c_lit; lit != 0; lit = lit->l_next) {
      if(--fix_lit == 0) {break;}}

/*-----------------------------------------*
* three types of modifications possible    *
* the first repairs a clause in which a    *
* literal was deleted.  the second repairs *
* a literal where a function was replaced  *
* by a variable.  the third simply         *
* repairs a literal in which a constant    *
* was replaced by a variable.              *
*-----------------------------------------*/

   if(change_array[ca_ptr][3] == -1) {
      c->c_nlit++;
      lit->l_id = lit->l_id & ~ELIMLIT;
      c->c_bits = c->c_bits & ~UNIT;}

   if((change_array[ca_ptr][5] & FCNMASK) &&
      (change_array[ca_ptr][5] & ARGMASK) > 0)  {


/*------------------------------------------*
* move characters down to make space        *
*------------------------------------------*/

      c->c_nv = weak[randy][4];
      p = &lit->l_id;
      p = p + change_array[ca_ptr][3];
      if(ca_ptr == 1) {
         for(i = 1; i <= deleted_terms[0]; i++) {
	    *p++ = deleted_terms[i];}}
      else {
         for(i = 101; i <= deleted_terms[100]; i++) {
	    *p++ = deleted_terms[i];}}}

   else {
      p = &lit->l_id;
      p = p + change_array[ca_ptr][3];
      *p = change_array[ca_ptr][5];
      c->c_nv--;}

   clean_var_names(c);

   if(c->c_nlit == 1) {
      for(lit2 = c->c_lit; ;lit2 = lit2->l_next) {
         if((lit2->l_id & ELIMLIT) == 0) break;}
      normalize_literal(lit2,c);
      calc_lit_hc(lit2,c);
      calc_clause_hashcode(c);}

/* printf("Reinstated base clause: ");print_clause(c);*/}
}

/*-------------------------------------------------------*
* bres_instantiated_proof() ...................#8        *
*                                                        *
* called by:    reconsider_proof()             modify.c  *
*                                                        *
* calls:        standardize()                  infer.c   *
*               clear_var_substitution_tble()  infer.c   *
*               unify()                        unify.c   *
*               generate_clause()              gen.c     *
*               generate_iclause()             gen.c     *
*                                                        *
*-------------------------------------------------------*/

int bres_instantiated_proof(m,ci1,ci2,ix1,ix2,c3)
clause *c3;
int m, ci1, ci2, ix1, ix2;
{
literal *l1, *l2;
int i;
clause *c, *c1, *c2;
int iz1, iz2, iq1, iq2;
int TCt;

/*--------------------------------------*
* get clauses from either ulist or clist*
*--------------------------------------*/

c1 = c2 = 0;

if(ci1 > org_base + 1) {
   if(questionable[ci1] == 0) {
      c1 = clist[ci1 - 1];}
   else {
      c1 = 0;
      for(i = 0; i < nextu; i++) {
         if(ulist[i]->c_id == ci1) {
	    c1 = ulist[i];
            break;}}}}
else {c1 = clist[ci1 - 1];}

if(ci2 > org_base + 1) {
   if(questionable[ci2] == 0) {
      c2 = clist[ci2 - 1];}
   else {
      c2 = 0;
      for(i = 0; i < nextu; i++) {
         if(ulist[i]->c_id == ci2) {
	    c2 = ulist[i];
	    break;}}}}
else {c2 = clist[ci2 - 1];}

if(c1 == 0 || c2 == 0) return(0);

iz1 = iq1 = 0;
for(l1 = c1->c_lit; l1 != 0; l1 = l1->l_next) {
   iq1++;
   if(l1->l_id & ELIMLIT) continue;
   iz1++;
   if(iz1 == ix1 || iz1 == c1->c_nlit) break;}

iz2 = iq2 = 0;
for(l2 = c2->c_lit; l2 != 0; l2 = l2->l_next) {
   iq2++;
   if(l2->l_id & ELIMLIT) continue;
   iz2++;
   if(iz2 == ix2 || iz2 == c2->c_nlit) break;}

c1->c_nv = biggest_var_in_clause(c1);
c2->c_nv = biggest_var_in_clause(c2);

standardize(c1,c2);             /* standardize variables apart  */

if(((l1->l_id ^ l2->l_id) & 0x8000FFFF) == NEGMASK) {
   clear_var_substitution_tble(c1->c_bits,c1->c_nv);
   clear_var_substitution_tble(c2->c_bits,c2->c_nv);
   if(unify(&l1->l_id,&l2->l_id)) {
      if(c = generate_iclause(c1,c2,l1,l2,iz1,iz2,c3)) {
          ulist[nextu]->c_id = m + 1;
          ulist[nextu++] = c;
          ulist[nextu] = c->c_end;
	  if(c->c_nlit == 0) {
	     questionable[c->c_id] = 0;}
          return(1);}}}

return(0);
}

/*-------------------------------------------------------*
* bfac_instantiated_proof() ...................#9        *
*                                                        *
* called by: reconsider_proof()             modify.c     *
*                                                        *
* calls:     clear_var_substitution_tble()  infer.c      *
*            unify()                        unify.c      *
*            generate_iclause()             gen.c        *
*                                                        *
* returns:   1 if clause added to ulist                  *
*            0 if clause not added                       *
*                                                        *
*-------------------------------------------------------*/

int bfac_instantiated_proof(ci1,ix1,ix2,c3)
clause *c3;
int ci1, ix1, ix2;
{
literal *l1, *l2;
int i, iz1, iz2;
clause *c, *c1;
int TCt;

/*--------------------------------------*
* get clauses from either ulist or clist*
*--------------------------------------*/

if(ci1 > org_base + 1) {
   if(questionable[ci1] == 0) {c1 = clist[ci1 - 1];}
   else {
      c1 = 0;
      for(i = 0; i < nextu; i++) {
         if(ulist[i]->c_id == ci1) {
            c1 = ulist[i];
            break;}}}}
else {c1 = clist[ci1 - 1];}

if(c1 == 0) return(0);

iz1 = 0;
for(l1 = c1->c_lit; l1 != 0; l1 = l1->l_next) {
   if(l1->l_id & ELIMLIT) continue;
   iz1++;
   if(iz1 == ix1 || iz1 == c1->c_nlit) break;}

iz2 = 0;
for(l2 = c1->c_lit; l2 != 0; l2 = l2->l_next) {
   if(l2->l_id & ELIMLIT) continue;
   iz2++;
   if(iz2 == ix2 || iz2 == c1->c_nlit) break;}

c1->c_nv = biggest_var_in_clause(c1);
clear_var_substitution_tble(c1->c_bits,c1->c_nv);

if((l1->l_id  & 0x8000FFFF) == (l2->l_id & 0x8000FFFF)) {
   if(unify(&l1->l_id,&l2->l_id)) {
      if((c = generate_iclause(c1,0,l1,l2,iz1,iz2,c3)) != 0) {
          ulist[nextu]->c_rpar = 0;
          ulist[nextu++] = c;
          ulist[nextu] = c->c_end;
          return(1);}}}
return(0);
}

/*--------------------------------------------------------*
* bsub_instantiated_proof() ...................#10       *
*                                                        *
* called by:  reconsider_proof()    modify,c             *
*                                                        *
* calls:      copy_clause_all()     modify.c             *
*             replace_vars()        hash.c               *
*             normalize_literal()   hash.c               *
*                                                        *
* returns:    1 if instantiated clause is added to ulist *
*             0 if not added                             *
*-------------------------------------------------------*/

int bsub_instantiated_proof(ci1)
int ci1;
{
int i;
clause *c1;
int var1, con1, var2, con2, v1_count, v2_count, substitution_info;

/*--------------------------------------*
* get clauses from either ulist or clist*
*--------------------------------------*/

if(ci1 > org_base + 1) {
   if(questionable[ci1] == 0) {
      c1 = clist[ci1 - 1];}
   else {
      c1 = 0;
      for(i = 0; i < nextu; i++) {
         if(ulist[i]->c_id == ci1) {
            c1 = ulist[i];
            break;}}}}
else {c1 = clist[ci1 - 1];}

if(c1 == 0) return(0);

substitution_info = clist[ci1]->c_change;
var1 = substitution_info & 0x000000FF;
con1 = (substitution_info >> 8) & 0x000000FF;
var2 = (substitution_info >> 16) & 0x000000FF;
con2 = (substitution_info >> 24) & 0x000000FF;

c1->c_nv = biggest_var_in_clause(c1);
copy_clause_all(c1);
ulist[nextu]->c_id = clist[ci1 - 1]->c_id + 1;
ulist[nextu]->c_lpar = clist[ci1]->c_lpar;
v1_count = 1;
if(var2 != 0) v2_count = 1;
replace_vars(ulist[nextu],substitution_info);

normalize_literal(ulist[nextu]->c_lit,ulist[nextu]);
nextu++;
ulist[nextu] = ulist[nextu - 1]->c_end;
return(1);
}

/*---------------------------------------------------------------*
* filter_ulist() .............................#11                *
*                                                                *
* called by:                                                     *
*                                                                *
* calls:      Nobody                                             *
*                                                                *
*---------------------------------------------------------------*/

void filter_ulist(utop)
int utop;
{
clause *c, *c1, *c2, *c4;
literal *lit, *lit1;
int id1, i, j, ii, k, kk, m, mm, ulist_min, ulist_min_loc;
int c1A, c2A, notEnd, line_end, del_pred, qi;
int b_main_line, e_main_line;
int print_axioms_once, print_nc_once;
int ulistA[1000];
int clause_modified;
int TCt;

if(o_learn == 0 || weakened_clause_count == 0) {return;}
if(questionableUsed == 0) excellent_weakened_proof = 1;

/*-------------------------------------------*
* intially make all clauses ACTIVE.  Then    *
* inactive clauses that have one more        *
* literal than their successor.              *
*-------------------------------------------*/

for(i = 0; i < nextu; i++) {ulist[i]->c_bits |= ACTIVE;}

/*-------------------------------------------*
* determine clauses at end of lines.  Then   *
* keep clause and its predecessor, but       *
* delete other clauses further up line.      *
*-------------------------------------------*/

for(i = 0; i < nextu; i++) {
   c = ulist[i];
   if(!(c->c_bits & ACTIVE)) {continue;}
   if(c->c_nlit == 0 && weak[randy][3] != -1) {
      c->c_bits &= ~ACTIVE;continue;}
   line_end = c->c_id;
   notEnd = 0;
   c1 = clist[line_end]; 
   c1A = c1->c_id;
   qi = 0;
   for(kk = 0; kk < nextu; kk++) {
     if(ulist[kk]->c_id == c1A) {qi = -1;break;}}
   if(c1A==line_end+1 && c1->c_lpar->c_id==line_end && qi==-1) {
      continue;}
   else {

/*-------------------------------------------*
* end of line found.                         *
* now search for predecessor.                *
*-------------------------------------------*/

      c1 = clist[line_end - 2];
      c1A = c1->c_id;
      if(c->c_lpar == 0) continue;
      if(c1A == line_end - 1 && c1->c_id == c->c_lpar->c_id) {
         
/*-------------------------------------------*
* predecessor found.  Now delete rest of     *
* line.                                      *
*-------------------------------------------*/

         del_pred = 1; 
	 c2 = clist[c1A - 2];
         c2A = c2->c_id;
         while(del_pred == 1) {
            del_pred = 0;
	    if(c2A <= org_base || c1->c_lpar == 0) continue;
	    if(c2A != c1->c_lpar->c_id) continue;
            for(mm = 0; mm < nextu; mm++) {
	       if(ulist[mm]->c_id == c2A) {
	          ulist[mm]->c_bits &= ~ACTIVE;
	          del_pred = 1;
		  break;}}
            c1A = c2A;
	    c1 = c2;
	    c2 = clist[c1A - 2];
	    c2A--;}}}}

/*--------------------------------------------*
* scan for and delete duplicates on ulist     *
*--------------------------------------------*/

for(i = 0; i < nextu; i++) {
   id1 = ulist[i]->c_id;
   for(j = i + 1; j < nextu; j++) {
      if(id1 == ulist[j]->c_id) {ulist[j]->c_bits &= ~ACTIVE;}}}

for(i = 0; i < nextu; i++) {
   if(ulist[i]->c_change != 0) {ulist[i]->c_bits &= ~ACTIVE;}}

for(i = ply - 2; i >= 0; i--) {if(cl[i][3] != 0) break;}
b_main_line = cl[i + 1][0] >> 16;
e_main_line = cl[ply - 1][0] >> 16;

/*---------------------------------------------------*
* calculate hash code of each lit and cls on ulist   *
*---------------------------------------------------*/

for(i = 0; i < nextu; i++) {
   c = ulist[i];
   for(lit = c->c_lit; lit != 0; lit = lit->l_next) {
      calc_lit_hc(lit,c);}
   calc_clause_hashcode(c);}

/*---------------------------------------------*
* delete tautologies                           *
*---------------------------------------------*/

for(i = 0; i < nextu; i++) {
   c = ulist[i];
   for(lit = c->c_lit; lit != 0 ; lit = lit->l_next) {
      for(lit1 = lit->l_next; lit1 != 0; lit1 = lit1->l_next) {
         if(lit->l_X == -lit1->l_X && lit->l_Y == -lit1->l_Y) {
            ulist[i]->c_bits &= ~ACTIVE;}}}}

if(excellent_weakened_proof == 1) {
   questionable[weak[randy][0]] = 0;
   if(split_pbs != 0) {
      printf(
      "C%d: END OF W-PROOF to Split Problem %d of %d (w-proof sufficient)\n",
         TC,this_split,split_pbs);
      if(TC!=0) fflush(stdout);
      OutPtr += sprintf(OutPtr,
      "C%d: END OF W-PROOF to Split Problem %d of %d (w-proof sufficient)\n",
         TC,this_split,split_pbs);}
   else {
      printf("C%d: END OF W-PROOF (w-proof sufficient)\n",TC);
      if(TC!=0) fflush(stdout);
      OutPtr += sprintf(OutPtr,
             "C%d: END OF W-PROOF (w-proof sufficient)\n",TC);}


   if(split_pbs == 0) {
      OutPtr += sprintf(OutPtr,"\nC%d: THE PROOF\n",TC);}
   else {
      if(split_proof_reported == 0) {
         split_proof_reported++;
         OutPtr += sprintf(OutPtr,
         "\nC%d: THE PROOF of Split Problem %d of %d\n",
         TC,this_split,split_pbs);}}

   print_axioms_once = print_nc_once = 0;
   for(i = 0; i <= org_base; i++) {
      if(org_base + 1 - nc - i) {
         if(print_axioms_once++ == 0) 
	    OutPtr += sprintf(OutPtr,"\nAxioms:\n");}
      else if(print_nc_once++ == 0) 
         OutPtr += sprintf(OutPtr,"\nNegated conclusion:\n");
      if(o_relevent == 0 || 
          (clist[i]->c_bits & LIST) != 0 || (clist[i]->c_bits & NC) != 0) { 
          print_clause(clist[i]);}}

   for(i = org_base + 1; i <= base; i++) {
      if(clist[i]->c_bits & LIST) {
        OutPtr += sprintf(OutPtr,"\nPhase 0 clauses in proof:\n"); 
	break;}}

   for(i = org_base + 1; i <= base; i++) {
      if(clist[i]->c_bits & LIST) {
         clause_modified = 0;
         if(questionable[i + 1]) {
            questionable[i + 1] = 0;
	    for(ii = 0; ii < nextu; ii++) {
	       if(ulist[ii]->c_id == i + 1) {
	          ulist[ii]->c_bits = clist[i]->c_bits;
                  if(o_relevent == 0 || (clist[i]->c_bits & LIST) != 0) { 
		     print_clause(ulist[ii]);}
		  clause_modified++;}}}
         if(clause_modified == 0) {
           if(o_relevent == 0 || (clist[i]->c_bits & LIST) != 0) { 
              print_clause(clist[i]);}}}}

   OutPtr += sprintf(OutPtr,"\nPhases 1 and 2 clauses in proof:\n"); 
	 
   for(i = 1; i <= ply; i++) {
      if(i > 1 && clist[depth[i].d_tc]->c_lpar->c_id !=
         clist[depth[i].d_tc]->c_id - 1) {
         OutPtr += sprintf(OutPtr,"\n");}
      clause_modified = 0;
      if(questionable[base + i + 1]) {
         questionable[base + i + 1] = 0;
         for(ii = 0; ii < nextu; ii++) {
            if(ulist[ii]->c_id == base + i + 1) {
               ulist[ii]->c_bits = clist[depth[i].d_tc]->c_bits;
               if((clist[depth[i].d_tc]->c_bits & LIST) !=0) { 
	          print_clause(ulist[ii]);}
	       clause_modified++;}}}
      if(clause_modified == 0) print_clause(clist[depth[i].d_tc]);}

   if(split_pbs == 0) {
      OutPtr += sprintf(OutPtr,"\nC%d: END OF PROOF\n",TC);}
   else {
      OutPtr += sprintf(OutPtr,
      "\nC%d: END OF PROOF of Split Problem %d of %d\n",
      TC,this_split,split_pbs);}}

else {

   if(split_pbs == 0) {
      OutPtr += sprintf(OutPtr,"\nC%d: END OF W-PROOF\n\n",TC);}
   else {
      OutPtr += sprintf(OutPtr,
      "\nC%d: END OF W-PROOF of Split Problem %d of %d\n",
      TC,this_split,split_pbs);}

   if(questionableUsed == 0 || excellent_weakened_proof == 1) 
      goto prepare_return;

/*-----------------------------------------*
* sort while printing                      *
*-----------------------------------------*/

   for(i = 0; i < 500; i++) {ulistA[i] = -1;}

   for(i = 0; i < nextu; i++) {
      ulist_min = 10000;
      ulist_min_loc = -1;
      for(k = 0; k < nextu; k++) {
         if(ulistA[k] == 1) {continue;}
         if(ulist[k]->c_id < ulist_min) {
            ulist_min_loc = k;
            ulist_min = ulist[k]->c_id;}}
      if(ulist_min == 10000) {break;}
      k = ulist_min_loc;
      ulist[k]->c_bits |= BASE;
      questionable[ulist[k]->c_id] = 0;
      for(m = 0; m < nextu; m++) {
         if(ulist[m]->c_id == ulist[k]->c_id) {ulistA[m] = 1;}}}}

loop_modify:
if(questionableUsed == 0 || excellent_weakened_proof == 1) 
   goto prepare_return;

for(i = 0; i < nextu; i++) {
   c = ulist[i];
   if((c->c_bits & ACTIVE) == 0) continue;
   if(c->c_nlit != 1) continue;
   for(lit = c->c_lit; lit->l_id & ELIMLIT; lit = lit->l_next) {}
   for(j = 0; j < nextu; j++) {
      if(i == j) continue;
      c1 = ulist[j];
      if((c1->c_bits & ACTIVE) == 0) continue;
    
      if(simple_subsume2(c,c1)) {
         c1->c_bits  &= ~ACTIVE;
         goto loop_modify;}
      if(c1->c_nlit == 1) continue;}}

for(i = 0; i < nextu; i++) {
   c = ulist[i];
   if((c->c_bits & ACTIVE) == 0) continue;
   if(c->c_nlit != 1) continue;
   for(lit = c->c_lit; lit->l_id & ELIMLIT; lit = lit->l_next) {}

   for(j = 0; j < org_base; j++) {
      c1 = clist[j];
      if((c1->c_bits & ACTIVE) == 0) continue;
      if(simple_subsume2(c,c1)) {
         c1->c_bits  &= ~ACTIVE;
         goto loop_modify;}
      if(c1->c_nlit == 1) continue;}}

if(questionableUsed != 0 && excellent_weakened_proof != 1) {
   if(nextu == 0) {
      OutPtr += sprintf(OutPtr,
      "\nC%d: No useful inferences derived from w-proof.\n",TC);
      goto prepare_return;}
   OutPtr += sprintf(OutPtr,"\nC%d: Final ulist\n\n",TC);
   for(i = 0; i < nextu; i++) {
      print_clause(ulist[i]);}
   OutPtr += sprintf(OutPtr,"\nC%d: End Final ulist\n\n",TC);}

prepare_return:
weakProofEndsHere = OutPtr;

}

/*---------------------------------------------------------------*
* skip_equal_all_vars() ..........................#12            *
*                                                                *
* called by:                                                     *
*                                                                *
* calls:                                                         *
*                                                                *
* don't use a clause of form equal(x,y) where x,y the same       *
* variable for learning.  The clause is of no help               *
*                                                                *
* return: 0 if clause is not of useless form                     *
* return: 1 if clause is useless                                 *
*---------------------------------------------------------------*/

int skip_equal_all_vars(c)
clause *c;
{
int *p;

if(c->c_nlit > 1) {return(0);}
if((c->c_lit->l_id & 0x8000FFFF) != (equal_pred | FCNMASK)) {return(0);}

/*---------------------------------*
* first test for equal(a,b)        *
*---------------------------------*/

p = &c->c_lit->l_id;
p++;
if((*p & FCNMASK) && ((*p & ARGMASK) == 0)) {
   p++;
   if((*p & FCNMASK) && ((*p & ARGMASK) == 0)) {return(1);}}

/*----------------------------------*
* now test for equal(xxxxx,constant)*
* (to be completed??)               *
*----------------------------------*/

p = &c->c_lit->l_id;
p++;
if((*p & FCNMASK) == 0) {
   p++;
   if((*p & FCNMASK) == 0) {return(1);}}
return(0);
}

/*---------------------------------------------------------------*
* skip_equal_first_last() ...........................#13         *
*                                                                *
* called by:                                                     *
*                                                                *
* calls:      nobody                                             *
*                                                                *
* don't use a clause of form equal(c,xxxxxy) or equal(xxxx,c)    *
*                                                                *
* return: 0 if clause is not of useless form                     *
* return: 1 if clause is useless                                 *
*---------------------------------------------------------------*/

int skip_equal_first_last(c,pp)
int *pp;
clause *c;
{
int *p, a;
int string_length, string_length2;
int *const_ptr, *const_ptr2;

if(c->c_nlit > 1) {return(0);}
if((c->c_lit->l_id & 0x8000FFFF) != (equal_pred | FCNMASK)) {return(0);}

/*---------------------------------*
* first test for equal(a,b)        *
*---------------------------------*/

string_length = 0; string_length2 = 1;

p = &c->c_lit->l_id;

for(a = *p++ & ARGMASK; a > 1; a--) {
   if(string_length == 0) {const_ptr = p;}
   string_length++;
   if(*p & FCNMASK) {a += *p++ & ARGMASK;}
   else {p++;}}

const_ptr2 = p;

if((*p & FCNMASK) == 0) {p++;}
else {
   for(a = *p++ & ARGMASK; a > 0; a--) {
      string_length2++;
      if(*p & FCNMASK) {a += *p++ & ARGMASK;}
      else {p++;}}}

if(string_length * string_length2 == string_length ||
   string_length * string_length2 == string_length2) {

   if(string_length == 1 && string_length2 > 1 && pp == const_ptr &&
            (*pp & FCNMASK)) {return(1);}

   if(string_length2 == 1 && string_length > 1 && pp == const_ptr2 &&
            (*pp & FCNMASK)) {return(1);}}

/*----------------------------------*
* now test for equal(xxxxx,constant)*
* (to be completed??)               *
*----------------------------------*/

return(0);
}

/*---------------------------------------------------------------**
* print_xl_array()  .................................... #14
*
*----------------------------------------------------------------*/

void print_xl_array()
{
int i;

printf("\nXL ARRAY  \n\n");
for(i = 0; xl[i][0] != 0; i++) {
   printf(
      "\n%d  L %4d,%d  R %4d,%d H %d  I: %d  ST %8x CH %x H %d,%d",
      (xl[i][0] >> 16) & CIDMASK2,                    /* block   */
      (xl[i][0] >> 4) & PARENT_MASK,                  /* l_par   */
      (xl[i][0] & INDEX_MASK),                        /* index1  */
      ((xl[i][1] >> 4) & PARENT_MASK),
      (xl[i][1] & INDEX_MASK),
      xl[i][3],
      xl[i][2],
      xl[i][4],
      xl[i][5],
      xl[i][6],
      xl[i][7]);
   printf(" HR %d H %d,%d",xl[i][8],xl[i][9],xl[i][10]);
   if(xl[i][11] == 1) {printf(" R");}
   if(xl[i][11] == 2) {printf(" E");}
   if(xl[i][12] != 0) {printf(" X%d",xl[i][12]);}
   if(xl[i][13] != 0) {printf(" S%x",xl[i][13]);}}
printf("\n");

OutPtr += sprintf(OutPtr,"\nXL ARRAY  \n\n");
for(i = 0; xl[i][0] != 0; i++) {
   OutPtr += sprintf(OutPtr,
      "\n%d  L %4d,%d  R %4d,%d H %d  I: %d  ST %8x CH %x H %d,%d",
      (xl[i][0] >> 16) & CIDMASK2,                    /* block   */
      (xl[i][0] >> 4) & PARENT_MASK,                  /* l_par   */
      (xl[i][0] & INDEX_MASK),                        /* index1  */
      ((xl[i][1] >> 4) & PARENT_MASK),
      (xl[i][1] & INDEX_MASK),
      xl[i][3],
      xl[i][2],
      xl[i][4],
      xl[i][5],
      xl[i][6],
      xl[i][7]);
   OutPtr += sprintf(OutPtr," HR %d H %d,%d",xl[i][8],xl[i][9],xl[i][10]);
   if(xl[i][11] == 1) {OutPtr += sprintf(OutPtr," R");}
   if(xl[i][11] == 2) {OutPtr += sprintf(OutPtr," E");}
   if(xl[i][12] != 0) {OutPtr += sprintf(OutPtr," X%d",xl[i][12]);}
   if(xl[i][13] != 0) {OutPtr += sprintf(OutPtr," S%x",xl[i][13]);}}
OutPtr += sprintf(OutPtr,"\n");
return;         
}
 

/*---------------------------------------------------------------**
*  det_dependencies() ................................... #15
*
*----------------------------------------------------------------*/

void det_dependencies(c)
clause *c;
{
int i, j;
int c_left,c_right;
int marked; 

i = 0;
while((xl[i][0] >> 16) != c->c_id && xl[i][0] != 0) {
   i++;}
if(xl[i][0] == 0) return;

/*------------------------------------------*
* mark clause                               *
*------------------------------------------*/

xl[i][13] = xl[i][0] >> 16; 

mark_more:
c_left = (xl[i][0] & 0x0000FFF0) >> 4;
for(j = 0; xl[j][0] != 0; j++) {
   if(((xl[j][0] & 0xFFFF0000) >> 16) == c_left && xl[j][13] == 0) {
      xl[j][13] = xl[j][0] >> 16;
      break;}}

if(xl[i][1] & 0x0000FFF0) {
   c_right = (xl[i][1] & 0x0000FFF0) >> 4;
   for(j = 0; xl[j][0] != 0; j++) {
      if(((xl[j][0] & 0xFFFF0000) >> 16) == c_right && xl[j][13] == 0) {
         xl[j][13] = xl[j][0] >> 16;
         break;}}}

xl[i][13] = - xl[i][13];
i = 0;
while(xl[i][13] <= 0 && xl[i][0] != 0) i++;
if(xl[i][0] == 0) return;
else goto mark_more;

}

/*--------------------------------------------------------------------**
* analyze_decomposeability()       ....................... #16
*
* called by:  main()                    theoM.c (theoS.c)
*
* calls:      calc_clause_hashcode()    hash.c 
*---------------------------------------------------------------------*/

void analyze_decomposeability(last_split)
int last_split;
{
clause *c, *c2;
literal *l, *lit, prev_lit;
int ij, i, ii, j, k, kk, m, mm, nn, pp, *p, a, var_rows;
int empty_col, empty_col_count;
int computer_temp;
int lit_count;
int lit_has_var, clause_has_var;
int q, jq;
int decomposible_size;
int dc_ptr2;
int max_split;

if(TC == 0) {printf("C%d: Enter analyze decomposibility\n",TC);}

most_dc1 = most_dc2 = most_dc3 = most_dc4 = -1;
pick_split_pbm1 = -1;
pick_split_pbm2 = -2;
pick_split_pbm3 = -3;
pick_split_pbm4 = -4;
dc_parts1 = dc_parts2 = dc_parts3 = dc_parts4 = 0;
lits_1 = lits_2 = lits_3 = lits_4 = 0;
split_pbs = 0;
unsolved_split_pbs = 0;
decompose2 = -1;
dc_ptr = 0;
dc_ptr2 = 0;

for(i = 0; i < 32; i++) {
   for(k = 0; k < 17; k++) {
      best_var_v1[i][k] = 0;
      best_var_v2[i][k] = 0;
      best_var_v3[i][k] = 0;
      best_var_v4[i][k] = 0;
      final_split_vector[i][k] = 0;}
   decomposible_clauses[i] = 0;
   solved_split_pbs[i] = 0;
   exists_split_pbs[i] = exists_split_lits[i] = 0;
   solved_split_lits[i] = 0;}

for(ij = 0; ij <= base; ij++) {
   if(ij <= base - nc) i = ij;
   else i = (2 * base)  + 1 - nc - ij;
   c = clist[i];
   if(c->c_nlit > 1 && c->c_nlit < 16) {
      for(k = 0; k < 32; k++) {for(j = 0; j < 17; j++) {var_vector[k][j] = 0;}}

      jq = 0;
      lit_count = 0;
      for(lit = c->c_lit; lit != 0;lit = lit->l_next) {
         lit_has_var = 0;
	 p = &lit->l_id;
	 for(a = *p++ & ARGMASK; a > 0; a--) {
	    if(*p & FCNMASK) {a += *p++ & ARGMASK;}
	    else {
	       lit_has_var = 1;
	       p++;}}

/*------------------------------------------*
* if lit has no var, add row                *
*------------------------------------------*/

         if(lit_has_var == 0) {var_vector[jq++][lit_count] = 1;}
         lit_count++;}

      for(j = jq; j < c->c_nv + jq; j++) {
         m = 0;
         for(lit = c->c_lit; lit != 0;lit = lit->l_next) {

/*-------------------------------------------*
* scan i-th literal for variable j           *
*-------------------------------------------*/

            p = &lit->l_id;
            for(a = *p++ & ARGMASK; a > 0; a--) {
               if(*p & FCNMASK) {a += *p++ & ARGMASK;}
               else {
                  if((*p & MAX_VARS_M1) == j - jq) {var_vector[j][m] = 1;}
                  p++;}} 
            m++;}}

/*--------------------------------------------*
* merge rows with common variables            *
*--------------------------------------------*/

      var_rows = c->c_nv + jq;
      for(q = 0; q < var_rows - 1; q++) {
pass_again:
         for(m = q + 1; m < var_rows; m++) {
            for(j = 0; j < c->c_nlit; j++) {
               if(var_vector[q][j]==1 && var_vector[m][j]==1) {
                  for(ii = 0; ii < c->c_nlit; ii++) {
                     var_vector[q][ii] |= var_vector[m][ii];}
                  for(nn = m; nn < var_rows - 1; nn++) {      
                     for(pp = 0; pp < c->c_nlit; pp++) {
                        var_vector[nn][pp]=var_vector[nn+1][pp];}}
                  var_rows--;
                  if(var_rows == q + 1) goto print_vector;
                  goto pass_again;}}}}
print_vector:

/*---------------------------------------------*
* delete empty rows (too many vars).           *
*---------------------------------------------*/

      for(mm = var_rows - 1; mm >= 0; mm --) {
         clause_has_var = 0;
         for(kk = 0; kk < c->c_nlit; kk++) {
	    if(var_vector[mm][kk] != 0) clause_has_var++;}
         if(clause_has_var == 0) {  /*delete row */
	    for(nn = mm; nn < var_rows - 1; nn++) {
	       for(pp = 0; pp < c->c_nlit; pp++) {
	          var_vector[nn][pp] = var_vector[nn+1][pp];}}
            var_rows--;}}
	    
      
/*---------------------------------------------*
* determine number of subclauses               *
*---------------------------------------------*/

      decompose2 = var_rows;
      empty_col_count = 0;

      if(dc_parts4 != 0 && dc_parts1 == decompose2 &&  c->c_nlit > lits_1) continue;

      if(TC == 0) { 
         for(mm = 0; mm < decompose2; mm++) {
            lit_count = 0;
            for(kk = 0; kk < c->c_nlit; kk++) {if(var_vector[mm][kk]) lit_count++;}
            var_vector[mm][kk] = lit_count;}}

/*----------------------------------------------*
* add another clause to split_clause list,
* but first back up last split clause on list
*----------------------------------------------*/

      if(dc_ptr == 17) dc_ptr--;
      if(dc_ptr < 17 && var_rows > 1) {
         
         decomposible_clauses[dc_ptr++] = c->c_id;

         final_split_vector[dc_ptr - 1][0] = c->c_id;
         for(mm = 0; mm < var_rows; mm++) {
            for(kk = 0; kk < c->c_nlit; kk++) {
	       if(var_vector[mm][kk] == 1) {
	          final_split_vector[dc_ptr - 1][kk + 1] = mm + 1;}}}

         dc_ptr2++;
         if(dc_ptr2 < 18) {
            dc_parts4 = dc_parts3;
            dc_parts3 = dc_parts2;
            dc_parts2 = dc_parts1;

            lits_4 = lits_3;
            lits_3 = lits_2;
            lits_2 = lits_1;

            most_dc4 = most_dc3;
            most_dc3 = most_dc2;
            most_dc2 = most_dc1;}

         dc_parts1 = decompose2;
         lits_1 = c->c_nlit; 
         most_dc1 = c->c_id;

	 for(mm = 0; mm < 32; mm++) {
            for(kk = 0; kk < 17; kk++) {
               best_var_v4[mm][kk]=best_var_v3[mm][kk];}} 

         for(mm = 0; mm < 32; mm++) {
            for(kk = 0; kk < 17; kk++) {
               best_var_v3[mm][kk]=best_var_v2[mm][kk];}} 

	 for(mm = 0; mm < 32; mm++) {
            for(kk = 0; kk < 17; kk++) {
               best_var_v2[mm][kk]=best_var_v1[mm][kk];}} 

	 for(mm = 0; mm < 32; mm++) {
            for(kk = 0; kk < 17; kk++) {
               best_var_v1[mm][kk] = 0;}} 

         for(mm = 0; mm < decompose2; mm++) {
            for(kk = 0; kk < c->c_nlit; kk++) {
               best_var_v1[mm][kk] = var_vector[mm][kk];}
            best_var_v1[mm][kk] = var_vector[mm][kk];}}}}

/*--------------------------------------*
* Print out the final vector array      *
* Clause called "most_decomposible" is  *
* the clause to decompose.              *
*--------------------------------------*/

if(dc_parts1 < 2) {
   if(pass_count == 0) {
      if(TC == 0) printf("C%d: No splitting!\n",TC);
      OutPtr += sprintf(OutPtr,"C%d: No splitting!\n",TC);}
   most_dc1 = most_dc2 = most_dc3 = most_dc4 = -1;
   goto exit_analyze;}

if(o_debug > 0) {
   printf("Split cls: %d %d %d %d\n",most_dc1,
                most_dc2,most_dc3,most_dc4);
   OutPtr += sprintf(OutPtr,"Split cls: %d %d %d %d\n",most_dc1,
                most_dc2,most_dc3,most_dc4);}

for(mm = 0; mm < dc_parts1; mm++) {
   for(kk=0; kk < clist[most_dc1-1]->c_nlit+1; kk++) {}}

/*---------------------------------------*
* sort FINAL VECTOR1                     *
*---------------------------------------*/

for(mm = 0; mm < dc_parts1 - 1; mm++) {
   for(nn = mm + 1; nn < dc_parts1; nn++) {
      if(best_var_v1[nn][kk-1] < best_var_v1[mm][kk-1]) {
         for(pp=0; pp < clist[most_dc1-1]->c_nlit + 1;pp++) {
	    split_temp[pp] = best_var_v2[mm][pp];
            best_var_v1[mm][pp] = best_var_v1[nn][pp];
	    best_var_v1[nn][pp] = split_temp[pp];}}}}

/*---------------------------------------*
* sort FINAL VECTOR2                     *
*---------------------------------------*/

for(mm = 0; mm < dc_parts2; mm++) {
   for(kk=0; kk < clist[most_dc2-1]->c_nlit+1; kk++) {}}

for(mm = 0; mm < dc_parts2 - 1; mm++) {
   for(nn = mm + 1; nn < dc_parts2; nn++) {
      if(best_var_v2[nn][kk-1] < best_var_v2[mm][kk-1]) {
         for(pp=0; pp < clist[most_dc2-1]->c_nlit + 1;pp++) {
	    split_temp[pp] = best_var_v2[mm][pp];
            best_var_v2[mm][pp] = best_var_v2[nn][pp];
	    best_var_v2[nn][pp] = split_temp[pp];}}}}

/*---------------------------------------*
* sort FINAL VECTOR3                     *
*---------------------------------------*/

for(mm = 0; mm < dc_parts3; mm++) {
   for(kk=0; kk < clist[most_dc3-1]->c_nlit+1; kk++) {}}

for(mm = 0; mm < dc_parts3 - 1; mm++) {
   for(nn = mm + 1; nn < dc_parts3; nn++) {
      if(best_var_v3[nn][kk-1] < best_var_v3[mm][kk-1]) {
         for(pp=0; pp < clist[most_dc3-1]->c_nlit + 1;pp++) {
	    split_temp[pp] = best_var_v3[mm][pp];
            best_var_v3[mm][pp] = best_var_v3[nn][pp];
	    best_var_v3[nn][pp] = split_temp[pp];}}}}

/*---------------------------------------*
* sort FINAL VECTOR4                     *
*---------------------------------------*/

for(mm = 0; mm < dc_parts4; mm++) {
   for(kk=0; kk < clist[most_dc4-1]->c_nlit+1; kk++) {}}

for(mm = 0; mm < dc_parts4 - 1; mm++) {
   for(nn = mm + 1; nn < dc_parts4; nn++) {
      if(best_var_v4[nn][kk-1] < best_var_v4[mm][kk-1]) {
         for(pp=0; pp < clist[most_dc4-1]->c_nlit + 1;pp++) {
	    split_temp[pp] = best_var_v4[mm][pp];
            best_var_v4[mm][pp] = best_var_v4[nn][pp];
	    best_var_v4[nn][pp] = split_temp[pp];}}}}

/*---------------------------------------*
* decide literals to keep for clause     *
*---------------------------------------*/

max_split = 17;
if(spawned < 17) max_split = spawned;

if(dc_parts4!=0 && dc_parts1*dc_parts2*dc_parts3*dc_parts4<max_split) {
   unsolved_exists_pbs = dc_parts1 * dc_parts2 * dc_parts3 * dc_parts4;}

else if(dc_parts3 != 0 && dc_parts1*dc_parts2*dc_parts3 < max_split) {
   most_dc4 = -1;
   unsolved_exists_pbs = dc_parts1 * dc_parts2 * dc_parts3;} 

else if(dc_parts2 != 0 && dc_parts1 * dc_parts2 < max_split) {
   most_dc4 = most_dc3 = -1;
   unsolved_exists_pbs = dc_parts1 * dc_parts2;}

else {
   most_dc4 = most_dc3 = most_dc2 = -1;
   unsolved_exists_pbs = dc_parts1;}

split_pbs = unsolved_split_pbs = unsolved_exists_pbs;

/*
if(pass_count == 0) {
   if(TC == 0) printf("\nC%d: FINAL_SPLIT_VECTOR:\n",TC);
   OutPtr += sprintf(OutPtr,"\nC%d: FINAL_SPLIT_VECTOR:\n",TC);

   for(i = 0; i < dc_ptr; i++) {
      if(TC == 0) printf("\n");
      sprintf(OutPtr++,"\n");
      for(k = 0; k < 17; k++) {
         if(TC == 0) printf("%2d ",final_split_vector[i][k]);
         OutPtr += sprintf(OutPtr,"%2d ",final_split_vector[i][k]);}}

   if(TC == 0) printf("\nC%d: Split candidates: ",TC);
   OutPtr += sprintf(OutPtr,"\nC%d: Split candidates: ",TC);
   for(mm = 0; mm < dc_ptr; mm++) {
      if(TC == 0) printf("%d ",decomposible_clauses[mm]);
      OutPtr += sprintf(OutPtr,"%d ",decomposible_clauses[mm]);}

   if(TC == 0) {printf("\n");}
   OutPtr += sprintf(OutPtr,"\n");}

if(TC == 0) {
   printf("Split counts: %d %d %d %d\n",dc_parts1,
                dc_parts2,dc_parts3,dc_parts4);
   OutPtr += sprintf(OutPtr,"Split counts: %d %d %d %d\n",dc_parts1,
                dc_parts2,dc_parts3,dc_parts4);}
*/

if(TC == 0 || spawned < split_pbs) goto master_decompose;

/*------------------------------------*
* slaves perform this - 4-way split?  *
*------------------------------------*/

if(dc_parts4 != 0 
    && dc_parts1 * dc_parts2 * dc_parts3 * dc_parts4 < max_split) {

   this_split=1 + ((TC + repeat_theorem - 1) %
        (dc_parts1*dc_parts2*dc_parts3*dc_parts4));

   pick_split_pbm1 = 1 + ( (int) ((TC + repeat_theorem -1)
                     / (dc_parts2 * dc_parts3 * dc_parts4))
                     % dc_parts1);

   pick_split_pbm2 = 1 + ( (int) ((TC + repeat_theorem - 1)
                     / (dc_parts3 * dc_parts4))
                     % dc_parts2);

   pick_split_pbm3 = 1 + ( (int) ((TC + repeat_theorem - 1)
                     / (dc_parts4))
                     % dc_parts3);

   pick_split_pbm4 = 1 + ((TC + repeat_theorem - 1)
                     % dc_parts4);}

/*------------------------------------*
* slaves perform this - 3-way split?  *
*------------------------------------*/

else if(dc_parts3 != 0 
     && dc_parts1 * dc_parts2 * dc_parts3 < max_split) {
   most_dc4 = -1;
   this_split=1+((TC + repeat_theorem - 1) %
                      (dc_parts1*dc_parts2 * dc_parts3));
   pick_split_pbm1 = 1 + ( (int) ((TC + repeat_theorem - 1)
                     / (dc_parts2 * dc_parts3))
	             % dc_parts1);

   pick_split_pbm2 = 1 + ( (int) ((TC + repeat_theorem - 1)
                     / dc_parts3)
                     % dc_parts2);

   pick_split_pbm3 = 1 + ((TC + repeat_theorem - 1)
                     % dc_parts3);}

/*------------------------------------*
* slaves perform this - 2-way split?  *
*------------------------------------*/

else if(dc_parts2!=0 && dc_parts1*dc_parts2 < max_split) {
   most_dc4 = most_dc3 = -1;
   this_split = 1 + ((TC + repeat_theorem - 1) % (dc_parts1 * dc_parts2));
   pick_split_pbm1 = 1 + ( (int) ((TC + repeat_theorem - 1)
                     / dc_parts2)
	             % dc_parts1);
   pick_split_pbm2 = 1 + ((TC + repeat_theorem - 1) % dc_parts2);}

/*------------------------------------*
* slaves perform this - 1-way split?  *
*------------------------------------*/

else {
   most_dc4 = most_dc3 = most_dc2 = -1;
   this_split = 1 + ((TC + repeat_theorem - 1) % (dc_parts1));
   pick_split_pbm1 = 1 + ((TC + repeat_theorem - 1) % dc_parts1);}

mm = 0;
c = clist[most_dc1 - 1];
for(lit = c->c_lit; lit != 0; lit = lit->l_next) {
   lit->l_id = lit->l_id & ~ELIMLIT;
   if(best_var_v1[pick_split_pbm1 - 1][mm++] == 0) {
      c->c_nlit--;
      lit->l_id = lit->l_id | ELIMLIT;}}
calc_clause_hashcode(c);

mm = 0;
if(dc_parts2 != 0 && dc_parts1 * dc_parts2 < max_split) {
   c2 = clist[most_dc2 - 1];
   for(lit = c2->c_lit; lit != 0; lit = lit->l_next) {
      lit->l_id = lit->l_id & ~ELIMLIT;
      if(best_var_v2[pick_split_pbm2 - 1][mm++] == 0) {
         c2->c_nlit--;
         lit->l_id = lit->l_id | ELIMLIT;}}
   calc_clause_hashcode(c2);}

mm = 0;
if(dc_parts3 != 0 && dc_parts1 * dc_parts2 * dc_parts3 < max_split) { 
   c2 = clist[most_dc3 - 1];
   for(lit = c2->c_lit; lit != 0; lit = lit->l_next) {
      lit->l_id = lit->l_id & ~ELIMLIT;
      if(best_var_v3[pick_split_pbm3 - 1][mm++] == 0) {
         c2->c_nlit--;
         lit->l_id = lit->l_id | ELIMLIT;}}
   calc_clause_hashcode(c2);}

mm = 0;
if(dc_parts4 != 0 && dc_parts1*dc_parts2*dc_parts3*dc_parts4 < max_split) { 
   c2 = clist[most_dc4 - 1];
   for(lit = c2->c_lit; lit != 0; lit = lit->l_next) {
      lit->l_id = lit->l_id & ~ELIMLIT;
      if(best_var_v4[pick_split_pbm4 - 1][mm++] == 0) {
         c2->c_nlit--;
         lit->l_id = lit->l_id | ELIMLIT;}}
   calc_clause_hashcode(c2);}

master_decompose:

//computer_temp = TC;TC = 0;
if(TC == 0) {
   printf("C%d: SPLIT 1: ",TC);print_clause(clist[most_dc1 - 1]);

   if(dc_parts2 != 0 && dc_parts1 * dc_parts2 < max_split) { 
      printf("C%d: SPLIT 2: ",TC);print_clause(clist[most_dc2 - 1]);}

   if(dc_parts3 != 0 && dc_parts1 * dc_parts2 * dc_parts3 < max_split) { 
      printf("C%d: SPLIT 3: ",TC);print_clause(clist[most_dc3 - 1]);}

   if(dc_parts4!=0 && dc_parts1*dc_parts2*dc_parts3*dc_parts4 < max_split) { 
      printf("C%d: SPLIT 4: ",TC);print_clause(clist[most_dc4 - 1]);}}
//TC = computer_temp;

exit_analyze:
if(pass_count == 0) {
   OutPtr += sprintf(OutPtr,
        "C%d: -----------------------------------------------------------\n\n",TC);}

if(TC != 0) return;

for(i = 1; i <= split_pbs; i++) {solved_split_lits[i] = 1;}
return;
} 
