module ExtSat 
    ( Solver      -- *
    , ExtSatM     -- * -> *
    , run         -- ExtSolver -> ExtSatN a -> IO a
    ) where

import Control.Monad.State as S
import Control.Monad.Trans as T   ( lift )
import Data.Set            as Set ( Set, empty, member, insert, fromList )

import SatWrappers
import SatMonad

data   Solver    = Solver [[Lit]] Lit ExtSolver (Either (Set Lit) [Lit])
type   ExtSatM a = StateT Solver IO a

solver :: ExtSolver -> Solver
solver solv =  Solver [[mkTrue]] (Lit 2) solv (Left Set.empty)

run :: ExtSolver -> ExtSatM a -> IO a
run s m = evalStateT m (solver s)

instance SatMonad (StateT Solver IO) where

    mkLit        = do Solver cs l solv res <- get
                      put (Solver cs (l+1) solv res)
                      return l

    value      l = do Solver cnf _ _ _ <- get
                      let units = [ x | [ x ] <- cnf ]
                      return $
                        if l `elem` units
                          then Just True
                          else if -l `elem` units
                                 then Just False
                                 else Nothing

    modelValue l = do Solver _ _ _ (Left m) <- get
                      return ((if l < 0 then not else id) $ Set.member (abs l) m)

    contr        = do Solver _ _ _ (Right c) <- get
                      return c

    okay         = do Solver _ _ _ res <- get
                      case res of
                        Right [] -> return False
                        _        -> return True

    clause c     = do b <- okay
                      if b
                         then do Solver cs l solv res <- get
                                 put (Solver (c:cs) l solv res)
                                 return True
                         else return False

    solve_ _ a   = do b <- okay
                      if b
                         then do Solver cs l solv _ <- get
                                 r <- T.lift $ solv ([ [ l ] | l <- a ] ++ cs)
                                 case r of
                                   Nothing -> do put (Solver cs l solv (Right (map negate a)))
                                                 return False
                                   Just m  -> do put (Solver cs l solv (Left (Set.fromList (filter (Lit 0 <) m))))
                                                 return True
                         else return False

    lift = T.lift
