module Interpreter where

import System.IO                  ( hFlush, stdout, isEOF )
import System.Environment         ( getArgs )

import ReadFile

import MiniSat  as Mini
import SatMonad as M
import Prop     as P

main = 
 do args <- getArgs
    case args of
       []     -> withSolver (lower (run interp))
       ["--log", log, file] -> withSolverLog log (lower (run (interpFile file)))
       [file] -> withSolver (lower (run (interpFile file)))
       ["--pipe", file] -> 
         do str <- readFile' file
            mapM_ parseLine (lines str)

io = P.lift . M.lift

interpFile :: String -> InterpT MiniSatM ()
interpFile file = io (readFile' file) >>= mapM_ interpLine . lines
     
interpLine :: String -> InterpT MiniSatM ()
interpLine line = 
  case parse line of
    Left  c -> interpret c >>= \r -> io (print r)
    Right e -> io (putStrLn ("error: " ++ e))

parseLine :: String -> IO ()
parseLine line =
  case parse line of
    Left  c -> return ()
    Right e -> putStrLn "parse error"

interp :: InterpT MiniSatM ()
interp = do
  io $ putStr "> "
  io $ hFlush stdout
  eof <- io $ isEOF
  if eof 
    then io $ putStrLn ""
    else io getLine >>= interpLine >> interp

