module SatWrappers
    ( ExtSolver   -- *
    , limmat      -- ExtSolver
    , minisat     -- ExtSolver
    , cleanup     -- IO ()
    ) where

import System.IO        ( FilePath )
import System.Cmd       ( system )
import System.Exit      ( ExitCode(..) )
import System.Directory ( removeFile )
import Data.List        ( intersperse, isPrefixOf )

import ReadFile

import SatMonad         ( Lit(..) )

type ExtSolver = [[Lit]] -> IO (Maybe [Lit])

limmat, minisat :: ExtSolver

dimacs :: [[Lit]] -> String
dimacs clauses = concat $ ["p cnf ", show vars, " ", show (length clauses), "\n"] ++ 
                 concatMap (\c -> intersperse " " (map show c) ++ [" 0\n"]) clauses
  where vars = maximum (0 : map (abs . (\ (Lit i) -> i)) (concat clauses))

tmpcnf = "tmp-external.temporary.cnf"
tmpres = "tmp-external.temporary.res"

sat2002solver solver cs =
    do --putStrLn ("Number of clauses: " ++ show (length cs))
       writeFile' tmpcnf (dimacs cs)
       system (solver ++ " " ++ tmpcnf ++ " >" ++ tmpres)
       result <- fmap lines (readFile' tmpres)
       case filter (isPrefixOf "s ") result of
         ["s SATISFIABLE"]   -> return (Just $ filter (Lit 0/=) $ map read $ concat $ map words $
                                               map (\('v':' ':xs) -> xs) $ 
                                               filter (isPrefixOf "v ") $ result)
         ["s UNSATISFIABLE"] -> return Nothing
         _                   -> error "s-line missing from result"
       

cleanup =
    do removeFile tmpcnf
       removeFile tmpres

limmat = sat2002solver "limmat"

minisat cs = 
    do putStrLn ("Number of clauses: " ++ show (length cs))
       writeFile' tmpcnf (dimacs cs)
       system ("minisat" ++ ">/dev/null " ++ tmpcnf ++ " " ++ tmpres)
       result <- fmap lines (readFile' tmpres)
       case result of
         ["SAT", rest] -> return (Just $ filter (Lit 0/=) $ map read $ words $ rest)
         ["UNSAT"]     -> return Nothing
         _             -> error "unexpected result from minisat"
