/*-------------------------------------------------------------*/
/*      theo.c
/*      copyrighted by Monty Newborn 2005
/*
/*      functions:
/*      time_now()..................... #1
/*      init_stats()................... #2
/*      init_options()................. #3
/*      grab_number().................. #4
/*      convert_n2s().................. #5
/*      fatal()........................ #6
/*      init_search_space()............ #7
/*      init_clause_hash_table()....... #8
/*      init_random_hash_array()....... #9           

/*      enter_theorem_name()........... #10
/*      prepare_clauses().............. #11
/*      read_buffer_word()............. #12
/*      sort_constants_fxns().......... #13
/*      is_fxn_a_constant()............ #14
/*      interchange_fxn_const()........ #15
/*      format_clause()................ #16
/*      get_pred_fxn_symbols()......... #17
/*      count_lit_characters()......... #18
/*      max_term_in_any_lit().......... #19
/*      hash_all_base_clauses()........ #20

/*      free_space_close_files()....... #21
/*      iterative_deepening().......... #22
/*      get_clause_stats()............. #23
/*      write_out_results()............ #24
/*      main()........>>>>>>>>>>>>>>>>> #25
/*      fix_proof().................... #26
/*      compile_clauses().............. #27
/*      final_clean().................. #28
/*      delete_literal()............... #29
/*      adjust_ht_size()............... #30
/*-------------------------------------------------------------*/
/*
/*      LINKAGE OF FUNCTIONS
/*
/*main()--------------<<<<<<<<<<<<<<<<<<<<<<<<<.........theo.c
/*
/* begin_proof:
/* 
/*       enter_theorem_name()...........................theo.c
/*       compile_clauses()..............................affc.c
/*       init_cl_array()................................pass2.c
/*       init_stats()...................................theo.c
/*       init_search_space()............................theo.c
/*       prepare_clauses()..............................theo.c
/*             get_pred_fxn_symbols()...................theo.c
/*             read_buffer_word().......................theo.c
/*             format_clause()..........................theo.c
/*                   read_buffer_word().................theo.c
/*             sort_constants_fxns()....................theo.c
/*                   is_fxn_a_constant()................theo.c
/*                   interchange_fxn_const()............theo.c
/*             max_term_in_any_lit()....................theo.c
/*                   count_lit_characters().............theo.c
/*       init_random_hash_array().......................theo.c
/*       hash_all_base_clauses()........................theo.c
/*             calc_lit_hc()............................hash.c
/*             calc_clause_hashcode()...................hash.c
/*             add_clause_to_ht().......................hash.c
/*             calc_alt_hashcodes().....................hash.c
/* 
/*----------------------------------------------------*/
/* PHASE 0                                            */
/* -------                                            */
/* simply input clauses if possible.                  */
/*----------------------------------------------------*/
/* 
/*        eliminate()...................................elim.c
/* 
/*----------------------------------------------------*/
/* PHASE 1                                            */
/* -------                                            */
/* begin searching for a proof.                       */
/*----------------------------------------------------*/
/* 
/*        iterative_deepening().........................theo.c
/*              init_clause_hash_table()................theo.c
/*              add_clause_to_ht()......................hash.c
/*              calc_alt_hashcodes()....................hash.c
/*              search_tree()...........................search.c
/*              save_line().............................pass2.c
/*              time_now()..............................theo.c
/* 
/*        stack_hash_codes()............................pass2.c
/*      
/*----------------------------------------------------*/
/* PHASE 2                                            */
/* -------                                            */
/* search for hash codes on hash_code_list.           */
/* put on stack new hash codes that must yet be       */
/* found on another pass through phase 2.             */
/*----------------------------------------------------*/
/* 
/*       iterative_deepening()..........................theo.c
/*       stack_hash_codes().............................pass2.c
/*      
/*----------------------------------------------------*/
/* PHASE 3                                            */
/* -------                                            */
/* set phase to 3, check proof, then print results    */
/*----------------------------------------------------*/
/* 
/*       rearrange_cl_array()...........................pass2.c
/*       search_proof_line()............................pass2.c
/*       write_out_results()............................theo.c
/*             print_clause()...........................tdump.c
/*             get_clause_stats().......................theo.c
/*
/*----------------------------------------------*/
/* free allocated space and close output file   */
/*----------------------------------------------*/
/* 
/*       free_space_close_files().......................theo.c
/*
/*------------------------------------------------------------------*/
			 

#include "const.h"
#include "clause.h"
#include "compwff.h"
 
#include <strings.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>
 
/*----------------------------------------------------------------*/
/* external functions called from this file                       */
/*----------------------------------------------------------------*/

extern void CompileTptp();               /* function in tptpi.c   */
extern int CompareStrings();

extern int search_tree();               /* functions in search.c  */
 
extern int calc_lit_hc();                /* functions in hash.c   */
extern int calc_alt_hashcodes();
extern void calc_clause_hashcode();
extern int add_clause_to_ht();
extern void normalize_literal();
extern int count_vars();

extern void eliminate();                 /* function in elim.c    */
extern void analyze_decomposeability(); 
extern int elim_equal_const();

extern void init_cl_array();             /* functions in pass2.c  */
extern void save_line(); 
extern void stack_hash_codes();
extern void rearrange_cl_array();
extern int search_proof_line();
extern void print_hash_code_list();

extern void print_clause();              /* function in tdump.c   */
extern void print_clause2();             /* function in tdump.c   */
extern void print0();
extern void print_continuation();
extern void print_substitutions();

extern void equal_predicate();

extern void prepare_subarray();             /* function in gen.c  */
extern void copy_clause3();
extern int test_equality();
extern int calc_alt_ehashcodes();
extern void rename_functions();
extern void modify_clause();
extern void mark_questionable();
extern void reconsider_proof();
extern void det_dependencies();
extern void print_xl_array();

extern void standardize();                 /* function in infer.c */
extern void standardize0();

extern int unify();

extern void print_buffers();             /* function in compwff.c */

/*----------------------------------------------------------------*/
/* hash_code_list_active is a bit mapping of the hash codes on    */
/* the hash code list that remain to be found.  bit i is a 1      */
/* if hash code i has yet to be found.                            */
/*----------------------------------------------------------------*/
 
extern int hash_code_active_number;
extern int killer_hashx, killer_hashy;

/*----------------------------------------------------------------*/
/* the hash code of the clause whose complement is found in the   */
/* clause_hash_table implying a contradiction; it is thus magic!  */
/*----------------------------------------------------------------*/

extern int magic_hashx, magic_hashy;
extern int alt_HX;

/*----------------------------------------------------------------*/
/* array for substitution strings                                 */
/*----------------------------------------------------------------*/

extern int sub_array[65536][6];
extern int sub_array_last;
extern int const_found_equal;

/*----------------------------------------------------------------*/
/* array for substitution strings used in proof                   */
/*----------------------------------------------------------------*/

extern int deleted_clauses;

extern int questionable[];

extern int change_array[50][10];
extern int change_level;

extern int weak[2000][7];
extern int p_change_number;
extern int randy, randy2;
extern int excellent_weakened_proof;

extern int theorem_rating;
extern int reinstate_parent[MAX_BASE_CLAUSES];

extern int most_dc1, most_dc2;
extern int c_split, l_split;
extern int o_maxterm_tempy, o_maxvar_tempy,o_maxlit_tempy;

extern int yyerrflag, yymorfg, yyprevious;

extern char mem_array[];

/*----------------------------------------------------------------*/
/* internally defined functions                                   */
/*----------------------------------------------------------------*/

int time_now();
void init_stats();
int init_options();
int grab_number();
int convert_n2s();
void fatal();
void init_search_space();
void init_clause_hash_table();
void init_random_hash_array();

void enter_theorem_name();
void free_space_close_files();
void prepare_clauses();
void read_buffer_word();
void sort_constants_fxns();
int is_fxn_a_constant();
void interchange_fxn_const();
int get_pred_fxn_symbols();
void format_clause();
int count_lit_characters();
void max_term_in_any_lit();
void hash_all_base_clauses();

int iterative_deepening();
void get_clause_stats();
void write_out_results();
int main();
 
int fix_proof();
void compile_clauses();
void adjust_ht_size();
void final_clean();
void delete_literal();

/*----------------------------------------------------------------*/
/* arrays used for reading in input file and converting           */
/* to format compatible with the search.                          */
/*----------------------------------------------------------------*/
 
char *theobufptr;
char theoCodeBuffer[2000000];
char *theoCodePtr;

char compiled_code_buffer[MAX_COMPILED_THM_SIZE];
char *comp_code_buffer_index;
int compiled_code_size;
 
pfnode pfspace[PF_SPACE_SIZE];
int kode[MAX_INPUT_THM_SIZE];
char name_buff[MAX_CHAR];
cnode pred_ptr_list[PRED_PTR_SIZE];

/*----------------------------------------------------------------*/
/* space for output, stored temporarily until output to disk file */
/*----------------------------------------------------------------*/

char *OutPtr2, *OutPtr2a;
char *OutPtr;
char results[1000000];
char *OutPtrZ;

/*----------------------------------------------------------------*/
/* space for clause list generated in elim.c for simplifying      */
/*----------------------------------------------------------------*/

int elim_infs[ELIM_ARRAY_SIZE];        /* clause list for elim.c  */

/*----------------------------------------------------------------*/
/* clist contains pointers to the clauses (base clauses and       */ 
/* clauses generated during the search.)                          */
/*----------------------------------------------------------------*/

clause *clist[CSPACE_CLAUSES];    /* first few are base clauses   */
int good_clauses[MAX_BASE_CLAUSES];
int good_inferred_cls;     /* # inferred clauses used in proof  */
clause *ulist[1500];

/*----------------------------------------------------------------*/
/* each entry in depth is the top, bottom, and current clause     */
/* under search as well as a pointer to the current clause.       */
/* each entry is thus four int  words.                            */
/*----------------------------------------------------------------*/

dnode  depth[DEPTH_PTR_SPACE];    /* depth ptr info               */

/*----------------------------------------------------------------*/
/* hr_hashcodes store the hashcode that was used in an            */
/* instantiation (in calc_inferences())                           */
/*----------------------------------------------------------------*/

int hr_hashcodes[3][50];          /* array to store hr hashcodes  */

int cl[PROOF_ARRAY_SIZE][14];
int xl[PROOF_ARRAY_SIZE][14];

/*----------------------------------------------------------------*/
/* the following two arrays occupy a large amount of memory       */
/* the space is dynamically allocated and is currently set to:    */
/* clause space: 256K words     clause_hash_table: 256K entries   */
/*----------------------------------------------------------------*/  

int *clause_space;
int nextc;                     /* location where next clause goes */
int *u_space;
int nextu;

hnode *clause_hash_table;           /* ptrs to clause hash tables */
unsigned int o_arraysize;
int array_sizem1;
unsigned int cspace_size;
int array_power;

/*----------------------------------------------------------------*/
/* rand_array is dynamically allocated. its size depends on the   */
/* particular theorem.                                            */
/*----------------------------------------------------------------*/
 
rnode *rand_array;

/*----------------------------------------------------------------*/
/*  arrays for storing name of input and outpout theorem          */
/*----------------------------------------------------------------*/

char *thm_ptr, *thm_out_ptr, *thm_out_ptr2;
char *orig_thm_ptr;
char orig_thm_name[256]; /* name of theorem read in               */
char thm_name[256];      /* name of theorem read in               */
char thm_out[256];         /* name of file where output is placed */
char thm_out2[256];         /* if sol found, new name of theorem  */
int  thm_out2end;          /* ptr to last char in output file name*/
int number_chars;     /* number of characters in name of theorem  */ 
int slash_char;            /* location of last slash of input thm */

char opus[30] = {'S','u','m','m','a','r','y','\0'};
char opus2[10];

/*----------------------------------------------------------------*/
/* variables that are concerned with recording time               */
/*----------------------------------------------------------------*/
 
int accumulated_time;
int elim_starting_time; 
int eliminate_time;
int search_start_time;
int s_statrttime;
int search_time;
int phase1_time;                      /* time for phase 1         */
 
int herby;

int base;                           /* ID base last base clause   */
int org_base;                 /* original number of base clauses  */
int clauses_alive;       /* count of the number of ACTIVE clauses */
int nc;                          /* # clauses in negated concl.   */
int iter_depth;                    /* current iteration depth     */
int solved_iteration;            /* iteration number when solved  */
int ply;                           /* current search depth        */
int lit_in_hash;
int sub_array_last1;
int rename_fxns, rename_preds;
int length_of_proof;
int outside_range, satisfiable, unknown, tough;
int weakened_ply, weakened_clause;
int weak_used, split1_used, split2_used;
int weakened_clause_count;
int numb_trivial_proofs;
int questionableUsed;
int pass_count;
int first_attempt;

/*----------------------------------------------------------------*/
/* number_rows,       numb_cols, numb_entries in the random array */
/*----------------------------------------------------------------*/
 
int numb_rows, numb_cols, numb_entries, rows_less_nfxns;
 
/*----------------------------------------------------------------*/
/* variables concerned with the state of the search               */
/*----------------------------------------------------------------*/
 
int in_elim;                    /* in_elim = 0 when in eliminate  */
int simple_sol;              /* simple_sol = 1 if elim finds sol  */
int extra_pass;     /* denotes whether more than 1 pass was done  */
int phase; /* search phase: elim = 0, 1st pass = 1, 2nd pass = 2  */
int big_bug;      /* set to 1 when a hash error misleads prover   */

FILE *infile, *outfile, *outfile2;
 
int memory_size;
char input_buffer[256]; /* buffer for input file name and options */ 
 
int  ww;  /* temp loc for reading chars from compiled_code_buffer */
 
unsigned int *var_substitution_tbl[MAX_VARS*4]; /* var subs table */
 
int  equal_pred;     /* =1 if "equal" or "Equal" pred is in thm   */
int  equal_clist_index;       /* the clause id of the eq identity */
int  units_begin;     /* the index of the first unit clause       */

int  equal_xyyzxz;

int max_var;
int number_preds;  /* number of different preds in input clauses  */
int number_fxns;   /* number of different fxns in input clauses   */
int number_constants; /* # of different constants in input cls    */
int number_const_nc;  /* number of constants in negated conclus   */
int number_const_hash; /* number of constants for hashing         */
int max_lit;
int max_term_in_lit;                /* maximum terms in a literal */
int max_term_in_nc;                 /* fattest literal in nc      */
int max_vars_in_lit;
int tab_amount;      /* used to keep track of the number of tabs  */

int  maxl, maxv, maxx; /* max lits, vars, trms in clause in proof */
 
char *pidtbl[MAX_PRED_IDS];   /* ptrs to strings for predicates   */
char *fidtbl[MAX_FXNS_IDS];   /* ptrs to strings for functions    */
int skolem[MAX_FXNS_IDS];     /* 1 if constant in nc only         */

cnode *ctbl[MAX_BASE_CLAUSES];

int begin_main_line, end_main_line, on_main_line;
 
int batch_start_time, batch_total_time;
int batch_tried, batch_solved, batch_satisfiable;
int batch_array[300][15];
char theorem_names[300][40];
int batch_ptr;

int proof_found_remember;
int weakened_proof_remember;
char *weakProofBeginsHere;
char *weakProofEndsHere;

int subsumed_clss;

/*----------------------------------------------------------------*/ 
/* The following are vars that record statistics on the search.   */
/*----------------------------------------------------------------*/
 
int  s_nod;                       /* # nodes on all iterations    */
int  s_nod0;                      /* # nodes at beginning of pass */
int  s_nod_last;
int  learn_nod;
int  s_res;              /* # resolvents from binary resolution   */
int  s_fac;              /* # factors inferred                    */
int  s_subsume;          /* # of s-subsumed clauses (see 6.14)    */
int  s_subfxn;           /* subsumed ending in fxn                */
 
int  s_htentry_phase1;    /* # of entries in cl ht on phase 1     */
int  s_hthit_phase1;      /* # of ht hits in phase 1              */
int  s_htelim_phase1;     /* # of clauses elim by ht hits in ph 1 */
int  s_htfail_phase1;     /* # of ht failures in phase 1          */

int  s_ltentry_phase1;    /* # of entries in lit ht on phase 1    */
int  s_lthit_phase1;      /* # of unit ht resolutions in phase 1  */
int  s_ltfail_phase1;     /* # of unit clause failures in phase 1 */
int  s_unit_res_phase1;   /* # of unit ht resolutions in phase 1  */
int  s_unit_sub_phase1;   /* # of unit clause subsump in phase 1  */
int  s_subsume_phase1;    /* # of s-subsumed clauses in phase 1   */

int  s_htentry; /* # of entries in clause hash table (see 6.10.3) */
int  s_hthit;   /* # of clause hash table hits                    */
int  s_htelim;  /* # of clauses elim by ht hits (see 6.10.4)      */
int  s_htfail;  /* # of failures to insert clause in hash table   */

int  s_ltentry;           /* # of single literal clauses          */
int  s_lthit;     /* # unit hash table resolutions (see 6.11.3)   */
int  s_ltfail;    /* # unit clause failures on entering ht        */
int  s_unit_res;  /* # of hash table resolutions (see 6.11.2)     */
int  s_unit_sub;  /* # of unit clause subsumptions (see 6.11.3)   */

int  s_default;   /* equals zero if default parameters are used   */
int  s_subs;      /* number of attempts to find a string subst.   */
int  s_subs_suc;  /* number of successful "   "   "   "    "      */

int  s_too_many_terms;
 
/*----------------------------------------------------------------*/
/* the following are variables that (in most cases)               */
/* the user can set as desired.  They control the search.         */
/*----------------------------------------------------------------*/
 
int o_elim;       /* o_elim = 1 => do elim, else do not do elim   */
int o_merge;      /* MERGE = ancestors must be merge clauses      */
int o_nc;         /* 1 = nc-strategy                              */
int o_hash;       /* 1 = clause hash table used in proof          */
int o_extend;     /* UNIT = allow unit extension past maxd        */
int o_subsume;    /* 1 = simple-subsumbtion used (0 => not used)  */
int o_maxlit;     /* n = max size of clauses (0 = no limit)       */
int o_maxvar;     /* o_maxvar = max # vars in a clause            */
int o_maxterm;    /* max terms in literal                         */
int o_maxtermM1;  /* max terms in literal - 1                     */
int o_maxsterm;   /* max term in literal in search phase          */
int o_maxtime;    /* max time to search for a proof               */
int o_batchtime;  /* max time to search for a proof               */
int s_starttime;
int inc_o_maxlit;
 
int o_step;       /* step depth                                   */
int o_print;      /* print depth                                  */
int o_print_low;  /* first node to print                          */
int o_print_high; /* last node to print                           */
int o_debug;      /* user is supposed to be unaware of this optn  */
int o_relevant;   /* = 1 => print only clauses involved in proof  */
int o_elimprint;  /* o_elimprint = 1 => print elim info           */
int o_tptp;       /* o_tptp = 1 => input file in tptp format      */
int o_batch;      /* o_batch = 1 if batch file is input           */
int o_learn;      /* 0 => no learning, 1 => learning              */
int o_learn_lits;
int o_learn_char;
int o_print_nodes;
int o_weakened_index;
int clauseToWeaken;

int x_maxterm;    /* x_maxterm = 0 => calc o_maxterm              */
int x_stop;       /* can be set to 1 by user in interactive mode  */
int x_cr;         /* when printing clause, add cr when o_cr = 1   */

int this_computer;
int theorem_number;
int spawned, to_spawn;
int fix_proof_once;
int add_time;
int SlaveFoundProof;
int TC = 0;
int SolvedBy;
int o_print_slave;
int repeat_theorem;
int slaves_killed;
int this_split;

int cade_format;
int mem_size;

/*------------------------------------------------------------*/
/* arrays uded for reading in input file and converting       */
/* to format compatible with the search.                      */
/*------------------------------------------------------------*/

static char alphabet32[] =
 {'0','1','2','3','4','5','6','7','8','9',
  'a','b','c','d','e','f','g','h','i','j',
  'k','l','m','n','o','p','q','r','s','t',
  'u','v','w','x','y','z','A','B','C','D',
  'E','F','G','H','I','J','K','L','M','N'};


/*------------------------------------------------------------*/
/*------------------------------------------------------------*/

static int term_column[] =
{0,0,0,0,0,0,1,1,2,2,3,3,3,4,4,4,5,5,5,6,6,6,7,7,7,7,8,
8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,
8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,
8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,
8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,
8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,
8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8};

static int tms3[] = {

/*------------------------------------------------------------*/
/* 5     7     9    12    15    18    21    25   +25 
/*------------------------------------------------------------*/
 
  10,   12,   14,   17,   20,   23,   28,   35,   50,  /*05*/
   0,   14,   16,   20,   23,   26,   32,   36,   50,  /*07*/
   0,    0,   18,   22,   25,   28,   32,   40,   50,  /*09*/
   0,    0,    0,   24,   27,   30,   33,   42,   50,  /*12*/
   0,    0,    0,    0,   27,   30,   33,   42,   50,  /*15*/
   0,    0,    0,    0,    0,   32,   35,   44,   50,  /*18*/
   0,    0,    0,    0,    0,    0,   36,   46,   50,  /*21*/
   0,    0,    0,    0,    0,    0,    0,   48,   50,  /*25*/
   0,    0,    0,    0,    0,    0,    0,    0,   50,  /*+25*/
};


static int tms4[] = {

/*------------------------------------------------------------*/
/* 5     7     9    12    15    18    21    25   +25 
/*------------------------------------------------------------*/
 
   9,   12,   14,   16,   18,   21,   24,   30,   40,  /*05*/
   0,   14,   16,   18,   18,   21,   24,   30,   40,  /*07*/
   0,    0,   18,   18,   20,   22,   24,   30,   40,  /*09*/
   0,    0,    0,   20,   20,   24,   25,   32,   40,  /*12*/
   0,    0,    0,    0,   20,   24,   25,   32,   40,  /*15*/
   0,    0,    0,    0,    0,   24,   26,   32,   40,  /*18*/
   0,    0,    0,    0,    0,    0,   26,   32,   40,  /*21*/
   0,    0,    0,    0,    0,    0,    0,   32,   40,  /*25*/
   0,    0,    0,    0,    0,    0,    0,    0,   40,  /*+25*/
};

static int tms5[] = {

/*------------------------------------------------------------*/
/* 5     7     9    12    15    18    21    25   +25 
/*------------------------------------------------------------*/
 
   9,   12,   14,   16,   18,   19,   22,   23,   24,  /*05*/
   0,   14,   16,   18,   18,   19,   22,   23,   24,  /*07*/
   0,    0,   18,   18,   20,   19,   22,   23,   24,  /*09*/
   0,    0,    0,   20,   20,   19,   22,   23,   24,  /*12*/
   0,    0,    0,    0,   20,   19,   22,   23,   24,  /*15*/
   0,    0,    0,    0,    0,   19,   22,   23,   24,  /*18*/
   0,    0,    0,    0,    0,    0,   22,   23,   24,  /*21*/
   0,    0,    0,    0,    0,    0,    0,   23,   24,  /*25*/
   0,    0,    0,    0,    0,    0,    0,    0,   24,  /*+25*/
};

static int tms[] = {

/*------------------------------------------------------------*/
/* 5     7     9    12    15    18    21    25   +25 
/*------------------------------------------------------------*/
 
  10,   10,   10,   10,   10,   10,   10,   10,   10,  /*05*/
   0,   10,   10,   11,   11,   11,   11,   11,   10,  /*07*/
   0,    0,   10,   11,   12,   12,   12,   12,   10,  /*09*/
   0,    0,    0,   12,   12,   12,   12,   12,   10,  /*12*/
   0,    0,    0,    0,   15,   15,   15,   12,   10,  /*15*/
   0,    0,    0,    0,    0,   14,   14,   12,   10,  /*18*/
   0,    0,    0,    0,    0,    0,   14,   12,   10,  /*21*/
   0,    0,    0,    0,    0,    0,    0,   12,   10,  /*25*/
   0,    0,    0,    0,    0,    0,    0,    0,   10,  /*+25*/
};

/*----------------------------------------------------------------*/
/*      time_now()                    ...................... #1
/*
/*      added to consolidate all system calls to theo 
/*
/*      called by:  main()                         theo.c
/*                  iterative_deepening            theo.c
/*                  halted_search_options()        tdump.c
/*                  eliminate()                    elim.c
/*
/*      calls:      nobody
/*
/*      returns:    time of day as a an integer
/*---------------------------------------------------------------*/
 
int time_now()
{
return(time(0));
}
 
/*--------------------------------------------------------------*/
/*      init_stats()                   ..................... #2
/*
/*      init statistical parameters gathered during the search
/*
/*      called from:    main()       theo.c
/*
/*      calls:          nobody
/*
/*      returns:        nil
/*---------------------------------------------------------------*/
 
void init_stats()
{
s_nod  = 0; 
s_fac  = 0;             
s_res  = 0;
s_nod0 = 0;
s_subs = 0;
s_subs_suc = 0;
solved_iteration = 0;
 
/*---------------------------------------------------------*/
/* stats for clause hash table and lit clause hash table   */
/*---------------------------------------------------------*/
	
s_htentry = 0;       
s_hthit   = 0;      
s_htfail  = 0;
s_ltentry = 0;       
s_lthit   = 0;      
s_ltfail  = 0; 
s_htelim  = 0;
s_unit_res = 0;
s_unit_sub = 0;
s_subsume = 0;
s_subfxn = 0;
s_too_many_terms = 0;
 
tab_amount = 0;        
	
}
 
/*----------------------------------------------------------------*/
/*      init_options()                          .............. #3
/*
/*      reads the search options that the user requests in the
/*      command line.
/*
/*      called from:  enter_theorem_name()    theo.c
/*
/*      calls:        grab_number()           theo.c
/*
/*      returns:      the number of chars in the name of the theorem
/*                    or a -1 if help was requested
/*----------------------------------------------------------------*/
 
int init_options()
{
int i, j;
unsigned int k;
 
/*-----------------------------------------------------------------*/
/* initialize options to default values                            */
/*-----------------------------------------------------------------*/
    
o_elim      = 1;  /* Option 1... elim phase: 1/0/2 => do/skip/only */
o_merge     = 1;  /* Option 2... merge strat: 1/0 => off/on        */
o_nc        = 0;  /* Option 3... linear-nc: 0/1 => off/on          */
o_hash      = 1;  /* Option 4... clause ht on: 1/0=> cht off/on    */
o_extend    = 1;  /* Option 5... extended search: 1/0=> on/off     */
o_subsume   = 1;  /* Option 6... s-subsumption: 1/0 => use/don't   */
o_maxlit    = 0;  /* Option 7... maximum # lits in a cls in proof  */
o_maxterm   = 16; /* Option 8... maximum # terms in a literal      */
x_maxterm   = 0;
o_maxsterm  = 0;
 
o_maxvar    = 0;   /* Option 9... maximum # diff vars in a clause  */

cspace_size = 2097152;

o_tptp      = 3;   /* Option 11.. 0/1 not in/in tptp format        */

o_batchtime = 3600;/* Option 12.. maximum time for proof..seconds  */

o_elimprint = 0;   /* Option 13.. 0/1 do/ do not give phase 0 info */
o_step      = 0;   /* Option 14.. step through tree to ply o_step  */ 
o_print     = 0;   /* Option 15.. print tree to ply o_step         */
o_print_low = 0;
o_print_high= 1000000000;
o_print_nodes = 0;

o_debug     = 0;   /* Option 16.. debug option: 0/1 => off/on      */
o_relevant  = 0;   /* Option 17.. 0/1 do/do not prt all proof cls  */

spawned     = 0;

/*-----------------------------------------------------------------*/
/*        END OF OPTIONS                                           */
/*-----------------------------------------------------------------*/

s_default   = 0;
big_bug     = 0;       /* set to 1 if a hash error found inphase 3 */
x_cr        = 0;       /* set to 1 => print cr when printing cls   */

/*------------------------------------------*/
/* scan the arguments passed to the program */
/* if a "?" appears anywhere on the line    */
/* print out a "help" message               */
/*------------------------------------------*/
 
for(i = 76 ; i > 1; i--) {
   if(input_buffer[i - 2] == '?') {
	
/*--------------------------------------------------------------*/
/* a help message, triggered by a "?" anywhere on input line    */
/*--------------------------------------------------------------*/

printf("\n--------------------------------------------------------");
printf("\n| Enter theorem name and then options as in this example:");
      printf(
      "\n| ");
      printf(
      "\n| STARK103.THM k1 m0 z1 h0 n0 b0 l4 x6 v6 t7 e1 p3 s2 d0");
      printf(
      "\n| ");
      printf(
      "\n| where:");
      printf(
      "\n|   k0: skip elim; k1 carry out elim; k2 carry out only elim");
      printf(
      "\n|   m0: turn off linear-merge strategy");
      printf(
      "\n|   z1: turn on linear-nc search strategy");
      printf(
      "\n|   h0: turn off hash table for clauses");
      printf(
      "\n|   n0: turn off extended search strategy");
      printf(
      "\n|   b0: turn off s-subsumption strategy");
      printf(
      "\n|   l4 x6 v6: set max # of lits, terms and vars to 4,6,6");
      printf(
      "\n|   t7: set maximum search time to 7 seconds");
      printf(
      "\n|   e1: print info on elim fxn; e2: include resolutions");
      printf(
      "\n|   p3: print three levels of search tree");
      printf(
      "\n|   s2: step two levels of search tree");
      printf(
      "\n|   d1: turn on debugger to obtain information on search");
      printf(
      "\n|   r1: when proof is found, list only relevant clauses");
      printf(
      "\n|   a524288: set clause_hash_table to 512K entries");
      printf(
      "\n| By default, k=m=1, z=0, h=n=b=1, t=3600, e=p=s=d=r=0.");
      printf(
      "\n| Default values for l,x,v depend on the theorem.");
      printf(
      "\n| To prove a batch of theorems add 'batch' to command line");
      printf(
      "\n| and the redirection symbol '<' and a file where theorems");
      printf(
      "\n| are listed.");
      printf(
      "\n------------------------------------------------------");
		
      for(i = 0; i < 256; i++) {
         input_buffer[i] = ' ';}           /* clr input buffer  */
      return(-1);}                       /* -1 => re-enter      */
			
   if(input_buffer[i] == ' ') {
			
/*------------------------------------------------------*/
/* L0 => no Learn strategy; L1 => uses Learn startegy   */
/*------------------------------------------------------*/
		
      if(input_buffer[i + 1] == 'L') {
	 o_learn = grab_number(i + 2);
	 if(o_learn > 2 || o_learn < 0) {
	    printf("\nbad argument 1(L)\n");return(-1);}}
 
/*------------------------------------------------------*/
/* N0 =>  print_time  N1 => print nodes!                */
/*------------------------------------------------------*/
		
      if(input_buffer[i + 1] == 'N') {
	 o_print_nodes = grab_number(i + 2);
	 if(o_print_nodes > 2 || o_print_nodes < 0) {
	    printf("\nbad argument 2(N)\n");return(-1);}}
 
/*------------------------------------------------------*/
/* k0 => skip elim phase, k1 => carry out elim phase    */
/* k2 => carry out only elim phase                      */
/*------------------------------------------------------*/
		
      if(input_buffer[i + 1] == 'k') {
         s_default = 1;
	 o_elim = grab_number(i + 2);
	 if(o_elim > 3 || o_elim < 0) {
	    printf("\nbad argument 3(k)\n");return(-1);}}
	
/*------------------------------------------------------*/
/* m0 => no merge strategy; m1 => uses merge startegy   */
/*------------------------------------------------------*/
		
      if(input_buffer[i + 1] == 'm') {
         s_default = 1;
	 o_merge = grab_number(i + 2);
	 if(o_merge > 1 || o_merge < 0) {
	    printf("\nbad argument 4(m)\n");return(-1);}}
 
/*------------------------------------------------------*/
/* z0 => no nc strategy; z1 => uses nc strategy         */
/*------------------------------------------------------*/
		
      if(input_buffer[i + 1] == 'z') {
         s_default = 1;
	 o_nc = grab_number(i + 2);
         if(o_nc > 1 || o_nc < 0) {
	    printf("\nbad argument 5(z)\n");return(-1);}}
 
/*------------------------------------------------------------------*/
/* h0 => turn off clause_hash_table h1 => turn on clause_hash_table */
/*------------------------------------------------------------------*/
 
      if(input_buffer[i + 1] == 'h') {
         s_default = 1;
	 o_hash = grab_number(i + 2);
	 if(o_hash > 1 || o_hash < 0) {
	    printf("\nbad argument 6(h)\n");return(-1);}}
 
/*--------------------------------------------------------------*/
/* n0 => no ext'd search str'gy; n1 => uses ext'd search str'gy */
/*--------------------------------------------------------------*/
		
      if(input_buffer[i + 1] == 'n') {
         s_default = 1;
	 o_extend = grab_number(i + 2);
	 if(o_extend > 1 || o_extend < 0) {
	    printf("\nbad argument 7(n)\n");return(-1);}}

/*--------------------------------------------------------------*/
/* batch => set o_batch to 1, for a batch file
/* b0 => no simple subsumptions; b1 => yes..simple subsumptions 
/*--------------------------------------------------------------*/
		
      if(input_buffer[i + 1] == 'b') {
	 if(input_buffer[i + 2] == 'a' &&
	    input_buffer[i + 3] == 't' &&
	    input_buffer[i + 4] == 'c' &&
            input_buffer[i + 5] == 'h') {
	    o_batch = 1;
	    continue;}
         o_subsume = grab_number(i + 2);
         if(o_subsume  > 1 || o_subsume  < 0) {
            printf("\nbad argument 8(b)\n");return(-1);}}
	    
/*--------------------------------------------------------------*/
/* if lx is found, set max number of literals in a clause to x  */
/* x is restricted to fall between 2 and 20                     */
/*--------------------------------------------------------------*/
						
      if(input_buffer[i + 1] == 'l') {
         s_default = 1;
	 o_maxlit = grab_number(i + 2);
         if(o_maxlit < 1 || o_maxlit > 20) {
	    printf("\nbad argument 9(l)\n");return(-1);}}
			
/*------------------------------------------------------------------*/
/* if xX is found, set max number of terms in a literal to X        */
/*------------------------------------------------------------------*/
			
      if(input_buffer[i + 1] == 'x') {
         s_default = 1;
         o_maxsterm = grab_number(i + 2);
         x_maxterm = 1;
         if(o_maxsterm < 2 || o_maxsterm > 200) {
            printf("\nbad argument 10(x)\n");return(-1);}}
			
/*------------------------------------------------------------------*/
/* if vX is found, set max number of variables in a clause to X     */
/*------------------------------------------------------------------*/
			
      if(input_buffer[i + 1] == 'v') {
         s_default = 1;
         o_maxvar = grab_number(i + 2);
         if(o_maxvar < 2 || o_maxvar > 32) {
            printf("\nbad argument 11(v)\n");return(-1);}}
 
/*------------------------------------------------------------------*/
/* if tptp is found, set o_tptp to 1................................*/
/* if tX is found, set max search time to X                         */
/*------------------------------------------------------------------*/
			
      if(input_buffer[i + 1] == 't') {
	 if(input_buffer[i + 2] == 'p' &&
	    input_buffer[i + 3] == 't' &&
	    input_buffer[i + 4] == 'p' &&
	    input_buffer[i + 5] == '2') {
	    o_tptp = 2;
	    continue;}

	 if(input_buffer[i + 2] == 'p' &&
	    input_buffer[i + 3] == 't' &&
	    input_buffer[i + 4] == 'p' &&
	    input_buffer[i + 5] == '3') {
	    o_tptp = 3;
	    continue;}

	 if(input_buffer[i + 2] == 'p' &&
	    input_buffer[i + 3] == 't' &&
	    input_buffer[i + 4] == 'p' &&
	    input_buffer[i + 5] == '0') {
	    o_tptp = 0;
	    continue;}

         o_batchtime = grab_number(i + 2);
         if(o_batchtime < 1 || o_batchtime > 200000) {
            printf("\nbad argument\n");return(-1);}}

/*-------------------------------------------------------------------*/
/* fist order formula format => o_tptp = 4                           */
/*-------------------------------------------------------------------*/

      if(input_buffer[i + 1] == 'f') {
	 if(input_buffer[i + 2] == 'o' &&
	    input_buffer[i + 3] == 'f') {
	    o_tptp = 4;}}
			
/*----------------------------------------------*/
/* if e0 => turn elim info printout off,        */
/* if e1 => set eliminate info printout on      */
/* if e2 => set elim info printout on,          */
/*          and list resolutions                */
/*----------------------------------------------*/
			
      if(input_buffer[i + 1] == 'e') {
         o_elimprint = grab_number(i + 2);
         if(o_elimprint < 0 || o_elimprint > 20) {
            printf("\nbad argument\n");return(-1);}}

/*----------------------------------------------*/
/* if px is found, set printing depth to x      */
/*----------------------------------------------*/
			
      if(input_buffer[i + 1] == 'p') {
         o_print = grab_number(i + 2);
         if(o_print < o_step) {
            o_print = o_step;}
         if(o_print < 0 || o_print > 40) {
            printf("\nbad argument\n");return(-1);}}
			
/*----------------------------------------------*/
/* if Hx is found, print until node x is found  */
/*----------------------------------------------*/
			
      if(input_buffer[i + 1] == 'H') {
         o_print_high = grab_number(i + 2);
         if(o_print_high < 0 || o_print_high > 1,000,000,000) {
            printf("\nbad argument\n");return(-1);}}

/*----------------------------------------------*/
/* if Bx is found, begin printing  after node x */
/*----------------------------------------------*/
			
      if(input_buffer[i + 1] == 'B') {
         o_print_low = grab_number(i + 2);
         if(o_print_low < 0) {
            printf("\nbad argument\n");return(-1);}}

/*--------------------------------------------------------------*/
/* if sx is found, set depth of stepping during printing to x   */
/*--------------------------------------------------------------*/
			
      if(input_buffer[i + 1] == 's') {
         o_step = grab_number(i + 2);
         if(o_print < o_step) {
            o_print = o_step;}
         if(o_step < 0 || o_step > 40) {
            printf("\nbad argument\n");return(-1);}}
			
/*----------------------------------*/
/* if d is found, set debugger on?  */
/*----------------------------------*/
			
      if(input_buffer[i + 1] == 'd') {
         o_debug = grab_number(i + 2);
         if(o_debug  > 3 || o_debug  < 0) {
            printf("\nbad argument\n");return(-1);}}
				
/*------------------------------------------------------------------*/
/* if r1 is found , set o_relevant=1 to print only relevant clauses */
/*------------------------------------------------------------------*/
			
      if(input_buffer[i + 1] == 'r') {
         o_relevant = grab_number(i + 2);
         if(o_relevant > 1 || o_relevant < 0) {
            printf("\nbad argument\n");return(-1);}}

/*------------------------------------------------------------------*/
/* if wX is found , set weakened wersion index to X                 */
/*------------------------------------------------------------------*/
			
      if(input_buffer[i + 1] == 'w') {
         o_weakened_index = grab_number(i + 2);
         if(o_weakened_index < 0) {
            printf("\nbad argument\n");return(-1);}}}}

/*---------------------------------------------------*/
/* end of option list:                               */ 
/* blank out input_buffer except for name of theorem */
/*---------------------------------------------------*/
	
for(i = 0; i < 255; i++) {
   if(input_buffer[i] == ' ' || input_buffer[i] == '\n') {
      j = i;
      for(j = i; j < 256 ; j++) input_buffer[j] = ' ';
      break;}}

/*--------------------------------------------------*/ 
/* return the number of characters in the name      */
/*--------------------------------------------------*/

if(o_learn == 2) {
   o_learn_lits = o_maxlit;
   o_learn_char = o_maxsterm;}
return(i);
}
 
/*-----------------------------------------------------------------*/
/*      grab_number(s)                       .............. #4
/*
/*      reads a string of characters and converts them to a
/*      integer equivalent
/*
/*      called from:  init_options()              theo.c
/*                    halted_search_options()     tdump.c
/*
/*      calls:        nobody
/*
/*      returns:      an integer correspond to the string s
/*-----------------------------------------------------------------*/
 
int grab_number(s)
int s;
{
int n;
	
for(n = 0; ; s++) {
   if(input_buffer[s] >= '0' && input_buffer[s] <= '9')
      n = n * 10 + ( (int) (input_buffer[s] - '0') );
   else
      return(n);}
}                                                       

/*-----------------------------------------------------------------*/
/*      convert_n2s()                       ................. #5
/*
/*      called by:
/*
/*      calls:
/*-----------------------------------------------------------------*/


/*-----------------------------------------------------------------*/ 
/*      fatal(msg)                          ................. #6
/*
/*      fatal will print an error message on the screen and quit
/*
/*      called by:    main()                theo.c
/*                    prepare_clauses()     theo.c
/*                    format_clause()       theo.c
/*                    eliminate()           elim.c
/*                    enter_theorem_name()  theo.c
/*
/*      calls:        time_now()            theo.c
/*
/*      returns:      to the operating system after reporting
/*                    error
/*-----------------------------------------------------------------*/
 
void fatal(msg)
char *msg;
{
printf("....%s\n",msg);
OutPtr += sprintf(OutPtr,"....%s\n",msg); 

if(*msg == 'D') {
  printf(
        "Theorems Tried %d\n Theorems Solved %d\n Satisfiable %d\n",
	 batch_tried,batch_solved,batch_satisfiable);
  OutPtr += sprintf(OutPtr,
        "Theorems Tried %d\n Theorems Solved %d\n Satisfiable %d\n",
	 batch_tried,batch_solved,batch_satisfiable);

  batch_total_time = time_now() - batch_start_time;

  printf("\n Total Time (in seconds) %d\n\n",
            batch_total_time);
  OutPtr += sprintf(OutPtr,
	     "\n Total Time (in seconds) %d\n\n",
            batch_total_time);}
   exit(0);
}
 
/*-------------------------------------------------------------------*/
/*      init_search_space()            ......................#7
/*
/*      allocates space for clauses.
/*      allocates space for clause hash table.
/*      allocates space for pointers at each level in tree
/*
/*      called from:   main()                 theo.c
/*
/*      calls:
/*
/*      returns:       nil
/*-------------------------------------------------------------------*/
 
void init_search_space()
{       

clause_space = (unsigned int *) calloc(cspace_size,4);   
nextc = 0;
clist[nextc] = (clause *) clause_space;

if(nextu == 0) {
   u_space = (unsigned int *) calloc(200000,4);   
   ulist[nextu] = (clause *) u_space;}
 
clause_hash_table =  (hnode *) calloc(3 * o_arraysize,4);
}
 
/*-----------------------------------------------------------------*/
/*      init_clause_hash_table()          .................#8
/*
/*      initialize the clause_hash_table to all 0's
/*
/*      called by:      main()                             theo.c
/*                      iterative_deepening()              theo.c
/*
/*      calls:
/*
/*      returns:        nil
/*-----------------------------------------------------------------*/
 
void init_clause_hash_table()
{
int i, j;

/*----------------------------------------------*/
/* allocate 4 * HT_SIZE int  words for the hash */
/* table.  initially set all words to 0.        */
/*----------------------------------------------*/
 
clause_hash_table =  (hnode *) calloc(3 * o_arraysize,4);

/*----------------------------------------------*/
/* clear hr_hashcodes array                     */
/*----------------------------------------------*/

for(i = 0; i < 3; i++) {
   for(j = 0; j < 50; j++) {hr_hashcodes[i][j] = 0;}}

return;
}
 
/*-----------------------------------------------------------------*/
/*      init_random_hash_array()    ........#9
/*
/*      initialize the rand_array to random numbers 
/*
/*      called from:   main()                      theo.c 
/*
/*      calls:
/*
/*      returns:       nil
/*-----------------------------------------------------------------*/
 
void init_random_hash_array()
{
rnode *r1;
int i;
unsigned int num_ent;
	
/*--------------------------------------------------------------*/
/* the number of rows in rand_array = MAX_VARS + numb_fxns      */
/* the number of columns in rand_array = numb_chars             */
/*--------------------------------------------------------------*/
 
numb_rows = MAX_VARS + number_fxns + number_preds;
numb_cols = 200;
rows_less_nfxns = numb_rows - number_fxns - 1;
 
/*--------------------------------------------------------------*/
/* the number of entries in the array rand_array is the         */
/* product of the numb_rows * numb_cols                         */ 
/*--------------------------------------------------------------*/
 
numb_entries = numb_rows * numb_cols;
num_ent = (unsigned int) (numb_entries * 2 + 1); 
if(num_ent < 50) {
   numb_entries = 25;
   num_ent = 51;}

/*--------------------------------------------------------------*/
/* allocate space for rand_array                                */
/*--------------------------------------------------------------*/

rand_array = (rnode *) calloc(num_ent * 2,4);

/*--------------------------------------------------------------*/
/* make 32-bit random number for each element in rand_array     */
/*--------------------------------------------------------------*/
 
srand(2);
 
for(i = 0 ; i < (int) numb_entries; i++) {
   r1 = &rand_array[i];

   r1->r_hash1 = rand()*16384*4 + rand();
   r1->r_hash2 = rand()*16384*4 + rand();}
   
}
 
/*-------------------------------------------------------------*/
/*      enter_theorem_name()           ................#6
/*
/*      read in the name of the theorem to be proved
/*
/*      called by:   main()                    theo.c
/*
/*      calls:       CompareStrings()          tptpi.c
/*                   init_options()            theo.c
/*
/*      returns:     nil
/*--------------------------------------------------------------*/
 
void enter_theorem_name(argc,argv)
int argc;
char *argv[];
{
int  bad_name;       /* bad_name = 0 until theorem found in file   */
int i, j, jj, jjj, m;
int  ask_for_help;    /* normally 0, set to 1 if "?" on input line */
char zz;

bad_name = ask_for_help = 0;

if(weakened_ply != 0 && excellent_weakened_proof == 0 && !satisfiable) {

   for(j = 0 ; j < 256 ; j++) {
      thm_name[j] = thm_out[j] = thm_out2[j] = 0;
      input_buffer[j] = ' ';}
   for(j = 0; j < 256; j++) {input_buffer[j] = orig_thm_name[j];}
   j = 0;
   learn_nod += s_nod;
   first_attempt = 1;
   //printf("FIRST_ATTEMPT=1\n");
   goto get_args;}

++argv;++argv;
if((*argv)[0] == '-' && (*argv)[1] == 'l') {
   (*argv)[0] = ' ';
   (*argv)[1] = ' ';
   cade_format = 1;}
else cade_format = 0;

argv--;argv--;
length_of_proof = 0;

excellent_weakened_proof = 0;
weakened_ply = 0;
weakened_clause_count = 1;
weakened_proof_remember = 0;
satisfiable = 0;
learn_nod = 0;
accumulated_time = 0;
numb_trivial_proofs = 0;
weak_used = split1_used = split2_used = -1;
c_split = -1;
l_split = -1;
inc_o_maxlit = 0;

if(repeat_theorem == 0) {
   for(i = 0; i < MAX_BASE_CLAUSES; i++) {good_clauses[i] = 0;}}

/*------------------------------------*/
/* read in the name of the theorem.   */
/* loop if name is not found.         */
/*------------------------------------*/

while(!bad_name) {
   for(j = 0 ; j < 256 ; j++) {
      orig_thm_name[j] = 0;
      thm_name[j] = 0;
      thm_out[j] = 0;
      thm_out2[j] = 0;
      input_buffer[j] = ' ';}
   ask_for_help = 0;
   while(!ask_for_help) {
      j = 0;
      if(cade_format == 0) {
      printf("\nEnter name of theorem (type '?' for help): ");

/*-------------------------------------*/
/* read from the terminal if no        */
/* srguments on command line           */
/*-------------------------------------*/

      while((zz = getchar()) != '\n') { 
         if(zz == '\b') {
            if(j > 0) {
	       j--; printf("\b \b");}}
         else if(zz == (char) -1) {fatal("DONE!");}
         else {
	    input_buffer[j] = orig_thm_name[j] = zz;
	    j++;}} 
      input_buffer[j] = orig_thm_name[j] = ' ';
      j++;}

      nextu = 0;

      for(jj = 1; jj < argc; jj++) {
         ++argv;
         jjj = 0;
         while((orig_thm_name[j] = input_buffer[j] = (*argv)[jjj++])
            !='\0') {j++;} 
         input_buffer[j] = orig_thm_name[j] = ' ';
	 j++;}

get_args:
      number_chars = init_options();       /* set search options */
      if(learn_nod && o_learn == 2) o_learn = 1;
      printf("\n");
      for(jj = 0; jj < number_chars; jj++) {
         printf("%c",input_buffer[jj]);}
      printf("\n");
      if(number_chars != -1) {
         ask_for_help++;
         printf("\n");}}         

   m = 0; 
   for(j = 0 ; j < number_chars ; j++) {  
      if(input_buffer[j] == '.') {m = j;}
      if(input_buffer[j] == '+') {if(!m) m = j;}
      if(input_buffer[j] == '/') {m = 0;}
      thm_name[j] = input_buffer[j];
      orig_thm_name[j] = thm_name[j];}

   for(j = number_chars ; j < 256 ; j++ ) {thm_name[j] = 0;}

/*--------------------------------------------------------------*/
/* give output file the same as input file with a suffix ".u"   */    
/*--------------------------------------------------------------*/

   if(m) { 
      for(j = 0; j < m; j++)  {
         thm_out[j] = input_buffer[j];
         thm_out2[j] = input_buffer[j];
         if(j == m) {break;}}
      input_buffer[m] = '.';
      thm_out[m] = thm_out2[m] = '.';}
   else {
      for(j = 0 ; j < number_chars + 1 ; j++)  {
         thm_out[j] = input_buffer[j];
         thm_out2[j] = input_buffer[j];
         if(input_buffer[j] == ' ') {
            thm_out[j] = '.';
            thm_out2[j] = '.';
	    break;}}}
			
   thm_ptr = &thm_name[0];
   thm_out_ptr = &thm_out[0];
   thm_out_ptr2 = &thm_out2[0];

/*------------------------------------*/
/* open input file of named theorem   */
/*------------------------------------*/

   if((infile = fopen(thm_ptr,"r")) == 0) {
      printf("Can't open theorem. Please retype name.\n\n");
      if(argc > 1) {fatal(" ");}}
   else bad_name = 1;}

if(weakened_ply == 0) {
  if(cade_format == 0) {
   if((outfile = fopen(thm_out_ptr,"w")) == 0) {
      fclose(outfile);
      fatal("\nCannot open output file:\nType cr and rerun");}}
   OutPtr = &results[0];
   s_starttime = time_now();
   first_attempt = 0;
   //printf("FIRST_ATTEMPT=0\n");
   clauseToWeaken = o_weakened_index;
   if(o_learn == 1) {
      o_maxtime = o_batchtime;
      weakProofBeginsHere = 0;}
   else o_maxtime = o_batchtime;}

/*-----------------------------------*/
/* scan for "/" in input_buffer      */
/* indicating theorems was in a      */
/* subdirectory.                     */
/*-----------------------------------*/

for(m = number_chars - 1; m >= 0; m--) {
   if(input_buffer[m] == '/') {break;}}
if(m == 0) {m = -1;}
slash_char = m;

/*------------------------------------*/
/* m+1 begins the name of theorem     */
/* save name of theorem in output     */
/* file.                              */
/*------------------------------------*/

for(jj = m + 1; jj < number_chars; jj++) {
   OutPtr += sprintf(OutPtr,"%c",input_buffer[jj]);}

sprintf(OutPtr++,"\n");
thm_out2end = j + 1;
}
 
/*--------------------------------------------------------------*/
/*      prepare_clauses()                .............. #19
/*
/*      prepare_clauses will read the clauses currently stored
/*      in the input_buffer and format them, placing the resulting
/*      clauses in cspace with pointers to each in the clist
/*
/*      NOTES:
/*      - we assume that all variables are numbered from 0..n so
/*        that n+1 is the number of distinct variables
/*
/*      - we pass the predicate and function symbol tables on to
/*        the output directly since we have no need for them
/*
/*      called from:    main()                          theo.c
/*
/*      calls:          get_pred_fxn_symbols()          theo.c
/*                      read_buffer_word()              theo.c
/*                      format_clause()                 theo.c
/*                      max_term_in_any_lit()           theo.c
/*                      sort_constants_fxns()           theo.c
/*                      equality_predicate()            equal.c
/*                      fatal()                         theo.c
/* 
/*      returns:        nil
/* -------------------------------------------------------------*/
 
void prepare_clauses()
{
static int characters[] = {0,3,6,9,12,15,18,21,24,26,28,30,32};
int i, n, x1, x2;
int min_lit;

/*--------------------------------------------------*/
/* set pointer to beginning of compiled_code_buffer */
/*--------------------------------------------------*/
 
comp_code_buffer_index = compiled_code_buffer;

/*--------------------------------------------------------------*/
/* We first pass the symbol table on to the output file.        */
/* This symbol table is two sets of non-empty lines followed    */
/* by an empty line (the predicate and function symbol tables   */
/*--------------------------------------------------------------*/
    
print0("Predicates: ");
number_preds = get_pred_fxn_symbols(pidtbl); /* pred sym table   */
print0("\n");

print0("Functions:");
number_fxns = get_pred_fxn_symbols(fidtbl);/* function sym table */
read_buffer_word(); 

n = ww; 
base = ww - 1;               /* clist[base] is last base clause  */
org_base = base;              /* original number of base clauses */
 
read_buffer_word();nc = ww;   /* # clauses in negated conclusion */
	
if(n < 2) {fatal("theorem has one clause; proof impossible. Quitting.");}
 
/*--------------------------------------------*/
/* read the clauses and put them in the clist */
/*--------------------------------------------*/
  
subsumed_clss = 0;
while(n-- > 0) {
   format_clause();}

sort_constants_fxns();
 
max_term_in_any_lit();

max_lit = 0;
min_lit = 1000;
for(n = 0; n <= base; n++) {
   if(clist[n]->c_nlit > max_lit) {max_lit = clist[n]->c_nlit;}
   if(clist[n]->c_nlit < min_lit) {min_lit = clist[n]->c_nlit;}}

//printf("Maximum lits in a clause = %d\n",max_lit);
 
/*-----------------------------------------------*/
/* given the maximum number of terms in          */
/* any literal in any base clause,               */
/* determine o_maxsterm, the maximum number of   */
/* terms in any literal kept during the          */
/* search.  Using the default value:             */
/*                                               */
/*       max_term_in_lit    o_maxsterm           */
/*           0                  0                */
/*           1                  3                */
/*           2                  6                */
/*           3                  9                */
/*           4                 12                */
/*           5                 15                */
/*           6                 18                */
/*           7                 21                */
/*           8                 24                */
/*           9                 26                */
/*          >10        2 * max_term_in_lit + 8   */
/*                                               */
/* the user can set o_maxsterm using the xi      */
/* option.                                       */
/*-----------------------------------------------*/
 
if(x_maxterm == 0) {
   x1 = term_column[max_term_in_nc];
   x2 = term_column[max_term_in_lit];
   if(x1 > 25) {x1 = 25;}
   if(x2 > 25) {x2 = 25;}

   if(max_lit <= 3) {o_maxterm = tms3[x1 * 9 + x2];}
   if(max_lit == 4) {o_maxterm = tms4[x1 * 9 + x2];}
   if(max_lit == 5) {o_maxterm = tms5[x1 * 9 + x2];}
   if(max_lit > 5) {o_maxterm = tms[x1 * 9 + x2];}
   o_maxtermM1 = o_maxterm - 1;}

if(x_maxterm) {
   o_maxterm = o_maxsterm;
   o_maxtermM1 = o_maxterm - 1;
   if(o_maxsterm < max_term_in_lit) {o_maxsterm = 2*max_term_in_lit+8;}}
 
/*--------------------------------------------------*/
/* determine the max number of literals in any      */
/* base clause and the maximum number of            */
/* different variable names in any base literal     */
/*--------------------------------------------------*/
 
max_var = 0;
	
for(n = 0; n <= base; n++) {
   if(clist[n]->c_nv > max_var)  {max_var = clist[n]->c_nv;}}

if(o_maxvar == 0) {
   o_maxvar = 2 * max_var;
   if(o_maxvar < max_vars_in_lit) {o_maxvar = max_vars_in_lit + 3;}}
      
if(o_maxlit == 0) {
   o_maxlit = max_lit + min_lit - 2;
   if(o_maxlit <= 3 && base < 30) o_maxlit = 4;
   if(o_maxlit < 3) {o_maxlit = 3;}}
if(max_lit < o_maxlit) {max_lit = o_maxlit;} 

/*--------------------------------------*/
/* search for equal predicate           */
/*--------------------------------------*/

rename_functions();

if(rename_preds) {

/*----------------------------------------*/
/* if predicates have been renamed, list  */
/*----------------------------------------*/

   print0("Renamed predicates: ");
   for(i = 1; i <= number_preds; i++) {
      printf(" %s",pidtbl[i]);
      OutPtr += sprintf(OutPtr," %s",pidtbl[i]);}
   print0("\n");} 

if(rename_fxns) {

/*----------------------------------------*/
/* if functions have been renamed, list   */
/*----------------------------------------*/

   print0("Renamed functions: ");
   for(i = 1; i <= number_fxns; i++) {
      printf(" %s",fidtbl[i]);
      OutPtr += sprintf(OutPtr," %s",fidtbl[i]);
      if(i == number_constants) {print0(" :");}
      if(i == number_const_nc) {print0(" .");}}
   print0("\n");}

equal_predicate();

if(equal_pred && number_preds == 1 && x_maxterm != 1) {
   if(o_maxterm < 50) {
      o_maxterm = o_maxterm * 1.5;
      o_maxtermM1 = o_maxterm - 1;
      o_maxsterm = o_maxsterm;}}

/*--------------------------------------*/
/* prepare the function sub. array      */
/*--------------------------------------*/

printf("\nOPT:");
printf(" k%d",o_elim);
printf(" m%d",o_merge);
printf(" z%d",o_nc);
printf(" h%d",o_hash);
printf(" n%d",o_extend);
printf(" b%d",o_subsume);
printf("  x%d",o_maxterm);
printf(":X%d:N%d",max_term_in_lit,max_term_in_nc);
printf(" v%d",o_maxvar);
printf(" l%d",o_maxlit);
printf(" t%d/%d",o_batchtime,o_maxtime);

OutPtr += sprintf(OutPtr,"\nOPT:");
OutPtr += sprintf(OutPtr," k%d",o_elim);
OutPtr += sprintf(OutPtr," m%d",o_merge);
OutPtr += sprintf(OutPtr," z%d",o_nc);
OutPtr += sprintf(OutPtr," h%d",o_hash);
OutPtr += sprintf(OutPtr," n%d",o_extend);
OutPtr += sprintf(OutPtr," b%d",o_subsume);
OutPtr += sprintf(OutPtr,"  x%d",o_maxterm);
OutPtr += sprintf(OutPtr,":X%d:N%d",max_term_in_lit,max_term_in_nc);
OutPtr += sprintf(OutPtr," v%d",o_maxvar);
OutPtr += sprintf(OutPtr," l%d",o_maxlit);
OutPtr += sprintf(OutPtr," t%d/%d",o_batchtime,o_maxtime);

return;
}

/*--------------------------------------------------------------*/
/*      read_buffer_word()               .............. #11
/*
/*      read a word (integer) from the input file
/*
/*      called by:    prepare_clauses()     theo.c
/*                    format_clause()       theo.c
/*
/*      calls:        nobody
/*
/*      returns:      nil
/*----------------------------------------------------------------*/
 
void read_buffer_word()
{
union u_tag {
   int w; 
   char ttext[4];} 
   temp;
			
int j;
	
for(j = 0; j < 4; j++) {
   temp.ttext[j] = *comp_code_buffer_index++;}
ww = temp.w;
}
 
/*--------------------------------------------------------------*/
/*      sort_constants_fxns()           .............. #14
/*
/*      sort the constants and functions so that the constants
/*      are placed at the top of the list of functions and constants
/*      once that is done, place the constants in the nc at the
/*      head of the constants list.
/*
/*      called by:   prepare_clauses()          theo.c
/*
/*      calls:       is_fxn_a_constant()        theo.c
/*                   interchange_fxn_const()    theo.c
/*
/*      returns:     nil
/*---------------------------------------------------------------*/
 
void sort_constants_fxns()
{
int i, k;
 
/*-----------------------------------------------*/
/*    put the constants at the head of the list  */
/*-----------------------------------------------*/
 
number_constants = 0;
for(i = 1; i <= number_fxns ; i++) {
   if(is_fxn_a_constant(i,0,base)) {
      number_constants++;       
      if(i >= number_fxns) break;}
   else {
      for(k = number_fxns; k > i; k--) {
	     if(is_fxn_a_constant(k,0,base)) {
	        interchange_fxn_const(i,k);
	        number_constants++;
	        break;}}}}  

/*-----------------------------------------------*/
/*   now resort the constants in the nc to the   */
/*   top of the list.                            */
/*-----------------------------------------------*/

number_const_nc = 0;
for(i = 1; i <= number_constants; i++) {
   if(is_fxn_a_constant(i,base - nc + 1,base)) { 
      number_const_nc++; 
      continue;}
   else {
      for(k = i + 1; k <= number_constants; k++) {
	 if(is_fxn_a_constant(k,base - nc + 1,base)) {
            number_const_nc++;
	    interchange_fxn_const(i,k);
	    break;}}}} 

if(number_constants <= 5) {number_const_hash = number_constants;}
else {number_const_hash = 5;}

for(i = 1; i <= number_constants; i++) {
   skolem[i] = 1;
   if(is_fxn_a_constant(i,0, base - nc)) {skolem[i] = 0;}}

for(i = 1; i <= number_fxns; i++) {
   if(o_debug > 0) {
      printf(" %d",i);
      OutPtr += sprintf(OutPtr," %d",i);}
      printf(" %s ",fidtbl[i]);
      OutPtr += sprintf(OutPtr," %s",fidtbl[i]);
   if(skolem[i] == 1) print0("(S)");
   if(i == number_constants) {print0(" :");}
   if(i == number_const_nc) {print0(" .");}}
print0("\n");
}
 
/*--------------------------------------------------------------*/
/*      is_fxn_a_constant()                .............. #13
/*
/*      called by:      sort_constants_fxns()       theo.c
/*
/*      calls:          nobody
/*
/*      returns:        1 if fxn is a constant, 0 otherwise
/*-----------------------------------------------------------*/
 
int is_fxn_a_constant(i,top_clause,bottom_clause)
int i,top_clause,bottom_clause;
{
unsigned int *p;
int k, a;
clause *c;
literal *l;
	
for(k = top_clause; k <= bottom_clause ; k++) { 
   c = clist[k];
   for(l = c->c_lit; l != 0; l = l->l_next) {  /* scan lits  */
      p = &l->l_id;
      for(a = (*p++ & ARGMASK); a > 0; a--) {
	     if(*p & FCNMASK) {
	        if(((*p & IDMASK) >> ARGSHIFT) == i) {
	           if((*p & ARGMASK) == 0) return(1);
	           else return(0);}
	        a += *p++ & ARGMASK;}
	     else {
	        p++;}}}}
return(0);
}
/*--------------------------------------------------------------*/
/*      interchange_fxn_const()          ............. #12
/*
/*       called by:      sort_constants_fxns()      theo.c
/*
/*       calls:          nobody
/*
/*       returns:         nil
/*--------------------------------------------------------------*/
 
void interchange_fxn_const(i,j)
int i,j;
{
unsigned int *p;
int k, a;
clause *c;
literal *l;
char *fide;
	
for(k = 0; k <= base ; k++) {
   c = clist[k];
   for(l = c->c_lit; l != 0; l = l->l_next) {
      p = &l->l_id;
      for(a = (*p++ & ARGMASK); a > 0; a--) {
	     if(*p & FCNMASK) {
	        if(((*p & IDMASK) >> ARGSHIFT) == i) {
	           *p = (*p & ~IDMASK) + (j << ARGSHIFT);}
	        else if(((*p & IDMASK) >> ARGSHIFT) == j) {
	           *p = (*p & ~IDMASK) + (i << ARGSHIFT);}
	        a += *p++ &ARGMASK;}
	     else {
	        p++;}}}}
			
fide = fidtbl[i];
fidtbl[i] = fidtbl[j];
fidtbl[j] = fide;
}
 
/*--------------------------------------------------------------*/ 
/*      get_pred_fxn_symbols()        .............. #18
/*
/*      get_pred_fxn_symbols will read in a symbol table and write
/*      it to the output file (and to the log file if there is one
/*
/*      called from:   prepare_clauses()          theo.c
/*
/*      calls:         nobody
/*
/*      returns:       the number of symbols (pred/fxns)
/*---------------------------------------------------------------*/
 
int get_pred_fxn_symbols(qtbl)
char *qtbl[];
{
int a, k;
char *ii, *jj;
static char Eq[] = "Equal";
static char eq[] = "equal";
static char Eqish[] = "equalish";

for(a = 1; ;a++) {  

/*----------------------------------------*/
/* check first if at least one pred, fxn  */           
/*----------------------------------------*/

   if(*comp_code_buffer_index == 0xa) {
      comp_code_buffer_index++; 
      return(0);}

/*----------------------------------------*/
/* set qtbl[a] to point to beginning of   */
/* the a-th predicate or function         */
/*----------------------------------------*/

   qtbl[a] = ii = comp_code_buffer_index;

   while(*comp_code_buffer_index != 0xa) {comp_code_buffer_index++;} 
 
/*----------------------------------------*/
/* replace terminating '0xa' with a 0     */
/*----------------------------------------*/
  
   *comp_code_buffer_index++ = 0;
 
/*--------------------------------------------------*/
/* print string of characters representing pred     */    
/*--------------------------------------------------*/
    
if(qtbl == pidtbl) {
   printf("%s ",qtbl[a]);
   OutPtr += sprintf(OutPtr,"%s ",qtbl[a]);}
 
/*--------------------------------------------------*/
/* test for equality predicate                      */
/*--------------------------------------------------*/
   jj = ii;
	
   for(k = 0; k < 6 ; k++) {
      if(*jj++ != Eq[k]) {k = 0; break;}}
   if(k) {equal_pred =  (a << ARGSHIFT) + 2;} 

   jj = ii;
   for(k = 0; k < 6 ; k++) {
      if(*jj++ != eq[k]) {k = 0; break;}}
   if(k) {equal_pred =  (a << ARGSHIFT) + 2;} 

   if(*comp_code_buffer_index == 0xa) {
      comp_code_buffer_index++;
      return(a);}}                      

}                                                       
/*--------------------------------------------------------------*/
/*      format_clause()                   .............. #15
/*
/*      format_clause will read clause from the input and stuff
/*      it into the clist
/*
/*      called from:    prepare_clauses()   theo.c
/*
/*      calls:          read_buffer_word()  theo.c
/*                      fatal()             theo.c
/* 
/*      returns:        nil
/*--------------------------------------------------------------*/
 
void format_clause()
{
clause *c;
literal *l, *l1, *l2, *l3, *l6, *l7, *l_prev,  *previous_literal;
unsigned int *ppp,*p, *q;
unsigned int *pend, *qend;
int i, a;
	
ply = 0;       
read_buffer_word();read_buffer_word();/* these are leading 0's */
 
/*------------------------------------------------------*/
/* initialize clause header (this is not part of the    */
/* input file, we add it as we read in the clause)      */
/*------------------------------------------------------*/
 
c = clist[nextc++];                  /* the new clause         */
c->c_id = nextc;
c->c_bits = ACTIVE | MERGE | BASE; /* base clauses are merges  */
if(nextc > base - nc + 1) {c->c_bits |= NC;}

c->c_rpar = c->c_lpar = 0;       /* no parents (base clause)   */
c->c_lit = 0;                        /* ^ first literal        */
c->c_nlit = 0;                             /* # literals       */
c->c_nv   = 0;                       /* # distinct variables   */
	
p = (unsigned int  *) (c + 1);  /* where predicate code goes   */
previous_literal = 0;                 /* literal chain         */
 
/*----------------------------------------*/
/* read in each literal, end clause at -1 */
/*----------------------------------------*/    
 
read_buffer_word(); 
while(ww != -1) {
   if(!previous_literal) {               /* link last literal  */
      c->c_lit = (literal *) p;}
   else {
      previous_literal->l_next = (literal *) p;}
      
   previous_literal = (literal *) p;
   previous_literal->l_id = ww;
   previous_literal->l_next = 0;
 
   p = (unsigned int *) (previous_literal + 1);
 
   for(a = ww & ARGMASK; a > 0; a--) {
      read_buffer_word();
      if(ww & FCNMASK) {                          /* function? */
	     *p++ = ww;                            /* yes       */
	     a += ww & ARGMASK;}                   /* end if    */ 
      else {
	     ww >>= ARGSHIFT;                      /* var ID    */
	     *p++ = ww;
	     if(ww > c->c_nv) {c->c_nv = ww;}}}      /* end for */
   
/*----------------------------------------------------------------*/
/* transform equality predicates to standard form                 */
/*----------------------------------------------------------------*/
 
   c->c_nlit++;
   read_buffer_word();}                       /* end while  */
 
/*-----------------------------------------------------------------*/
/* set up a ptr to the next clause, pointed to by clist[nextc]     */
/*-----------------------------------------------------------------*/
 
c->c_end = (clause *) p;
clist[nextc] = c->c_end;
	
/*--------------------------------------------------*/
/* if clause has only one literal, set UNIT flag on */
/*--------------------------------------------------*/
 
if(c->c_nlit == 1) {c->c_bits |= UNIT;}
 
if(++c->c_nv > MAX_VARS) {
   printf("Clause %d has too many variables!!\n",nextc);
   OutPtr += sprintf(OutPtr,
	   "Clause %d has too many variables!\n",nextc);
   fatal("Too many variables in clause");}

/*--------------------------------------------------*/
/* eliminate tautologies  and delete identical lits */
/*--------------------------------------------------*/

del_taut:
//print_clause(c);getchar();
for(l1 = c->c_lit; l1->l_next != 0; l1 = l1->l_next) {
   for(l2 = l1->l_next; l2 != 0; l2 = l2->l_next) {
      p = &l1->l_id;
      q = &l2->l_id;
      if(((unsigned int) *p ^ (unsigned int) *q) == FCNMASK) {
         pend = l1->l_next;
	 if(pend == 0) {goto format1;}
         qend = l2->l_next;
	 if(qend == 0) qend = c->c_end;
	 p++;q++;
         while(*p == *q && (unsigned int) p < (unsigned int) pend) {
            p++;q++;}
         if(p == pend) {
            if(c->c_bits & NC) nc--;
	    nextc--;base--;org_base--;return;}}
      else {
         if(*p == *q) {
            pend = l1->l_next;
            qend = l2->l_next;
	    if(qend == 0) {
               l_prev = l1;
               for(l3 = l1->l_next; l3 != l2; l3 = l3->l_next) {
                  l_prev = l3;}
	       qend = c->c_end;}
	    p++;q++;
            while(*p == *q && (unsigned int) p < (unsigned int) pend) {
               p++;q++;}
            if(p == pend) {

/*----------------------------------------------*/
/* identical literals found, delete one.        */
/*----------------------------------------------*/

	       c->c_nlit--;
               if(l2->l_next == 0) {
                  l_prev->l_next = 0;
		  c->c_end = &l2->l_next;
		  goto del_taut;}

/*----------------------------------------------*/
/* remove non-terminal literal                  */
/*----------------------------------------------*/

	       p = &l2->l_next;
	       q = l2->l_next;

               while((unsigned int)q < (unsigned int) c->c_end) {
                  *p = *q;
		  if(*p & 0x7FFF0000) {
                     *p = *p - ((unsigned int) q - (unsigned int) p);}
		  p++;q++;}

	       c->c_end = p;
               goto del_taut;}}}}}

/*--------------------------------------------------------------*/
/* delete subsumed clauses                                      */
/*--------------------------------------------------------------*/ 

format1:
for(a = nextc - 2; a >= 0; a--) {
   q = &c->c_nv;
   for(p = &clist[a]->c_nv ; (unsigned int) p < (unsigned int) clist[a]->c_end; p++) {
      if((*p & 0x7FFF0000) || (*p == 0 && (*q & 0x7FFF0000))) q++;
      else if(*p != *q) {
	      goto formatA;}
         else q++;}
   if(c->c_bits & NC) nc--;
   nextc--;base--;org_base--;
   subsumed_clss++;
   return;
formatA: continue;}

start_compare:
//printf("Compare...\n");
//print_clause(c);
if(nextc < 2) goto format2;
//print_clause(clist[nextc - 2]);
q=&clist[nextc - 2]->c_nv;
for(p=&c->c_nv;(unsigned int)p<(unsigned int)c->c_end; p++) {
   if(p == &c->c_nv) {q++;continue;}
   if((*p & 0x7FFF0000) || (*p == 0 && (*q & 0x7FFF0000))) q++;
    else if(*p != *q) {
	      goto format2;}
         else q++;}

subsume_check:
//printf("Subsumed...\n");print_clause(clist[nextc - 2]);getchar();
l = clist[nextc - 2]->c_lit;
for(a = 1; a < c->c_nlit; a++) {
   l = l->l_next;}
clist[nextc - 2]->c_nlit = c->c_nlit;
clist[nextc - 2]->c_end = l->l_next;
l->l_next = 0;
//printf("Revised to...\n");print_clause(clist[nextc - 2]);getchar();
if(c->c_bits & NC) nc--;
nextc--;base--;org_base--;

c = clist[nextc - 1];
goto start_compare;

format2:
if(c->c_nlit == 0) {fatal("Empty clause in base");}
}
 
/*---------------------------------------------------------------*/
/*      count_lit_characters()               .............. #16
/*
/*      count_lit_characters counts terms in lit l
/*
/*      called from:  max_term_in_any_lit()        theo.c
/*
/*      calls:        nobody
/*
/*      returns:      number of terms in literal l
/*----------------------------------------------------------------*/
 
int count_lit_characters(l)
literal *l;                    /* l is a literal in clause c      */
{
int a;
unsigned int *p;
int literal_terms;         /* count of the number of terms in lit */
int var_count;
	
p = &l->l_id;        
literal_terms = 0;     
var_count = 0;
	
for(a = *p++ & ARGMASK; a > 0; a--) {
   if(*p & FCNMASK) { 
      literal_terms++;  /* function; inc count; det. # arguments   */
      a += *p++ & ARGMASK;}
   else {               
      literal_terms++;
      var_count++;
      p++;}}              /* variable; inc count                   */

if(var_count > max_vars_in_lit) {max_vars_in_lit = var_count;}

return(literal_terms);
}
 
/*--------------------------------------------------------------*/
/*      max_term_in_any_lit()           .............. #17
/*
/*      calculates the maximum number of terms in any literal
/*      o_maxsterm records this maximum number
/*
/*      called from:    prepare_clauses()         theo.c
/*
/*      calls:          count_lit_characters()    theo.c
/*
/*      returns:        nil
/*--------------------------------------------------------------*/
 
void max_term_in_any_lit()
{
clause *c;
literal *lit;
int i, ncc;
int lit_term_cnt;

max_term_in_nc = 0;
max_term_in_lit = 0;
max_vars_in_lit = 0;

if(nc == 0) ncc = 1; else ncc = nc; 

for(i = 0; i <= base; i++) {
   c = clist[i];
   for(lit = c->c_lit; lit != 0; lit = lit->l_next) {
      lit_term_cnt = count_lit_characters(lit);
      if(lit_term_cnt > max_term_in_nc && 
                      i >= base - ncc + 1 &&
                      i <= base) {
         max_term_in_nc = lit_term_cnt;}
      if(lit_term_cnt > max_term_in_lit) {
	     max_term_in_lit = lit_term_cnt;}}}

return;
}
			
/*---------------------------------------------------------------*/
/*      hash_all_base_clauses()            .............. #20
/*
/*      calculates a hash code for each literal and a hash code
/*      for each base clause.  It adds the clause to the literal
/*      hash table if there is only one literal and to the
/*      the clause hash table in any case
/*
/*      called from: main()                      theo.c
/*
/*      calls:       calc_lit_hc()               hash.c
/*                   calc_clause_hashcode()      hash.c
/*                   add_clause_to_ht()          hash.c
/*                   calc_alt_hashcodes()        hash.c
/*		     normalize_literal()         hash.c
/*
/*      returns:     nil
/*--------------------------------------------------------------*/
 
void hash_all_base_clauses()
{
clause *c;
literal *l;
int i;
int l_cnt, o_maxterm_tmp;
	
o_maxterm_tmp = o_maxterm;
o_maxterm = 200; 
 
for(i = 0; i <= base; i++) {
   c = clist[i];
   for(l = c->c_lit; l != 0; l = l->l_next) {
      if((l->l_id & ELIMLIT) != 0) {continue;}
      l_cnt = calc_lit_hc(l,c);}
   calc_clause_hashcode(c);
   add_clause_to_ht(c->c_X,c->c_Y,c->c_nlit);
   if(c->c_nlit == 1) {
      for(l = c->c_lit; l != 0; l = l->l_next) {
         if((l->l_id & ELIMLIT) == 0) {break;}}
      normalize_literal(l,c);
      add_clause_to_ht(l->l_X,l->l_Y,1);
      if(c->c_nv != 0) {calc_alt_hashcodes(l,c);}
      calc_alt_ehashcodes(c,l);}}  

o_maxterm = o_maxterm_tmp;

/*-----------------------------------*/
/* find first unit clause            */
/*-----------------------------------*/

units_begin = 0;
for(i = 0; i <= base; i++) {
   c = clist[i];
   if(c->c_nlit > 1) continue;
   else {
      units_begin = i;
      break;}}
return;
}       

/*-------------------------------------------------------------*/
/*      free_space_close_files()  ...................#10
/*
/*      frees space and closes files
/*
/*      called by:   main()                 theo.c
/*
/*      calls:       nobody
/*
/*      returns:     nil
/*--------------------------------------------------------------*/
 
void free_space_close_files()
{
free(clause_hash_table);
free(rand_array);
free(clause_space);  
fclose(infile);
if(herby == 1 || weakened_ply == 0 
           || satisfiable || excellent_weakened_proof) {
   if(cade_format == 0) {
      fclose(outfile);}} 
}

/*----------------------------------------------------------------*/
/*      iterative_deepening()      .............. #21
/*
/*      prove will actually perform the iteratively deepening
/*      depth first search.  We return the depth at which we stop
/*
/*      called by:      main()                     theo.c
/*
/*      calls:          init_clause_hash_table()   theo.c
/*                      add_clause_to_ht()         hash.c
/*                      calc_alt_hashcodes()       hash.c
/*                      normalize_literal()        hash.c
/*                      save_line()                pass2.c
/*                      print_continuation()       tdump.c
/*                      search_tree()              search.c
/*                      time_now()                 theo.c
/*
/*      returns:        integer regarding the result of search
/*---------------------------------------------------------------*/
 
int iterative_deepening()
{
int s;              /* normally s = 0; if proof found, s = 1     */
int i; 
clause *c;
literal *l;
int total_nodes, total_nodesM1, total_nodesM2;
int l_cnt;
int units_last_iteration;
int units_this_iteration;

total_nodes = total_nodesM1 = total_nodesM2 = 0;

if(o_print) {print0("\n\nSearch dump: \n");}

/*---------------------------------------------------------------*/
/*                                                               */
/* Main iterative deepening loop begins here                     */
/*                                                               */
/*---------------------------------------------------------------*/

	
iter_depth = 0;
units_last_iteration = 0;
units_this_iteration = 0;
s_nod0 = s_nod;
 
/*----------------------------------------------------------------*/
/* if the number of base clauses is less than 9, turn off         */
/* merge option                                                   */
/*----------------------------------------------------------------*/

if(org_base < 9) {o_merge = 0;}

/*----------------------------------------------------------------*/
/* before beginning search, clear out hash tables                 */
/*----------------------------------------------------------------*/ 
 
free(clause_hash_table);
init_clause_hash_table();       
 
/*------------------------------------------*/
/* add clauses to clause_hash_table         */
/* and unit clauses to literal_hash_table   */
/*------------------------------------------*/

while(iter_depth < MAX_ITERATIONS) {                    

for(i = 0; i <= base; i++) {
   c = clist[i];
   for(l = c->c_lit; l != 0; l = l->l_next) {
      if((l->l_id & ELIMLIT) != 0) {continue;}
      l_cnt = calc_lit_hc(l,c);}
   calc_clause_hashcode(c);

   add_clause_to_ht(c->c_X,c->c_Y,c->c_nlit);
   if(c->c_nlit == 1) {
      for(l = c->c_lit; l != 0; l = l->l_next) {
         if((l->l_id & ELIMLIT) == 0) {break;}}
      normalize_literal(l,c);
      add_clause_to_ht(l->l_X,l->l_Y,1);
      if(c->c_nv) {calc_alt_hashcodes(l,c);}
      calc_alt_ehashcodes(c,l);}}

/*--------------------------*/
/* inc iteration depth      */    
/*--------------------------*/
	
   iter_depth++;

   if(o_print)  {
      printf("\nDepth %d\n\n",iter_depth);
      OutPtr += sprintf(OutPtr,"\nDepth %d\n\n",iter_depth);}
		
/*--------------------------*/
/* Begin search!!!!         */
/*--------------------------*/

   s = search_tree(); 

/*----------------------------------------*/
/* save the current time                  */
/*----------------------------------------*/

   search_time = time_now();
    
/*----------------------------------------------------------*/
/* extended printing routine: on each iteration print:      */
/*          1. time so far                                  */
/*          2. nodes searched on this pass                  */
/*          3. clause hash table entries                    */
/*          4. clause hash table failures                   */
/*          5. unit clause hash table entries               */
/*          6. unit clause hash table failures              */
/*----------------------------------------------------------*/
	
   printf("%d {T%d N%d-%d [C%d H%d h%d U%d u%d f%d] E%d s%d t%d L%d}",
      iter_depth,(search_time - search_start_time),
      s_nod,s_nod - s_nod0,
      s_htentry,s_hthit,s_htfail,
      s_ltentry,s_lthit,s_ltfail,
      sub_array_last,s_subs_suc,s_subfxn,s_too_many_terms);
   OutPtr += sprintf(OutPtr,
     "%d {T%d N%d-%d [C%d H%d h%d U%d u%d f%d] E%d s%d t%d L%d}",
      iter_depth,(search_time - search_start_time),
      s_nod,s_nod - s_nod0,
      s_htentry,s_hthit,s_htfail,
      s_ltentry,s_lthit,s_ltfail,
      sub_array_last,s_subs_suc,s_subfxn,s_too_many_terms);
 
/*------------------------------*/
/* if proof found, print *, ret */
/*------------------------------*/

   if(s != 1) {print0("\n");}

/*----------------------------------------*/
/* if proof found (s != 0), and in        */
/* phase 1, then save line                */
/*----------------------------------------*/
 
   if(s == 1 && phase == 1) {print0("* ");save_line(0,0);}
   if(s == 1) {
      stack_hash_codes();
      if(hash_code_active_number == 0) {
	 print0("$\n");
	 if(weakened_ply == 1) {
	    if(proof_found_remember == 0) { 
	       proof_found_remember++;
               if(weakened_clause_count == 0) {
                  print0("\nProof Found; details soon!\n\n");}
               else {
                  print0(
                  "\n\nWeakened proof obtained; details soon!!\n\n");}}}
         else {
	    if(weakened_proof_remember == 0) {
	       weakened_proof_remember++;
	       print0("\n\nWeakened proof obtained; details soon!!\n\n");}}}
      else if(phase == 1) {
	 proof_found_remember = 1;
	 if(weakened_clause_count == 0) {
	     print0("\n\nProof Found; details coming!\n\n");}
         else {
	    if(weakened_proof_remember == 0) {
	       weakened_proof_remember++;
	       print0("\n\nWeakened proof obtained; details soon!!\n\n");}}}

      else print0("\n");
      return(1);} 

/*------------------------------*/
/* too much time,...quit        */
/*------------------------------*/
 
   if(s == 2)  {
      print0("\n\nTime has expired; no proof obtained\n");
      print0("Last continuation searched was:");
      print_continuation();
      ply = 0;
      return(2);}
      

/*------------------------------*/
/* compare size of trees        */
/* searched, and if appropriate */
/* widen search parameters      */
/*------------------------------*/

total_nodesM2 = total_nodesM1;
total_nodesM1 = total_nodes;
total_nodes = s_nod - s_nod0;
units_this_iteration = s_ltentry;

if(units_this_iteration == units_last_iteration) {
   o_maxterm = 1.5*o_maxterm;
   printf("X%d l%d ",o_maxterm,o_maxlit);
   OutPtr += sprintf(OutPtr,"X%d l%d ",o_maxterm,o_maxlit);
   if(o_maxterm > 100) o_maxterm = 100;}

else if((iter_depth> 3 && (total_nodes - total_nodesM1) == 
         (total_nodesM1 - total_nodesM2)) ||
	 total_nodes == 0) {
   o_maxlit = 1 + o_maxlit;
   if(o_maxlit > 15) o_maxlit = 15;
   o_maxterm = 1.5*o_maxterm;
   printf("X%d l%d ",o_maxterm,o_maxlit);
   OutPtr += sprintf(OutPtr,"X%d l%d ",o_maxterm,o_maxlit);
   if(o_maxterm > 100) o_maxterm = 100;}

units_last_iteration = units_this_iteration;

/*------------------------------*/
/* giving up search,....quit    */
/*------------------------------*/
 
   if(s == 3)  {
      print0("\n\nGiving up the search; no proof obtained\n");
      ply = 0;
      return(2);}
   
   if(s_nod == 0 && iter_depth > 10) {
      print0("\n\nNo proof obtained; check your clauses!\n");
      ply = 0;
      return(2);}}
	       
/*----------------------*/              
/* return with no proof */
/*----------------------*/
 
return(4);
}

/*-------------------------------------------------------------*/
/*      get_clause_stats()         .............. #22
/*
/*      determine the number of chars, vars, and lits
/*      in clause c
/*
/*      called by:     write_out_results    theo.c
/*
/*      calls:         nobody
/*
/*      returns:       nil
/*
/*--------------------------------------------------------------*/

void get_clause_stats(c)
clause *c;
{
literal *l;
unsigned int lit_length;

if(c->c_nlit > maxl) maxl = c->c_nlit;
if(c->c_nv   > maxv) maxv = c->c_nv;
for(l = c->c_lit; l != 0; l = l->l_next) {
   if(l->l_next) {lit_length = ((unsigned int) l->l_next - 
		   (unsigned int) l)/4 - 4;}
   else {lit_length = ((unsigned int) c->c_end - 
		   (unsigned int) l)/4 - 4;}
   if(lit_length > maxx) {maxx = lit_length;}}
}

/*------------------------------------------------------------*/
/*      write_out_results()         .............. #24
/*
/*      write_out_results writes out the cl's that form the proof
/*      (if a proof found) and write statistics out. it writes
/*      the results on the screen and to a file "in-theorem.u".
/*
/*      called by:      main()              theo.c
/*
/*      calls:          print_clause()      tdump.c
/*                      get_clause_stats()  theo.c
/*
/*      returns:        nil
/*---------------------------------------------------------------*/
 
void write_out_results(proof_found)
int proof_found;
{
clause *c, *ci, *c_next;
int i,j,jj,k,m;         
int total_search_time;     /* total search time                 */
int yes_inferred_cls;
int t_temp, t1_temp, t_power;
int x_temp;
char t1_char;
int leading_zeros;
int cls_id;
char *OutPtr2;
int used_nc_clauses, used_axioms, new_clauses, mm;
char *wpbh, *wpehP1, *wp_ptr;
int printAxiomsOnce, printNcOnce;
int weakened_clauses_used;
clause *c_pseudo1, *c_pseudo2;
int c_pseudo_indices;

/*--------------------------------------------------*/
/* set screen printing to print to a depth of 1000. */
/*--------------------------------------------------*/

if(proof_found != 1 && weakened_ply == 1) 
   print0("No Proof Obtained.\n\n");

o_print = 1000; o_step = -1;
maxx = maxl = maxv = 0;

/*----------------------------------------------------------------*/
/* determine which of the inferred clauses were used in the proof.*/  
/* fix c->c_bits by adding LIST bit for inferred clauses that     */ 
/* should be printed. for each clause in the proof found in phases*/
/* 1,2, check LIST bit of parents                                 */
/*----------------------------------------------------------------*/

for(i = 1; i <= ply; i++) {
   c = clist[depth[i].d_tc];
   if(c->c_lpar != 0) {
      ci = c->c_lpar;
      ci->c_bits |= LIST;} /* check lft par of clist[depth[i].d_tc*/
   if(c->c_rpar != 0)  {           /* check if c is a factor      */
      ci = c->c_rpar;
      ci->c_bits |= LIST;}}
      
/*-------------------------------------------------------------*/
/* now check LIST bit of other inferred cl's to det parents    */
/*-------------------------------------------------------------*/

yes_inferred_cls = 0; 
for(i = base; i > org_base; i--) {
   c = clist[i];
   if(c->c_nlit == 0) c->c_bits |= LIST; /* null cls is inc */
   if((c->c_bits & LIST) == LIST) {
      for(m = 0; m < nextu; m++) {
         if(ulist[m]->c_X == c->c_X && ulist[m]->c_Y == c->c_Y) {
	    det_dependencies(ulist[m]);
//	    print_xl_array();
	    break;}}
      yes_inferred_cls++;
      if(reinstate_parent[i + 1] != 0) {
	 clist[reinstate_parent[i + 1] - 1]->c_bits |= LIST;}
      if(c->c_lpar != 0) ci = c->c_lpar;
      ci->c_bits |= LIST;
      if(c->c_rpar != 0) ci = c->c_rpar;
      ci->c_bits |= LIST;}
   if(reinstate_parent[i] != 0) {
      clist[reinstate_parent[i] - 1]->c_bits |= LIST;}}
			 
good_inferred_cls = 0;        /* count # of inferred cls used   */

/*--------------------------------------------------------------*/
/* write out the base clauses and the clauses used from phase 0 */
/*--------------------------------------------------------------*/

if(proof_found == 1 && excellent_weakened_proof == 0
    && weakened_clause_count == 0)
{print0("\nTHE PROOF\n\n");}

else if(proof_found == 1) {print0("\nTHE W-PROOF\n\n");}

tab_amount = 0;
questionableUsed = 0;

printAxiomsOnce = printNcOnce = 0;
for(i = 0; i <= base; i++) {
   if(org_base + 1 - nc - i) {
      if(printAxiomsOnce++ == 0 && printNcOnce == 0) print0("\nAxioms:\n");}
   else {if(printNcOnce++ == 0) print0("\nNegated conclusion:\n");}
   c = clist[i];
   if(c->c_id <= org_base + 1) {
      if((c->c_bits & LIST) == LIST) {good_clauses[i] = 1;}}
   if(i <= org_base) {
      if(o_relevant == 0) {
         if((c->c_bits & LIST) == LIST) {get_clause_stats(c);}
         print_clause(c);}
      else if((c->c_bits & LIST) == LIST) {
         get_clause_stats(c);
         print_clause(c);}}
   else {
      if(o_debug > 0) {print_clause(c);}
      if(((c->c_bits & LIST) == LIST) || proof_found > 1) {
         get_clause_stats(c);
         print_clause(c);
	 if(questionable[c->c_id] == 0) {
	    copy_clause3(c);
	    if(c->c_nlit != 1) {
	       nextu--;
	       ulist[nextu++]->c_id &= ~LIST;}}
	 if(questionable[c->c_id] != 0) questionableUsed++;
	 good_inferred_cls++;}}         

   if(i == org_base) print0("\n---------------\n");
   if(i == org_base && org_base != base && yes_inferred_cls) {
      weakened_clauses_used = 0;
      for(j = 0; xl[j][0] != 0; j++) {
         if(xl[j][13] != 0) {
	    weakened_clauses_used++;}}
      if(weakened_clauses_used != 0) {
         print0("\nWeakened clauses used in proof:\n");}
      else {
         print0("\nPhase 0 clauses used in proof:\n");}

      for(j = 0; xl[j][0] != 0; j++) {
         if(xl[j][13] < 0) {
            weakened_clauses_used++;
            //for(jj = 0; xl[jj][0] != 0; jj++) {
            for(jj = 0; ulist[jj] != 0; jj++) {
               if(ulist[jj]->c_id == -xl[j][13]) { 
                  ulist[jj]->c_lpar = &c_pseudo1; 
		  if(ulist[jj]->c_rpar != 0) { 
		     ulist[jj]->c_rpar = &c_pseudo2;}
		  c_pseudo1 = (xl[j][0] & 0x0000FFF0) >> 4;
		  c_pseudo2 = (xl[j][1] & 0x0000FFF0) >> 4;
	          print_clause(ulist[jj]);
	          break;}}}} 
      if(weakened_clauses_used != 0) {
         print0("\nPhase 0 clauses used in proof:\n");}}}

/*----------------------------------*/
/* write out the clauses from       */
/* phases 1 and 2 used in proof     */
/* if proof not found in phase 0    */
/*----------------------------------*/
	
//print_xl_array();
for(i = 0; i < PROOF_ARRAY_SIZE; i++) {
   for(j = 0; j < 14; j++) {xl[i][j] = 0;}}

for(i = ply - 2; i >= 0; i--) {if(cl[i][3] != 0) break;}
begin_main_line = cl[i + 1][0] >> 16;
end_main_line   = cl[ply - 1][0] >> 16;

if(clist[base]->c_nlit != 0 && ply != 0) {
   print0("\nPhases 1 and 2 clauses used in proof:\n");
   for(i = 1; i <= ply; i++)  {
      c = clist[depth[i].d_tc];
      if(c->c_id >= begin_main_line && c->c_id <= end_main_line - 1) {
         on_main_line = 1;}
      else on_main_line = 0;

      if(c->c_lpar->c_id != c->c_id - 1 && i != 1) {print0("\n");}
      get_clause_stats(c);
      print_clause(c);
      if(questionable[c->c_id] != 0) questionableUsed++;
      if(questionable[c->c_id] == 0 && c->c_nlit != 0) {
	 if(depth[i + 1].d_tc) c_next = clist[depth[i + 1].d_tc];
	    copy_clause3(c);
	    nextu--;
	    if(ulist[nextu]->c_nlit != 1) 
	       ulist[nextu]->c_bits = (ulist[nextu]->c_bits & ~LIST);
	    nextu++;}

      if(on_main_line == 1 && questionable[c->c_id] == 0) {
         copy_clause3(c);}

      if(i > 1) {
	 cls_id = c->c_id;
	 if((questionable[cls_id] == 0) 
	    && questionable[cls_id + 1] == 1) {
	    copy_clause3(c);}}}}

statistics:
if(proof_found == 1 && weakened_clause_count == 1) {
   print0("\nEND OF W-PROOF\n");}

weakened_ply = 0;
for(i = 0; i < 2000; i++) {
   if(questionable[i] != 0) {weakened_ply = 1;}}

if(weakened_ply == 1) {
   reconsider_proof(proof_found);
   for(i = 0; i <2000; i++) {questionable[i] = 0;}
   if(excellent_weakened_proof == 1) goto done_analysis;

/*---------------------------------------------------------------*/
/* ANALYZE WEAKENED PROOF
/*---------------------------------------------------------------*/

   print0("\n---------------------------------------------------*/\n");
   print0("ANALYSIS OF WEAKENED PROOF\n\n");

   printf("Clause %d was weakened:  ",change_array[1][1]);
   OutPtr += sprintf(OutPtr,"Clause %d was weakened:  ",change_array[1][1]);
   print_clause(clist[change_array[1][1] - 1]);

/*---------------------------------------------------------------*/
/* type of weakening (for first weakened clause)
/*---------------------------------------------------------------*/

   if(weak[randy][6] ==3 || weak[randy][6] == 6) {
      printf("The clause was weakened by deleting literal %d\n\n",
         change_array[1][2]);
      OutPtr += sprintf(OutPtr,
	     "The clause was weakened by deleting literal %d\n\n",
                               change_array[1][2]);}
	    
   else if(weak[randy][6]==1 
                        || weak[randy][6] == 4) {
      print0("The clause was weakened by replacing a constant with a variable.\n\n");}

   else if(weak[randy][6] == 2 
                        || weak[randy][6] == 5) {
      print0("The clause was weakened by replacing a function with a variable.\n\n");}

/*---------------------------------------------------------------*/
/* type of weakening (for 2nd weakened clause if there was one)
/*---------------------------------------------------------------*/

   if(weakened_clause_count == 2) { 
      printf("Clause %d was also weakened:  ",change_array[2][1]);
      OutPtr += sprintf(OutPtr,"Clause %d was also weakened:  ",change_array[2][1]);
      print_clause(clist[change_array[2][1] - 1]);

      if(weak[randy2][6] ==3 || weak[randy2][6] == 6) {
         printf("The clause was weakened by deleting literal %d\n\n",
              change_array[2][2]);
         OutPtr += sprintf(OutPtr,"The clause was weakened by deleting literal %d\n\n",
              change_array[2][2]);}
	    
      else if(weak[randy2][6]==1 
                        || weak[randy2][6] == 4) {
         print0("The clause was weakened by replacing a constant with a variable.\n\n");}

      else if(weak[randy2][6] == 2 
                        || weak[randy2][6] == 5) {
         print0("The clause was weakened by replacing a function with a variable.\n\n");}}

/*--------------------------------------------------------------*/
/* print length of weakened proof, nodes
/*--------------------------------------------------------------*/

   printf("Length of weakened proof is %d inferences.\n",good_inferred_cls + ply);
   OutPtr += sprintf(OutPtr,"Length of weakened proof is %d inferences.\n",good_inferred_cls + ply);

   printf("Nodes searched for weakened proof is %d.\n",s_nod);
   OutPtr += sprintf(OutPtr,"Nodes searched for weakened proof is %d.\n",s_nod);

   if(good_inferred_cls + ply > 0 && good_inferred_cls + ply < 4) {
      print0("Weakened theorem too easy!\n");}
 
/*---------------------------------------------------------------*/
/* print number of weakened inference are good to use in
/* proof of theorem, then the number of axioms to add to sos. 
/*---------------------------------------------------------------*/

   printf("%d clauses in w-proof usable to prove theorem.\n",nextu);
   OutPtr += sprintf(OutPtr,
          "%d clauses in w-proof usable to prove theorem.\n",nextu);

   new_clauses = 0;
   for(mm = 0; mm < nextu; mm++) {
      if(ulist[mm]->c_bits & ACTIVE) {new_clauses++;}}
   printf("Of these, %d will be added to the base.\n",new_clauses);
   OutPtr += sprintf(OutPtr,
	  "Of these, %d will be added to the base.\n",new_clauses);

   used_axioms = 0;
   for(mm = 0; mm <= org_base - nc; mm++) {
      if(clist[mm]->c_bits & LIST) {used_axioms++;}}
   printf("%d of given %d axioms used in w-proof placed in sos.\n",
      used_axioms,org_base - nc + 1);
   OutPtr += sprintf(OutPtr,
          "%d of given %d axioms used in w-proof placed in sos.\n",
      used_axioms,org_base - nc + 1);

   used_nc_clauses = 0;
   for(mm = org_base - nc + 1; mm <= org_base; mm++) {
      if(clist[mm]->c_bits & LIST) {used_nc_clauses++;}}
   printf("%d of given %d sos clauses used in w-proof.\n\n",
	     used_nc_clauses,nc);
   OutPtr += sprintf(OutPtr,
          "%d of given %d sos clauses used in w-proof.\n\n",
	     used_nc_clauses,nc);
   if(good_inferred_cls + ply > 3) {
      print0("Proof of w-version may be useful.\n\n");}
   else {
      print0("Proof of w-version too short; not likely useful.\n\n");}
   
   if((clist[change_array[1][1] - 1]->c_bits & LIST) == 0) {
      print0("Weakend clause not used in proof!\n");
      weak_used = 0;}
   else {
      print0("Weakend clause used in proof.\n");
      weak_used = change_array[1][1];} 

   if(most_dc1 > 0) {
      if((clist[most_dc1 - 1]->c_bits & LIST) == 0) {
         print0("Split clause #1 not used in proof.\n");
	 split1_used = 0;}
      else {
         print0("Split clause #1 used in proof.\n");
         split1_used = most_dc1;}}

   if(most_dc2 > 0) {
      if((clist[most_dc2 - 1]->c_bits & LIST) == 0) {
         print0("Split clause #2 not used in proof.\n");
	 split2_used = 0;}
      else {
         print0("Split clause #2 used in proof.\n");
         split2_used = most_dc2;}}

   print0("\n-----------------------------------------------------*/\n");}

/*---------------------------------------------------------------*/
/* END OF ANALYSIS OF WEAKENED PROOF
/*---------------------------------------------------------------*/

done_analysis:
total_search_time = search_time - search_start_time; 

if(max_lit > 16) {
   thm_out2[thm_out2end++] = 'z';
   goto rename_theorem;}

if(satisfiable == 1) {
   thm_out2[thm_out2end++] = 's';
   goto rename_theorem;}

if(weakened_ply == 0 || excellent_weakened_proof == 1) {
   if(tough == 1 && unknown == 0) {
      thm_out2[thm_out2end++] = 'x';}

   if(tough == 1 && unknown == 1) {
      thm_out2[thm_out2end++] = 'X';}

   if(tough != 1) {
      thm_out2[thm_out2end++] = 48+theorem_rating/10;
      thm_out2[thm_out2end++] = 48+(theorem_rating%10);}

   if(proof_found == 1) {
   
      batch_tried++; batch_solved++;
      if(o_batch == 1) {
         printf(" (%d Solved, %d Tried, %d Satisfiable)",
	           batch_solved,batch_tried,batch_satisfiable);
         OutPtr += sprintf(OutPtr," (%d Solved, %d Tried, %d Satisfiable)",
	           batch_solved,batch_tried,batch_satisfiable);
         print0("\n");}
      thm_out2[thm_out2end++] = 'T';}

   if(proof_found == 1 && excellent_weakened_proof == 0) {
      print0("\nEND OF PROOF\n");}

   if(proof_found != 1) {
      batch_tried++;
      if(o_batch == 1) {
         printf(" (%d Solved, %d Tried, %d Satisfiable)",
	              batch_solved,batch_tried,batch_satisfiable);
         OutPtr += sprintf(OutPtr," (%d Solved, %d Tried, %d Satisfiable)",
	              batch_solved,batch_tried,batch_satisfiable);
         print0("\n");}
      thm_out2[thm_out2end++] = 't';}

   if(excellent_weakened_proof == 1) {
      thm_out2[thm_out2end++] = 'W';}

/*-----------------------------------------------------*/
/* convert time (in t_temp) to string which is then    */
/* suffixed to the file name.                          */
/*-----------------------------------------------------*/

if(simple_sol == 1) {phase1_time = 0;}

leading_zeros = 0;
t_temp = eliminate_time + phase1_time + eliminate_time + accumulated_time;
for(t_power = 100000; ; t_power = t_power/10) {
   if(t_temp >= t_power) {
      t1_temp = t_temp/t_power;
      t1_char = t1_temp + 48; 
      thm_out2[thm_out2end++] =  t1_char; 
      leading_zeros = 1;
      t_temp = t_temp - t1_temp*t_power;}
   else {
      if(leading_zeros == 1 || t_power == 1)  {
         t1_char = 48; 
         thm_out2[thm_out2end++] = t1_char;
         leading_zeros = 1;}}
   if(t_power == 1) {break;}}

if(proof_found == 1) {
   thm_out2[thm_out2end++] = '.';
   leading_zeros = 0;
   t_temp = good_inferred_cls + ply;
   for(t_power = 100000; ; t_power = t_power/10) {
      if(t_temp >= t_power) {
         t1_temp = t_temp/t_power;
         t1_char = t1_temp + 48; 
         thm_out2[thm_out2end++] =  t1_char; 
         leading_zeros = 1;
         t_temp = t_temp - t1_temp*t_power;}
      else {
         if(leading_zeros == 1 || t_power == 1)  {
            t1_char = 48; 
            thm_out2[thm_out2end++] = t1_char;
            leading_zeros = 1;}}
      if(t_power == 1) {break;}}}

thm_out2[thm_out2end++] = '.';

if(o_learn != 0) {
   if(weak[randy][3] == -1) thm_out2[thm_out2end++] = 'l';
   else if(weak[randy][3] == 0) thm_out2[thm_out2end++] = 'v';
   else if((weak[randy][3] & ARGMASK) == 0) thm_out2[thm_out2end++] = 'c';
   else thm_out2[thm_out2end++] = 'f';

   leading_zeros = 0;
   t_temp = randy;
   for(t_power = 100000; ; t_power = t_power/10) {
      if(t_temp >= t_power) {
         t1_temp = t_temp/t_power;
         t1_char = t1_temp + 48; 
         thm_out2[thm_out2end++] =  t1_char; 
         leading_zeros = 1;
         t_temp = t_temp - t1_temp*t_power;}
      else {
         if(leading_zeros == 1 || t_power == 1)  {
            t1_char = 48; 
            thm_out2[thm_out2end++] = t1_char;
            leading_zeros = 1;}}
      if(t_power == 1) {break;}}

   thm_out2[thm_out2end++] = '.';}

t_temp = s_nod + learn_nod;
leading_zeros = 0;
for(t_power = 10000000; ; t_power = t_power/10) {
   if(t_temp >= t_power) {
      t1_temp = t_temp/t_power;
      t1_char = t1_temp + 48; 
      thm_out2[thm_out2end++] =  t1_char; 
      leading_zeros = 1;
      t_temp = t_temp - t1_temp*t_power;}
   else {
      if(leading_zeros == 1 || t_power == 1)  {
         t1_char = 48; 
         thm_out2[thm_out2end++] = t1_char;
         leading_zeros = 1;}}
   if(t_power == 1) {break;}}

if(c->c_nlit) {print0("\nNo Proof Obtained !!\n");}

/*------------------------------------------------------*/
/* if proof found, rename theorem to "theorem.Utime     */
/*------------------------------------------------------*/

rename_theorem:
rename(thm_out_ptr,thm_out_ptr2);}
	
/*------------------------------------------------------------------*/
/* first print info on time for elim, phase 1, phase 2, sum of 1,2  */
/*------------------------------------------------------------------*/
   
if(weakened_ply == 1) {
   if(nextu > 0 && questionableUsed != 0) {
      for(i = 0; i < nextu; i++) {ulist[i]->c_bits |= BASE;}}}

printf("\nPHASE 0: %d s,",eliminate_time);
if(simple_sol == 0) printf(" PHASE 1: %d s,",phase1_time);
else printf(" PHASE 1: -NA- ");
if(extra_pass == 1)  {printf(" PHASE 2: %d s ",total_search_time - phase1_time);}
else printf(" PHASE 2: -NA- ");
printf("ACCUM: %d",accumulated_time);
if(simple_sol == 0) {
   printf("  Total Time: %d s\n",
             accumulated_time + eliminate_time + total_search_time);
   accumulated_time += eliminate_time + total_search_time;}
else printf("\n");
 
/*----------------------------------*/
/* print statistics on the screen   */
/*----------------------------------*/    
	
s_nod_last = s_nod;
printf("NOD: %-10d\t",s_nod);               /* nodes in tree    */
printf("RES: %-10d\t",s_res);            /* binary resolv'ts    */
printf("FAC: %-10d\t",s_fac);            /* # factors           */
printf("T: %d  ",o_maxterm_tempy);        /* max terms in lit    */
printf("V: %d  ",o_maxvar_tempy);         /* max vars in clause  */
printf("L: %d\n",o_maxlit_tempy);         /* max lits in clause  */

printf("CTE: %-10d\t",s_htentry_phase1); /* clause tble entries */
printf("CTH: %-10d\t",s_hthit_phase1);   /* hash hits           */
printf("CTF: %-10d\t",s_htfail_phase1);  /* hash failures       */
printf("CSZ: %-10d\n",o_arraysize);              /* size of ht  */
 
printf("UTE: %-10d\t",s_ltentry_phase1); /* # unit cl ht phase1 */
printf("UTH: %-10d\t",s_lthit_phase1);   /* unit cls hits       */
printf("UTF: %-10d\t",s_ltfail_phase1); /* unit clause failures */
printf("SBA: %-10d\n",sub_array_last1);

printf("BAS: %-10d\tRED: %-10d\t",org_base + 1,clauses_alive);
 
/*------------------------------------------------------------------*/
/* print proof length: number of inferred clauses that are used     */
/* plus the number of clauses found in the search (phases 1,2)      */
/*------------------------------------------------------------------*/
	
if(proof_found == 1 || excellent_weakened_proof == 1) {
   length_of_proof = good_inferred_cls + ply;
   printf("LEN: %d+%d=%d\t",good_inferred_cls,ply,length_of_proof);}
	
/*----------------------------------------------------------*/
/* print options: (1) elim strategy, (2) merge strategy     */
/* (3) linear-nc strat. (4) clause hash table               */
/* (5) literal hash table, and (6) extended search,         */
/* (7) s-subsumption strategy.                              */
/* options are only printed if they are not default values. */
/*----------------------------------------------------------*/
	
printf("OPT:");
printf(" k%d",o_elim);
printf(" m%d",o_merge);
printf(" z%d",o_nc);
printf(" h%d",o_hash);
printf(" n%d",o_extend);
printf(" b%d",o_subsume);
printf(" t%d/%d",o_batchtime,o_maxtime);
printf("\n");   

/*--------------------------------------------------*/        
/* output to theorem.p (same as output to screen)   */
/*--------------------------------------------------*/
	
OutPtr += sprintf(OutPtr,"\nPHASE 0: %d s",eliminate_time);

if(simple_sol == 0) {
   OutPtr += sprintf(OutPtr," PHASE 1: %d s ",phase1_time);}
else OutPtr += sprintf(OutPtr," PHASE 1: -NA-  ");
	
if(extra_pass == 1) { 
   OutPtr += sprintf(OutPtr,
     " PHASE 2: %d s ",total_search_time - phase1_time);}
else {
   OutPtr += sprintf(OutPtr," PHASE 2: -NA-  ");}
OutPtr += sprintf(OutPtr,"ACCUM: %d",accumulated_time);
if(simple_sol == 0) {
   OutPtr += sprintf(OutPtr,
   " Total Time: %d s\n",accumulated_time + eliminate_time + total_search_time);}
else OutPtr += sprintf(OutPtr,"\n");

OutPtr += sprintf(OutPtr,"NOD: %-10d\t",s_nod);  
OutPtr += sprintf(OutPtr,"RES: %-10d\t",s_res);
OutPtr += sprintf(OutPtr,"FAC: %-10d\t",s_fac);       
OutPtr += sprintf(OutPtr,"T: %d  ",o_maxterm_tempy);
OutPtr += sprintf(OutPtr,"V: %d  ",o_maxvar_tempy);
OutPtr += sprintf(OutPtr,"L: %d\n",o_maxlit_tempy);        

OutPtr += sprintf(OutPtr,"CTE: %-10d\t",s_htentry_phase1);  
OutPtr += sprintf(OutPtr,"CTH: %-10d\t",s_hthit_phase1);      
OutPtr += sprintf(OutPtr,"CTF: %-10d\t",s_htfail_phase1);      
OutPtr += sprintf(OutPtr,"CSZ: %-10d\n",o_arraysize);     
 
OutPtr += sprintf(OutPtr,"UTE: %-10d\t",s_ltentry_phase1);      
OutPtr += sprintf(OutPtr,"UTH: %-10d\t",s_lthit_phase1);     
OutPtr += sprintf(OutPtr,"UTF: %-10d\t",s_ltfail_phase1);      
OutPtr += sprintf(OutPtr,"SBA: %-10d\n",sub_array_last1);

OutPtr += sprintf(OutPtr,
   "BAS: %-10d\tRED: %-10d\t",org_base + 1,clauses_alive);
if(proof_found == 1) {
   OutPtr += sprintf(OutPtr,
      "LEN: %d+%d=%d\t",
           good_inferred_cls,ply,length_of_proof);}
	
OutPtr += sprintf(OutPtr,"OPT:");
OutPtr += sprintf(OutPtr," k%d",o_elim);
OutPtr += sprintf(OutPtr," m%d",o_merge);
OutPtr += sprintf(OutPtr," z%d",o_nc);
OutPtr += sprintf(OutPtr," h%d",o_hash);
OutPtr += sprintf(OutPtr," n%d",o_extend);
OutPtr += sprintf(OutPtr," b%d",o_subsume);
OutPtr += sprintf(OutPtr," t%d/%d",o_batchtime,o_maxtime);
OutPtr += sprintf(OutPtr,"\n");

if(weakened_ply == 0 || excellent_weakened_proof || satisfiable) {
   OutPtr2 = &results[0];
// printf("SIZE = %d\n",(int)OutPtr - (int)OutPtr2);
   if(cade_format == 0) {
      while(OutPtr2 <= OutPtr) {fprintf(outfile,"%c",*OutPtr2++);}}}

if(proof_found == 1) {weakened_proof_remember = 1;}
//printf("Dump Done\n");
 
}
 
/*-----------------------------------------------------------------*/
/*      main()                             .............. #25
/*
/*      The theorem prover enters here. We set the options specified
/*      by the  user, read in the clause system, prove the theorem,
/*      and write the proof out.
/*
/*      calls:   fatal()                        theo.c
/*               init_options()                 theo.c
/*               compile_clauses()              affc.c
/*               init_cl_array()                pass2.c
/*               init_clause_space()            theo.c
/*               init_stats()                   theo.c
/*               prepare_clauses()              theo.c
/*               init_random_hash_array()       theo.c
/*               hash_all_base_clauses()        theo.c
/*          (PHASE 0)
/*               eliminate()                    elim.c
/*          (PHASE 1)
/*               init_stats()                   theo.c
/*               iterative_deepening()          theo.c
/*               stack_hash_codes()             pass2.c
/*          (PHASE 2)
/*               iterative_deepening()          theo.c
/*               stack_hash_codes()             pass2.c
/*               rearrange_cl_array()           pass2.c
/*               search_proof_line()            pass2.c
/*          (PHASE 3)
/*               write_out_results()            theo.c
/*               free_space_close_files()       theo.c
/*-----------------------------------------------------------------*/
 
int main(argc,argv)
int argc;
char *argv[];

{
int zz;
int proof_found;
char decision_char;   /* char to decide whether to do another thm  */
int i,j;
int o_maxtermZ, o_maxvarZ, o_maxlitZ;
int phase2_count;

adjust_ht_size();

herby = 0;
this_computer = 0;
o_batch = 0;
o_weakened_index = 0;
o_learn = 0;
batch_start_time = time_now();
batch_tried = 0;
batch_solved = 0;
batch_satisfiable = 0;
weakened_ply = 0;
excellent_weakened_proof = 0;

/*-----------------------------------------*/
/* change_array used in modify.c to keep   */
/* track of changes made to the given thm  */
/*-----------------------------------------*/

for(i = 0; i < 50; i++) {
   for(j = 0; j < 10; j++) {change_array[i][j] = 0;}}
change_level = 1;

/*-----------------------------------------*/
/* potential_change_array is used to store */
/* potential vars and constants  that may  */
/* be changed for a weakened version of thm*/
/*-----------------------------------------*/

for(i = 0; i < 2000; i++) {
   for(j = 0; j < 7; j++) {weak[i][j] = 0;}}
p_change_number = 0;
	
begin_proof:
proof_found  = 0;
killer_hashx = 0;
killer_hashy = 0;
magic_hashx  = 0;
magic_hashy  = 0;
extra_pass   = 0;
phase        = 0;
iter_depth   = 0;
equal_pred   = 0;
lit_in_hash  = 0;
alt_HX       = 0;
unknown      = 0;
tough        = 0;
equal_xyyzxz = -2;
theorem_rating = 0;
equal_clist_index = -2;
questionableUsed = 0;

/*------------------------------------------------------*/
/* attempt to determine theorem's name                  */
/*------------------------------------------------------*/
 
enter_theorem_name(argc,argv);

/*--------------------------------------------------*/
/* if input file is in tptp format, call            */
/* CompileTptp()                                    */
/*--------------------------------------------------*/

if(o_tptp != 0) {CompileTptp(o_tptp);}

/*--------------------------------------------------*/
/* allocate space for data structures in affc.c     */
/*                                                  */
/* read in clauses and format for theorem prover    */
/* This is done in the function compile_clauses()   */
/* which is found in affc.c.                        */
/*                                                  */
/* free space used by compile; close input file     */
/*                                                  */
/*--------------------------------------------------*/

compile_clauses();

/*--------------------------------------------------*/
/* init arrays for phase 0, then fix up clauses     */
/*--------------------------------------------------*/

restart_proof:
for(i = 0; i < 65536; i++) {
   for(j = 0; j < 6; j++) {sub_array[i][j] = 0;}}

for(i = 0; i < 2000; i++) {questionable[i] = 0;}

init_cl_array();          /* init cl list in pass2.c         */
init_search_space();      /* init search space               */
init_stats();             /* zero our statistics variables   */
prepare_clauses();

final_clean(1);

proof_found_remember = 0;

if(max_lit > 16) {
   proof_found = 0;
   goto write_results;}

modify_clause();

/*---------------------------------------------------*/
/* initialize the random_hash_array and then hash    */
/* all base clauses                                  */
/*---------------------------------------------------*/

init_random_hash_array();
hash_all_base_clauses();

if(satisfiable == 1) {
   if(o_batch == 1) batch_satisfiable++;
   proof_found = 0;
   goto write_results;}

/*----------------------------------------------------*/
/* PHASE 0                                            */
/* -------                                            */
/* simplify input clauses if possible.                */
/* the array elim_inferences stores clauses generated */
/* in elim.                                           */
/*----------------------------------------------------*/

if(o_learn == 2) {
   o_learn_lits = o_maxlit;
   o_learn_char = o_maxterm;}

in_elim = 0;
if(o_elim) {
   simple_sol = 0;
   eliminate();
   if(o_debug > 1 && base != org_base) {
      printf("Phase 0 clauses:\n");
      for(i = org_base + 1; i <= base; i++) {
         print_clause(clist[i]);}
      printf("\n\n");}}
in_elim = 1;
		
/*----------------------------------------*/
/* if proof found, don't carry out search */
/*----------------------------------------*/
 
if(o_elim == 2) simple_sol = 2;
	
/*----------------------------------------------------*/
/* PHASE 1                                            */
/* -------                                            */
/* begin searching for a proof.                       */
/*----------------------------------------------------*/

phase = 1;      
ply = 0;
lit_in_hash = 0;

search_start_time = time_now(); /* start time for search  */ 
 
init_stats();           /* zero our statistics variables  */
	
if(simple_sol == 0) {
   prepare_subarray();
   /*print_substitutions();*/
   o_maxlitZ = o_maxlit;o_maxtermZ = o_maxterm;o_maxvarZ = o_maxvar;
   proof_found = iterative_deepening();
   lit_in_hash = 0;
   if(proof_found == 4) {
      o_maxlit++;
      o_maxterm = o_maxterm * 2;
      o_maxvar = 30;
      if(o_maxterm > 80) {o_maxterm = 80;}
      printf("X%d v%d l%d ",o_maxterm,o_maxvar,o_maxlit);
      OutPtr += sprintf(OutPtr,
	     "X%d v%d l%d ",o_maxterm,o_maxvar,o_maxlit);
      proof_found = iterative_deepening();}
   if(o_debug > 2) {print_substitutions();}
   phase1_time = search_time - search_start_time;
   sub_array_last1 = sub_array_last;
   solved_iteration = iter_depth;
   s_ltentry_phase1  = s_ltentry;
   s_htentry_phase1  = s_htentry;
   s_hthit_phase1    = s_hthit;
   s_lthit_phase1    = s_lthit;
   s_htfail_phase1   = s_htfail;
   s_ltfail_phase1   = s_ltfail;
   s_htelim_phase1   = s_htelim;
   s_subsume_phase1  = s_subsume;
   s_unit_res_phase1 = s_unit_res;
   s_unit_sub_phase1 = s_unit_sub;
   s_ltentry = s_htentry = s_hthit = s_lthit = 0;
   s_htfail = s_ltfail = s_subsume = s_unit_res = s_unit_sub = 0;
   s_htelim = 0;
	
/*----------------------------------------------------*/
/* PHASE 2                                            */
/* -------                                            */
/* search for hash codes on hash_code_list.           */
/*----------------------------------------------------*/

   phase = 2;
   //phase2_count = 0;
   while(hash_code_active_number != 0 && proof_found < 2) {
      ply = 0; 
      if(o_debug > 3) {print_substitutions();}
      for(i = 0; i < 65536; i++) {
         for(j = 0; j < 6; j++) {sub_array[i][j] = 0;}}
      prepare_subarray();
      extra_pass = 1;
      o_maxlit = o_maxlitZ;o_maxterm = o_maxtermZ;o_maxvar = o_maxvarZ;
      proof_found = iterative_deepening();
      lit_in_hash = 0;
      s_ltentry = s_htentry = s_hthit = s_lthit = 0;
      s_htfail = s_ltfail = s_subsume = s_unit_res = s_unit_sub = 0;
      s_htelim = 0;}}

/*----------------------------------------------------*/
/* PHASE 3                                            */
/* -------                                            */
/* check proof.  if proof is  ok, then print results  */
/*----------------------------------------------------*/

phase = 3;
if(simple_sol == 0 && proof_found < 2) {search_proof_line();}

mark_questionable();

write_results:
write_out_results(proof_found);

/*----------------------------------------------*/
/* free allocated space and close output file   */
/*----------------------------------------------*/

free_space_close_files();
 
/*----------------------------------------------*/
/* prove another theorem?                       */
/*----------------------------------------------*/

another_theorem:
if(o_batch == 1 || (weakened_ply == 1 && excellent_weakened_proof == 0)){
   goto begin_proof;}

if(cade_format == 1) fatal("\nOK, Quitting Now\n");

printf("Another theorem(Y,N)? ");
decision_char = getchar();printf("%c",decision_char);
if(decision_char == '\n') {goto another_theorem;}

zz = getchar();printf("%c",zz);
if(zz != '\n') {printf("%c",'\n');goto another_theorem;}
if(decision_char == 'y' || decision_char == 'Y') goto begin_proof;
else if(decision_char == 'n' || decision_char == 'N') {
   fatal("\nOK, Quitting Now\n");}
else goto another_theorem;
}

/*---------------------------------------------------------------*/
/*     fix_proof()  ...................................#28
/*
/*     called by:
/*
/*     calls:
/*
/*     returns:  
/*--------------------------------------------------------------*/

int fix_proof()
{
return(0);
}

/*------------------------------------------------------------------*/ 
/*      compile_clauses()         .......................#28
/*
/*      called by:  main()                  theo.c
/*
/*      calls:      pf_initsym()            affc.c
/*                  c_init()                affc.c
/*                  yyparse()               affc.c
/*
/*      returns:    nil
/*------------------------------------------------------------------*/
 
void compile_clauses()
{
int i;

yyerrflag = 0;                  /* error recovery flag        */
yymorfg = 0;
yyprevious = YYNEWLINE;

theobufptr = theoCodeBuffer;
nc = 0;
pf_initsym();
c_init();

//printf("compile_clauses: enter yyparse\n");
yyparse();
//printf("compile_clauses: exit yyparse\n");

/*------------------------------------------------------------*/
/*  determine the # of chars in compiled_code_buffer
/*------------------------------------------------------------*/

compiled_code_size = comp_code_buffer_index -
	     compiled_code_buffer;
}

/*-----------------------------------------------------------------*/
/* void final_clean()  .............................#26
/*
/* called by: prepare_clauses()
/*
/* calls: nobody
/*
/* returns: nil
/*-----------------------------------------------------------------*/

void final_clean(type)
int type;
{
clause *c1, *c2;
literal *l1, *l2, *l3, *l_prev;
int i, j;
unsigned int *pend, *qend, *p, *q;
int clean_count, tope;
int lit_cnt;
int subTotal1, sT1;
int subTotal2, sT2;
int equal_tries, tries;

if(type == 1) tope = base;
else tope = nextu - 1;

final_clean_top:

clean_count = 0;
if(type == 1) clean_count = elim_equal_const();

/*-----------------------------------------------------*/
/* scan clauses to eliminate ~Equal(a,a)
/*-----------------------------------------------------*/

equal_tries = 0;
for(i = 0; i <= tope; i++) { 
   if(type == 1) c1 = clist[i];
   else c1 = ulist[i];
   if(c1->c_nlit == 1) continue;
   lit_cnt = 0;
   //if(randy >= 0 && weak[randy][0] == i + 1) continue;
   for(l1 = c1->c_lit; l1 != 0; l1 = l1->l_next) {
      lit_cnt++;
      if((l1->l_id & 0x8000FFFF)==(equal_pred | FCNMASK)) {
         if(!test_equality(l1)) {continue;}
	 clean_count++;
	 equal_tries++;
	 delete_literal(c1,0,l1,lit_cnt);
	 normalize_literal(c1->c_lit,c1);
	 if(c1->c_nlit == 1) c1->c_bits |= UNIT;
	 if(type == 1) continue;
	 for(l3 = c1->c_lit; l3 != 0; l3 = l3->l_next) {
	    calc_lit_hc(l3,c1);}
	 calc_clause_hashcode(c1);}}}

/*-----------------------------------------------------*/
/* scan for units (c1) that have an identical
/* in clause c2.  Delete literal in c2.
/*-----------------------------------------------------*/


tries = 0;
for(i = 0; i <= tope ; i++) {
   if(type == 1) c1 = clist[i];
   else c1 = ulist[i];
   if(c1->c_nlit != 1) continue;
   //if(randy >= 0 && weak[randy][0] == i + 1) continue;
   l1 = c1->c_lit;
   c1->c_nv = biggest_var(l1);
   for(j = 0; j <= tope; j++) {
      if(i == j) continue;
delete2:
      if(type == 1) c2 = clist[j];
      else c2 = ulist[j];
      if(c2->c_nlit == 1) continue;
      lit_cnt = 0;
      //if(randy >= 0 && weak[randy][0] == j + 1) continue;
      for(l2 = c2->c_lit; l2 != 0; l2 = l2->l_next) {
         lit_cnt++;
         p = &l1->l_id;
         q = &l2->l_id;
         if(((unsigned int) *p ^ (unsigned int) *q) != FCNMASK) {
	    continue;}

         standardize0(c1);
	 standardize(c1,c2);
         clear_var_substitution_tble(c1->c_bits,c1->c_nv);
	 clear_var_substitution_tble(c2->c_bits,c2->c_nv);
	 if(unify(&l1->l_id,&l2->l_id)) {
//printf("VIDs: c1VID=%x c2VID=%x\n",c1->c_bits & VID,c2->c_bits&VID);
//printf("v1=%d v2=%d....  ",c1->c_nv, c2->c_nv);
	    standardize0(c2);
	    subTotal1 = subTotal2 = 0;
	    for(sT1 = 0; sT1 < c1->c_nv; sT1++) {
	    //   printf("*%x",
	    //   var_substitution_tbl[sT1]);
	       subTotal1 += var_substitution_tbl[sT1];}
	    for(sT2 = 0; sT2 < c2->c_nv; sT2++) {
	     //  printf(".%x",
	     //  var_substitution_tbl[MAX_VARS+sT2]);
	       subTotal2 += var_substitution_tbl[MAX_VARS+sT2];}
//printf(" s1=%d s2=%d \n",subTotal1,subTotal2);
//print_clause(c1);print_clause(c2);getchar();
	    if(c1->c_nv == 0 &&
	         (c2->c_nv == 0 || subTotal2 == 0)) goto LitsSame;
            if(subTotal2 == 0 && c1->c_nv > 0) goto LitsSame;
	    continue;}
         else {
	    standardize0(c2);
	    continue;}

/*----------------------------------------------*/
/* identical comp. literals found, delete       */
/*----------------------------------------------*/

         LitsSame:
         clean_count++;
	 tries++;
	 delete_literal(c2,c1,l2,lit_cnt);
	 if(c2->c_nlit == 1) c2->c_bits |= UNIT;
	 normalize_literal(c2->c_lit,c2);
	 if(type == 1) goto delete2;
	 for(l3 = c2->c_lit; l3 != 0; l3 = l3->l_next) {
	    calc_lit_hc(l3,c2);}
	 calc_clause_hashcode(c2);
         goto delete2;}}}

if(clean_count != 0) goto final_clean_top;

fclean:
max_lit = 0;
for(i = 0; i <= base; i++) {
   if(clist[i]->c_nlit > max_lit) { max_lit = clist[i]->c_nlit;}}
//printf("C%d: Exit final clean %d %d\n",TC,equal_tries,tries);
return;
}

/*-----------------------------------------------------*/
/* void delete_literal()
/*
/* called by: final_clean()
/*
/* calls:     nobody
/*
/* returns:   nil
/*-----------------------------------------------------*/

void delete_literal(c,c2,l,lit_number)
clause *c, *c2;
literal *l;
int lit_number;
{
literal *l_prev;
unsigned int *p, *q;

/*
if(c2 == 0) {
   printf("Delete literal %d in clause %d\n",
   lit_number,c->c_id);
   OutPtr += sprintf(OutPtr,
   "Delete literal %d in clause %d\n",
   lit_number,c->c_id);
   printf("Given: ");
   OutPtr += sprintf(OutPtr,"Given: ");
   print_clause(c);}

else {
   printf("Resolve away literal %d in clause %d\n",
          lit_number, c->c_id);
   OutPtr += sprintf(OutPtr,
          "Resolve away literal %d in clause %d\n",
	  lit_number,c2->c_id);
   printf("Given: "); 
   OutPtr += sprintf(OutPtr,"Given: ");
   print_clause(c);
   printf("Resolver: (%d)\n",c2->c_id);
   OutPtr += sprintf(OutPtr,"Resolver: ");
   print_clause(c2);}
*/

c->c_nlit--;
if(l->l_next == 0) {
   for(l_prev = c->c_lit; l_prev != 0; l_prev = l_prev->l_next) {
      if(l_prev->l_next == l) {
          l_prev->l_next = 0;
	  c->c_end = &l->l_next;
	  goto delete_end;}}}

/*----------------------------------------------*/
/* remove non-terminal literal                  */
/*----------------------------------------------*/

p = &l->l_next;
q = l->l_next;
while((unsigned int) q < (unsigned int) c->c_end) {
   *p = *q;
   if(*p & 0x1FFF0000) {
       *p = *p - ((unsigned int) q - (unsigned int) p);}
   p++;q++;}
c->c_end = p;

delete_end:

/*
printf("Resulting: \n");
OutPtr += sprintf(OutPtr,"Resulting: \n");
print_clause(c);
printf("\n");
sprintf(OutPtr++,"\n");
getchar();
*/

return;
}

/*-------------------------------------------------------------------*/
/*      adjust_ht_size()            ......................#6a
/*
/*      called from:   main()                 theo.c
/*
/*      calls:         nobody
/*
/*      returns:       nil
/*-------------------------------------------------------------------*/

void adjust_ht_size()
{
char *mz;
int i, j, k;
char hname[64];
int mem_dynamic_size;

mem_size = 512000;
mem_dynamic_size = mem_size/2;

for(i = 0; i < 20 ; i++) {hname[i] = -1;}
gethostname(hname,64);
//if(cade_format == 0) system("who");
for(i = 0; i < 20; i++) {if(hname[i] == -1) break;}

o_arraysize = 2*4194304; 
array_sizem1 = o_arraysize - 1;  
array_power = 23;
return;

mz = &mem_array;
for(j = 0; *mz != '#'; j++) {
    if(strncmp(mz,&hname[0],i - 1) == 0) {
	while( *mz != '.')  {mz++;}
	mz++;
        mem_size = 0;
        while(*mz >= '0' && *mz <= '9')
           mem_size = mem_size * 10 + ( (int) (*mz++ - '0') );
	while( *mz != '.')  {mz++;}
	mz++;
        mem_dynamic_size = 0;
        while(*mz >= '0' && *mz <= '9')
           mem_dynamic_size=mem_dynamic_size * 10 + ((int) (*mz++ - '0') );
        break;}
    mz++;}

while((o_arraysize * 12) + 60000000 > mem_dynamic_size * 1000) {
   o_arraysize = o_arraysize/2;
   array_sizem1 = o_arraysize - 1;
   array_power = array_power - 1;}

while((o_arraysize * 12) + 60000000 < (1000 * mem_dynamic_size)/2) {
printf("increasing %d\n",o_arraysize);
   o_arraysize = 2 * o_arraysize;
   array_sizem1 = o_arraysize - 1;
   array_power = array_power + 1;}

printf("RAM = %d  D-RAM = %d  HT = %d (%d)  FREE = %d\n",
		mem_size,mem_dynamic_size,(o_arraysize*12)/1000,
		array_power,mem_dynamic_size - (o_arraysize*12)/1000);
}

