open Lib
open Printf

let ()= out_str "\n---------------- iProver v0.1 ---------------\n"

type options = 
    {mutable problem_path : string; 
     mutable include_path : string; 
     mutable problem_files : string list}

(* --problem_path /home/TPTP-v3.0.1/Problems/ --input_file  Problems/ANA/ANA002-1.p*)      
(* --include_path /home/TPTP-v3.0.1/Problems/  "include('Axioms/SET003-0.ax')." *)

let default_options = 
  {problem_path  = "";
   include_path  = ""; 
   problem_files = []}

let options  = default_options
    
(*files are added in reverse order *)

let add_file_options options file_name = 
  options.problem_files <- file_name::options.problem_files
					
let default_arg_fun = add_file_options options

let input_file_fun  = default_arg_fun     

let problem_path_fun path = 
  options.problem_path <- path 

let include_path_fun path = 
  options.include_path <- path 

let spec_list = 
  [
   ("--problem_path", Arg.String(problem_path_fun), "problem path prefix");
   ("--include_path", Arg.String(include_path_fun), "include path prefix");
   ("--input_file", Arg.String(input_file_fun), "input file name")
 ]
        
let usage_msg = 
  "iproveropt [--problem_path path_prefix] [--include_path path_prefix] [--input_file file] input_file_1 ... input_file_n"
    
let read_args() = 
  Arg.parse spec_list default_arg_fun usage_msg
    
let () = read_args(); 
  if options.problem_files = [] then failwith "No problem files to solve"

let () = 
  let add_path fn = options.problem_path^fn in
  options.problem_files <- (List.map add_path options.problem_files)

(*
let problem_files_out = 
  out_str_debug ("%Problem files: "^(list_of_str_to_str options.problem_files ",")^"\n")
*) 

let parsed_all = 
  ParseFiles.parse_files options.include_path options.problem_files

let _ = Parsed_input_to_db.parsed_input_to_db parsed_all

(*let ()=out_str_debug ("debug iprover:\n "
		^(Clause.clause_list_to_string !Parsed_input_to_db.clause_list_ref))
*) 

(****************************************)
let () = Instantiation.start_instantiation ()
(****************************************)
